/*
*       Copyright (c) 2011-2022 by esd electronics gmbh
*
*  This software is copyrighted by and is the sole property of
*  esd gmbh.  All rights, title, ownership, or other interests
*  in the software remain the property of esd gmbh. This
*  software may only be used in accordance with the corresponding
*  license agreement.  Any unauthorized use, duplication, transmission,
*  distribution, or disclosure of this software is expressly forbidden.
*
*  This Copyright notice may not be removed or modified without prior
*  written consent of esd gmbh.
*
*  esd gmbh, reserves the right to modify this software without notice.
*
*  esd electronics gmbh                   Tel. +49-511-37298-0
*  Vahrenwalder Str 207                   Fax. +49-511-37298-68
*  30165 Hannover                         http://www.esd.eu
*  Germany                                sales@esd.eu
*
*************************************************************************
*
* Description:
* =============================================================================
* This module contains the updater for ECS-PCIe/FPGA, ECS-XMC/FPGA and ECS-PMC/FPGA cards.
*
*/




/* Includes: */
/*************/
#include "ess.h" /* includes essSystem.h which usually includes all standard stuff like <stdio.h> etc. */




/* Application/Slave configuration macros: */
/*******************************************/

#if defined BOARD_ECS_XMC_FPGA
#define BOARD_ORDER_NUMBER  "E.1102.02"
#define BOARD_NAME          "ECS-XMC/FPGA"
#elif defined BOARD_ECS_PMC_FPGA
#define BOARD_ORDER_NUMBER  "E.1104.02"
#define BOARD_NAME          "ECS-PMC/FPGA"
#elif defined BOARD_ECS_PCIE_FPGA
#define BOARD_ORDER_NUMBER  "E.1106.02"
#define BOARD_ORDER_NUMBER2 "E.1106.04"  /* Low profile variant */
#define BOARD_NAME          "ECS-PCIe/FPGA"
#endif


/* The SLAVE_ defines below might be overwritten at runtime, see Init_GenericObjects() */

#ifndef SLAVE_SOFTWAREVERSION                           /* Your version string, used in CoE Object 0x100a and debug output */
# define SLAVE_SOFTWAREVERSION                          "esd " BOARD_NAME " Updater (built: " __DATE__ " " __TIME__ ")"
#endif

/* Other macros: */
/****************/
                                                        /* Used to get No. of elements in array */
#define ARRAY_LEN(arr)                                  ((uint32_t)(sizeof(arr) / sizeof((arr)[0])))
#define PRINT(x)                                        do { ESS_PRINTF x; ESS_PRINTF("\n"); } while (0)
#define MEM_DUMP(title, pData, len)                     do { uint32_t _currOffset = 0; ESS_PRINTF("%s\n0x%.4x: ", (title), _currOffset); if (pData) { while (_currOffset < (uint32_t)(len)) { ESS_PRINTF("%.2x ", ((uint8_t*)(pData))[_currOffset]); _currOffset++; if ((_currOffset % 16 == 0)) ESS_PRINTF("\n0x%.4x: ", _currOffset); else if ((_currOffset % 8) == 0) ESS_PRINTF(" "); } } ESS_PRINTF("\n"); } while (0)

#if defined(_MSC_VER)
# pragma warning(disable:4127)                          /* avoid "conditional expression is constant" warnings... */
#endif

/* Declarations: */
/*****************/

static const unsigned char ecs_fpga_data[]={
#if defined BOARD_ECS_XMC_FPGA || defined BOARD_ECS_PCIE_FPGA
#include "ecs_xmc_fpgadata.c"
#elif defined BOARD_ECS_PMC_FPGA
#include "ecs_pmc_fpgadata.c"
#else
#error Unknown board!
#endif
};

#ifdef __VXWORKS__
#define main __ecsUpdater   /* Redefine main for to be called from helper */
int main(int argc, char **argv);
#endif


static const unsigned int ecs_fpga_len = sizeof(ecs_fpga_data);

static int ShowVersionInfos(void)
{
    ESS_DEVICE_INDEX i;
    ESS_HANDLE h;
    int devcnt;

    PRINT(("Application: %s", SLAVE_SOFTWAREVERSION));
    PRINT(("Stack: %s", essGetVersion()));

    PRINT(("Trying to open all devices... (CFG_ESS_MAX_DEVICES = %u)\n", CFG_ESS_MAX_DEVICES));
    devcnt=0;

    for (i = 0; i < CFG_ESS_MAX_DEVICES; ++i) {
        ESS_RESULT res;
        char bufstr[200];
        uint32_t buf32;
        unsigned int esc_type,esc_revi,esc_build;

        res = essOpen(i, NULL, &h);
        if (res != ESS_RESULT_SUCCESS)
            continue;

        /* All these IOCTLs are not necessarily supported, so no error if the're no infos: */
        PRINT(("Device 0 found:"));

        res = essIoctl(h, ESS_IOCTL_GET_HAL_BUILDINFOSTR, bufstr, (uint32_t)sizeof(bufstr));
        if (res == ESS_RESULT_SUCCESS)
            PRINT((" HAL: %s", bufstr));
        res = essIoctl(h, ESS_IOCTL_GET_DEV_DRIVERINFOSTR, bufstr, (uint32_t)sizeof(bufstr));
        if (res == ESS_RESULT_SUCCESS)
            PRINT((" Driver: %s", bufstr));

        res = essIoctl(h, ESS_IOCTL_GET_DEV_PRODNAMESTR, bufstr, (uint32_t)sizeof(bufstr));
        if (res == ESS_RESULT_SUCCESS)
            PRINT((" Product Name: %s", bufstr));
        res = essIoctl(h, ESS_IOCTL_GET_DEV_ORDERNOSTR, bufstr, (uint32_t)sizeof(bufstr));
        if (res == ESS_RESULT_SUCCESS)
            PRINT((" Order No.: %s", bufstr));
        res = essIoctl(h, ESS_IOCTL_GET_DEV_SERIALNOSTR, bufstr, (uint32_t)sizeof(bufstr));
        if (res == ESS_RESULT_SUCCESS)
            PRINT((" Serial No.: %s", bufstr));

        res = essIoctl(h, ESS_IOCTL_GET_DEV_HWVERSION32, &buf32, (uint32_t)sizeof(buf32));
        if (res == ESS_RESULT_SUCCESS)
            PRINT((" HW Version: 0x%.8x", buf32));
        res = essIoctl(h, ESS_IOCTL_GET_DEV_HWFEATURES32, &buf32, (uint32_t)sizeof(buf32));
        if (res == ESS_RESULT_SUCCESS)
            PRINT((" HW Features: 0x%.8x", buf32));

        res = essIoctl(h, ESS_IOCTL_GET_DEV_TEMP16, &buf32, (uint32_t)sizeof(buf32));
        if (res == ESS_RESULT_SUCCESS)
            PRINT((" Temperature: %u.%u deg. C", ((*((uint16_t*)&buf32) * 10) >> 4) / 10, ((*((uint16_t*)&buf32) * 10) >> 4) % 10));

        esc_type = essESCRead8(h, ESC_REG_TYPE);
        esc_revi = essESCRead8(h, ESC_REG_REVISION);
        esc_build = essESCRead16(h, ESC_REG_BUILD);
        PRINT((" TYPE REV BUILD: 0x%.2x 0x%.2x 0x%.4x\n", esc_type, esc_revi, esc_build));
        essClose(h);
        devcnt++;
    }
    if (devcnt==0) {
        PRINT(("No devices found!"));
    }
    return devcnt;
}

/* Just a wrapper for ESS_IOCTL_DEV_RAWWRITE */
static ESS_RESULT ECS_XMC_FPGA_FlashWrite(ESS_HANDLE h, const void* data, uint32_t offset, uint32_t len)
{
    ESS_IOCTL_RAW_DATA ioctlData;

    memset(&ioctlData, 0, sizeof(ioctlData));
    ioctlData.data = (void*)(uintptr_t)data;
    ioctlData.length = len;
    ioctlData.target = 0x30697073;
    ioctlData.offset = offset;

    return essIoctl(h, ESS_IOCTL_DEV_RAWWRITE, &ioctlData, sizeof(ioctlData));
}

/* Just a wrapper for ESS_IOCTL_DEV_RAWREAD */
static ESS_RESULT ECS_XMC_FPGA_FlashRead(ESS_HANDLE h, void* dest, uint32_t offset, uint32_t len)
{
    ESS_IOCTL_RAW_DATA ioctlData;

    memset(&ioctlData, 0, sizeof(ioctlData));
    ioctlData.data = dest;
    ioctlData.length = len;
    ioctlData.target = 0x30697073;
    ioctlData.offset = offset;

    return essIoctl(h, ESS_IOCTL_DEV_RAWREAD, &ioctlData, sizeof(ioctlData));
}


/* Returns 0 if image was sucessfully written, <0 for error and 1 if current flash data already has same CRC32 as file */
static int ECS_XMC_FPGA_FlashUpdate(ESS_DEVICE_INDEX devIdx, int force)
{

    unsigned char *readData;
    int writeAttempts = 0;
    int updateRes = -1;
    ESS_RESULT res;
    ESS_HANDLE h;

    readData = malloc((size_t)ecs_fpga_len);
    if (readData == NULL) {
        PRINT(("%s(): malloc(%u) failed!", __FUNCTION__, ecs_fpga_len));
        return -10;
    }

    res = essOpen(devIdx, NULL, &h);
    if (res != ESS_RESULT_SUCCESS) {
        PRINT(("%s(): essOpen() returned %s!", __FUNCTION__, essFormatResult(res)));
        return -11;
    }

    {
        char orderNo[32];

        res = essIoctl(h, ESS_IOCTL_GET_DEV_ORDERNOSTR, orderNo, (uint32_t)sizeof(orderNo));
        if (res != ESS_RESULT_SUCCESS) {
            PRINT(("%s(): ESS_IOCTL_GET_DEV_ORDERNOSTR failed (%d)!", __FUNCTION__, (int)(int16_t)res));
#ifndef BOARD_ORDER_NUMBER2
            PRINT(("%s(): This update is only for %s %s!", __FUNCTION__,BOARD_ORDER_NUMBER,BOARD_NAME));
#else
            PRINT(("%s(): This update is only for %s/%s %s!", __FUNCTION__,BOARD_ORDER_NUMBER,BOARD_ORDER_NUMBER2,BOARD_NAME));
#endif
            updateRes = -13;
            goto cleanup_and_return;
        }

        PRINT(("%s(): read card order No.: \"%s\"", __FUNCTION__, orderNo));
#ifndef BOARD_ORDER_NUMBER2
        if (strcmp(BOARD_ORDER_NUMBER, orderNo) != 0) {
            PRINT(("%s(): This update is only for %s %s!", __FUNCTION__,BOARD_ORDER_NUMBER,BOARD_NAME));
#else
        if ((strcmp(BOARD_ORDER_NUMBER, orderNo) != 0)&&(strcmp(BOARD_ORDER_NUMBER2, orderNo) != 0)) {
            PRINT(("%s(): This update is only for %s/%s %s!", __FUNCTION__,BOARD_ORDER_NUMBER,BOARD_ORDER_NUMBER2,BOARD_NAME));
#endif
            updateRes = -15;
            goto cleanup_and_return;
        }
    }

    PRINT(("%s(): reading current flash data...", __FUNCTION__));
    res = ECS_XMC_FPGA_FlashRead(h, readData, 0, (uint32_t)ecs_fpga_len);
    if (res != ESS_RESULT_SUCCESS) {
        PRINT(("%s(): ECS_XMC_FPGA_FlashRead() returned %s!", __FUNCTION__, essFormatResult(res)));
        updateRes = -31;
        goto cleanup_and_return;
    }
    if (memcmp(readData,ecs_fpga_data,ecs_fpga_len) == 0) {
        PRINT(("%s(): Updater FPGA image matches the image on the card. No update needed!", __FUNCTION__));
        if (force == 0) {
            updateRes = 1; /* 1 = no update needed */
            goto cleanup_and_return;
        }
        PRINT(("%s(): Forcemode, updating anyway! (you should never have to do so...)", __FUNCTION__));
    }

    /* 5. Write update image to flash and recalc current flash data's CRC32: */
    do {
        PRINT(("%s(): Writing image to flash...", __FUNCTION__));
        res = ECS_XMC_FPGA_FlashWrite(h, ecs_fpga_data, 0, (uint32_t)ecs_fpga_len);
        if (res != ESS_RESULT_SUCCESS) {
            PRINT(("%s(): ECS_XMC_FPGA_FlashWrite() returned %s!", __FUNCTION__, essFormatResult(res)));
        } else {
            PRINT(("%s(): Update image written.", __FUNCTION__));
            break;
        }
    } while (writeAttempts++ < 3);

    PRINT(("%s(): re-reading current flash data...", __FUNCTION__));
    res = ECS_XMC_FPGA_FlashRead(h, readData, 0, (uint32_t)ecs_fpga_len);
    if (res != ESS_RESULT_SUCCESS) {
        PRINT(("%s(): ECS_XMC_FPGA_FlashRead() returned %s!", __FUNCTION__, essFormatResult(res)));
        updateRes = -51; /* Unknown state then, maybe update was successful */
        goto cleanup_and_return;
    }
    if (memcmp(readData,ecs_fpga_data,ecs_fpga_len) != 0) {
        PRINT(("%s(): *** FAILED: data does not match update image! ***", __FUNCTION__));
        PRINT(("%s(): *** But flash data was modified.              ***", __FUNCTION__));
        PRINT(("%s(): *** Please retry the update immediately!      ***", __FUNCTION__));
        PRINT(("%s(): *** DO NOT POWER OFF YOUR BOARD UNTIL SUCCESS ***", __FUNCTION__));
        updateRes = -33; /* Unknown state then, maybe update was successful */
        goto cleanup_and_return;
    }
    PRINT(("%s(): *** OK: update image successfully written. ***", __FUNCTION__));
    updateRes = 0; /* 0 = success */

cleanup_and_return:
    essClose(h);
    free(readData);

    return updateRes;
}
    
    
#if defined(__VXWORKS__)

#define CT_MAX_ARGS 30 /* Maximum arguments for command line */
/*
 * Helper code for VxWorks limited capability to deal with many cmd line args
 */
int ecsUpdater(char *args)
{
	char *token;
	int   argc = 1;
	char *argv[CT_MAX_ARGS] = {"ecsUpdater"};

	/*
	 * Make command line from given string
	 */
	if(args != NULL) {
		for (token = strtok(args, " "); token && argc < (CT_MAX_ARGS - 1); argc++) {
			printf("%s %d token %s\n", __FUNCTION__, __LINE__, token);
			argv[argc] = token;
			token = strtok(NULL, " ");
		}
	}

	/*
	 * Call real test program. Main is redefined for VxWorks
	 */
	return(main(argc, argv));
}
#endif /* __VXWORKS__ */
    
int main(int argc, char **argv)
{
    ESS_DEVICE_INDEX devIdx = 0;
    int forcemode = 0;
    int argpos = 1;

    if (ShowVersionInfos() == 0) return 1; /* number of devices is zero, no board found */

    if (argc <= 1) {
        PRINT(("Call with device number as parameter to upgrade a board,"));
        PRINT(("e.g. %s 0",argv[0]));
        return 2;
    }
    if (strcmp(argv[argpos],"-f") == 0) {
        forcemode = 1;
        argpos++;
    }
    devIdx = (ESS_DEVICE_INDEX)atoi(argv[argpos++]);
    if (argc > argpos) {
        PRINT(("Unknown excess parameter(s)."));
        return 2;
    }

    if (ECS_XMC_FPGA_FlashUpdate(devIdx,forcemode) < 0) return 1;
    return 0;
}

/* Just to make sure the source is used as it should: */
#ifdef ESS_HAL_INCLUDED
# error essHAL.h should not be included by Application!
#endif

#ifdef ESS_PRIVATE_INCLUDED
# error essPrivate.h should not be included by Application!
#endif

