/*
*       Copyright (c) 2011-2013 by esd electronic system design gmbh
*
*  This software is copyrighted by and is the sole property of
*  esd gmbh.  All rights, title, ownership, or other interests
*  in the software remain the property of esd gmbh. This
*  software may only be used in accordance with the corresponding
*  license agreement.  Any unauthorized use, duplication, transmission,
*  distribution, or disclosure of this software is expressly forbidden.
*
*  This Copyright notice may not be removed or modified without prior
*  written consent of esd gmbh.
*
*  esd gmbh, reserves the right to modify this software without notice.
*
*  electronic system design gmbh          Tel. +49-511-37298-0
*  Vahrenwalder Str 207                   Fax. +49-511-37298-68
*  30165 Hannover                         http://www.esd.eu
*  Germany                                sales@esd.eu
*
*************************************************************************
*
* Description:
* =============================================================================
* This module contains code that shows how to use the EoE functions.
* (It omits the CoE dictionary handling, therefore it's not intended to be
* used directly - copy the EoE specific functions to your application instead)
*
* Remember: Used slave code, slave ESI and Slave EEPROM content must all match!
*
*/



/* Includes: */
/*************/
#include "ess.h" /* includes essSystem.h which includes all standard stuff like #include <stdio.h> etc. */

#if APP_EOE_USE_ESDTAP
# if defined(ESS_PLATFORM_WINDOWS) || defined(ESS_PLATFORM_LINUX)
#  include "esdtap.h"
# else
# endif
#endif



/* App configuration: */
/**********************/
                                                        /* Set to 1 to implement eoeEthernetStart(), eoeEthernetStop(), eoeEthernetRead() and eoeEthernetWrite() by esd's TAP module */
#define APP_EOE_USE_ESDTAP                              0

                                                        /* Which network adapter the EoE implementation shall use. */
#define APP_EOE_ADAPTER_IDX                             0

                                                        /* Number of frames queued from Ethernet to EtherCAT (must be at least 2) */
#define APP_EOE_TX_QUEUE_SIZE                           8

#if defined(_MSC_VER)
# pragma warning(disable:4127)                          /* avoid "conditional expression is constant" warnings... */
#endif



/* Macros: */
/*******************/
                                                        /* Used to get No. of elements in array */
#define ARRAY_LEN(arr)                                  ((uint32_t)(sizeof(arr) / sizeof((arr)[0])))
#define PRINT(x)                                        do { ESS_PRINTF("[Application] "); ESS_PRINTF x; ESS_PRINTF("\n"); } while (0)
#define MEM_DUMP(title, pData, len)                     do { uint32_t _currOffset = 0; ESS_PRINTF("%s\n0x%.4x: ", (title), _currOffset); if (pData) { while (_currOffset < (uint32_t)(len)) { ESS_PRINTF("%.2x ", ((uint8_t*)(pData))[_currOffset]); _currOffset++; if ((_currOffset % 16 == 0)) ESS_PRINTF("\n0x%.4x: ", _currOffset); else if ((_currOffset % 8) == 0) ESS_PRINTF(" "); } } ESS_PRINTF("\n"); } while (0)
#define MEMCOPY(pDst, pSrc, len)                        memcpy((pDst), (pSrc), (size_t)(len))
#define MEMCOMP(p1, p2, len)                            memcmp((p1), (p2), (size_t)(len))



/* Declarations: */
/*****************/
static void cbInOutputsActivate(ESS_CBDATA_INOUTPUTS_ACTIVATE* cbData);
static void cbEoESetIPParam(ESS_CBDATA_EOE_SETIPPARAM* cbData);
static void cbEoESetAddrFilter(ESS_CBDATA_EOE_SETADDRFILTER* cbData);
static void cbEoEFrameReceived(ESS_CBDATA_EOE_FRAME* cbData);
static void cbDCEvent(ESS_CBDATA_DC* cbData);
static void cbCyclic(ESS_CBDATA_CYCLIC* cbData);
static void cbStateRequest(ESS_CBDATA_STATE_REQUEST* cbData);



/* Constants: */
/**************/
static const ESS_SM_CONFIGURATION smConfigs[] =
{
    { /* SM0: MBoxOut */
        46, /* minSize */
        522, /* defSize */
        1024, /* maxSize */
        0x1000, /* startAddr */
        SM_TYPE_MBXOUT | REG_MASK_SMCONTROL_PDIINT /* contrByte */
    },

    { /* SM1: MBoxIn */
        46, /* minSize */
        522, /* defSize */
        1024, /* maxSize */
        0x1400, /* startAddr */
        SM_TYPE_MBXIN | REG_MASK_SMCONTROL_PDIINT /* contrByte */
    },

    { /* SM2: Outputs */
        0, /* minSize */
        0, /* defSize */
        0, /* maxSize */
        0x1800, /* startAddr */
        SM_TYPE_OUTPUTS | REG_MASK_SMCONTROL_PDIINT /* contrByte */
    },

    { /* SM3: Inputs */
        0, /* minSize */
        0, /* defSize */
        0, /* maxSize */
        0x2400, /* startAddr */
        SM_TYPE_INPUTS | REG_MASK_SMCONTROL_PDIINT /* contrByte */
    }

};

static const ESS_CALLBACKS callbacks = {
    /* ESS_CB_CYCLIC */
    &cbCyclic,

    /* ESS_CB_STATE_REQUEST */
    &cbStateRequest,

    /* ESS_CB_SYNCMANAGER */
    NULL,

    /* ESS_CB_INOUTPUTS_ACTIVATE */
    &cbInOutputsActivate,

    /* ESS_CB_OUTPUTS_UPDATED */
    NULL,

    /* ESS_CB_INPUTS_UPDATED */
    NULL,

    /* ESS_CB_COE_READWRITE */
    NULL,

    /* ESS_CB_COE_EVENT */
    NULL,

    /* ESS_CB_FOE_OPEN */
    NULL,

    /* ESS_CB_FOE_CLOSE */
    NULL,

    /* ESS_CB_FOE_DATA */
    NULL,

    /* ESS_CB_EOE_SETIPPARAM */
    &cbEoESetIPParam,

    /* ESS_CB_EOE_SETADDRFILTER */
    &cbEoESetAddrFilter,

    /* ESS_CB_EOE_FRAME */
    &cbEoEFrameReceived,

    /* ESS_CB_VOE */
    NULL,

    /* ESS_CB_EEPROM_EMULATION */
    NULL,

    /* ESS_CB_DC */
    &cbDCEvent,

    /* ESS_CB_SOE */
    NULL,

    /* ESS_CB_AOE */
    NULL,
};

/* Needed to initialize the stack with essOpen(): */
static const ESS_CONFIGURATION config = {
    ESS_ABI_VERSION,

    /* Config flags */
    ESS_CONFIG_FLAGS_USE_ISR, /* ESS_CONFIG_FLAGS_NONE */

    /* Timer interval for cyclic stack call / callback. In microseconds (Accuracy, etc. depends on hardware/platform. essStart() might also return ESS_RESULT_INVALID_TIMER_CONFIG when this value is too high or too low) */
    1 * 1000, /* us */

    /* Custom Tag, for later retrieval by essGetTag() */
    NULL,

    /* All stack callbacks */
    &callbacks,

    /* See smConfigs[] */
    smConfigs,
    ARRAY_LEN(smConfigs),
    0,

    /* Reserved, must be 0 */
    { 0, 0 },

    /* Statistics */
    NULL,

    /* Reserved */
    NULL,
};



/* Variables: */
/**************/
/* Our handle to the ESC. (Stack needs this e.g. to distinguish between different ESCs) */
static ESS_HANDLE hDev;
/* Needed for our outputs watchdog */
static ESC_STATE currESCState = ESC_STATE_INIT;


/* Other variables: */
/********************/

/* Set to ESS_TRUE when master sent IP configuration */
static ESS_BOOL eoeIPPortMode;
/* Own MAC address, set in cbEoESetIPParam() */
static uint8_t eoeMACAddr[6];
/* Own IP address, set in cbEoESetIPParam() */
static uint32_t eoeIPAddr;







#if APP_EOE_USE_ESDTAP

static TAP_ADAPTER eoeAdapter;

static void eoeEthernetStop(void)
{
    if (eoeAdapter) {
        tap_close(eoeAdapter);
        eoeAdapter = NULL;
    }

    eoeIPPortMode = ESS_FALSE;
}

/* Returns 0 on success */
static int eoeEthernetStart(int adapterIdx)
{
    eoeEthernetStop(); /* Just to be sure we're not opening/starting anything multiple times */

    if (tap_open(adapterIdx, &eoeAdapter, TAP_FLAGS_NON_BLOCKING) != TAP_RESULT_SUCCESS) {
        PRINT(("EoE: ERROR: Opening adapter failed!"));
        eoeAdapter = NULL;
        return -1;
    }

    return 0;
}

/* Returns > 0 on success, 0 when nothing to receive and < 0 on error */
static int eoeEthernetRead(void* dest, uint32_t sizeOfDest, uint16_t* len)
{
    int read = tap_read_frame(eoeAdapter, dest, (int)sizeOfDest);

    if (read <= 0)
        return read;

    *len = (uint16_t)read;

    return 1;
}

/* Returns 0 on success */
static int eoeEthernetWrite(const void* packetData, uint16_t dataLen)
{
    return tap_write_frame(eoeAdapter, packetData, dataLen);
}

#else

# ifdef __GNUC__
#  warning TODO: implement EoE functions. (Or set APP_EOE_USE_ESDTAP when esd TAP is available/wanted)
# endif

static void eoeEthernetStop(void)
{

}

/* Return 0 on success */
static int eoeEthernetStart(int adapterIdx)
{
    (void)adapterIdx;

    return -1;
}

/* Return > 0 on success, 0 when nothing to receive and < 0 on error */
static int eoeEthernetRead(void* dest, uint32_t sizeOfDest, uint16_t* len)
{
    (void)dest;
    (void)sizeOfDest;
    (void)len;

    return -1;
}

/* Return 0 on success */
static int eoeEthernetWrite(const void* packetData, uint16_t dataLen)
{
    (void)packetData;
    (void)dataLen;

    return -1;
}

#endif







/* Called when Inputs/Outputs shall be started/stopped: (Even when no outputs exist, etc.) */
static void cbInOutputsActivate(ESS_CBDATA_INOUTPUTS_ACTIVATE* cbData)
{
    (void)cbData;
}

/* Sends a gratuitous ARP request with values from eoeIPAddr/eoeMACAddr. Called when going to OP (and EoE IP addr, etc. were set up previously) */
static void eoeARPSendGratuitous(void)
{
    uint8_t data[42];
    uint8_t* arp = &data[14];

    PRINT(("EoE: Sending Gratuitous ARP request for our IP address %d.%d.%d.%d with MAC-Id %.2x-%.2x-%.2x-%.2x-%.2x-%.2x", HI_BYTE(HI_WORD(eoeIPAddr)), LO_BYTE(HI_WORD(eoeIPAddr)), HI_BYTE(LO_WORD(eoeIPAddr)), LO_BYTE(LO_WORD(eoeIPAddr)), eoeMACAddr[0], eoeMACAddr[1], eoeMACAddr[2], eoeMACAddr[3], eoeMACAddr[4], eoeMACAddr[5]));

    /* Dest: */
    MEMCOPY(&data[0], "\xff\xff\xff\xff\xff\xff", 6);
    /* Src: */
    MEMCOPY(&data[6], eoeMACAddr, 6);
    /* Eth.Type: */
    data[12] = 0x08; data[13] = 0x06;

    /* ARP: */
    /* HW addr. type: */
    arp[0] = 0; arp[1] = 1;
    /* Prot. addr. type: */
    arp[2] = 0x08; arp[3] = 0x00;
    /* HW size: */
    arp[4] = 6;
    /* Prot. size: */
    arp[5] = 4;
    /* Opcode: */
    arp[6] = 0;
    arp[7] = 1;
    /* Src MAC: */
    MEMCOPY(&arp[8], eoeMACAddr, 6);
    /* Src IP: */
    arp[14] = HI_BYTE(HI_WORD(eoeIPAddr)); arp[15] = LO_BYTE(HI_WORD(eoeIPAddr)); arp[16] = HI_BYTE(LO_WORD(eoeIPAddr)); arp[17] = LO_BYTE(LO_WORD(eoeIPAddr));
    /* Dst MAC: */
    MEMCOPY(&arp[18], eoeMACAddr, 6);
    /* Dst IP: */
    arp[24] = arp[14]; arp[25] = arp[15]; arp[26] = arp[16]; arp[27] = arp[17];

    essEoESendFrame(hDev, data, sizeof(data), NULL);
}

/* Called cyclically when in Op and eoeIPPortMode==0. Ethernet frames are received and sent to EtherCAT here. */
static void eoeGetEthernetFrames(void)
{
    struct QueueItem {
        uint8_t data[1540];
        uint16_t dataLen;
    };
    static struct QueueItem queue[APP_EOE_TX_QUEUE_SIZE];
    static int qHead = 0;
    static int qTail = 0;
    int frameRead;

    /* First: Try to send queued items: */
    if (qHead != qTail) {
        ESS_RESULT res = essEoESendFrame(hDev, queue[qTail].data, queue[qTail].dataLen, NULL);
        if (res == ESS_RESULT_SUCCESS) {
            qTail = (qTail + 1) % APP_EOE_TX_QUEUE_SIZE;

        } else if (res != ESS_RESULT_TRY_AGAIN) {
            qTail = (qTail + 1) % APP_EOE_TX_QUEUE_SIZE;
            PRINT(("EoE: ERROR: essEoESendFrame() failed!? (%s)", essFormatResult(res)));

        }
        /* else: do nothing => try again next time */
    }

    /* Second: Read new frame from Ethernet and append it to queue: */
    frameRead = eoeEthernetRead(queue[qHead].data, sizeof(queue[qHead].data), &queue[qHead].dataLen);
    if (frameRead <= 0)
        return; /* Either no data available or an error occurred */

    if (((qHead + 1) % APP_EOE_TX_QUEUE_SIZE) == qTail) {
        PRINT(("EoE: queue full!, last frame overwritten..."));
    } else {
        qHead = (qHead + 1) % APP_EOE_TX_QUEUE_SIZE;
    }
}

static void cbEoESetIPParam(ESS_CBDATA_EOE_SETIPPARAM* cbData)
{
    PRINT(("------> EoE: Set IP Parameter..."));

    if (cbData->macAddress) {
        PRINT(("Set MAC Address to: %.2x-%.2x-%.2x-%.2x-%.2x-%.2x", cbData->macAddress[0], cbData->macAddress[1], cbData->macAddress[2], cbData->macAddress[3], cbData->macAddress[4], cbData->macAddress[5]));

        MEMCOPY(eoeMACAddr, cbData->macAddress, 6);
    }

    if (cbData->ipAddress) {
        PRINT(("Set IP Address to: %d.%d.%d.%d", cbData->ipAddress[3], cbData->ipAddress[2], cbData->ipAddress[1], cbData->ipAddress[0]));
        eoeIPAddr = MAKE_DWORD(cbData->ipAddress[0], cbData->ipAddress[1], cbData->ipAddress[2], cbData->ipAddress[3]);
    } else {
        eoeIPAddr = 0;
    }

    if (cbData->subnetMask)
        PRINT(("Set Subnetmask to: %d.%d.%d.%d", cbData->subnetMask[3], cbData->subnetMask[2], cbData->subnetMask[1], cbData->subnetMask[0]));

    if (cbData->defaultGateway)
        PRINT(("Set Default Gateway to: %d.%d.%d.%d", cbData->defaultGateway[3], cbData->defaultGateway[2], cbData->defaultGateway[1], cbData->defaultGateway[0]));

    if (cbData->dnsServerIP)
        PRINT(("Set DNS Server to: %d.%d.%d.%d", cbData->dnsServerIP[3], cbData->dnsServerIP[2], cbData->dnsServerIP[1], cbData->dnsServerIP[0]));

    if (cbData->dnsName)
        PRINT(("Set DNS Name to: \"%s\"", cbData->dnsName));

    eoeIPPortMode = (eoeIPAddr != 0);
    cbData->result = EOE_RESULTCODE_SUCCESS;
}

/* Ususally this should not be needed (Only for gateways) */
static void cbEoESetAddrFilter(ESS_CBDATA_EOE_SETADDRFILTER* cbData)
{
    int i;

    PRINT(("EoE: Set Address filter..."));

    for (i = 0; i < cbData->macAddrCount; ++i) {
        int j = i * 6;
        PRINT(("    MAC Address: %.2x-%.2x-%.2x-%.2x-%.2x-%.2x", cbData->macAddresses[j], cbData->macAddresses[j + 1], cbData->macAddresses[j + 2], cbData->macAddresses[j + 3], cbData->macAddresses[j + 4], cbData->macAddresses[j + 5]));
    }

    for (i = 0; i < cbData->macAddrFilterCount; ++i) {
        int j = i * 6;
        PRINT(("    Filter: %.2x-%.2x-%.2x-%.2x-%.2x-%.2x", cbData->macAddrFilter[j], cbData->macAddrFilter[j + 1], cbData->macAddrFilter[j + 2], cbData->macAddrFilter[j + 3], cbData->macAddrFilter[j + 4], cbData->macAddrFilter[j + 5]));
    }

    PRINT(("    Filter broadcasts = %s", cbData->filterBroadcasts ? "TRUE" : "FALSE"));

    cbData->result = EOE_RESULTCODE_SUCCESS;
}

/* Called when a frame from EtherCAT was received */
static void cbEoEFrameReceived(ESS_CBDATA_EOE_FRAME* cbData)
{
    if (currESCState != ESC_STATE_OP)
        return; /* Only in OP EoE was successfully started */

    if (eoeIPPortMode) {

        /* "IP Port" mode: Master sent us an IP address, etc. so we can handle the frames ourself,
           i.e. we will reply to ARP requests in this example: */

        if (cbData->dataLen >= 14) {
            uint8_t* dstMAC = &cbData->data[0];
            uint8_t* srcMAC = &cbData->data[6];
            uint16_t ethType = MAKE_WORD(cbData->data[13], cbData->data[12]);
            uint8_t* data = &cbData->data[14];

            if (ethType == 0x8100) { /* VLAN tagged */
                ethType = MAKE_WORD(cbData->data[17], cbData->data[16]);
                data += 4;
            }

            PRINT(("    EthType = 0x%.4x, from %.2x-%.2x-%.2x-%.2x-%.2x-%.2x to %.2x-%.2x-%.2x-%.2x-%.2x-%.2x", ethType, srcMAC[0], srcMAC[1], srcMAC[2], srcMAC[3], srcMAC[4], srcMAC[5], dstMAC[0], dstMAC[1], dstMAC[2], dstMAC[3], dstMAC[4], dstMAC[5]));
            /* MEM_DUMP("Complete data:", cbData->data, cbData->dataLen); */

            /* Answer ARP requests for own IP address */
            if ((ethType == 0x0806) && (MEMCOMP(dstMAC, "\xff\xff\xff\xff\xff\xff", 6) == 0)) { /* ARP to brodcast? */
                uint16_t hwAddrType = MAKE_WORD(data[1], data[0]), protAddrType = MAKE_WORD(data[3], data[2]), op = MAKE_WORD(data[7], data[6]);
                uint8_t hwAddrSize = data[4], protAddrSize = data[5];

                if ((op == 1) && (hwAddrType == 1) && (protAddrType == 0x0800) && (hwAddrSize == 6) && (protAddrSize == 4)) { /* IPv4 request? */
                    uint32_t fromIP = MAKE_DWORD(data[17], data[16], data[15], data[14]);
                    uint32_t askedIP = MAKE_DWORD(data[27], data[26], data[25], data[24]);

                    if (askedIP == eoeIPAddr) { /* Wants our IP address? */

                        data[7] = 2; /* ARP request to ARP answer */
                        MEMCOPY(&data[18], &data[8], 6); /* Source MAC to dest MAC */
                        MEMCOPY(&data[8], eoeMACAddr, 6); /* Our MAC as source MAC */
                        MEMCOPY(&data[24], &data[14], 4); /* Source IP to dest IP */
                        data[14] = HI_BYTE(HI_WORD(eoeIPAddr)); /* Our IP as source IP */
                        data[15] = LO_BYTE(HI_WORD(eoeIPAddr));
                        data[16] = HI_BYTE(LO_WORD(eoeIPAddr));
                        data[17] = LO_BYTE(LO_WORD(eoeIPAddr));

                        PRINT(("    Answering ARP request for our IP adress... (from %d.%d.%d.%d)", HI_BYTE(HI_WORD(fromIP)), LO_BYTE(HI_WORD(fromIP)), HI_BYTE(LO_WORD(fromIP)), LO_BYTE(LO_WORD(fromIP))));
                        essEoESendFrame(hDev, cbData->data, 42, NULL);
                    }
                }
            }
        }

    } else {
        /* "Switch Port" mode: Just send frame from EtherCAT to Ethernet: */

        if (eoeEthernetWrite(cbData->data, cbData->dataLen)) {
            PRINT(("EoE: an eoeEthernetWrite() failed."));
        }
    }
}

/* Called when a DC event occured (Make sure it's not disabled by InitALEventMask(), see ESC manual for other requirements) */
static void cbDCEvent(ESS_CBDATA_DC* cbData)
{
    switch (cbData->evtType)
    {
        case ESS_DC_EVENT_SYNC0:
            PRINT(("SYNC0 Interrupt!"));
            break;

        case ESS_DC_EVENT_LATCH:
            PRINT(("DC Latch Interrupt!"));
            break;

        default:
            break;
    }
}

/* Called cyclically by stack (after essStart() was called) */
static void cbCyclic(ESS_CBDATA_CYCLIC* cbData)
{
    (void)cbData;

    if (currESCState == ESC_STATE_OP) {
        if (eoeIPPortMode) {
            /* IP Port mode: nothing to do in this example. */

        } else {
            /* Switch port mode: */
            eoeGetEthernetFrames();  /* Just sends all frames from Ethernet to EtherCAT */
        }
    }
}

/* Called when AL EVENT signaled an AL CONTROL event

   cbData->statusCode shows errors detected by stack.
   cbData->newState shows the state that will be set

   Application can override cbData->statusCode and cbData->newState
   or set cbData->tryAgain to 1 if it can't be decided yet (callback is
   called again then)
*/
static void cbStateRequest(ESS_CBDATA_STATE_REQUEST* cbData)
{
    ESS_BOOL stackFoundNoError;

    if (cbData->statusCode != REG_VAL_ALSTATUSCODE_NOERROR) {
        PRINT(("Stack found error 0x%.4x when checking transition from %s to %s (new state: %s)", cbData->statusCode, ESC_STATE_TO_STRING(LO_BYTE(cbData->transition)), ESC_STATE_TO_STRING(HI_BYTE(cbData->transition)), ESC_STATE_TO_STRING(cbData->newState)));

        stackFoundNoError = ESS_FALSE;
        /* When stack already found an error we MUST NOT override cbData->statusCode in this callback */

    } else {

        stackFoundNoError = ESS_TRUE;
    }

    PRINT(("---> Requested transition from %s (0x%.2x) to %s (0x%.2x)...", ESC_STATE_TO_STRING(LO_BYTE(cbData->transition)), LO_BYTE(cbData->transition), ESC_STATE_TO_STRING(HI_BYTE(cbData->transition)), HI_BYTE(cbData->transition)));

    switch (cbData->transition) {

        case ESC_TRANSITION_II:
            break;

        case ESC_TRANSITION_IB:
            if (stackFoundNoError) {
                PRINT(("Entering BootStrap mode!"));
            }
            break;

        case ESC_TRANSITION_IP:
            break;

        case ESC_TRANSITION_PS:
            break;

        case ESC_TRANSITION_SO:
            if (stackFoundNoError) {
                if (eoeEthernetStart(APP_EOE_ADAPTER_IDX)) {
                    PRINT(("ERROR: eoeEthernetStart() failed!"));
                    cbData->newState = ESC_STATE_SAFEOP;
                    cbData->statusCode = REG_VAL_ALSTATUSCODE_MBXEOE;

                } else {
                    if (eoeIPPortMode)
                        eoeARPSendGratuitous();
                }
            }
            break;

        case ESC_TRANSITION_OS:
            break;

        case ESC_TRANSITION_OP:
            break;

        case ESC_TRANSITION_OI:
            break;

        case ESC_TRANSITION_SS:
            break;

        case ESC_TRANSITION_SP:
            break;

        case ESC_TRANSITION_SI:
            break;

        case ESC_TRANSITION_PP:
            break;

        case ESC_TRANSITION_PI:
            break;

        case ESC_TRANSITION_BI:
            break;

        case ESC_TRANSITION_OO:
            break;

        default:
            PRINT(("Unknown transition!, from %s (0x%.2x) to %s (0x%.2x)", ESC_STATE_TO_STRING(LO_BYTE(cbData->transition)), LO_BYTE(cbData->transition), ESC_STATE_TO_STRING(HI_BYTE(cbData->transition)), HI_BYTE(cbData->transition)));
            if (stackFoundNoError) {
                /* Change state from Op to SafeOp, else stay in current state */
                cbData->newState = (LO_BYTE(cbData->transition) == ESC_STATE_OP) ? ESC_STATE_SAFEOP : LO_BYTE(cbData->transition);
                cbData->statusCode = (ESC_STATE_KNOWN(HI_BYTE(cbData->transition))) ? REG_VAL_ALSTATUSCODE_INVALIDALCONTROL : REG_VAL_ALSTATUSCODE_UNKNOWNALCONTROL;
            }
            break;
    }


    currESCState = cbData->newState;
    PRINT(("---> new state is %s (0x%.2x)", ESC_STATE_TO_STRING(currESCState), currESCState));

    if (currESCState != ESC_STATE_OP)
        eoeEthernetStop();
}


int main(void)
{
    ESS_RESULT res;

    PRINT(("Stack: %s", essGetVersion()));

#if APP_EOE_USE_ESDTAP
    {
        int tapCnt = tap_get_count();
        int i;

        PRINT(("EoE: Found %d adapters:", tapCnt));
        for (i = 0; i < tapCnt; ++i) {
            TAP_ADAPTER_INFO info;

            if (tap_get_info(i, &info) == TAP_RESULT_SUCCESS) {
                PRINT(("EoE: Adapter%d: \"%s\", Name: \"%s\"\n\t\t\tMAC-Id: %.2x-%.2x-%.2x-%.2x-%.2x-%.2x, Version: \"%s\"", i, info.productName, info.name, info.macId[0], info.macId[1], info.macId[2], info.macId[3], info.macId[4], info.macId[5], info.driverVersion));

            } else {
                PRINT(("EoE: Failed to get infos for adapter %d!", i));
            }

        }
    }
#endif

    /* Open ESC/Init stack, resulting handle will be copied to hDev */
    res = essOpen(0, &config, &hDev);
    if (res != ESS_RESULT_SUCCESS) {
        PRINT(("Opening device failed with %s", essFormatResult(res)));
        return 10;
    }

    PRINT(("Starting..."));
    res = essStart(hDev);
    if (res != ESS_RESULT_SUCCESS)
        PRINT(("essStart() returned %s", essFormatResult(res)));

    essClose(hDev);
    return 0;
}




/* Just to make sure the source is used as it should: */
#ifdef ESS_HAL_INCLUDED
# error essHAL.h should not be included by Application!
#endif

#ifdef ESS_PRIVATE_INCLUDED
# error essPrivate.h should not be included by Application!
#endif
