/*
*       Copyright (c) 2011-2018 by esd electronics gmbh
*
*  This software is copyrighted by and is the sole property of
*  esd gmbh.  All rights, title, ownership, or other interests
*  in the software remain the property of esd gmbh. This
*  software may only be used in accordance with the corresponding
*  license agreement.  Any unauthorized use, duplication, transmission,
*  distribution, or disclosure of this software is expressly forbidden.
*
*  This Copyright notice may not be removed or modified without prior
*  written consent of esd gmbh.
*
*  esd gmbh, reserves the right to modify this software without notice.
*
*  esd electronics gmbh                   Tel. +49-511-37298-0
*  Vahrenwalder Strasse 207               Fax. +49-511-37298-68
*  30165 Hannover                         http://www.esd.eu
*  Germany                                sales@esd.eu
*
*************************************************************************
*
* Description:
* =============================================================================
* Minimalistic sample application for the esd EtherCAT Slave Stack to show
* the basic usage. See complex.c for a more comprehensive sample.
*
* (Testing to build this was done by "gcc minimal.c -o minimal -DWINDOWS -I../include -L../bin/win32 -less")
*
* When the output variable is changed, it's printed.
* Input variable is changed every second (when in "SafeOp" or "Op")
*
* Remember: Used slave code, slave ESI and Slave EEPROM content must all match!
*
*
* RxPDOs (Outputs):
*     0x1600 [Assigned by default]
*         Contains Obj. 0x2000: "Output1", 32 bit
*
* TxPDOs (Inputs):
*     0x1a00 [Assigned by default]
*         Contains Obj. 0x2010: "Input1", 32 bit
*
*
* Object dictionary, automatically added items:
***********************************************
*   Objects 0x1000, 0x1008, 0x1009, 0x100a, 0x1018
*     by essODAddGenericObjects()
*   Object 0x1c00
*     by flag ESS_OD_FLAGS_HANDLE_SM_TYPES
*   Objects 0x1c12, 0x1c13
*     by essODUpdatePDOAssignment()
*   Objects 0x1600, 0x1601, 0x1a00, 0x1a01
*     by essODUpdatePDOConfiguration()
*
* Object dictionary, manually added items:
******************************************
*   Objects 0x2000
*       Output variable
*
*   Objects 0x2010
*       Input variable
*
*/

#include "ess.h"

/* Macros: */
/***********/
#define ARRAY_LEN(arr)  ((uint32_t)(sizeof(arr) / sizeof((arr)[0])))
#define PRINT(x)        do { ESS_PRINTF("[Application] "); ESS_PRINTF x; ESS_PRINTF("\n"); } while (0)
#define CHECK_RES()     do { if (res) PRINT(("ess call in line %d failed!", __LINE__)); } while (0)

/* Forward declarations: */
/*************************/
static void cbInOutputsActivate(ESS_CBDATA_INOUTPUTS_ACTIVATE* cbData);
static void cbOutputsUpdated(ESS_CBDATA_SM_EVENT* cbData);
static void cbCyclic(ESS_CBDATA_CYCLIC* cbData);
static void cbDCEvent(ESS_CBDATA_DC* cbData);
static void cbStateRequest(ESS_CBDATA_STATE_REQUEST* cbData);
static void InitDictionary(void);

/* Variables and constants: */
/****************************/
static ESS_HANDLE hDev;
static ESC_STATE currESCState = ESC_STATE_INIT;
static uint32_t input1, output1;
static ESS_STATISTICS essStats;

/* PDOs and their contents: */
static const ESS_PDO_ENTRY PDO1600Entries[] = { ESS_MAP_ENTRY( 0x2000, 0, 32 ) };
static const ESS_PDO_ENTRY PDO1a00Entries[] = { ESS_MAP_ENTRY( 0x2010, 0, 32 ) };
static const uint16_t SM2DefPDOs[] = { 0x1600 };
static const uint16_t SM3DefPDOs[] = { 0x1a00 };

/* SyncManager configuration: */
static const ESS_SM_CONFIGURATION smConfigs[] = {
    { 46, 522, 1024, 0x1000, SM_TYPE_MBXOUT | REG_MASK_SMCONTROL_PDIINT },
    { 46, 522, 1024, 0x1400, SM_TYPE_MBXIN | REG_MASK_SMCONTROL_PDIINT },
    { 0, 4, 0, 0x1800, SM_TYPE_OUTPUTS | REG_MASK_SMCONTROL_PDIINT },
    { 0, 4, 0, 0x2400, SM_TYPE_INPUTS | REG_MASK_SMCONTROL_PDIINT }
};

/* Callbacks: (functions defined here that will be called by the stack) */
static const ESS_CALLBACKS callbacks =  { &cbCyclic, &cbStateRequest, NULL, &cbInOutputsActivate, &cbOutputsUpdated, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &cbDCEvent, NULL, NULL };

/* The complete configuration given to the stack: */
static const ESS_CONFIGURATION config = { ESS_ABI_VERSION, ESS_CONFIG_FLAGS_USE_ISR, 1 * 1000, NULL, &callbacks, smConfigs, ARRAY_LEN(smConfigs), 0, { 0, 0 }, &essStats, NULL };


int main(int argc, char *argv[])
{
    ESS_DEVICE_INDEX devIdx = 0;

    if(argc > 1) {
        devIdx = (ESS_DEVICE_INDEX)atoi(argv[1]);
    }

    ESS_RESULT res = essOpen(devIdx, &config, &hDev);

    if (res != ESS_RESULT_SUCCESS) {
        PRINT(("Opening device failed with %s", essFormatResult(res)));
        return 10;
    }

    InitDictionary();
    {
        uint32_t alEventMask = essESCRead32(hDev, ESC_REG_ALEVENTMASK);
        /* Disable unwanted interrupts: */
        alEventMask &= ~0x00000040u;  /* PD watchdog expired (Usually not needed) */
        alEventMask &= ~(0x00000100u << 3); /* Inputs Read SM3 (That interrupt is only cleared when new inputs are written to SM buffer, */
        /* but we don't do that every cycle, so we don't get any more interrupts because of the permanent inputs interrupt. */
        /* So we need to disable it. */
        essESCWrite32(hDev, ESC_REG_ALEVENTMASK, alEventMask);
    }
    res = essStart(hDev);
    if (res != ESS_RESULT_SUCCESS)
        PRINT(("essStart() returned %s", essFormatResult(res)));

    /* This line is usually not reached. essStart() never returns, the application runs only by its callbacks then. */

    essClose(hDev);
    return 0;
}


static void SetDefaultPDOs(void)
{
    ESS_RESULT res;

    res = essODUpdatePDOConfiguration(hDev, 0x1600, "RxPDO1", PDO1600Entries, ARRAY_LEN(PDO1600Entries), ESS_FALSE); CHECK_RES();
    res = essODUpdatePDOConfiguration(hDev, 0x1a00, "TxPDO1", PDO1a00Entries, ARRAY_LEN(PDO1a00Entries), ESS_FALSE); CHECK_RES();
    res = essODUpdatePDOAssignment(hDev, ESS_SM_2, SM2DefPDOs, ARRAY_LEN(SM2DefPDOs), ESS_TRUE); CHECK_RES();
    res = essODUpdatePDOAssignment(hDev, ESS_SM_3, SM3DefPDOs, ARRAY_LEN(SM3DefPDOs), ESS_TRUE); CHECK_RES();
}

static void InitDictionary(void)
{
    static uint32_t slaveProductCode = SWAP32_HOST2LE(0x00000005);
    static uint32_t slaveRevisionNo = SWAP32_HOST2LE(0x01000001);
    static uint32_t slaveDeviceType = SWAP32_HOST2LE(5001U);
    static uint32_t slaveVendorId = SWAP32_HOST2LE(0xe0000017);
    static uint32_t slaveSerial = SWAP32_HOST2LE(0);
    const char* slaveDeviceName = "N/A";
    const char* slaveHWVersion = "N/A";
    const char* slaveSWVersion = "ess minimalistic Sample";
    ESS_RESULT res;

    res = essODCreate(hDev, ESS_OD_FLAGS_HANDLE_SM_TYPES); CHECK_RES();

    res = essODAddGenericObjects(hDev, &slaveDeviceType, &slaveVendorId, &slaveProductCode, &slaveRevisionNo, &slaveSerial, slaveDeviceName, slaveHWVersion, slaveSWVersion); CHECK_RES();

    {
        static const ESS_OD_OBJECT_INFOS objInfos = { "Output1", COE_DATATYPE_UDINT, COE_CODE_VARIABLE };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "Output1", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };

        res = essODObjectAdd(hDev, 0x2000, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2000, 0x00, 32, &output1, COE_ACCESS_RXMAPPABLE | COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();
    }
    {
        static const ESS_OD_OBJECT_INFOS objInfos = { "Input1", COE_DATATYPE_UDINT, COE_CODE_VARIABLE };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "Input1", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };

        res = essODObjectAdd(hDev, 0x2010, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2010, 0x00, 32, &input1, COE_ACCESS_TXMAPPABLE | COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();
    }

    SetDefaultPDOs();
}

static void cbOutputsUpdated(ESS_CBDATA_SM_EVENT* cbData)
{
    static uint32_t output1Prev;
    (void)cbData;

    if (output1 != output1Prev) {
        PRINT(("*** output1 changed to %u", SWAP32_LE2HOST(output1)));
        output1Prev = output1;
    }
}

static void cbCyclic(ESS_CBDATA_CYCLIC* cbData)
{
    (void)cbData;

    if ((currESCState == ESC_STATE_SAFEOP) || (currESCState == ESC_STATE_OP)) {
        static uint32_t nextUpdate;

        if (essGetTime() >= nextUpdate) {
            input1 = SWAP32_HOST2LE(SWAP32_LE2HOST(input1) + 1);

            essSyncInputs(hDev, ESS_ISYN_FLAGS_NONE);

            nextUpdate = essGetTime() + 1000;
        }
    }
}

static void cbStateRequest(ESS_CBDATA_STATE_REQUEST* cbData)
{
    const ESS_BOOL stackFoundNoError = (cbData->statusCode == REG_VAL_ALSTATUSCODE_NOERROR);

    PRINT(("---> Transition from %s (0x%.2x) to %s (0x%.2x):", ESC_STATE_TO_STRING(LO_BYTE(cbData->transition)), LO_BYTE(cbData->transition), ESC_STATE_TO_STRING(HI_BYTE(cbData->transition)), HI_BYTE(cbData->transition)));
    if (!stackFoundNoError)
        PRINT(("     Error %s (0x%.4x)", REG_VAL_ALSTATUSCODE_TO_STRING(cbData->statusCode), cbData->statusCode));

    PRINT(("---> New state will be %s (0x%.2x)", ESC_STATE_TO_STRING(cbData->newState), cbData->newState));

    currESCState = cbData->newState;

    if (currESCState == ESC_STATE_INIT)
        SetDefaultPDOs();
}

static void cbInOutputsActivate(ESS_CBDATA_INOUTPUTS_ACTIVATE* cbData) { (void)cbData; }

static void cbDCEvent(ESS_CBDATA_DC* cbData) { (void)cbData; }
