/*
*       Copyright (c) 2011-2022 by esd electronics gmbh
*
*  This software is copyrighted by and is the sole property of
*  esd gmbh.  All rights, title, ownership, or other interests
*  in the software remain the property of esd gmbh. This
*  software may only be used in accordance with the corresponding
*  license agreement.  Any unauthorized use, duplication, transmission,
*  distribution, or disclosure of this software is expressly forbidden.
*
*  This Copyright notice may not be removed or modified without prior
*  written consent of esd gmbh.
*
*  esd gmbh, reserves the right to modify this software without notice.
*
*  esd electronics gmbh                   Tel. +49-511-37298-0
*  Vahrenwalder Str 207                   Fax. +49-511-37298-68
*  30165 Hannover                         http://www.esd.eu
*  Germany                                sales@esd.eu
*
*************************************************************************
*
* Description:
* =============================================================================
* ess HAL for Linux. (Also as sample to build own HAL)
*
*/

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <inttypes.h>
#include <assert.h>
#include <stdarg.h>
#include <signal.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/select.h>
#include <limits.h>
#define CFG_HAVE_STD_INT_TYPES          1

#include "../essHAL.h"
#include "../../include/ess.h"

#define SYSFS_UIO_PATH                  "/sys/class/uio/"

                                        /* Expected content of /sys/class/uio/io0X/name for our device */
#define SYSFS_EXPECTED_NAME             "pexesc"

#define SYSFS_MAX_MAPS                  5

#define MAX_UIO_DEVS_TO_CHECK           100



#define MORE_DEBUG                      0

#ifdef _DEBUG
# define PRINT(x)                       do { printf("[HAL] "); printf x; printf("\n"); } while (0)
# define MEMDUMP(p, length, displOffs)  JustAnotherMemdump((p), (length), (displOffs))
#else
# define PRINT(x)                       do { } while (0)
# define MEMDUMP(p, length, displOffs)  do { } while (0)
#endif

#if (MORE_DEBUG)
# define PRINT_MORE(x)                  PRINT(x)
#else
# define PRINT_MORE(x)                  do { } while (0)
#endif

#if 0
static void JustAnotherMemdump(const void* data, int len, uint32_t addrOffset)
{
    while (len > 0) {
        const unsigned char* d = (const unsigned char*)data;
        const int lineLen = (len >= 16) ? 16 : len;
        int i;

        printf("0x%.4x:", addrOffset);

        for (i = 0; i < lineLen; ++i)
            printf(" %.2x", d[i]);
        printf(" | ");

        for (i = 0; i < lineLen; ++i)
            printf("%c", ((d[i] >= 32) && (d[i] <= 127)) ? (char)d[i] : '.');
        printf("\n");

        addrOffset += lineLen;
        len -= lineLen;
        data = d + lineLen;
    }
}
#endif


/* For ecs_xmc_fpga_flash_read() and ecs_xmc_fpga_flash_write(): */
#include "../ecs_xmc_fpga/spiflash.c"

/* For ecs_xmc_fpga_get_temperature(): */
#include "../ecs_xmc_fpga/i2c.c"

enum CARD_TYPE { /* keep in sync with other HALs! */
    CARD_TYPE_UNKNOWN = 0,
    CARD_TYPE_ECSPCIe1100,
    CARD_TYPE_ECSXMCFPGA
};
#define ESD_CARD_TYPE_TO_STR(ct)    (\
            ((ct) == CARD_TYPE_ECSPCIe1100) ? "ECS-PCIe/1100" : \
            ((ct) == CARD_TYPE_ECSXMCFPGA) ? "ECS-xMC/FPGA" : \
            "Unknown Card!")

typedef struct {
    char name[32];
    size_t offset;
    size_t size;
    void* p;
} uio_map;

typedef struct {
    int uio_idx;
    /* -1 if uio dev not opened */
    int fd;
    uio_map uio_maps[SYSFS_MAX_MAPS];

    enum CARD_TYPE card_type;
    char driver_build_info[128];

    /* Pointers below created by map pointers: (see InitDevPointers()) */
    void* pESC;
    void* pLEDs;
    void* pSysId;
    void* pPIO_XMC;
    void* pPIO_PMC;
    void* pI2C;
    void* pPIORevision;
    void* pPIOLatchControl;
    void* pFlash;
    void* pPCIEp;
    void* pDMAMemory;
    void* pDMAControl;

    ECS_XMC_FPGA_FLASH_CONFIGDATA configData;
    /* 0 = not yet, 1 = yes, -1 = read failed, -2 = CRC error */
    int configDataRead;

    int started;
    ESS_HAL_CALLBACK* isr;
    ESS_HAL_CALLBACK* cyclic;
    uint32_t cyclicInterval; /* in us */
    void* cbData;
    int have_msi;
    int have_dma;
} DEV_INFOS;

static DEV_INFOS devInfos[CFG_ESS_MAX_DEVICES];

#if (CFG_ESS_MAX_DEVICES == 0)
# error "You don't mean that, do you!?"
#endif

static uint32_t my_get_utickcount(void)
{
#if ((defined(_POSIX_C_SOURCE)) && (_POSIX_C_SOURCE >= 199309L))
    struct timespec tp;

    if (clock_gettime(CLOCK_MONOTONIC, &tp) == -1) /* Usually needs to link with -lrt */
        return 0;

    return tp.tv_sec * 1000000 + tp.tv_nsec / 1000;

#else
    struct timeval tv;

    gettimeofday(&tv, NULL);
# warning "using gettimeofday() (= not using CLOCK_MONOTONIC)"

    return tv.tv_sec * 1000000 + tv.tv_usec;
#endif
}
/* Usually you just define _GNU_SOURCE to use the newer functions */
static void my_usleep(uint32_t us)
{
#if ((defined(_XOPEN_SOURCE) && (_XOPEN_SOURCE >= 600)) || ((defined(_POSIX_C_SOURCE)) && (_POSIX_C_SOURCE >= 200112L)))
    struct timespec ts;

    ts.tv_sec = (time_t)(us / 1000000);
    ts.tv_nsec = (us % 1000000) * 1000;

    clock_nanosleep(CLOCK_MONOTONIC, 0, &ts, NULL); /* Usually needs to link with -lrt */
#elif (defined(_POSIX_C_SOURCE) && (_POSIX_C_SOURCE >= 199309L))
    struct timespec ts;

    ts.tv_sec = (time_t)(us / 1000000);
    ts.tv_nsec = (us % 1000000) * 1000;

    nanosleep(&ts, NULL); /* Usually needs to link with -lrt */
# warning "using nanosleep()! (= not using CLOCK_MONOTONIC)"
#else
    usleep(us);
# warning "using usleep()! (= not using CLOCK_MONOTONIC)"
#endif
}

/* Content's trailing spaces/linebreaks are trimmed! */
static char* ReadFileContent(const char* fileName, char* dest, size_t sizeOfDest)
{
    FILE* f;
    size_t readBytes;
    char* res = NULL;

    if ((sizeOfDest < 1) || (dest == NULL))
        return res;

    dest[0] = '\0';

    f = fopen(fileName, "rb");
    if (f == NULL)
        return res;

    readBytes = fread(dest, 1, sizeOfDest, f);
    if (readBytes > 0) {
        dest[sizeOfDest - 1] = '\0';

        while ((readBytes-- > 0) && ((dest[readBytes] == '\n') || (dest[readBytes] == '\r') || (dest[readBytes] == '\t') || (dest[readBytes] == ' ')))
            dest[readBytes] = '\0';

        res = dest;
    }

    fclose(f);
    return res;
}

/* -1: Error, 0: Different, 1: equal */
static int CompareFileContent(const char* fileName, const char* with)
{
    char content[128];

    if ((with == NULL) || (ReadFileContent(fileName, content, sizeof(content)) == NULL))
        return -1;

    return strcmp(content, with) == 0;
}

/* -1: Error = dev not found */
static int devIdx_to_uio_idx(ESS_DEVICE_INDEX devIdx)
{
    int currDevIdx = -1;
    int i;

    for (i = 0; i < MAX_UIO_DEVS_TO_CHECK; ++i) {
        char fn[128];

        snprintf(fn, sizeof(fn), SYSFS_UIO_PATH "uio%d/name", i);
        fn[sizeof(fn) - 1] = '\0';

        if (CompareFileContent(fn, SYSFS_EXPECTED_NAME) <= 0)
            continue;

        currDevIdx++;
        if (currDevIdx == devIdx)
            return i;
    }

    return -1;
}

/* (size_t)-1 on error */
static size_t ReadFileContentAsSize(const char* fileName)
{
    char content[128];
    size_t res;

    if (ReadFileContent(fileName, content, sizeof(content)) == NULL)
        return (size_t)-1;

    errno = 0;
    res = strtoul(content, NULL, 0);

    if ((res == 0) && (errno))
        res = (size_t)-1;

    return res;
}

/* ...  */
static void read_card_type_and_driver_build_info(DEV_INFOS* dev)
{
    char fn[128];
    size_t val;

    snprintf(fn, sizeof(fn), SYSFS_UIO_PATH "uio%d/device/pexesc_cardtype", dev->uio_idx);
    fn[sizeof(fn) - 1] = '\0';
    val = ReadFileContentAsSize(fn);
    dev->card_type = (val == (size_t)-1) ? 0 : (int)val;

    snprintf(fn, sizeof(fn), SYSFS_UIO_PATH "uio%d/device/pexesc_buildinfo", dev->uio_idx);
    fn[sizeof(fn) - 1] = '\0';
    ReadFileContent(fn, dev->driver_build_info, sizeof(dev->driver_build_info));
    
    dev->have_msi=0;
    snprintf(fn, sizeof(fn), SYSFS_UIO_PATH "uio%d/device/msi_irqs/", dev->uio_idx);
    fn[sizeof(fn) - 1] = '\0';
    if(access(fn,F_OK)==0) {
        dev->have_msi=1;
        PRINT_MORE(("    device is using MSI"));
    }
}

/* Just reads all from uio%d/maps/map%d/{name|offset|...} setting members to default on error */
static void uio_init_maps(DEV_INFOS* dev)
{
    char fn[128];
    int i;

    for (i = 0; i < SYSFS_MAX_MAPS; ++i) {
        uio_map* const map = &dev->uio_maps[i];

        snprintf(fn, sizeof(fn), SYSFS_UIO_PATH "uio%d/maps/map%d/name", dev->uio_idx, i);
        fn[sizeof(fn) - 1] = '\0';
        ReadFileContent(fn, map->name, sizeof(map->name));

        snprintf(fn, sizeof(fn), SYSFS_UIO_PATH "uio%d/maps/map%d/offset", dev->uio_idx, i);
        fn[sizeof(fn) - 1] = '\0';
        map->offset = ReadFileContentAsSize(fn);

        snprintf(fn, sizeof(fn), SYSFS_UIO_PATH "uio%d/maps/map%d/size", dev->uio_idx, i);
        fn[sizeof(fn) - 1] = '\0';
        map->size = ReadFileContentAsSize(fn);

        map->p = MAP_FAILED;
    }
}

/* Returns MAP_FAILED on error */
static void* uio_map_get(DEV_INFOS* dev, const char* map_name, uint32_t offset)
{
    int i;

    for (i = 0; i < SYSFS_MAX_MAPS; ++i) {
        uio_map* const map = &dev->uio_maps[i];

        if (strcmp(map_name, map->name))
            continue;

        if ((map->p == MAP_FAILED) || (map->p == NULL))
            return MAP_FAILED;

        return (uint8_t*)map->p + map->offset + offset;
    }

    return MAP_FAILED;
}

/* Just handles dev->fd */
static int uio_open(DEV_INFOS* dev)
{
    char fn[128];

    snprintf(fn, sizeof(fn), "/dev/uio%d", dev->uio_idx);
    fn[sizeof(fn) - 1] = '\0';

    dev->fd = open(fn, O_RDWR | O_SYNC);
    if (dev->fd == -1) {
        perror(fn);
        return 1;
    }

    PRINT_MORE(("uio_open() for uio%d: fd = %d", dev->uio_idx, dev->fd));
    return 0;
}

/* Just handles dev->fd */
static void uio_close(DEV_INFOS* dev)
{
    if (dev->fd != -1) {
        PRINT_MORE(("uio_close() for uio%d: (fd was %d)", dev->uio_idx, dev->fd));
        close(dev->fd);
        dev->fd = -1;
    }
}

/* Tries to map all maps, just leaving them at MAP_FAILED on error */
static void uio_start(DEV_INFOS* dev)
{
    int i;

    PRINT_MORE(("uio_start() for uio%d called:", dev->uio_idx));
    for (i = 0; i < SYSFS_MAX_MAPS; ++i) {
        uio_map* const map = &dev->uio_maps[i];

        if (map->size == (size_t)-1)
            continue;

        map->p = mmap(NULL, map->size, PROT_READ | PROT_WRITE, MAP_SHARED, dev->fd, i * getpagesize());
        PRINT_MORE(("    map '%s': 0x%.x bytes at %p (offs: 0x%x)", map->name, (unsigned int)map->size, map->p, (unsigned int)map->offset));
    }
}

/* Just unmaps all maps that are not MAP_FAILED */
static void uio_stop(DEV_INFOS* dev)
{
    int i;

    PRINT_MORE(("uio_stop() for uio%d called", dev->uio_idx));

    for (i = 0; i < SYSFS_MAX_MAPS; ++i) {
        uio_map* const map = &dev->uio_maps[i];

        if (map->p != MAP_FAILED) {
            munmap((char*)map->p - map->offset, map->size);
            map->p = MAP_FAILED;
        }
    }
}

/* ... */
static int InitDevPointers(DEV_INFOS* dev)
{
    PRINT_MORE(("InitDevPointers() for uio%u called:", dev->uio_idx));

    dev->pESC = uio_map_get(dev, "esc", 0);
    if (dev->pESC == MAP_FAILED) {
        PRINT(("    uio map '%s' not found!", "esc"));
        return -1;
    }

    PRINT_MORE(("    pESC:     %p", dev->pESC));

    switch (dev->card_type) {
        case CARD_TYPE_ECSPCIe1100:
        {
            dev->pLEDs = uio_map_get(dev, "8111", 0);
            /* pLEDs is optional for ECS-PCIe/1100 */

            /* map "9056" not needed... */

            PRINT_MORE(("    pLEDs:    %p", dev->pLEDs));

            return 0;
        }

        case CARD_TYPE_ECSXMCFPGA:
        {
            dev->pSysId = uio_map_get(dev, "util", 0x0000);
//            dev->pLEDs = uio_map_get(dev, "util", 0x0020);
            dev->pLEDs = uio_map_get(dev, "esc", 0x20010);
            dev->pPIO_XMC = uio_map_get(dev, "util", 0x0040);
            dev->pPIO_PMC = uio_map_get(dev, "util", 0x0060);
            dev->pI2C = uio_map_get(dev, "util", 0x0080);
            dev->pPIORevision = uio_map_get(dev, "util", 0x00a0);
            dev->pPIOLatchControl = uio_map_get(dev, "util", 0x00c0);
            dev->pFlash = uio_map_get(dev, "util", 0x0800);
            dev->pPCIEp = uio_map_get(dev, "util", 0x4000);
            if ((dev->pSysId == MAP_FAILED) || (dev->pLEDs == MAP_FAILED) || (dev->pPIO_XMC == MAP_FAILED) ||
                    (dev->pPIO_PMC == MAP_FAILED) || (dev->pI2C == MAP_FAILED) || (dev->pPIORevision == MAP_FAILED) ||
                    (dev->pPIOLatchControl == MAP_FAILED) || (dev->pFlash == MAP_FAILED) || (dev->pPCIEp == MAP_FAILED)) {
                PRINT(("uio map '%s' not found!", "util"));
                return -1;
            }
            dev->pDMAMemory = uio_map_get(dev, "dma", 0);
            if (dev->pDMAMemory == MAP_FAILED) {
                PRINT(("uio map '%s' not found! Upgrade FPGA and/or UIO-driver for DMA support", "dma"));
            } else {
                dev->have_dma=1;
            }
            dev->pDMAControl = uio_map_get(dev, "esc", 0x20000);

            PRINT_MORE(("    pSysId:            %p", dev->pSysId));
            PRINT_MORE(("    pLEDs:             %p", dev->pLEDs));
            PRINT_MORE(("    pPIO_XMC:          %p", dev->pPIO_XMC));
            PRINT_MORE(("    pPIO_PMC:          %p", dev->pPIO_PMC));
            PRINT_MORE(("    pI2C:              %p", dev->pI2C));
            PRINT_MORE(("    pPIORev:           %p", dev->pPIORevision));
            PRINT_MORE(("    pPIOLatchControl:  %p", dev->pPIOLatchControl));
            PRINT_MORE(("    pFlash:            %p", dev->pFlash));
            PRINT_MORE(("    pPCIEp:            %p", dev->pPCIEp));
            PRINT_MORE(("    pDMAMemory:        %p", dev->pDMAMemory));
            PRINT_MORE(("    pDMAControl:       %p", dev->pDMAControl));

            {
                const uint32_t revNo = *((const uint32_t*)dev->pSysId);
                const uint8_t major = (uint8_t)(revNo >> 24);

                if (major != 0x02) {
                    PRINT(("FPGA version 0x%.8x (%u.%u.%u) not supported!", revNo,
                        (unsigned int)major,
                        (unsigned int)((revNo >> 12) & 0xfff), /* minor    */
                        (unsigned int)(revNo & 0xfff)          /* revision */
                    ));
#ifndef _DEBUG
                    return -1;
#endif
                }
            }

            return 0;
        }

        default:
            PRINT(("Unknown card! (%u)", dev->card_type));
            return -1;
    }
}

static void EnableInterrupt(DEV_INFOS* dev) {
#if 0
    if (dev->card_type == CARD_TYPE_ECSXMCFPGA) {
        int handled = 0;
        uint32_t cmd=0;
        (void)ecs_xmc_fpga_ioctl(dev, ESS_IOCTL_DEV_CPUCOMMAND, &cmd, sizeof(cmd), &handled);
    }
#endif
    {
        int32_t data;
        /* (re)enable interrupt: */
        data = 1;
        if (write(dev->fd, &data, sizeof(data)) == -1) {
            perror("write()");
            PRINT(("Enabling interrupt might have failed!"));
        }
    }
}

/* Returns 0 on timeout, -1 on error else it was actual interrupt */
static int WaitForInterrupt(DEV_INFOS* dev, uint32_t max)
{
    fd_set rfd;
    struct timeval tv;
    int res;
    int32_t data;

    if(!dev->have_msi) {
        /* (re)enable interrupt: */
        EnableInterrupt(dev);
    }

    /* then wait for it: */
    FD_ZERO(&rfd);
    tv.tv_sec = 0;
    tv.tv_usec = max;
    FD_SET(dev->fd, &rfd);

    res = select(dev->fd + 1, &rfd, NULL, NULL, &tv);
    if ((res == 0) || (res == -1))
        return res;

    if (FD_ISSET(dev->fd, &rfd) == 0)
        return -1;

    if (read(dev->fd, &data, sizeof(data)) == -1)
        PRINT(("read() failed!"));

#if 0
    PRINT(("INTERRUPT! (No. %d)\n", data));
#endif

    return 1;
}

/* Just calls dev->cyclic() in dev->cyclicInterval */
static ESS_RESULT RunMainLoop_CyclicOnly(DEV_INFOS* dev)
{
    dev->started = 1; /* May be reset in cyclic callback */

    do {
        my_usleep(dev->cyclicInterval);
        dev->cyclic(dev->cbData);

    } while (dev->started);

    return ESS_RESULT_SUCCESS;
}

/* Calls dev->cyclic() in dev->cyclicInterval and dev->isr() when an interrupt occurs */
static ESS_RESULT RunMainLoop(DEV_INFOS* dev)
{
    /* Cyclic and ISR: */
    uint32_t lastCyclic = my_get_utickcount();
    uint32_t maxWait = dev->cyclicInterval; /* us */

    /* (re-)enable interrupt: */
    EnableInterrupt(dev);

    dev->started = 1; /* May be reset in ISR/cyclic callback */
    do {
        const int waitRes = WaitForInterrupt(dev, maxWait);
        const uint32_t now = my_get_utickcount();

        if (waitRes == 0) { /* time out */
            //printf("RunMainLoop: Timeout!\n");
            dev->cyclic(dev->cbData);
            lastCyclic = now;
            maxWait = dev->cyclicInterval;

        } else {
            if (waitRes == 1)
                dev->isr(dev->cbData);

            if (ESS_TIMESTAMP_PASSED(lastCyclic + dev->cyclicInterval, now)) {
                dev->cyclic(dev->cbData);
                lastCyclic = now;
                maxWait = dev->cyclicInterval;
            } else {
                maxWait = dev->cyclicInterval - (uint32_t)(now - lastCyclic);
            }
        }

    } while (dev->started);

    return ESS_RESULT_SUCCESS;
}







static int global_init_done;
static int devices_opened;
/* Just to handle global_init() and global_finish()  (No thread safety required: parallel access to essOpen() and essClose() is impossible by definition) */


/* Called for first essHALOpen() call. Might be called multiple times if opening is unsuccessful */
static int global_init(void)
{
    int i;

    if (sizeof(ECS_XMC_FPGA_FLASH_CONFIGDATA) != 256)
        PRINT(("BUG!: sizeof(ECS_XMC_FPGA_FLASH_CONFIGDATA) == %u!", (unsigned int)sizeof(ECS_XMC_FPGA_FLASH_CONFIGDATA)));

    if (global_init_done)
        return 0;

    PRINT_MORE(("global_init()..."));

    for (i = 0; i < CFG_ESS_MAX_DEVICES; ++i) {
        devInfos[i].uio_idx = -1;
        devInfos[i].fd = -1;
    }

    global_init_done = 1;
    return 0;
}

/* Called for last essHALClose() call */
static void global_finish(void)
{
    if (global_init_done == 0)
        return;

    /* Nothing to be done here... */
}


/* Called by essOpen(), no other HAL function for this devIdx is used before. */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALOpen(ESS_DEVICE_INDEX devIdx)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    PRINT_MORE(("\n\n*** essHALOpen() for device %u:", devIdx));

    if (devices_opened == 0)
        if (global_init())
            return ESS_RESULT_UNKNOWN;

    dev = &devInfos[devIdx];
    if (dev->fd != -1) {
        PRINT_MORE(("    already opened!"));
        return ESS_RESULT_ALREADY_ACTIVE;
    }

    dev->uio_idx = devIdx_to_uio_idx(devIdx);
    if (dev->uio_idx < 0) {
        PRINT_MORE(("    uio index not found!"));
        return ESS_RESULT_INVALID_INDEX;
    }
    PRINT_MORE(("    device %u is uio%d", devIdx, dev->uio_idx));

    read_card_type_and_driver_build_info(dev);
    uio_init_maps(dev);

    if (uio_open(dev)) {
        PRINT_MORE(("    uio_open() failed!"));
        return ESS_RESULT_CANT_OPEN;
    }

    uio_start(dev);

    if (InitDevPointers(dev)) {
        PRINT_MORE(("    InitDevPointers() failed!"));
        uio_stop(dev);
        uio_close(dev);
        return ESS_RESULT_UNKNOWN;
    }

    devices_opened++;

    PRINT_MORE(("    essHALOpen() for device %u successful.", devIdx));
    return ESS_RESULT_SUCCESS;
}

/* Called by essClose() for a device */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALClose(ESS_DEVICE_INDEX devIdx)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    PRINT_MORE(("*** essHALClose() for device %u:", devIdx));

    dev = &devInfos[devIdx];
    if (dev->fd == -1) {
        PRINT_MORE(("    wasn't opened!"));
        return ESS_RESULT_NOT_READY;
    }

    /* essClose() includes call to essHALStop()... */

    uio_stop(dev);
    uio_close(dev);

    devices_opened--;
    if (devices_opened == 0)
        global_finish();

    PRINT_MORE(("    essHALClose() for device %u successful.", devIdx));
    return ESS_RESULT_SUCCESS;
}

/* Called by essOpen() */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALMapESC(ESS_DEVICE_INDEX devIdx, void** escPointer)
{
    if ((devIdx >= CFG_ESS_MAX_DEVICES) || (devInfos[devIdx].uio_idx < 0))
        return ESS_RESULT_INVALID_INDEX;

    *escPointer = devInfos[devIdx].pESC;

    return ((*escPointer != NULL) && (*escPointer != (void*)-1)) ? ESS_RESULT_SUCCESS : ESS_RESULT_UNKNOWN;
}

/* Called by essClose() */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALUnmapESC(ESS_DEVICE_INDEX devIdx)
{
    if ((devIdx >= CFG_ESS_MAX_DEVICES) || (devInfos[devIdx].uio_idx < 0))
        return ESS_RESULT_INVALID_INDEX;

    return ESS_RESULT_SUCCESS;
}

/* Time in milliseconds (must wrap at 0xffffffff) */
HAL_EXPORT ESS_TIMESTAMP HAL_CALLTYPE essHALGetTime(void)
{
#if ((defined(_POSIX_C_SOURCE)) && (_POSIX_C_SOURCE >= 199309L))
    struct timespec tp;

    if (clock_gettime(CLOCK_MONOTONIC, &tp) == -1) /* Usually needs to link with -lrt */
        return 0;

    return tp.tv_sec * 1000 + tp.tv_nsec / 1000000;
#else
    struct timeval tv;

    gettimeofday(&tv, NULL);

    return tv.tv_sec * 1000 + tv.tv_usec / 1000;
#endif
}

/* Called by essStart(). HAL takes control here and stack only runs by cyclic/isr callback then */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALStart(ESS_DEVICE_INDEX devIdx, ESS_HAL_CALLBACK* isr, ESS_HAL_CALLBACK* cyclicCallback, ESS_TIMESTAMP cyclicInterval, void* cbData)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];
    if (dev->started)
        return ESS_RESULT_ALREADY_ACTIVE;

    if ((cyclicCallback == NULL) || (cyclicInterval >= 1000000))
        return ESS_RESULT_INVALID_TIMER_CONFIG;

    dev->isr = isr;
    dev->cbData = cbData;
    dev->cyclic = cyclicCallback;
    dev->cyclicInterval = cyclicInterval;

    return (dev->isr) ? RunMainLoop(dev) : RunMainLoop_CyclicOnly(dev);
}

/* Called by essStop() to abort the loop started by essHALStart() */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALStop(ESS_DEVICE_INDEX devIdx)
{
    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    if (devInfos[devIdx].started == 0)
        return ESS_RESULT_NOT_READY;

    devInfos[devIdx].started = 0;

    return ESS_RESULT_SUCCESS;
}

HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALSetLEDState(ESS_DEVICE_INDEX devIdx, ESC_LED_TYPE ledId, ESS_BOOL ledOn)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];

    if (dev->fd == -1)
        return ESS_RESULT_NOT_READY;

    switch (dev->card_type) {
        case CARD_TYPE_ECSPCIe1100:
        {
            if (ledId != ESC_LED_TYPE_ERR) /* no other LED is supported */
                return ESS_RESULT_INVALID_ARG;

            if (dev->pLEDs != MAP_FAILED) {
                volatile uint32_t* pxlCntrl = (volatile uint32_t*)((char*)dev->pLEDs + 0x6c);

                *pxlCntrl = (ledOn) ? (*pxlCntrl & ~0x00010000) : (*pxlCntrl | 0x00010000);
            }
            /* No error if LEDs not mapped (prototypes exist where it couldn't be mapped) */

            return ESS_RESULT_SUCCESS;
        }

        case CARD_TYPE_ECSXMCFPGA:
        {
            volatile uint32_t* const pLEDs = dev->pLEDs;

            if (ledId > ESC_LED_TYPE_CUSTOM2) /* no other LED is supported */
                return ESS_RESULT_INVALID_ARG;

            pLEDs[ledId] = 0x10000u | (ledOn?1:0);

            return ESS_RESULT_SUCCESS;
        }

        default:
            return ESS_RESULT_UNKNOWN;
    }
}

/* For ecs_xmc_fpga_ioctl(): */
#include "../ecs_xmc_fpga/common_ioctls.c"

HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALIoctl(ESS_DEVICE_INDEX devIdx, ESS_IOCTL fn, void* data, uint32_t sizeOfData)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];
    if (dev->fd == -1)
        return ESS_RESULT_NOT_READY;

    if (dev->card_type == CARD_TYPE_ECSXMCFPGA) {
        int handled = 0;
        ESS_RESULT res;

        res = ecs_xmc_fpga_ioctl(dev, fn, data, sizeOfData, &handled);
        if (handled)
            return res;
    }

    switch (fn) {
        /*
        case ESS_IOCTL_GET_HAL_BUILDINFOSTR:
            only useful if HAL is not included in stack (CFG_HAL_EXTERNAL)
        */

        case ESS_IOCTL_GET_DEV_DRIVERINFOSTR:
            if (dev->card_type == CARD_TYPE_ECSXMCFPGA) {
                return ecs_xmc_fpga_ioctl_getdrvinfostr(dev, dev->driver_build_info, data, sizeOfData);

            } else {
                const size_t len = strlen(dev->driver_build_info) + 1;

                if ((data == NULL) || (sizeOfData < len))
                    return ESS_RESULT_INSUFFICIENT_BUFFER;

                memcpy(data, dev->driver_build_info, len);

                return ESS_RESULT_SUCCESS;
            }


        default:
            return ESS_RESULT_NOT_IMPLEMENTED;
    }
}

HAL_EXPORT void HAL_CALLTYPE essHALReadESCMem(ESS_DEVICE_INDEX devIdx, uint16_t address, void* dst, uint16_t len)
{
    DEV_INFOS *dev;
    volatile uint32_t *pCnt;
    uint32_t cnt,cmd;
    
    dev = &devInfos[devIdx];
    if (dev->have_dma) {
        pCnt=(void*)(((uint8_t *)dev->pDMAMemory)+0x10000);
        cnt=*pCnt;
        cmd=(((uint32_t)address)<<16)|len;
        ((volatile uint32_t *)dev->pDMAControl)[0]=cmd;
        PRINT_MORE(("    essHALReadESCMem() sending cmd %08x.", (unsigned int)cmd));
        do {} while(*pCnt==cnt);
        memcpy(dst,(void*)(((uint8_t *)dev->pDMAMemory)+address),len);
    } else {
        memcpy(dst,(void*)(((uint8_t *)dev->pESC)+address),len); /* TBD */
    }
}

#if defined USE_MEMCPY_2_2_5 && USE_MEMCPY_2_2_5 != 0
/*
 * Solution to prevent that the minimum libc version is forced to 2.14
 * using memcpy(). See https://gist.github.com/nicky-zs/7541169
 *
 * This solution requires the linker option: -Wl,--wrap=memcpy
 */
/* NOTE: memcpy@GLIBC_2.2.5 is only present in x86_64 glibc! */
void *__memcpy_glibc_2_2_5(void *, const void *, size_t);
 
asm(".symver __memcpy_glibc_2_2_5, memcpy@GLIBC_2.2.5");
void *__wrap_memcpy(void *dest, const void *src, size_t n)
{
     return __memcpy_glibc_2_2_5(dest, src, n);
}
#endif  /* #if defined USE_MEMCPY_2_2_5 && USE_MEMCPY_2_2_5 != 0 */
