/*
*       Copyright (c) 2011-2017 by esd electronics gmbh
*
*  This software is copyrighted by and is the sole property of
*  esd gmbh.  All rights, title, ownership, or other interests
*  in the software remain the property of esd gmbh. This
*  software may only be used in accordance with the corresponding
*  license agreement.  Any unauthorized use, duplication, transmission,
*  distribution, or disclosure of this software is expressly forbidden.
*
*  This Copyright notice may not be removed or modified without prior
*  written consent of esd gmbh.
*
*  esd gmbh, reserves the right to modify this software without notice.
*
*  esd electronics gmbh                   Tel. +49-511-37298-0
*  Vahrenwalder Strasse 207               Fax. +49-511-37298-68
*  30165 Hannover                         http://www.esd.eu
*  Germany                                sales@esd.eu
*
*************************************************************************
*
* Description:
* =============================================================================
* EtherCAT Slave Stack (ESS) HAL for Windows.
*
*/

#include <windows.h>
#include <setupapi.h>
#include <public.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "essHAL.h"
#include "ess.h"


    /* Exceptions from the esd PC-lint/FlexeLint policy for this file */
/*lint -e717  [Info: do ... while(0)] */

#ifdef _DEBUG
# define esdDbgPrint(x)                                          \
    do { DWORD _t = GetTickCount();                              \
         printf("%3u.%02u [essHAL: %s()] ",                      \
         (unsigned)((_t % 1000000) / 1000),                      \
         (unsigned)((_t % 1000) / 10) , __FUNCTION__);           \
         printf x;                                               \
         printf("\r\n");                                         \
        } while (0)
#else
# define esdDbgPrint(x)             do {} while (0)
#endif

#define PRINT(x)                    esdDbgPrint(x)

#if !defined(MORE_DEBUG)
# define MORE_DEBUG                 0
#endif

#if (MORE_DEBUG)
# define PRINT_MORE(x)              PRINT(x)
#else
# define PRINT_MORE(x)              do { } while (0)
#endif

#ifdef _MSC_VER
# pragma warning(disable:4127) /* avoid "...conditional expression is constant" warnings */
# pragma warning(disable:4996) /* avoid "...unsafe" warnings */
#endif

/* For ecs_xmc_fpga_flash_read() and ecs_xmc_fpga_flash_write(): */
#include "../ecs_xmc_fpga/spiflash.c"

/* For ecs_xmc_fpga_get_temperature(): */
#include "../ecs_xmc_fpga/i2c.c"

typedef struct {
    HANDLE hFile;
    volatile int started;

    enum CARD_TYPE card_type;
    char driver_build_info[128];

    ESS_HAL_CALLBACK* isr;
    ESS_HAL_CALLBACK* cyclic;
    uint32_t cyclicInterval;

    /* These 3 directly correspond to a BAR: */
    void* pESC;
    void* pOther1;
    void* pOther2;
    /* And these pointers are just the ptrs above with some offsets: (see InitDevPointers()) */
    void* pLEDs;
    void* pSysId;
    void* pPIO_XMC;
    void* pPIO_PMC;
    void* pI2C;
    void* pPIORevision;
    void* pPIOLatchControl;
    void* pFlash;
    void* pPCIEp;
    void* pDMA;
    void* pDMACtrl;

    ECS_XMC_FPGA_FLASH_CONFIGDATA configData;
    /* 0 = not yet, 1 = yes, -1 = read failed, -2 = CRC error */
    int configDataRead;

    /* 0: WaitableTimer, 1: Event (Interrupt) */
    HANDLE events[2];

    void* cbData;

    int have_dma;   /* 1 = DMA support */
} DEV_INFOS;

typedef struct
{
    CHAR szCompDescription[MAX_PATH];
    CHAR szFriendlyName[MAX_PATH];
    CHAR szDevPath[MAX_PATH];
} DriverInterfaces;

static DEV_INFOS devInfos[CFG_ESS_MAX_DEVICES];

#if (CFG_ESS_MAX_DEVICES == 0)
# error ...
#endif




static PSP_DEVICE_INTERFACE_DETAIL_DATA AllocAndGetDevDetails(HDEVINFO hdi, PSP_DEVICE_INTERFACE_DATA did)
{
    PSP_DEVICE_INTERFACE_DETAIL_DATA deviceInterfaceDetailData;
    ULONG requiredLength = 0;
    ULONG len = 0;

    if (!SetupDiGetDeviceInterfaceDetail(hdi, did, NULL, 0, &requiredLength, NULL)) {
        if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
            esdDbgPrint(("SetupDiGetDeviceInterfaceDetail() failed with 0x%lx", GetLastError()));
            return NULL;
        }
    }

    deviceInterfaceDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)malloc(requiredLength);
    if (deviceInterfaceDetailData == NULL) {
        esdDbgPrint(("malloc() for %lu bytes failed!", requiredLength));
        return NULL;
    }

    deviceInterfaceDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

    if (!SetupDiGetDeviceInterfaceDetail(hdi, did, deviceInterfaceDetailData, requiredLength, &len, NULL)) {
        esdDbgPrint(("SetupDiGetDeviceInterfaceDetail() failed with 0x%lx", GetLastError()));
        free(deviceInterfaceDetailData);
        return NULL;
    }

    return deviceInterfaceDetailData;
}

/* returns number of devices written to devs, or < 0 for errors... */
static int EnumDevices(const GUID* interfaceGUID, DriverInterfaces* devs, int maxDevs)
{
    SP_DEVICE_INTERFACE_DATA deviceInterfaceData;
    SP_DEVINFO_DATA deid;
    DWORD dwRegType;
    HDEVINFO hdi;
    DWORD devIdx = 0;
    int found = 0;

    hdi = SetupDiGetClassDevs((LPGUID)interfaceGUID, NULL, NULL, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);
    if (hdi == INVALID_HANDLE_VALUE) {
        esdDbgPrint(("SetupDiGetClassDevs() failed with 0x%lx", GetLastError()));
        return -1;
    }

    deviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

    for (;;) {

        if (SetupDiEnumDeviceInterfaces(hdi, 0, (LPGUID)interfaceGUID, devIdx, &deviceInterfaceData)) {

            PSP_DEVICE_INTERFACE_DETAIL_DATA details = AllocAndGetDevDetails(hdi, &deviceInterfaceData);
            if (details != NULL) {

                memset(&devs[found], 0, sizeof(DriverInterfaces));
                deid.cbSize = sizeof(SP_DEVINFO_DATA);
                if (SetupDiEnumDeviceInfo(hdi, devIdx, &deid)) {
                    (void)SetupDiGetDeviceRegistryProperty(hdi, &deid, SPDRP_DEVICEDESC, &dwRegType,
                        (BYTE*)devs[found].szCompDescription, MAX_PATH, NULL);
                    (void)SetupDiGetDeviceRegistryProperty(hdi, &deid, SPDRP_FRIENDLYNAME, &dwRegType,
                        (BYTE*)devs[found].szFriendlyName, MAX_PATH, NULL);
                } else {
                    esdDbgPrint(("Failed to get infos for dev %lu (friendly name/description will be missing)", devIdx));
                }

                strncpy(devs[found].szDevPath, details->DevicePath, MAX_PATH - 1);
                devs[found].szDevPath[MAX_PATH - 1] = '\0';

                free(details);
                found++;
                if (found >= maxDevs)
                    break;
            } else {
                esdDbgPrint(("Failed to get details for dev %lu", devIdx));
            }

        } else {
            if (GetLastError() == ERROR_NO_MORE_ITEMS)
                break;
            else
                esdDbgPrint(("SetupDiEnumDeviceInterfaces() failed for dev %lu", devIdx));
        }

        devIdx++;
    }

    (void)SetupDiDestroyDeviceInfoList(hdi);
    return found;
}

/* ... */
static void* ioctl_get_ptr(HANDLE h, int ioctl)
{
    PVOID ptr;
    DWORD len;

    if (!DeviceIoControl(h, (DWORD)ioctl, NULL, 0, &ptr, sizeof(ptr), &len, NULL) || (len != sizeof(ptr))) {
        /* esdDbgPrint(("ioctl 0x%.8x failed (0x%x / %u)", ioctl, GetLastError(), len)); */
        return NULL;
    }

    return ptr;
}

/* ... */
static int InitDevPointers(DEV_INFOS* dev)
{
    if (dev->pESC == NULL) {
        esdDbgPrint(("pESC == NULL"));
        return 1;
    }

    switch (dev->card_type) {
        case CARD_TYPE_ECSPCIe1100:
            dev->pLEDs = dev->pOther1;
            /* pLEDs is optional for ECS-PCIe/1100 */
            /* pOther2 is unused for ECS-PCIe/1100 */
            return 0;

        case CARD_TYPE_ECSXXXFPGA:
            if (dev->pOther1 == NULL) {
                esdDbgPrint(("pOther1 == NULL"));
                return 1;
            }
            if (dev->pOther2 == NULL) {
                esdDbgPrint(("pOther2 == NULL"));
                return 1;
            }
            dev->pSysId           = (uint8_t*)dev->pOther1;
            dev->pDMA             = (uint8_t*)dev->pOther2;
            dev->pDMACtrl         = (uint8_t*)dev->pESC   + 0x20000;
            dev->pLEDs            = (uint8_t*)dev->pESC   + 0x20010;
            dev->pPIO_XMC         = (uint8_t*)dev->pOther1 + 0x0040;
            dev->pPIO_PMC         = (uint8_t*)dev->pOther1 + 0x0060;
            dev->pI2C             = (uint8_t*)dev->pOther1 + 0x0080;
            dev->pPIORevision     = (uint8_t*)dev->pOther1 + 0x00a0;
            dev->pPIOLatchControl = (uint8_t*)dev->pOther1 + 0x00c0;
            dev->pFlash           = (uint8_t*)dev->pOther1 + 0x0800;
            dev->pPCIEp           = (uint8_t*)dev->pOther1 + 0x4000;

            /* DMA is only supported since FPGA V2.x */
            {
                const uint32_t revNo = *((const uint32_t*)dev->pSysId);
                const uint8_t major  = (uint8_t)(revNo >> 24);
                
                if (major != 0x02) {
                    PRINT(("FPGA version 0x%.8x (%u.%u.%u) not supported!", revNo,
                            (unsigned int)major,
                            (unsigned int)((revNo >> 12) & 0xfff), /* minor    */
                            (unsigned int)(revNo & 0xfff)          /* revision */
                          ));
#ifndef _DEBUG
                        return -1;
#endif
                }
                dev->have_dma = 1;
            }

            return 0;

        default:
            esdDbgPrint(("Unknown card type! (%u)", (unsigned int)dev->card_type));
            return -1;
    }
}

/* Opens dev->hFile and calls initCard */
static int openDevice(int devIdx, DEV_INFOS* dev)
{
    DriverInterfaces devs[CFG_ESS_MAX_DEVICES];
    int numDevs;
    DWORD temp;

    /* esdDbgPrint(("Trying to open ess device %u...", devIdx)); */

    numDevs = EnumDevices(&GUID_DEVINTERFACE_ESD_ETHERCAT1, devs, CFG_ESS_MAX_DEVICES);
    if (numDevs < 0) {
        esdDbgPrint(("Failed to enumerate devices!"));
        return -1;
    }

    if ((devIdx < 0) || (devIdx >= numDevs) || (devIdx >= CFG_ESS_MAX_DEVICES)) {
        esdDbgPrint(("Invalid index / not found"));
        return -1;
    }

    dev->hFile = CreateFile(devs[devIdx].szDevPath, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
    if (dev->hFile == INVALID_HANDLE_VALUE) {
        esdDbgPrint(("CreateFile() failed with 0x%lx", GetLastError()));
        return -1;
    }

    if (!DeviceIoControl(dev->hFile, (DWORD)IOCTL_GET_CARD_TYPE, NULL, 0, (LPVOID)&dev->card_type, sizeof(dev->card_type), &temp, NULL) || (temp != sizeof(dev->card_type))) {
        esdDbgPrint(("IOCTL_GET_CARD_TYPE failed (0x%lx / %lu)", GetLastError(), temp));
        (void)CloseHandle(dev->hFile);
        dev->hFile = INVALID_HANDLE_VALUE;
        return -1;
    }

    if (DeviceIoControl(dev->hFile, (DWORD)IOCTL_GET_BUILDINFO, NULL, 0, dev->driver_build_info,
                        (DWORD)sizeof(dev->driver_build_info), &temp, NULL) == 0) {
        esdDbgPrint(("IOCTL_GET_BUILDINFO failed (0x%lx / %lu)", GetLastError(), temp));
        (void)CloseHandle(dev->hFile);
        dev->hFile = INVALID_HANDLE_VALUE;
        return -1;
    }

    dev->pESC = ioctl_get_ptr(dev->hFile, IOCTL_MAP_ESC);
    dev->pOther1 = ioctl_get_ptr(dev->hFile, IOCTL_MAP_OTHER1);
    dev->pOther2 = ioctl_get_ptr(dev->hFile, IOCTL_MAP_OTHER2);

    if (InitDevPointers(dev)) {
        (void)CloseHandle(dev->hFile);
        dev->hFile = INVALID_HANDLE_VALUE;
        return -1;
    }

    return 0;
}


static DWORD WINAPI IOControlWaitThread(void* args)
{
    DEV_INFOS* dev = (DEV_INFOS*)args;
    DWORD ioctlLen;
    DWORD ioctlTimeout = 500; /* in ms */

    for(;;) {
        if (DeviceIoControl(dev->hFile, (DWORD)IOCTL_WAIT_FOR_INT, &ioctlTimeout, sizeof(ioctlTimeout), NULL, 0, &ioctlLen, NULL)) {
            if (dev->started == 0)
                break;

            (void)SetEvent(dev->events[1]);
        } else {
            if (dev->started == 0)
                break;
        }
    }

    return 0;
}

static ESS_RESULT RunMainLoop(DEV_INFOS* dev)
{
    LARGE_INTEGER timerValue;
    HANDLE thrdHandle = NULL;

    timerValue.QuadPart = -(dev->cyclicInterval * 10LL); /* in 100 ns intervals */

    dev->events[0] = CreateWaitableTimer(NULL, TRUE, NULL);
    if (dev->events[0] == NULL)
    {
        esdDbgPrint(("CreateWaitableTimer() failed! GetLastError() = 0x%.8lx", GetLastError()));
        return ESS_RESULT_UNKNOWN;
    }

    dev->events[1] = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (dev->events[1] == NULL)
    {
        (void)CloseHandle(dev->events[0]);
        esdDbgPrint(("CreateEvent() failed! GetLastError() = 0x%.8lx", GetLastError()));
        return ESS_RESULT_UNKNOWN;
    }

    if (!SetWaitableTimer(dev->events[0], &timerValue, 0, NULL, NULL, 0))
    {
        (void)CloseHandle(dev->events[0]);
        (void)CloseHandle(dev->events[1]);
        esdDbgPrint(("SetWaitableTimer() failed! GetLastError() = 0x%.8lx", GetLastError()));
        return ESS_RESULT_UNKNOWN;
    }

    dev->started = 1;
    if (dev->isr) {
        thrdHandle = CreateThread(NULL, 0, IOControlWaitThread, dev, 0, NULL);
        if (thrdHandle == NULL) {
            dev->started = 0;
            (void)CloseHandle(dev->events[0]);
            (void)CloseHandle(dev->events[1]);
            esdDbgPrint(("CreateThread() failed! GetLastError() = 0x%.8lx", GetLastError()));
            return ESS_RESULT_UNKNOWN;
        }
    }

    do {
        const DWORD dwEvent = WaitForMultipleObjects(2, dev->events, FALSE, 500);
        switch (dwEvent)
        {
            case WAIT_OBJECT_0 + 0:
                if(!SetWaitableTimer(dev->events[0], &timerValue, 0, NULL, NULL, 0)) {
                    esdDbgPrint(("FATAL: SetWaitableTimer() failed! GetLastError() = 0x%.8lx",
                                GetLastError()));
                }
                dev->cyclic(dev->cbData);
                break;

            case WAIT_OBJECT_0 + 1:
                dev->isr(dev->cbData);
                break;

            case WAIT_TIMEOUT:
                break;

            default:
                esdDbgPrint(("WaitForMultipleObjects() failed!? GetLastError() = 0x%.8lx", GetLastError()));
                break;
        }
    } while (dev->started);

    if (thrdHandle) {
        if (WaitForSingleObject(thrdHandle, 3000) == WAIT_TIMEOUT)
            esdDbgPrint(("IOControl waiter thread did not finish!"));
        (void)CloseHandle(thrdHandle);
    }

    (void)CloseHandle(dev->events[0]);
    (void)CloseHandle(dev->events[1]);

    return ESS_RESULT_SUCCESS;
}



static int global_init_done;
static int devices_opened;
/* Just to handle global_init() and global_finish()  (No thread safety required: parallel access to essOpen() and essClose() is impossible by definition) */

/* Called for first essHALOpen() call. Might be called multiple times if opening is unsuccessful */
static int global_init(void)
{
    int i;

    if (global_init_done)
        return 0;

    for (i = 0; i < CFG_ESS_MAX_DEVICES; ++i) {
        devInfos[i].hFile = INVALID_HANDLE_VALUE;
    }

    /* Set system clock period to 1 ms */
    (void)timeBeginPeriod(1);

    global_init_done = 1;
    return 0;
}

/* Called for last essHALClose() call */
static void global_finish(void)
{
    if (global_init_done) {
        /* Restore original clock period */
        (void)timeEndPeriod(1);

        global_init_done = 0;
    }
}



/* Called by essOpen(), no other HAL function for this devIdx is used before. */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALOpen(ESS_DEVICE_INDEX devIdx)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];

    if (devices_opened == 0)
        if (global_init())
            return ESS_RESULT_UNKNOWN;

    if (dev->hFile != INVALID_HANDLE_VALUE)
        return ESS_RESULT_ALREADY_ACTIVE;

    if (openDevice(devIdx, dev))
        return ESS_RESULT_CANT_OPEN;

    devices_opened++;
    return ESS_RESULT_SUCCESS;
}

/* Called by essClose() for a device */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALClose(ESS_DEVICE_INDEX devIdx)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];
    if (dev->hFile == INVALID_HANDLE_VALUE)
        return ESS_RESULT_NOT_READY;

    /* essClose() includes call to essHALStop()... */
    while (dev->started)
        Sleep(100);

    (void)CloseHandle(dev->hFile);
    dev->hFile = INVALID_HANDLE_VALUE;

    devices_opened--;
    if (devices_opened == 0)
        global_finish();

    return ESS_RESULT_SUCCESS;
}

HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALMapESC(ESS_DEVICE_INDEX devIdx, void** escPointer)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];
    if (dev->hFile == INVALID_HANDLE_VALUE)
        return ESS_RESULT_NOT_READY;

    *escPointer = dev->pESC;
    return (*escPointer) ? ESS_RESULT_SUCCESS : ESS_RESULT_UNKNOWN;
}

HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALUnmapESC(ESS_DEVICE_INDEX devIdx)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];
    if (dev->hFile == INVALID_HANDLE_VALUE)
        return ESS_RESULT_NOT_READY;

    /* Nothing to do... */

    return ESS_RESULT_SUCCESS;
}

HAL_EXPORT ESS_TIMESTAMP HAL_CALLTYPE essHALGetTime(void)
{
    return timeGetTime();
}

HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALStart(ESS_DEVICE_INDEX devIdx, ESS_HAL_CALLBACK* isr, ESS_HAL_CALLBACK* cyclicCallback, uint32_t cyclicInterval, void* cbData)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];
    if (dev->hFile == INVALID_HANDLE_VALUE)
        return ESS_RESULT_NOT_READY;

    if (dev->started)
        return ESS_RESULT_ALREADY_ACTIVE;

    if (cyclicCallback == NULL)
        return ESS_RESULT_INVALID_TIMER_CONFIG;

    dev->isr = isr;
    dev->cbData = cbData;
    dev->cyclic = cyclicCallback;
    dev->cyclicInterval = cyclicInterval;

    return RunMainLoop(dev);
}

HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALStop(ESS_DEVICE_INDEX devIdx)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];
    if (dev->hFile == INVALID_HANDLE_VALUE)
        return ESS_RESULT_NOT_READY;

    if (dev->started == 0)
        return ESS_RESULT_NOT_READY;

    devInfos[devIdx].started = 0;

    return ESS_RESULT_SUCCESS;
}


HAL_EXPORT void HAL_CALLTYPE essHALReadESCMem(ESS_DEVICE_INDEX devIdx, uint16_t address, void* dst, uint16_t len)
{
    DEV_INFOS *dev;
    volatile uint32_t *pCnt;
    uint32_t cnt,cmd;

    dev = &devInfos[devIdx];
    if (dev->have_dma) {
        pCnt=(void*)(((uint8_t *)dev->pDMA)+0x10000);
        cnt=*pCnt;
        cmd=(((uint32_t)address)<<16)|len;
        ((volatile uint32_t *)dev->pDMACtrl)[0]=cmd;
        PRINT_MORE(("    essHALReadESCMem() sending cmd %08x.", (unsigned int)cmd));
        do {} while(*pCnt==cnt);
        memcpy(dst,(void*)(((uint8_t *)dev->pDMA)+address),len);
    } else {
        memcpy(dst,(void*)(((uint8_t *)dev->pESC)+address),len); /* TBD */
    }
}

/* HAL offers pointer (CFG_ESC_HAVE_POINTER), so stack does not use this. */
HAL_EXPORT void ESS_CALLTYPE essHALWriteESCMem(ESS_DEVICE_INDEX devIdx, uint16_t address, const void* src, uint16_t len)
{
    (void)devIdx;
    (void)address;
    (void)src;
    (void)len;
}

HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALSetLEDState(ESS_DEVICE_INDEX devIdx, ESC_LED_TYPE ledId, ESS_BOOL ledOn)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];

    switch (dev->card_type) {
        case CARD_TYPE_ECSPCIe1100:
        {
            if (ledId != ESC_LED_TYPE_ERR) /* no other LED is supported */
                return ESS_RESULT_INVALID_ARG;

            if (dev->pLEDs != NULL) {
                volatile uint32_t* pxlCntrl = (volatile uint32_t*)((char*)dev->pLEDs + 0x6c);

                *pxlCntrl = (ledOn) ? (*pxlCntrl & ~0x00010000) : (*pxlCntrl | 0x00010000);
            }
            /* No error if LEDs not mapped (prototypes exist where it couldn't be mapped) */

            return ESS_RESULT_SUCCESS;
        }

        case CARD_TYPE_ECSXXXFPGA:
        {
            volatile uint32_t* const pLEDs = dev->pLEDs;
            
            if (ledId > ESC_LED_TYPE_CUSTOM2)     /* No other LED is supported */
                return ESS_RESULT_INVALID_ARG;

            pLEDs[ledId] = 0x10000u | (ledOn ? 1 : 0);

            return ESS_RESULT_SUCCESS;
        }

        default:
            return ESS_RESULT_UNKNOWN;
    }
}

HAL_EXPORT void HAL_CALLTYPE essHALStackEvent(ESS_EVENT_DATA* e)
{
    (void)e;
}




/* For ecs_xmc_fpga_ioctl(): */
#include "../ecs_xmc_fpga/common_ioctls.c"

HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALIoctl(ESS_DEVICE_INDEX devIdx, ESS_IOCTL fn, void* data, uint32_t sizeOfData)
{
    DEV_INFOS* dev;
    DWORD temp;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];
    if (dev->hFile == INVALID_HANDLE_VALUE)
        return ESS_RESULT_NOT_READY;

    if (dev->card_type == CARD_TYPE_ECSXXXFPGA) {
        int handled = 0;
        ESS_RESULT res;

        res = ecs_xmc_fpga_ioctl(dev, fn, data, sizeOfData, &handled);
        if (handled) {
            return res;
        }
    }

    switch (fn) {
        case ESS_IOCTL_GET_DEV_DRIVERINFOSTR:
            if (dev->card_type == CARD_TYPE_ECSXXXFPGA) {
                return ecs_xmc_fpga_ioctl_getdrvinfostr(dev, dev->driver_build_info, data, sizeOfData);
            } else {
                const size_t len = strlen(dev->driver_build_info) + 1;

                if ((data == NULL) || (sizeOfData < len))
                    return ESS_RESULT_INSUFFICIENT_BUFFER;

                memcpy(data, dev->driver_build_info, len);

            }

            return ESS_RESULT_SUCCESS;

        case ESS_IOCTL_GET_DEV_ORDERNOSTR:
            if(CARD_TYPE_ECSPCIe1100 == dev->card_type) {
                if ((data == NULL) || (sizeOfData < 10)) {
                    return ESS_RESULT_INSUFFICIENT_BUFFER;
                }

                strcpy(data, "E.1100.02");

                return ESS_RESULT_SUCCESS;
            }
            
            return ESS_RESULT_NOT_IMPLEMENTED;

        case ESS_IOCTL_DEV_RESET:
            return (DeviceIoControl(dev->hFile, (DWORD)IOCTL_RESET_BOARD, NULL, 0, NULL, 0, &temp, NULL)) ? ESS_RESULT_SUCCESS : ESS_RESULT_UNKNOWN;


        default:
            return ESS_RESULT_NOT_IMPLEMENTED;
    }
}


