/*
*       Copyright (c) 2011-2020 by esd electronics gmbh
*
*  This software is copyrighted by and is the sole property of
*  esd gmbh.  All rights, title, ownership, or other interests
*  in the software remain the property of esd gmbh. This
*  software may only be used in accordance with the corresponding
*  license agreement.  Any unauthorized use, duplication, transmission,
*  distribution, or disclosure of this software is expressly forbidden.
*
*  This Copyright notice may not be removed or modified without prior
*  written consent of esd gmbh.
*
*  esd gmbh, reserves the right to modify this software without notice.
*
*  esd electronics gmbh                   Tel. +49-511-37298-0
*  Vahrenwalder Strasse 207               Fax. +49-511-37298-68
*  30165 Hannover                         http://www.esd.eu
*  Germany                                sales@esd.eu
*
*************************************************************************
*
* Description:
* =============================================================================
* This file contains declarations/macros that are needed only internally,
* i.e. by ess.c and essOD.c. NOT visible to HAL and application.
*
*/


#ifndef ESS_PRIVATE_INCLUDED
#define ESS_PRIVATE_INCLUDED

#include "essConfig.h"
#include "essTypes.h"


#if !defined(ESS_PLATFORM_WINDOWS) && !defined(ESS_PLATFORM_LINUX) && !defined(ESS_PLATFORM_AM335X) && \
    !defined(ESS_PLATFORM_DUMMY) && !defined(ESS_PLATFORM_QNX) && !defined(ESS_PLATFORM_RIN32M3) &&    \
	!defined(ESS_PLATFORM_XMC4800) && !defined(ESS_PLATFORM_VXWORKS) && !defined(ESS_PLATFORM_STM32CUBE)
# error "Unknown or undefined ESS_PLATFORM! Define it, or make sure essPrivate.h works for your platform (essSystem.h must be checked then, too)"
#endif

#if defined(ESS_PLATFORM_WINDOWS) && defined(_MSC_VER)
# pragma warning(disable:4127)                          /* avoid "conditional expression is constant" warnings... */
#endif

/*****************************************************************************/
/****** MACROS THAT HAVE TO BE CHECKED WHEN IMPLEMENTING A NEW PLATFORM ******/
/*****************************************************************************/

/* Just define your macros here to override the defaults */

#ifndef INLINE
# define INLINE                                         __inline
#endif

#ifdef _DEBUG
# ifndef ESS_DBG
#  define ESS_DBG(x)                                    do { ESS_PRINTF("[Stack: %s()] ", __FUNCTION__); ESS_PRINTF x; ESS_PRINTF("\n"); ESS_PRINTFFLUSH(); } while (0)
# endif

# ifndef ESS_ASSERT
/*lint -emacro( 506, ESS_ASSERT) [Constant value boolean]                    */
/*lint -emacro( 774, ESS_ASSERT) [Boolean within if always evaluates to ...] */
/*lint -emacro( 778, ESS_ASSERT) [Constant expression evaluates to 0]        */
#  define ESS_ASSERT(e, x)                              do { if (!(e)) { ESS_PRINTF("ERROR: Assertion failed! (%s:%u)\n", __FILE__, (unsigned)__LINE__); ESS_PRINTF x; ESS_PRINTF("\n"); ESS_PRINTFFLUSH(); } } while (0)
# endif

# ifndef ESS_MEM_DUMP
#  define ESS_MEM_DUMP(title, pData, len)               do { uint32_t _currOffset = 0; ESS_DBG(("%s", (title))); ESS_PRINTF("0x%.4x: ", _currOffset); if (pData) { while (_currOffset < (uint32_t)(len)) { ESS_PRINTF("%.2x ", ((uint8_t*)(pData))[_currOffset]); _currOffset++; if ((_currOffset % 16 == 0)) { ESS_PRINTF("\n"); ESS_PRINTF("0x%.4x: ", _currOffset); } else if ((_currOffset % 8) == 0) ESS_PRINTF(" "); } } ESS_PRINTF("\n"); ESS_PRINTFFLUSH(); } while (0)
# endif
#endif /* _DEBUG */

#ifndef ESS_ALLOC
# define ESS_ALLOC(result, cast, byteLen)               do { (result) = (cast)malloc((size_t)(byteLen)); } while (0)
#endif

#ifndef ESS_REALLOC
# define ESS_REALLOC(result, cast, ptr, newLen)         do { (result) = (cast)realloc((ptr), (size_t)(newLen)); /* if ((ptr) != NULL) { ESS_DBG(("REALLOC() for %lu bytes at %p result: %p", (unsigned long)(newLen), (void*)(ptr), (void*)(result))); } */ } while (0)
#endif

#ifndef ESS_FREE
# define ESS_FREE(ptr)                                  do { free(ptr); /* ESS_DBG(("ESS_FREE called for %p", (void*)(ptr))); */ } while (0)
#endif

#ifndef ESS_MEMCOPY
# define ESS_MEMCOPY(pDst, pSrc, len)                   memcpy((pDst), (pSrc), (size_t)(len)) /* Never used to copy from/to ESC memory */
#endif

#ifndef ESS_MEMCMP
# define ESS_MEMCMP(pMem1, pMem2, len)                  memcmp((pMem1), (pMem2), (size_t)(len)) /* Never used with ESC memory, must return 0 when memory is identical */
#endif

#ifndef ESS_MEMZERO
# define ESS_MEMZERO(pMem, len)                         memset((pMem), 0, (size_t)(len))
#endif

#ifndef ESS_DBG
# define ESS_DBG(x)                                     do { } while (0)
#endif

#ifndef ESS_ASSERT
# define ESS_ASSERT(e, x)                               do { } while (0)
#endif

#ifndef ESS_MEM_DUMP
# define ESS_MEM_DUMP(title, pData, len)                do { } while (0)
#endif

/*****************************************************************************/
/****************** END OF MACROS HAT HAVE TO BE CHECKED *********************/
/*****************************************************************************/





#ifdef _DEBUG
# ifndef _ESS_GETTHREADID
#  define _ESS_GETTHREADID()                            ((void*)0)
# endif
# define _ESSHANDLE_IS_INVALID(h)                       (((h) == NULL) || ((h) == (ESS_HANDLE)(uintptr_t)-1) || (((ESC_DATA*)(h))->magic != (uint32_t)0x07bb0b04UL))
# define _ESSHANDLE_THRD_IS_INVALID(h)                  (((ESC_DATA*)(h))->tIdOnOpen != _ESS_GETTHREADID())
# define _ESSHANDLE_INVALID_WARNING(h)                  do { ESS_PRINTF("%s() called with invalid handle! (%p)\n", __FUNCTION__, (h)); ESS_PRINTFFLUSH(); } while (0)
# define _ESSHANDLE_INVALID_THRD_WARNING()              do { ESS_PRINTF("%s() called from different thread! (%p)\n", __FUNCTION__, _ESS_GETTHREADID()); ESS_PRINTFFLUSH(); } while (0)
# define _ESSHANDLE_IF_INVALID_RETURN_(h, _ret)         do { if (_ESSHANDLE_IS_INVALID(h)) { _ESSHANDLE_INVALID_WARNING(h); return (_ret); } if (_ESSHANDLE_THRD_IS_INVALID(h)) { _ESSHANDLE_INVALID_THRD_WARNING(); return (_ret); } } while (0)
# define _ESSHANDLE_IF_INVALID_RETURN(h)                do { if (_ESSHANDLE_IS_INVALID(h)) { _ESSHANDLE_INVALID_WARNING(h); return; } if (_ESSHANDLE_THRD_IS_INVALID(h)) { _ESSHANDLE_INVALID_THRD_WARNING(); return; } } while (0)
#else
# define _ESSHANDLE_IF_INVALID_RETURN_(h, _ret)         do { } while (0)
# define _ESSHANDLE_IF_INVALID_RETURN(h)                do { } while (0)
#endif


#if (CFG_ESS_COPY_CALLBACKS)
# define ED_CALLBACK(ped, cbName)                       (ped)->callbacks.cbName
# define ED_DO_CALLBACK(ped, cbName, arg)               (ped)->callbacks.cbName(arg)
#else
# define ED_CALLBACK(ped, cbName)                       (ped)->config->cb->cbName
# define ED_DO_CALLBACK(ped, cbName, arg)               (ped)->config->cb->cbName(arg)
#endif


#if (CFG_HAL_NEEDS_EVENTS)
# define ED_DO_EVENT(ped, ev)                           do { (ped)->stackEvent.devIdx = (ped)->devIdx; (ped)->stackEvent.event = (ev); essHALStackEvent(&((ped)->stackEvent)); } while (0)
# define ED_DO_EVENT1(ped, ev, a1)                      do { (ped)->stackEvent.devIdx = (ped)->devIdx; (ped)->stackEvent.event = (ev); (ped)->stackEvent.arg1 = (a1); essHALStackEvent(&((ped)->stackEvent)); } while (0)
# define ED_DO_EVENT2(ped, ev, a1, a2)                  do { (ped)->stackEvent.devIdx = (ped)->devIdx; (ped)->stackEvent.event = (ev); (ped)->stackEvent.arg1 = (a1); (ped)->stackEvent.arg2 = (a2); essHALStackEvent(&((ped)->stackEvent)); } while (0)
# define ED_DO_EVENT3(ped, ev, a1, a2, a3)              do { (ped)->stackEvent.devIdx = (ped)->devIdx; (ped)->stackEvent.event = (ev); (ped)->stackEvent.arg1 = (a1); (ped)->stackEvent.arg2 = (a2); (ped)->stackEvent.arg3 = (a3); essHALStackEvent(&((ped)->stackEvent)); } while (0)
# define ED_EVENT_RESULT(ped)                           ((ped)->stackEvent.result)
#else
# define ED_DO_EVENT(ped, ev)                           do { } while (0)
# define ED_DO_EVENT1(ped, ev, a1)                      do { } while (0)
# define ED_DO_EVENT2(ped, ev, a1, a2)                  do { } while (0)
# define ED_DO_EVENT3(ped, ev, a1, a2, a3)              do { } while (0)
# define ED_EVENT_RESULT(ped)                           /* should be unused */
#endif


#if (CFG_ESS_SUPPORT_COE)

/* Don't forget the public ESS_OD_OBJECT_FLAGS_ flags defined in essTypes.h! */
#define ESS_OD_OBJECT_FLAGS_PDO                         ((ESS_OD_OBJECT_FLAGS)0x8000)
#define ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT              ((ESS_OD_OBJECT_FLAGS)0x4000)
#define ESS_OD_OBJECT_FLAGS_ARRAY                       ((ESS_OD_OBJECT_FLAGS)0x2000)
#define ESS_OD_OBJECT_FLAGS_PDO_PARAMETER               ((ESS_OD_OBJECT_FLAGS)0x1000)
                                                        /* To mark objects of which infos must be free'd */
#define ESS_OD_OBJECT_FLAGS_ALLOCEDINFOS                ((ESS_OD_OBJECT_FLAGS)0xf000)

#define ESS_OD_OBJECT_FLAGS_PRIVATE_MASK                ((ESS_OD_OBJECT_FLAGS)0xff00)

    /*
     * TRUE if ESS_PDO_ENTRY *p defines a padding object to map objects with
     * more than 31 bytes according to ETG.1020
     */
#define ESS_IS_EXTEND_OBJECT(p) \
    ((ESS_PDO_SUB_PADDING == (p)->objIdx) && (ESS_PDO_SUB_EXTEND  == (p)->subIdx))

	/*
     * Configure an optimized copy PDO mapping
	*/
#define ESS_FLAG_OPT_COPY   (1 << 7)

#define ESS_SET_OPTIMZED_COPY_SIZE(mapInfo, pdoBitSize) {                              \
	 uint16_t pdoByteSize = (uint16_t)(((pdoBitSize) >> 3) & 0x7FF);                   \
 	 (mapInfo)->addrBit &= 0x0F;                                                       \
	 (mapInfo)->addrBit |= (uint8_t)((ESS_FLAG_OPT_COPY | (pdoByteSize >> 4)) & 0xF0); \
	 (mapInfo)->bitLen   = (uint8_t)(pdoByteSize);}

#define ESS_GET_OPTIMZED_COPY_SIZE(mapInfo)                               \
	 (((mapInfo)->addrBit & 0xF0) ?                                       \
      (((mapInfo)->bitLen + ((mapInfo)->addrBit >> 4) * 256) & 0x7FF) : 0)

typedef struct _ESS_OD_ENTRY {
    void* infos;
    uint8_t* data;
    uint32_t bitLen;
    COE_ACCESS access;
    ESS_OD_ENTRY_FLAGS flags;
    uint8_t subIdx;
} ESS_OD_ENTRY;

typedef struct _ESS_OD_OBJECT {
    void* infos;
    ESS_OD_ENTRY** entries;
    uint16_t index;
    uint16_t entriesCapacity;
    uint16_t entriesCount;
    ESS_OD_OBJECT_FLAGS flags;
} ESS_OD_OBJECT;

typedef struct _ESS_OBJECT_DICT {
    ESS_OD_OBJECT** objects;
    uint16_t objectsCapacity;
    uint16_t objectsCount;
    ESS_OD_FLAGS flags;
} ESS_OBJECT_DICT;

/* When the stack requests an entry from the dictionary, it will only get this - to allow the dict serving entries that do not actually exist. */
typedef struct _ESS_OD_VIRTUAL_ENTRY {
    ESS_OD_ENTRY* entry;
    ESS_OD_OBJECT* obj;
    uint8_t subIdx;
    ESS_BOOL complAcc;
} ESS_OD_VIRTUAL_ENTRY;

typedef struct _ESS_OD_OBJECT_INFOS_PDO {
    ESS_PDO_ENTRY items[CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT];
    const char* objName;
    uint8_t itemsUsed;
    ESS_PDO_CFG_FLAGS flags;
} ESS_OD_OBJECT_INFOS_PDO;

typedef struct _ESS_OD_OBJECT_INFOS_PDOASSIGN {
    uint16_t items[CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT];
    const char* objName;
    uint8_t itemsUsed;
    ESS_BOOL writableInPreOp;
} ESS_OD_OBJECT_INFOS_PDOASSIGN;

typedef struct _ESS_OD_OBJECT_INFOS_ARRAY {
    uint8_t* data;
    const char* objName;
    uint32_t entryBitLen;
    COE_DATATYPE entriesDataType;
    COE_ACCESS entriesAcc;
    uint16_t entriesDataStride;
    uint8_t entriesCount;
    uint8_t  maxSubIdx;
} ESS_OD_OBJECT_INFOS_ARRAY;

typedef struct _ESS_OD_OBJECT_INFOS_PDOPARAM {
    const char* name;
    const void* exclData;
    ESS_OD_PDOPARAM_FLAGS flags;
    uint16_t exclDataLen;
    uint8_t  maxSubIdx;
    ESS_BOOL state;
    ESS_BOOL control;
    ESS_BOOL toggle;
} ESS_OD_OBJECT_INFOS_PDOPARAM;


/* To store infos about SDO up/downloads in progress */
typedef struct _ESS_SDOTRANSFER_STATUS {
    ESS_OD_VIRTUAL_ENTRY entry;
    uint32_t totalBytes;
    uint32_t bytesDone;
    ESS_TIMESTAMP lastPacket;
    uint16_t objIdx;
    uint8_t subIdx;
    uint8_t nextToggleBit;
    ESS_BOOL active;
} ESS_SDOTRANSFER_STATUS;

/* To store infos about SDO info transfer (segmented item list) */
typedef struct _ESS_SDOINFOTRANSFER_STATUS {
    uint16_t listType;
    uint16_t objCount;
    uint16_t nextItem;
    ESS_BOOL active;
} ESS_SDOINFOTRANSFER_STATUS;

typedef struct _ESS_PDO_MAP_INFO {
    uint8_t* data;
    uint16_t bufferOffset;
    uint8_t addrBit;    /* Bit 0..3 = bit offset / Bit 4..7 Optimized copy info */
    uint8_t bitLen;
} ESS_PDO_MAP_INFO;

typedef struct _ESS_SM_MAPPINGS {
    /* Array of ESS_PDO_MAP_INFO */
    ESS_PDO_MAP_INFO* entries;
    /* Number of items allocated in entries[] */
    uint16_t entriesCapacity;
    /* Number of items used in entries[] */
    uint16_t entriesCount;
} ESS_SM_MAPPINGS;

#endif /* CFG_ESS_SUPPORT_COE */

/* To store infos about current blink state of a LED, etc... */
typedef struct _ESS_LED_STATEINFOS {
    ESC_LED_STATE state;
    int8_t lastWritten;
} ESS_LED_STATEINFOS;

#if (CFG_ESS_SUPPORT_FOE)
/* To store infos about current FoE transfer */
typedef struct _ESS_FOE_STATEINFOS {
    void* appHandle;
    ESS_TIMESTAMP lastPacket;
    uint32_t packetNo;
    /* FOE_OPCODE_NONE, FOE_OPCODE_READ or FOE_OPCODE_WRITE */
    uint8_t operation;
    ESS_BOOL uploadDone;
} ESS_FOE_STATEINFOS;
#endif /* CFG_ESS_SUPPORT_FOE */


#if (CFG_ESS_SUPPORT_EOE)
/* To store infos about current EoE transfer */
typedef struct _ESS_EOE_RCV_STATES {
    /* Must be a multiple of 32 (49 * 32 = 1568) */
    uint8_t data[49 * 32];
    uint16_t currDataLen;
    uint8_t lastFragmNo;
    uint8_t frameNo;
    ESS_BOOL timeAppended;
} ESS_EOE_RCV_STATES;

/* To store infos about current EoE transfer */
typedef struct _ESS_EOE_SEND_STATES {
    /* Must be a multiple of 32 (49 * 32 = 1568) */
    uint8_t data[49 * 32];
    uint16_t dataLen;
    uint16_t dataSent;
    uint8_t fragmNo;
    uint8_t frameNo;
    ESS_BOOL active;
} ESS_EOE_SEND_STATES;
#endif /* CFG_ESS_SUPPORT_EOE */

    /* Flags of ESC_DATA */
#define ESS_FLAG_SYNC_ACK_MODE        0x01       /* Acknowledge SYNC signals             */

/* This is the data behind the device ESS_HANDLE */
typedef struct _ESC_DATA {
    /* Used to verify handle and to determine if indexed dev is already opened */
    uint32_t magic;
#ifdef _DEBUG
    /* Used to ensure stack functions aren't called from different threads... (not necessarily implemented on all platforms) */
    void* tIdOnOpen;
#endif
    void* escPointer;
    const ESS_CONFIGURATION* config;

#if (CFG_ESS_COPY_CALLBACKS)
    ESS_CALLBACKS callbacks;
#endif

    ESS_CBDATA_STATE_REQUEST cbDataStateRequest;

    /* Needed cause HAL works only by index */
    uint8_t devIdx;

    /* Current EtherCAT State, e.g. ESC_STATE_PREOP */
    ESC_STATE currState;

    /* Configuration flags */
    uint8_t  flags;

    /* Copy of the Sync/Latch PDI configuration register 0x151 */
    uint8_t  escRegPdiCfg;

    /* Data about the SMs, also used as callback data */
    ESS_CBDATA_SM cbDataSM[CFG_ESS_MAX_SM_COUNT];
    /* Items to use in cbDataSM[], read from ESC_REG_SMCOUNT / by SM changed interrupt */
    uint8_t smCount;
    /* Index of MBoxOut SM (Receive from master, usually SM0) in cbDataSM, determined by buffer type and direction (SM_IDX_NONE if none exists) */
    ESS_SM smIdxMBoxOut;
    /* Index of MBoxIn SM (Send to master, usually SM1) in cbDataSM, determined by buffer type and direction (SM_IDX_NONE if none exists) */
    ESS_SM smIdxMBoxIn;
    /* True when Outputs SM exists (determined by buffer type and direction) */
    ESS_BOOL smOutputsExist;

    /* = SM length - 6 bytes mbx header */
    uint16_t mbxOutDataLen;
    /* = SM length - 6 bytes mbx header */
    uint16_t mbxInDataLen;

    /* True when outputs are mapped / outputs size is > 0 */
    ESS_BOOL outputsExist;
    /* True when in Op and output SM exists */
    ESS_BOOL outputsActive;
    /* True when inputs are mapped / inputs size is > 0 */
    ESS_BOOL inputsExist;
    /* True when in SafeOp or Op and input SM exists */
    ESS_BOOL inputsActive;
    /* True when mailbox SMs active (usually all states but Init) */
    ESS_BOOL mbxActive;

    ESS_CBDATA_SM_EVENT cbDataInputsUpdated;
    ESS_CBDATA_SM_EVENT cbDataOutputsUpdated;

    /* Last own mbx service counter (1..7) */
    uint8_t mbxCnt;
    /* Last mailbox service counter received from master (1..7, or 0 for old master which does not support this) */
    uint8_t mbxCntMaster;
    /* Temp storage to receive and send */
    ESS_MBX_PACKET mbxTempPacket;
    /* Last packet that was stored in our send mailbox */
    ESS_MBX_PACKET mbxLastStoredPacket;
    /* Last packet that was read from our send mailbox */
    ESS_MBX_PACKET mbxLastSentPacket;
    ESS_BOOL mbxRepeatBitMaster;
    ESS_BOOL mbxSendFull;
    ESS_BOOL mbxOutInterruptPending;

#if (CFG_ESS_SUPPORT_COE)
    ESS_OBJECT_DICT dict;
    ESS_SDOTRANSFER_STATUS coeSDOUploadState;
    ESS_SDOTRANSFER_STATUS coeSDODownloadState;
    /* Queued emergency, when coeEmergencyReady != 0 it will be sent next time send mailbox is free */
    ESS_COE_EMERGENCY coeEmergency;
    ESS_BOOL coeEmergencyReady;
    /* Set by essODSyncInputMappedEntries() */
    ESS_BOOL smUpdateInputs;
    /* Entries mapped to each SM */
    ESS_SM_MAPPINGS smMappedEntries[CFG_ESS_MAX_SM_COUNT];
    /* For input SMs we keep a copy of the complete buffer. (For easier copying the entries in the buffer) */
    uint8_t* smBuffer[CFG_ESS_MAX_SM_COUNT];
# if (!CFG_ESS_SMCOPYINUSERBUFFER)
    /* Stores how much space is allocated at each smBuffer[] pointer */
    uint16_t smBufferAllocSize[CFG_ESS_MAX_SM_COUNT];
# endif
# if (CFG_ESS_OD_ALLOW_HANDLE_SM_TYPE)
    uint8_t smObj0x1c00Data[CFG_ESS_MAX_SM_COUNT];
# endif
# if (CFG_ESS_COE_SUPPORT_SDO_INFO_SERVICE)
    ESS_SDOINFOTRANSFER_STATUS sdoInfoState;
# endif
#endif /* CFG_ESS_SUPPORT_COE */

#if (CFG_ESS_SUPPORT_FOE)
    ESS_FOE_STATEINFOS foeState;
#endif

#if (CFG_ESS_SUPPORT_EOE)
    ESS_EOE_RCV_STATES eoeReceive;
    ESS_EOE_SEND_STATES eoeSend;
#endif

#if (CFG_ESS_SUPPORT_VOE)
    ESS_BOOL voeMoreFragments;
#endif

#if (CFG_ESS_SUPPORT_SOE)
    ESS_BOOL soeMoreFragments;
#endif

#if (CFG_ESS_SUPPORT_AOE)
    ESS_BOOL aoeMoreFragments;
#endif

#if (CFG_HAL_NEEDS_EVENTS)
    ESS_EVENT_DATA stackEvent;
#endif

#if (CFG_ESS_SERVE_RUN_LED || CFG_ESS_SERVE_ERR_LED || CFG_ESS_SERVE_CUSTOM_LED)
    ESS_LED_STATEINFOS ledStates[ESC_LED_TYPE_NUMBER_OF_LEDS];
#endif
} ESC_DATA;




/* Below function declarations for functions in other files (should be used by ess.c only) */

#if (CFG_ESS_SUPPORT_COE)

COE_ABORTCODE essOD_DoEntryEventCallback(ESC_DATA* ed, uint16_t objIdx, uint8_t subIdx, ESS_OD_ENTRY_CALLBACK cbType);

COE_ABORTCODE essOD_GetEntry(const ESC_DATA* ed, ESS_OD_VIRTUAL_ENTRY* dest, uint16_t objIdx, int16_t subIdx);

uint32_t essOD_GetEntrySize(const ESC_DATA* ed, ESS_OD_VIRTUAL_ENTRY* e, ESS_BOOL forUpload);

COE_ABORTCODE essOD_DoEntryUpload(ESC_DATA* ed, ESS_OD_VIRTUAL_ENTRY* e, uint8_t* dest, uint32_t offset, uint16_t copyLen, ESS_BOOL isLastPart);

COE_ABORTCODE essOD_DoEntryDownload(ESC_DATA* ed, ESS_OD_VIRTUAL_ENTRY* e, uint8_t* src, uint32_t totalDownloadLen, uint32_t offset, uint16_t copyLen, ESS_BOOL isLastPart);

uint16_t essOD_GetPDOAssignInfos(ESC_DATA* ed, ESS_SM sm, const ESS_OD_OBJECT_INFOS_PDOASSIGN** infos);

uint16_t essOD_GetPDOInfos(ESC_DATA* ed, uint16_t pdo, const ESS_OD_OBJECT_INFOS_PDO** infos);

uint16_t essOD_GetEntryData(const ESC_DATA* ed, uint16_t objIdx, uint8_t subIdx, uint8_t** data);


# if (CFG_ESS_COE_SUPPORT_SDO_INFO_SERVICE)

void essOD_SDOInfo_GetListCounts(const ESS_OBJECT_DICT* dict, uint16_t* dest);

uint16_t essOD_SDOInfo_GetObjIndex(const ESS_OBJECT_DICT* dict, uint16_t listType, uint16_t objListPos);

COE_ABORTCODE essOD_SDOInfo_GetObjInfos(const ESS_OBJECT_DICT* dict, uint16_t objIdx, ESS_OD_OBJECT_INFOS* dest, uint8_t* maxSubIdx, COE_DATATYPE* dataType);

COE_ABORTCODE essOD_SDOInfo_GetEntryInfos(const ESS_OBJECT_DICT* dict, uint16_t objIdx, uint8_t subIdx, ESS_OD_ENTRY_INFOS* dest, COE_ACCESS* acc, uint32_t* bitLen);

# endif /* CFG_ESS_COE_SUPPORT_SDO_INFO_SERVICE */


#endif /* CFG_ESS_SUPPORT_COE */


#endif /* ESS_PRIVATE_INCLUDED */
