/*
*       Copyright (c) 2011-2020 by esd electronics gmbh
*
*  This software is copyrighted by and is the sole property of
*  esd gmbh.  All rights, title, ownership, or other interests
*  in the software remain the property of esd gmbh. This
*  software may only be used in accordance with the corresponding
*  license agreement.  Any unauthorized use, duplication, transmission,
*  distribution, or disclosure of this software is expressly forbidden.
*
*  This Copyright notice may not be removed or modified without prior
*  written consent of esd gmbh.
*
*  esd gmbh, reserves the right to modify this software without notice.
*
*  esd electronics gmbh                   Tel. +49-511-37298-0
*  Vahrenwalder Strasse 207               Fax. +49-511-37298-68
*  30165 Hannover                         http://www.esd.eu
*  Germany                                sales@esd.eu
*
*************************************************************************
*
* Description:
* =============================================================================
* This module contains code to handle the CoE object dictionary
*
*/

#include "../include/essSystem.h"
#include "../include/essPrivate.h"
#include "../hal/essHAL.h"
#include "../include/ess.h"

/*
 * According to ETG.2010 the stack should return an 'empty' entry description
 * for any sub index which does not exist but also does not exceeds the maximum
 * possible sub index of a dynamic array.
 * Currently this seems to break the conformance test.
 */ 
#define ESS_NO_SDO_INFO_CTT_ERROR 

    /* Exceptions from the esd PC-lint/FlexeLint policy for this file */
/*lint -e717  [Info: do ... while(0)]            */
/*lint -e750  [Info: local macro not referenced] */

#if (CFG_ESS_SUPPORT_COE)

#if (CFG_ESS_COE_SUPPORT_SDO_INFO_SERVICE)
# define DECLARE_OBJECT_INFOS(varName, objName, objDataType, objCode) static const ESS_OD_OBJECT_INFOS varName[1] = { { objName, objDataType, objCode } }
# define DECLARE_ENTRY_INFOS(varName, entryName, entryDataType) static const ESS_OD_ENTRY_INFOS varName[1] = { { entryName, NULL, NULL, NULL, 0, entryDataType } }
#else
# define DECLARE_OBJECT_INFOS(varName, objName, objDataType, objCode) static const ESS_OD_OBJECT_INFOS* varName = NULL
# define DECLARE_ENTRY_INFOS(varName, entryName, entryDataType) static const ESS_OD_ENTRY_INFOS* varName = NULL
#endif


/* Macros used for special "code injection" - shouldn't be needed! */
#ifndef ESS_DO_OD_COMPLACC_ENTRYUPLOAD
# define ESS_DO_OD_COMPLACC_ENTRYUPLOAD()
#endif
#ifndef ESS_DO_OD_COMPLACC_GETENTRYSIZE
# define ESS_DO_OD_COMPLACC_GETENTRYSIZE()
#endif

#define ESS_SUB0_NAME           "Subindex 000"

static uint32_t StrLen(const uint8_t* s)
{
    uint32_t len = 0;

    while (*s++)
        len++;

    return len;
}

static void DecToAscii(char *p, uint8_t value)
{
  *p++ = '0' + (value / 100);
  value %= 100;
  *p++ = '0' + (value / 10);
  value %= 10;
  *p++ = '0' + value;
  *p = '\0';
}

/* Performs a binary search in dict->objects[], returns NULL when not found */
static ESS_OD_OBJECT* FindObj(const ESS_OBJECT_DICT* dict, uint16_t objIdx)
{
    uint32_t start = 0;
    uint32_t end = dict->objectsCount;

    while (end > start) {
        uint32_t i = start + ((end - start) >> 1);

        if (dict->objects[i]->index == objIdx)
            return dict->objects[i];

        if (objIdx > dict->objects[i]->index) {
            start = i + 1;
        } else {
            end = i;
        }
    }

    return NULL;
}

/* Performs a binary search in obj->entries[], returns -1 when not found */
static ESS_OD_ENTRY* FindEntry(const ESS_OD_OBJECT* obj, uint8_t subIdx)
{
    uint32_t start = 0;
    uint32_t end = obj->entriesCount;

    while (end > start) {
        uint32_t i = start + ((end - start) >> 1);

        if (obj->entries[i]->subIdx == subIdx)
            return obj->entries[i];

        if (subIdx > obj->entries[i]->subIdx) {
            start = i + 1;
        } else {
            end = i;
        }
    }

    return NULL;
}

/* Insert the object into the dict (sorted) */
static void InsertObject(ESS_OBJECT_DICT* dict, ESS_OD_OBJECT* obj)
{
    uint16_t i;
    uint16_t j;

    for (i = 0; i < dict->objectsCount; i++) {
        if (dict->objects[i]->index > obj->index)
            break;
    }
    /* i is now dest idx of new object */

    for (j = dict->objectsCount; j > i; j--) {
        dict->objects[j] = dict->objects[j - 1];
    }
    /* other entries now moved away */

    dict->objects[i] = obj;
    dict->objectsCount++;
}

/* Insert the entry into the object (sorted) */
static void InsertEntry(ESS_OD_OBJECT* obj, ESS_OD_ENTRY* entry)
{
    uint16_t i;
    uint16_t j;

    for (i = 0; i < obj->entriesCount; i++) {
        if (obj->entries[i]->subIdx > entry->subIdx)
            break;
    }
    /* i is now dest idx of new entry */

    for (j = obj->entriesCount; j > i; j--) {
        obj->entries[j] = obj->entries[j - 1];
    }
    /* other entries now moved away */

    obj->entries[i] = entry;
    obj->entriesCount++;
}

/* Makes sure that obj list capacity is greater than obj count */
static ESS_BOOL VerifyDictCapacity(ESS_OBJECT_DICT* dict)
{
    if (dict->objectsCount >= dict->objectsCapacity) {
        ESS_OD_OBJECT** pNew;

        ESS_REALLOC(pNew, ESS_OD_OBJECT**, dict->objects, sizeof(ESS_OD_OBJECT*) * (dict->objectsCapacity + CFG_ESS_COE_OBJECT_LIST_INCREMENT));
        if (pNew == NULL) {
            ESS_DBG(("ERROR: realloc for %lu byte failed!", (unsigned long)(sizeof(ESS_OD_OBJECT*) * (dict->objectsCapacity + CFG_ESS_COE_OBJECT_LIST_INCREMENT))));
            return ESS_TRUE;
        }

        dict->objects = pNew;
        dict->objectsCapacity += CFG_ESS_COE_OBJECT_LIST_INCREMENT;
    }

    return ESS_FALSE;
}

/* Makes sure that entry list capacity is greater than entry count */
static ESS_BOOL VerifyObjectCapacity(ESS_OD_OBJECT* obj)
{
    if (obj->entriesCount >= obj->entriesCapacity) {
        ESS_OD_ENTRY** pNew;

        ESS_REALLOC(pNew, ESS_OD_ENTRY**, obj->entries, sizeof(ESS_OD_ENTRY*) * (obj->entriesCapacity + CFG_ESS_COE_ENTRY_LIST_INCREMENT));
        if (pNew == NULL) {
            ESS_DBG(("ERROR: realloc for %lu byte failed!", (unsigned long)(sizeof(ESS_OD_ENTRY*) * (obj->entriesCapacity + CFG_ESS_COE_ENTRY_LIST_INCREMENT))));
            return ESS_TRUE;
        }

        obj->entries = pNew;
        obj->entriesCapacity += CFG_ESS_COE_ENTRY_LIST_INCREMENT;
    }

    return ESS_FALSE;
}

/* Just calls cb with the given arguments and returns the abort code from the callbackData */
static COE_ABORTCODE DoEntryRWCallback(ESC_DATA* ed, ESS_BOOL isRead, void* destSrc, uint16_t objIdx, uint8_t entryIdx, ESS_BOOL isLast, uint32_t len, uint32_t offset)
{
    if (ED_CALLBACK(ed, cbCoEReadWrite)) {
        ESS_CBDATA_COE_READWRITE cbrw;

        cbrw.hDev = (ESS_HANDLE)ed;
        cbrw.isRead = isRead;
        cbrw.destSrc = (uint8_t*)destSrc;
        cbrw.entryIdx = entryIdx;
        cbrw.isLastPart = isLast;
        cbrw.len = len;
        cbrw.objIdx = objIdx;
        cbrw.offset = offset;
        cbrw.abortCode = COE_ABORTCODE_SUBINDEX;

        ED_DO_CALLBACK(ed, cbCoEReadWrite, &cbrw);

        return cbrw.abortCode;

    } else {
        ESS_DBG(("ERROR: CoE RWCallback is missing!"));
        return COE_ABORTCODE_GENERAL;
    }
}

/* Calls ed->...cbCoEEvent() with the given arguments and returns its abortCode */
COE_ABORTCODE essOD_DoEntryEventCallback(ESC_DATA* ed, uint16_t objIdx, uint8_t subIdx, ESS_OD_ENTRY_CALLBACK cbType)
{
    if (ED_CALLBACK(ed, cbCoEEvent)) {
        ESS_CBDATA_COE_EVENT cbData;

        cbData.hDev = (ESS_HANDLE)ed;
        cbData.abortCode = COE_ABORTCODE_NONE;
        cbData.cbType = cbType;
        cbData.objIdx = objIdx;
        cbData.entryIdx = subIdx;

        ED_DO_CALLBACK(ed, cbCoEEvent, &cbData);

        return cbData.abortCode;
    }

    return COE_ABORTCODE_NONE;
}


#if (CFG_ESS_COE_SUPPORT_SDO_INFO_SERVICE)

/* Writes number of objects in each of the 5 object lists to dest */
void essOD_SDOInfo_GetListCounts(const ESS_OBJECT_DICT* dict, uint16_t* dest)
{
    int32_t i;

	/*lint -e778 (Constant expression evaluates to 0) */
    dest[MBX_COE_SDOINFO_ODLISTREQ_ALLOBJECTS - 1] = dict->objectsCount;
    dest[MBX_COE_SDOINFO_ODLISTREQ_RXPDOS - 1] = 0;
    dest[MBX_COE_SDOINFO_ODLISTREQ_TXPDOS - 1] = 0;
    dest[MBX_COE_SDOINFO_ODLISTREQ_REPLACEMOBJECTS - 1] = 0;
    dest[MBX_COE_SDOINFO_ODLISTREQ_STARTUPOBJECTS - 1] = 0;

    for (i = 0; i < dict->objectsCount; i++) {
        const ESS_OD_OBJECT* obj = dict->objects[i];

        /* Special object?: */
        if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO) {
            /* Nothing, this object can't be anything list specific */

        } else if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT) {
            /* Nothing, this object can't be anything list specific */

#if (CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS)
        } else if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO_PARAMETER) {
            const ESS_OD_OBJECT_INFOS_PDOPARAM* infos = (const ESS_OD_OBJECT_INFOS_PDOPARAM*)obj->infos;

            if ((infos->flags & ESS_OD_PDOPARAM_FLAGS_STATE_EXISTS) || ((obj->index < 0x1a00) && (obj->index >= 0x1600) && (infos->flags & ESS_OD_PDOPARAM_FLAGS_TOGGLE_EXISTS)))
                dest[MBX_COE_SDOINFO_ODLISTREQ_TXPDOS - 1]++;

            if ((obj->index < 0x1600) && (obj->index >= 0x1400) && (infos->flags & (ESS_OD_PDOPARAM_FLAGS_CONTROL_EXISTS | ESS_OD_PDOPARAM_FLAGS_TOGGLE_EXISTS)))
                dest[MBX_COE_SDOINFO_ODLISTREQ_RXPDOS - 1]++;
#endif

        } else if (obj->flags & ESS_OD_OBJECT_FLAGS_ARRAY) {
            const ESS_OD_OBJECT_INFOS_ARRAY* infos = (const ESS_OD_OBJECT_INFOS_ARRAY*)obj->infos;

            if (infos->entriesAcc & COE_ACCESS_RXMAPPABLE)
                dest[MBX_COE_SDOINFO_ODLISTREQ_RXPDOS - 1]++;

            if (infos->entriesAcc & COE_ACCESS_TXMAPPABLE)
                dest[MBX_COE_SDOINFO_ODLISTREQ_TXPDOS - 1]++;

            if (infos->entriesAcc & COE_ACCESS_BACKUPOBJECT)
                dest[MBX_COE_SDOINFO_ODLISTREQ_REPLACEMOBJECTS - 1]++;

            if (infos->entriesAcc & COE_ACCESS_SETTINGSOBJECT)
                dest[MBX_COE_SDOINFO_ODLISTREQ_STARTUPOBJECTS - 1]++;

        } else {
            /* Normal object: */
            COE_ACCESS acc = 0;
            uint16_t j;

            for (j = 0; j < obj->entriesCount; j++) {
                const ESS_OD_ENTRY* entry = obj->entries[j];

                if (entry->flags & ESS_OD_ENTRY_FLAGS_NOSDOINFO)
                    continue;

                acc |= entry->access;
            }

            if (acc & COE_ACCESS_RXMAPPABLE)
                dest[MBX_COE_SDOINFO_ODLISTREQ_RXPDOS - 1]++;

            if (acc & COE_ACCESS_TXMAPPABLE)
                dest[MBX_COE_SDOINFO_ODLISTREQ_TXPDOS - 1]++;

            if (acc & COE_ACCESS_BACKUPOBJECT)
                dest[MBX_COE_SDOINFO_ODLISTREQ_REPLACEMOBJECTS - 1]++;

            if (acc & COE_ACCESS_SETTINGSOBJECT)
                dest[MBX_COE_SDOINFO_ODLISTREQ_STARTUPOBJECTS - 1]++;
        }
    }
}

/* Returns the index of the object, used with CoE_Handle_ListRequest() */
uint16_t essOD_SDOInfo_GetObjIndex(const ESS_OBJECT_DICT* dict, uint16_t listType, uint16_t objListPos)
{
#   define ACCESS_MATCHES_LISTTYPE(acc, lt) ((((lt) == MBX_COE_SDOINFO_ODLISTREQ_TXPDOS) && ((acc) & COE_ACCESS_TXMAPPABLE)) || (((lt) == MBX_COE_SDOINFO_ODLISTREQ_RXPDOS) && ((acc) & COE_ACCESS_RXMAPPABLE)) || (((lt) == MBX_COE_SDOINFO_ODLISTREQ_REPLACEMOBJECTS) && ((acc) & COE_ACCESS_BACKUPOBJECT)) || (((lt) == MBX_COE_SDOINFO_ODLISTREQ_STARTUPOBJECTS) && ((acc) & COE_ACCESS_SETTINGSOBJECT)))
    uint16_t i;
    uint16_t cnt = 0;

    if (listType == MBX_COE_SDOINFO_ODLISTREQ_ALLOBJECTS)
        return dict->objects[objListPos]->index;

    for (i = 0; i < dict->objectsCount; i++) {
        const ESS_OD_OBJECT* obj = dict->objects[i];
        ESS_BOOL objMatchesListType = ESS_FALSE;
        uint16_t j;

        /* Special object?: */
        if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO) {
            /* Nothing, this object can't be anything list specific */

        } else if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT) {
            /* Nothing, this object can't be anything list specific */

#if (CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS)
        } else if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO_PARAMETER) {
            const ESS_OD_OBJECT_INFOS_PDOPARAM* infos = (const ESS_OD_OBJECT_INFOS_PDOPARAM*)obj->infos;
            COE_ACCESS acc = 0;

            if ((infos->flags & ESS_OD_PDOPARAM_FLAGS_STATE_EXISTS) || ((obj->index < 0x1a00) && (obj->index >= 0x1600) && (infos->flags & ESS_OD_PDOPARAM_FLAGS_TOGGLE_EXISTS)))
                acc |= COE_ACCESS_TXMAPPABLE;

            if ((obj->index < 0x1600) && (obj->index >= 0x1400) && (infos->flags & (ESS_OD_PDOPARAM_FLAGS_CONTROL_EXISTS | ESS_OD_PDOPARAM_FLAGS_TOGGLE_EXISTS)))
                acc |= COE_ACCESS_RXMAPPABLE;

            objMatchesListType = ACCESS_MATCHES_LISTTYPE(acc, listType);
#endif

        } else if (obj->flags & ESS_OD_OBJECT_FLAGS_ARRAY) {
            const ESS_OD_OBJECT_INFOS_ARRAY* infos = (const ESS_OD_OBJECT_INFOS_ARRAY*)obj->infos;
            objMatchesListType = ACCESS_MATCHES_LISTTYPE(infos->entriesAcc, listType);

        } else {
            /* Normal object: */
            for (j = 0; j < obj->entriesCount; j++) {
                const ESS_OD_ENTRY* entry = obj->entries[j];

                if (entry->flags & ESS_OD_ENTRY_FLAGS_NOSDOINFO)
                    continue;

                if (ACCESS_MATCHES_LISTTYPE(entry->access, listType)) {
                    objMatchesListType = ESS_TRUE;
                    break;
                }
            }
        }

        if (objMatchesListType) {
            if (cnt++ == objListPos)
                return obj->index;
        }
    }

    return 0xffff;
#   undef ACCESS_MATCHES_LISTTYPE
}

/* Returns object infos to handle SDO info object request */
COE_ABORTCODE essOD_SDOInfo_GetObjInfos(const ESS_OBJECT_DICT* dict, uint16_t objIdx,
                                        ESS_OD_OBJECT_INFOS* dest, uint8_t* maxSubIdx,
                                        COE_DATATYPE* dataType)
{
    const ESS_OD_OBJECT* obj;

    /* Return "Unsupported access" for requests with index < 0x1000 according to ETG.1020 */
    if(objIdx < 0x1000)
        return COE_ABORTCODE_ACCESS;

    obj = FindObj(dict, objIdx);
    if (obj == NULL)
        return COE_ABORTCODE_INDEX;

    /* Special object?: */
    if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO) {
        const ESS_OD_OBJECT_INFOS_PDO* infos = (const ESS_OD_OBJECT_INFOS_PDO*)obj->infos;
        dest->code = COE_CODE_RECORD;
        dest->name = infos->objName;
        if(dict->flags & ESS_OD_SDO_INFO_MODE_ETG1020)
            *maxSubIdx = CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT;
        else
            *maxSubIdx = infos->itemsUsed;
        *dataType = COE_DATATYPE_PDOMAPPING;

    } else if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT) {
        const ESS_OD_OBJECT_INFOS_PDOASSIGN* infos = (const ESS_OD_OBJECT_INFOS_PDOASSIGN*)obj->infos;
        dest->code = COE_CODE_ARRAY;
        dest->name = infos->objName;
        if(dict->flags & ESS_OD_SDO_INFO_MODE_ETG1020)
            *maxSubIdx = CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT;
        else
            *maxSubIdx = infos->itemsUsed;
        *dataType = COE_DATATYPE_UINT;

#if (CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS)
    } else if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO_PARAMETER) {
        const ESS_OD_OBJECT_INFOS_PDOPARAM* infos = (const ESS_OD_OBJECT_INFOS_PDOPARAM*)obj->infos;
        dest->code = COE_CODE_RECORD;
        dest->name = infos->name;
        *maxSubIdx = infos->maxSubIdx;
        *dataType = COE_DATATYPE_INVALID; /* COE_DATATYPE_PDOCOMMPAR? */

#endif

    } else if (obj->flags & ESS_OD_OBJECT_FLAGS_ARRAY) {
        const ESS_OD_OBJECT_INFOS_ARRAY* infos = (const ESS_OD_OBJECT_INFOS_ARRAY*)obj->infos;
        dest->code = COE_CODE_ARRAY;
        dest->name = infos->objName;
        if(dict->flags & ESS_OD_SDO_INFO_MODE_ETG1020)
            *maxSubIdx = infos->maxSubIdx;
        else
            *maxSubIdx = infos->entriesCount;

        *dataType = infos->entriesDataType;

    } else {
        /* Normal object: */
        const ESS_OD_OBJECT_INFOS* infos = (const ESS_OD_OBJECT_INFOS*)obj->infos;
        uint16_t i;

        *maxSubIdx = 0;
        for (i = 0; i < obj->entriesCount; ++i) {
            const ESS_OD_ENTRY* entry = obj->entries[i];

            if (entry->flags & ESS_OD_ENTRY_FLAGS_USE_DATA_AS_MAXSUBIDX) {
                *maxSubIdx = entry->data[0];
                break;
            }

            if (entry->flags & ESS_OD_ENTRY_FLAGS_NOSDOINFO)
                continue;

            if (entry->subIdx > *maxSubIdx)
                *maxSubIdx = entry->subIdx;
        }

        if (infos) {
            dest->code = infos->code;
            dest->name = infos->name;
            *dataType = infos->dataType;
        } else {
            /* This should not happen, app shall give us infos when enabling SDO infos!: */
            dest->code = (obj->entriesCount == 1) ? COE_CODE_VARIABLE : COE_CODE_RECORD;
            dest->name = NULL;
            *dataType = COE_DATATYPE_INVALID;
        }
    }

    return COE_ABORTCODE_NONE;
}

/* Returns entry infos to handle SDO info entry request */
COE_ABORTCODE essOD_SDOInfo_GetEntryInfos(const ESS_OBJECT_DICT* dict, uint16_t objIdx, uint8_t subIdx, ESS_OD_ENTRY_INFOS* dest, COE_ACCESS* acc, uint32_t* bitLen)
{
    const ESS_OD_OBJECT* obj;

    /* Return "Unsupported access" for requests with index < 0x1000 following ETG.1020 */
    if(objIdx < 0x1000)
        return COE_ABORTCODE_ACCESS;

    obj = FindObj(dict, objIdx);
    if (obj == NULL)
        return COE_ABORTCODE_INDEX;

    /* Preset to 'empty' entry */
    *bitLen = * acc = 0;
    dest->defaultValue = NULL;
    dest->maxValue = NULL;
    dest->minValue = NULL;
    dest->unitType = 0;
    dest->dataType = COE_DATATYPE_INVALID;
    dest->name = NULL;

    /* Handle sub index 0 of arrays which control structures are managed by the stack */
    if ((subIdx == 0) && (obj->flags & ESS_OD_OBJECT_FLAGS_ALLOCEDINFOS)) {
        *bitLen = 8;
        dest->name = ESS_SUB0_NAME;
        dest->dataType = COE_DATATYPE_USINT;
        switch(obj->flags & ESS_OD_OBJECT_FLAGS_ALLOCEDINFOS)
        {
            case ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT:
                *acc = (((const ESS_OD_OBJECT_INFOS_PDOASSIGN*)obj->infos)->writableInPreOp) ?
                    COE_ACCESS_R | COE_ACCESS_W_PREOP : COE_ACCESS_R;
                break;
            case ESS_OD_OBJECT_FLAGS_ARRAY:
                *acc = ((const ESS_OD_OBJECT_INFOS_ARRAY*)obj->infos)->entriesAcc & ~COE_ACCESS_W;
                break;
            default: /* ESS_OD_OBJECT_FLAGS_PDO and ESS_OD_OBJECT_FLAGS_PDO_PARAMETER */
                *acc = COE_ACCESS_R;    /* CHECKME: May be cond. writable for ESS_OD_OBJECT_FLAGS_PDO */
                break;
        }
        return COE_ABORTCODE_NONE;
    }

    /* Special object?: */
    if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO) {
        const ESS_OD_OBJECT_INFOS_PDO* infos = (const ESS_OD_OBJECT_INFOS_PDO*)obj->infos;
	static char objectName[16] = "PDO Mapping XXX";

#if CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT<255 /* subIdx is 8 bit unsigned so if MAX is 255, there is no need to check */
        if(subIdx > CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT) {
            return COE_ABORTCODE_SUBINDEX;
        }
#endif

        if(subIdx > infos->itemsUsed) {
            /* Return 'Empty' entries (see ETG.1020 >= V1.1.0) */
            return(COE_ABORTCODE_NONE);
        }

        *acc = (infos->flags & ESS_PDO_CFG_FLAGS_WRITABLE_IN_PREOP) ? COE_ACCESS_R | COE_ACCESS_W_PREOP : COE_ACCESS_R;
        *bitLen = 32;
        DecToAscii(objectName + 12, subIdx); 
        dest->name = objectName;
        dest->dataType = COE_DATATYPE_UDINT;

        return(COE_ABORTCODE_NONE);

    } else if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT) {
        const ESS_OD_OBJECT_INFOS_PDOASSIGN* infos = (const ESS_OD_OBJECT_INFOS_PDOASSIGN*)obj->infos;


#if CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT<255 /* subIdx is 8 bit unsigned so if MAX is 255, there is no need to check */
        if(subIdx > CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT) {
            return COE_ABORTCODE_SUBINDEX;
        }
#endif

#ifdef ESS_NO_SDO_INFO_CTT_ERROR
        if(dict->flags & ESS_OD_SDO_INFO_MODE_ETG1020) {
#else
        /* Return 'Empty' entries (see ETG.1020 >= V1.1.0) */
        if(subIdx > infos->itemsUsed) {
            return(COE_ABORTCODE_NONE);
#endif                    
        } else {
            if(subIdx > infos->itemsUsed)
                return(COE_ABORTCODE_SUBINDEX);
        }
        
        *acc = (infos->writableInPreOp) ? COE_ACCESS_R | COE_ACCESS_W_PREOP : COE_ACCESS_R;
        *bitLen = 16;
        dest->name = "PDO Assignment";
        dest->dataType = COE_DATATYPE_UINT;

        return(COE_ABORTCODE_NONE);

#if (CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS)
    } else if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO_PARAMETER) {
        const ESS_OD_OBJECT_INFOS_PDOPARAM* infos = (const ESS_OD_OBJECT_INFOS_PDOPARAM*)obj->infos;
        const ESS_BOOL isRxParam = ((obj->index < 0x1600) && (obj->index >= 0x1400));

        switch (subIdx) {
            case 6:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_EXCLPDOS_EXISTS) {
                    *bitLen = infos->exclDataLen << 3;
                    *acc = COE_ACCESS_R;
                    dest->name = (isRxParam) ? "Exclude RxPDOs" : "Exclude TxPDOs";
                    dest->dataType = COE_DATATYPE_OCTETSTRING;
                    break;
                }
                return COE_ABORTCODE_NONE;  /* Return 'empty' entry for gaps (see ETG.1020) */
            case 7:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_STATE_EXISTS) {
                    *acc = COE_ACCESS_R | COE_ACCESS_TXMAPPABLE;
                    dest->name = (isRxParam) ? "RxPDO State" : "TxPDO State";
                    break;
                }
                return COE_ABORTCODE_NONE;  /* Return 'empty' entry for gaps (see ETG.1020) */
            case 8:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_CONTROL_EXISTS) {
                    *acc = COE_ACCESS_RW | COE_ACCESS_RXMAPPABLE; /* Write access does not seem to be useful, but it's not clearly defined. */
                    dest->name = "RxPDO Control";
                    break;
                }
                return COE_ABORTCODE_NONE;  /* Return 'empty' entry for gaps (see ETG.1020) */
            case 9:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_TOGGLE_EXISTS) {
                    if (isRxParam) {
                        *acc = COE_ACCESS_RW | COE_ACCESS_RXMAPPABLE; /* Write access does not seem to be useful, but it's not clearly defined. */
                        dest->name = "RxPDO Toggle";
                    } else {
                        *acc = COE_ACCESS_R | COE_ACCESS_TXMAPPABLE;
                        dest->name = "TxPDO Toggle";
                    }
                    break;
                }
                return COE_ABORTCODE_NONE;  /* Return 'empty' entry for gaps (see ETG.1020) */
            default:
                if(subIdx < infos->maxSubIdx) {
                    return COE_ABORTCODE_NONE;  /* Return 'empty' entry for gaps (see ETG.1020) */
                }
                return COE_ABORTCODE_SUBINDEX;
        }

        *bitLen = 1;
        dest->dataType = COE_DATATYPE_BOOL;
#endif

    } else if (obj->flags & ESS_OD_OBJECT_FLAGS_ARRAY) {
        const ESS_OD_OBJECT_INFOS_ARRAY* infos = (const ESS_OD_OBJECT_INFOS_ARRAY*)obj->infos;

        if(subIdx <= infos->maxSubIdx) {

#ifdef ESS_NO_SDO_INFO_CTT_ERROR        
            if(dict->flags & ESS_OD_SDO_INFO_MODE_ETG1020) {            
#else
            /* Return 'Empty' entries (see ETG.1020 >= V1.1.0) */
            if(subIdx > infos->entriesCount) {
                return(COE_ABORTCODE_NONE);     
#endif                    
            } else {
                if(subIdx > infos->entriesCount)
                    return(COE_ABORTCODE_SUBINDEX);
            }

            *bitLen = infos->entryBitLen;
            dest->dataType = infos->entriesDataType;
            /* CHECKME: name for essODAddArrayObject() entrys? ("SubIndex nnn") */
            *acc = infos->entriesAcc;
            return(COE_ABORTCODE_NONE);
        }

        return COE_ABORTCODE_SUBINDEX;
    } else {
        /* Normal object: */
        const ESS_OD_ENTRY* entry = FindEntry(obj, subIdx);

        if (NULL == entry) {
	  if(COE_CODE_RECORD == ((const ESS_OD_OBJECT_INFOS*)obj->infos)->code) {
                entry = FindEntry(obj, 0);
                /* Return 'Empty' entries (see ETG.1020 >= V1.1.0) */
                if((entry != NULL) && (subIdx <= *(uint8_t *)entry->data)) {
                    return COE_ABORTCODE_NONE;
                }
            }
            return COE_ABORTCODE_SUBINDEX;
        }
        
        if (entry->flags & ESS_OD_ENTRY_FLAGS_NOSDOINFO) {
            return COE_ABORTCODE_SUBINDEX;
        }


        if ((entry->flags & ESS_OD_ENTRY_FLAGS_UPLOAD_TRIMMED) && (entry->data)) {
            /* So, we're trimming even for SDO Info. Seems odd, but e.g. Beckhoff's EL6692 does it too with its 0x10ff entries. */
            /* (Not to mention 0xffff as entry len for variable length items (ETG.1000.6, "Get Entry Description Response") is not used and unsupported by TwinCAT 2.x...) */
            uint32_t len = (entry->bitLen + 7) >> 3;

            while ((len > 0) && (entry->data[len - 1] == 0x00))
                len--;

            *bitLen = len << 3;
        }
#if 0		
		else if((entry->flags & ESS_OD_ENTRY_FLAGS_DYNAMIC_SIZE) != 0) {
            *bitLen = 0xFFFF;   /* Say: Dynamic size according to ETG1000.6 */
        }
#endif			
        else {
            *bitLen = entry->bitLen;
        }
        *acc = entry->access;

        if (entry->infos) {
            const ESS_OD_ENTRY_INFOS* infos = (const ESS_OD_ENTRY_INFOS*)entry->infos;

            dest->dataType = infos->dataType;
            dest->defaultValue = infos->defaultValue;
            dest->maxValue = infos->maxValue;
            dest->minValue = infos->minValue;
            dest->unitType = infos->unitType;
            if((subIdx != 0) && (obj->flags & ESS_OD_OBJECT_FLAGS_UNIQUE_ENTRY_NAMES)) {
                static char objectName[16];
                int idx;

                for(idx = 0;;idx++) {
                    if(('\0' == infos->name[idx]) || ((sizeof(objectName) - 4)  == idx)) {
                        DecToAscii(objectName + idx, subIdx);
                        break;
                    }
                    objectName[idx] = infos->name[idx];
                }
                dest->name = objectName;
            } else {
                dest->name = infos->name;
            }
        }
    }

    return COE_ABORTCODE_NONE;
}

#endif /* CFG_ESS_COE_SUPPORT_SDO_INFO_SERVICE */

/* Returns a handle to the entry in "dest" which should no be touched by the stack, this allows serving virtual items to the stack! */
COE_ABORTCODE essOD_GetEntry(const ESC_DATA* ed, ESS_OD_VIRTUAL_ENTRY* dest, uint16_t objIdx, int16_t subIdx)
{
    ESS_OD_ENTRY* entry;
    ESS_OD_OBJECT* obj;

    obj = FindObj(&ed->dict, objIdx);
    if (obj == NULL)
        return COE_ABORTCODE_INDEX;

    dest->obj = obj;

#if (CFG_ESS_COE_SUPPORT_COMPLETE_ACCESS)
    if (subIdx < 0) {
        dest->entry = NULL;
        dest->subIdx = (uint8_t)(-subIdx - 1);
        dest->complAcc = ESS_TRUE;

        return COE_ABORTCODE_NONE;
    }
#endif

    /* Special object?: */
    if (obj->flags & (ESS_OD_OBJECT_FLAGS_PDO | ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT | ESS_OD_OBJECT_FLAGS_PDO_PARAMETER | ESS_OD_OBJECT_FLAGS_ARRAY)) {
        entry = NULL;

    } else {
        /* Normal object: */
        entry = FindEntry(obj, (uint8_t)subIdx);
        if (entry == NULL)
            return COE_ABORTCODE_SUBINDEX;
    }

    dest->entry = entry;
    dest->subIdx = (uint8_t)subIdx;
    dest->complAcc = ESS_FALSE;

    return COE_ABORTCODE_NONE;
}

/* Returns total entry size (max len for downloads, len to use for uploads) */
uint32_t essOD_GetEntrySize(const ESC_DATA* ed, ESS_OD_VIRTUAL_ENTRY* e, ESS_BOOL forUpload)
{
    (void)ed;

#if (CFG_ESS_COE_SUPPORT_COMPLETE_ACCESS)
    if (e->complAcc) {
        /* CHECKME: implement CompleteAccess object size calculation */
        ESS_DO_OD_COMPLACC_GETENTRYSIZE();
        return 0;
    }
#endif

    /* Special object?: */
    if (e->obj->flags & ESS_OD_OBJECT_FLAGS_PDO) {
        return (e->subIdx == 0) ? 1 : 4;

    } else if (e->obj->flags & ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT) {
        return (e->subIdx == 0) ? 1 : 2;

#if (CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS)
    } else if (e->obj->flags & ESS_OD_OBJECT_FLAGS_PDO_PARAMETER) {
        const ESS_OD_OBJECT_INFOS_PDOPARAM* infos = (const ESS_OD_OBJECT_INFOS_PDOPARAM*)e->obj->infos;
        return (e->subIdx == 0) ? 1 : (e->subIdx == 6) ? infos->exclDataLen : 1;
#endif

    } else if (e->obj->flags & ESS_OD_OBJECT_FLAGS_ARRAY) {
        const ESS_OD_OBJECT_INFOS_ARRAY* infos = (const ESS_OD_OBJECT_INFOS_ARRAY*)e->obj->infos;

        return (e->subIdx == 0) ? 1 : (infos->entryBitLen + 7) >> 3;

    } else {
        if ((e->entry->flags & ESS_OD_ENTRY_FLAGS_UPLOAD_TRIMMED) && (forUpload) && (e->entry->data)) {
            uint32_t len = (e->entry->bitLen + 7) >> 3;

            while ((len > 0) && (e->entry->data[len - 1] == 0x00))
                len--;

            return len;
        }

        if (((e->entry->flags & ESS_OD_ENTRY_FLAGS_DYNAMIC_SIZE) != 0) && (forUpload) && (NULL == (e->entry->data))) {
            return 0; /* Say: Dynamic size for CoE Upload */
        }

        /* Normal object: */
        return (e->entry->bitLen + 7) >> 3;
    }

}

/*
 * FSoE safety input/output/parameter assumed to be stored in memory as a sequence
 *  of 16-bit data followed by 16-bit CRC
 */
INLINE static void essOD_UploadSafetyData(uint8_t *dest, uint8_t *src, uint32_t offset, uint16_t copyLen)
{
    /* while offset counts 0 1 2 3 4 5 we need to read at src-offsets 0 1 4 5 8 9 and so on */
    for(; copyLen > 0; copyLen--, offset++) {
        *(dest++) = src[(offset & 1u)+((offset & ~1u)<<1u)];
    }
}

/* Checks access/state and copies "copyLen" bytes from entry to dest */
COE_ABORTCODE essOD_DoEntryUpload(ESC_DATA* ed, ESS_OD_VIRTUAL_ENTRY* e, uint8_t* dest, uint32_t offset, uint16_t copyLen, ESS_BOOL isLastPart)
{
#if (CFG_ESS_COE_SUPPORT_COMPLETE_ACCESS)
    if (e->complAcc) {
        /* CHECKME: implement SDO Upload with CompleteAccess */
        ESS_DO_OD_COMPLACC_ENTRYUPLOAD();
        return COE_ABORTCODE_ACCESS;
    }
#endif

    /* Special object?: */
    if (e->obj->flags & ESS_OD_OBJECT_FLAGS_PDO) {
        const ESS_OD_OBJECT_INFOS_PDO* infos = (const ESS_OD_OBJECT_INFOS_PDO*)e->obj->infos;

        if (e->subIdx > infos->itemsUsed) {
            return COE_ABORTCODE_SUBINDEX;

        } else if (e->subIdx == 0) {
            (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
            *dest = infos->itemsUsed;

        } else {
            const ESS_PDO_ENTRY* p = (const ESS_PDO_ENTRY*)&infos->items[e->subIdx - 1];

            (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
            dest[0] = p->len;
            dest[1] = p->subIdx;
            dest[2] = LO_BYTE(p->objIdx);
            dest[3] = HI_BYTE(p->objIdx);
            if(ESS_IS_EXTEND_OBJECT(p)) {
                dest[1] = ESS_PDO_SUB_PADDING;  /* Return standard padding object */
            }
        }
        return COE_ABORTCODE_NONE;

    } else if (e->obj->flags & ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT) {
        const ESS_OD_OBJECT_INFOS_PDOASSIGN* infos = (const ESS_OD_OBJECT_INFOS_PDOASSIGN*)e->obj->infos;

        if (e->subIdx > infos->itemsUsed) {
            return COE_ABORTCODE_SUBINDEX;

        } else if (e->subIdx == 0) {
            (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
            *dest = infos->itemsUsed;

        } else {
            (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
            dest[0] = LO_BYTE(infos->items[e->subIdx - 1]);
            dest[1] = HI_BYTE(infos->items[e->subIdx - 1]);
        }
        return COE_ABORTCODE_NONE;

#if (CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS)
    } else if (e->obj->flags & ESS_OD_OBJECT_FLAGS_PDO_PARAMETER) {
        const ESS_OD_OBJECT_INFOS_PDOPARAM* infos = (const ESS_OD_OBJECT_INFOS_PDOPARAM*)(e->obj->infos);
        const ESS_BOOL isRxParam = ((e->obj->index < 0x1600) && (e->obj->index >= 0x1400));

        switch (e->subIdx) {
            case 0:
                (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
                *dest = infos->maxSubIdx;
                return COE_ABORTCODE_NONE;

            case 6:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_EXCLPDOS_EXISTS) {
                    if (offset == 0)
                        (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
                    ESS_MEMCOPY(dest, (const uint8_t*)infos->exclData + offset, copyLen);
                    return COE_ABORTCODE_NONE;
                }
                break;

            case 7:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_STATE_EXISTS) {
                    (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
                    *dest = (uint8_t)infos->state;
                    return COE_ABORTCODE_NONE;
                }
                break;

            case 8:
                if ((isRxParam) && (infos->flags & ESS_OD_PDOPARAM_FLAGS_CONTROL_EXISTS)) {
                    (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
                    *dest = (uint8_t)infos->control;
                    return COE_ABORTCODE_NONE;
                }
                break;

            case 9:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_TOGGLE_EXISTS) {
                    (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
                    *dest = (uint8_t)infos->toggle;
                    return COE_ABORTCODE_NONE;
                }
                break;
			default:
				break;
        }

        return COE_ABORTCODE_SUBINDEX;
#endif

    } else if (e->obj->flags & ESS_OD_OBJECT_FLAGS_ARRAY) {
        const ESS_OD_OBJECT_INFOS_ARRAY* infos = (const ESS_OD_OBJECT_INFOS_ARRAY*)e->obj->infos;

        if (e->subIdx > infos->entriesCount) {
            return COE_ABORTCODE_SUBINDEX;

        } else if (e->subIdx == 0) {
            (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
            *dest = infos->entriesCount;

        } else {
            if ((infos->entriesAcc & COE_ACCESS_R) == 0)
                return COE_ABORTCODE_WRITEONLY;
            if (!COE_ACCESS_R_ALLOWED(infos->entriesAcc, ed->currState)) {
                return COE_ABORTCODE_STORE_DEVICE_STATE;
            }

            if (offset == 0)
                (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);

            if (infos->data) {
                if((infos->entriesAcc & COE_ACCESS_SAFEVARIABLE) != 0) {
                    essOD_UploadSafetyData(dest, &infos->data[(e->subIdx - 1) * infos->entriesDataStride], offset, copyLen);
                } else {
                    ESS_MEMCOPY(dest, &infos->data[(e->subIdx - 1) * infos->entriesDataStride + offset], copyLen);
                }
            } else {
                return DoEntryRWCallback(ed, ESS_TRUE, dest, e->obj->index, e->subIdx, isLastPart, copyLen, offset);
            }
        }
        return COE_ABORTCODE_NONE;

    } else {
        /* Normal object: */
        /* Check if read access is allowed in current state: */
        if ((e->entry->access & COE_ACCESS_R) == 0)
            return COE_ABORTCODE_WRITEONLY;
        if (!COE_ACCESS_R_ALLOWED(e->entry->access, ed->currState)) {
            return COE_ABORTCODE_STORE_DEVICE_STATE;
        }

        if (offset == 0) {
            const COE_ABORTCODE ab = essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_UPLOAD);
            if (ab != COE_ABORTCODE_NONE)
                return ab;
        }

        /* Then get data: */
        if (e->entry->data) {
            if(e->entry->access & COE_ACCESS_SAFEVARIABLE) {
                essOD_UploadSafetyData(dest, &e->entry->data[0], offset, copyLen);
            } else {
                ESS_MEMCOPY(dest, &e->entry->data[offset], copyLen);
            }
            return COE_ABORTCODE_NONE;
        } else {
            return DoEntryRWCallback(ed, ESS_TRUE, dest, e->obj->index, e->entry->subIdx, isLastPart, copyLen, offset);
        }
    }
}

/* Checks access/state and copies "copyLen" bytes from src to entry */
COE_ABORTCODE essOD_DoEntryDownload(ESC_DATA* ed, ESS_OD_VIRTUAL_ENTRY* e, uint8_t* src, uint32_t totalDownloadLen, uint32_t offset, uint16_t copyLen, ESS_BOOL isLastPart)
{
#if (CFG_ESS_COE_SUPPORT_COMPLETE_ACCESS)
    if (e->complAcc) {
        /* CHECKME: implement SDO Download with CompleteAccess */
        return COE_ABORTCODE_ACCESS;
    }
#endif

    /* Special object?: */
    if (e->obj->flags & ESS_OD_OBJECT_FLAGS_PDO) {
        ESS_OD_OBJECT_INFOS_PDO* infos = (ESS_OD_OBJECT_INFOS_PDO*)e->obj->infos;

        if (0 == (infos->flags & ESS_PDO_CFG_FLAGS_WRITABLE_IN_PREOP))
            return COE_ABORTCODE_READONLY;
        else if (ed->currState != ESC_STATE_PREOP) {
            return COE_ABORTCODE_STORE_DEVICE_STATE;
        }

#if (CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT < 255)
        if (e->subIdx > CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT)
            return COE_ABORTCODE_SUBINDEX;
#endif

        if (e->subIdx == 0) {
            if (totalDownloadLen > 1)
                return COE_ABORTCODE_LEN_TOO_HIGH;

#if (CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT < 255)
            if (*src > CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT)
                return COE_ABORTCODE_DATA_TOO_HIGH;
#endif

            (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_DOWNLOAD);
            infos->itemsUsed = *src;
        } else {
            ESS_PDO_ENTRY* p = (ESS_PDO_ENTRY*)&infos->items[e->subIdx - 1];

            if (totalDownloadLen > 4)
                return COE_ABORTCODE_LEN_TOO_HIGH;
            else if (totalDownloadLen < 4)
                return COE_ABORTCODE_LEN_TOO_LOW;

            (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_DOWNLOAD);
            p->len = src[0];
            p->subIdx = src[1];
            p->objIdx = MAKE_WORD(src[2], src[3]);
        }
        return COE_ABORTCODE_NONE;

    } else if (e->obj->flags & ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT) {
        ESS_OD_OBJECT_INFOS_PDOASSIGN* infos = (ESS_OD_OBJECT_INFOS_PDOASSIGN*)e->obj->infos;

        if (infos->writableInPreOp == ESS_FALSE)
            return COE_ABORTCODE_READONLY;
        else if (ed->currState != ESC_STATE_PREOP)  {
            return COE_ABORTCODE_STORE_DEVICE_STATE;
        }

#if (CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT < 255)
        if (e->subIdx > CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT)
            return COE_ABORTCODE_SUBINDEX;
#endif

        if (e->subIdx == 0) {
            if (totalDownloadLen > 1)
                return COE_ABORTCODE_LEN_TOO_HIGH;

#if (CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT < 255)
            if (*src > CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT)
                return COE_ABORTCODE_DATA_TOO_HIGH;
#endif

            (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_DOWNLOAD);
            infos->itemsUsed = *src;
        } else {
            if (totalDownloadLen > 2)
                return COE_ABORTCODE_LEN_TOO_HIGH;
            else if (totalDownloadLen < 2)
                return COE_ABORTCODE_LEN_TOO_LOW;

            if (infos->itemsUsed != 0)
                return COE_ABORTCODE_SUBINDEX;

            (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_DOWNLOAD);
            infos->items[e->subIdx - 1] = MAKE_WORD(src[0], src[1]);
        }
        return COE_ABORTCODE_NONE;

#if (CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS)
    } else if (e->obj->flags & ESS_OD_OBJECT_FLAGS_PDO_PARAMETER) {
        ESS_OD_OBJECT_INFOS_PDOPARAM* infos = (ESS_OD_OBJECT_INFOS_PDOPARAM*)e->obj->infos;
        const ESS_BOOL isRxParam = ((e->obj->index < 0x1600) && (e->obj->index >= 0x1400));

        switch (e->subIdx) {
            case 0:
                return COE_ABORTCODE_READONLY;

            case 6:
                return (infos->flags & ESS_OD_PDOPARAM_FLAGS_EXCLPDOS_EXISTS) ? COE_ABORTCODE_READONLY : COE_ABORTCODE_SUBINDEX;

            case 7:
                return (infos->flags & ESS_OD_PDOPARAM_FLAGS_STATE_EXISTS) ? COE_ABORTCODE_READONLY : COE_ABORTCODE_SUBINDEX;

            case 8:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_CONTROL_EXISTS) {
                    if (totalDownloadLen > 1) return COE_ABORTCODE_LEN_TOO_HIGH; else if (totalDownloadLen < 1) return COE_ABORTCODE_LEN_TOO_LOW;
                    (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_DOWNLOAD);
                    infos->control = (*src != 0);
                    return COE_ABORTCODE_NONE;
                }
                return COE_ABORTCODE_SUBINDEX;

            case 9:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_TOGGLE_EXISTS) {
                    if (!isRxParam) return COE_ABORTCODE_READONLY;
                    if (totalDownloadLen > 1) return COE_ABORTCODE_LEN_TOO_HIGH; else if (totalDownloadLen < 1) return COE_ABORTCODE_LEN_TOO_LOW;
                    (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_DOWNLOAD);
                    infos->toggle = (*src != 0);
                    return COE_ABORTCODE_NONE;
                }
                return COE_ABORTCODE_SUBINDEX;

            default:
                return COE_ABORTCODE_SUBINDEX;
        }

#endif

    } else if (e->obj->flags & ESS_OD_OBJECT_FLAGS_ARRAY) {
        const ESS_OD_OBJECT_INFOS_ARRAY* infos = (const ESS_OD_OBJECT_INFOS_ARRAY*)e->obj->infos;

        if (e->subIdx > infos->entriesCount) {
            return COE_ABORTCODE_SUBINDEX;

        } else if (e->subIdx == 0) {
            return COE_ABORTCODE_READONLY; /* CHECKME: add support for dynamic arrays objects */

        } else {
            const uint32_t entryByteLen = (infos->entryBitLen + 7) >> 3;

            if ((infos->entriesAcc & COE_ACCESS_W) == 0)
                return COE_ABORTCODE_READONLY;
            if (!COE_ACCESS_W_ALLOWED(infos->entriesAcc, ed->currState)) {
                return COE_ABORTCODE_STORE_DEVICE_STATE;
            }

            if (totalDownloadLen > entryByteLen)
                return COE_ABORTCODE_LEN_TOO_HIGH;
            else if (totalDownloadLen < entryByteLen)
                return COE_ABORTCODE_LEN_TOO_LOW;

            if (offset == 0)
                (void)essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_DOWNLOAD);

            if (infos->data) {
                ESS_MEMCOPY(&infos->data[(e->subIdx - 1) * infos->entriesDataStride + offset], src, copyLen);
            } else {
                return DoEntryRWCallback(ed, ESS_FALSE, src, e->obj->index, e->subIdx, isLastPart, copyLen, offset);
            }
        }
        return COE_ABORTCODE_NONE;

    } else {
        /* Normal object: */
        uint32_t entrySize;

        /* Check if write access is allowed in current state: */
        if ((e->entry->access & COE_ACCESS_W) == 0)
            return COE_ABORTCODE_READONLY;
        if (!COE_ACCESS_W_ALLOWED(e->entry->access, ed->currState))  {
            return COE_ABORTCODE_STORE_DEVICE_STATE;
        }

        /* Check if len is OK: */
        entrySize = essOD_GetEntrySize(ed, e, ESS_FALSE);
        if (totalDownloadLen > entrySize)
            return COE_ABORTCODE_LEN_TOO_HIGH;

        if (totalDownloadLen < entrySize)
            if ((e->entry->flags & ESS_OD_ENTRY_FLAGS_ALLOWPARTIALDOWNLOAD) == 0)
                return COE_ABORTCODE_LEN_TOO_LOW;

        if (offset == 0) {
            const COE_ABORTCODE ab = essOD_DoEntryEventCallback(ed, e->obj->index, e->subIdx, ESS_OD_ENTRY_CALLBACK_STARTING_DOWNLOAD);
            if (ab != COE_ABORTCODE_NONE)
                return ab;
        }

        /* Then get data: */
        if (e->entry->data) {
            ESS_MEMCOPY(&e->entry->data[offset], src, copyLen);

            if (isLastPart) {
                uint32_t i;
                /* Fill bytes that weren't downloaded with 0 (smaller downloads only for string type entries - checked above) */
                for (i = copyLen + offset; i < ((e->entry->bitLen + 7) >> 3); ++i)
                    e->entry->data[i] = 0;
            }

            return COE_ABORTCODE_NONE;

        } else {
            return DoEntryRWCallback(ed, ESS_FALSE, src, e->obj->index, e->entry->subIdx, isLastPart, copyLen, offset);
        }
    }
}

/* Looks for object (0x1c10 + sm) and returns its infos (which contains an array of uint16_t - the PDOs assigned to this SM) */
uint16_t essOD_GetPDOAssignInfos(ESC_DATA* ed, ESS_SM sm, const ESS_OD_OBJECT_INFOS_PDOASSIGN** infos)
{
    const ESS_OD_OBJECT* obj = FindObj(&ed->dict, 0x1c10 + sm);

    if (obj == NULL)
        return REG_VAL_ALSTATUSCODE_ESD_MAP_ASSIGNMISSING;

    if ((obj->flags & ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT) == 0)
        return REG_VAL_ALSTATUSCODE_ESD_MAP_ASSIGNOBJERR;

    *infos = (const ESS_OD_OBJECT_INFOS_PDOASSIGN*)obj->infos;

    return REG_VAL_ALSTATUSCODE_NOERROR;
}

/* Looks for the object "pdo" and returns its infos */
uint16_t essOD_GetPDOInfos(ESC_DATA* ed, uint16_t pdo, const ESS_OD_OBJECT_INFOS_PDO** infos)
{
    const ESS_OD_OBJECT* obj = FindObj(&ed->dict, pdo);

    if (obj == NULL)
        return REG_VAL_ALSTATUSCODE_ESD_MAP_PDOMISSING;

    if ((obj->flags & ESS_OD_OBJECT_FLAGS_PDO) == 0)
        return REG_VAL_ALSTATUSCODE_ESD_MAP_PDOOBJERR;

    *infos = (const ESS_OD_OBJECT_INFOS_PDO*)obj->infos;

    return REG_VAL_ALSTATUSCODE_NOERROR;
}

/* When an entry is to be mapped, this is used to get the pointer. So no virtual items with callback, etc. possible here! */
uint16_t essOD_GetEntryData(const ESC_DATA* ed, uint16_t objIdx, uint8_t subIdx, uint8_t** data)
{
    ESS_OD_ENTRY* entry;
    ESS_OD_OBJECT* obj;

    obj = FindObj(&ed->dict, objIdx);
    if (obj == NULL)
        return REG_VAL_ALSTATUSCODE_ESD_MAP_MAPPEDOBJMISSING;

    /* Special object?: */
    if (obj->flags & (ESS_OD_OBJECT_FLAGS_PDO | ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT)) {
        return REG_VAL_ALSTATUSCODE_ESD_MAP_MAPPEDOBJERR;

#if (CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS)
    } else if (obj->flags & ESS_OD_OBJECT_FLAGS_PDO_PARAMETER) {
        ESS_OD_OBJECT_INFOS_PDOPARAM* infos = (ESS_OD_OBJECT_INFOS_PDOPARAM*)obj->infos;

        switch (subIdx) {
            case 7:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_STATE_EXISTS) {
                    *data = (uint8_t*)&infos->state;
                    return REG_VAL_ALSTATUSCODE_NOERROR;
                }
                break;
            case 8:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_CONTROL_EXISTS) {
                    *data = (uint8_t*)&infos->control;
                    return REG_VAL_ALSTATUSCODE_NOERROR;
                }
                break;
            case 9:
                if (infos->flags & ESS_OD_PDOPARAM_FLAGS_TOGGLE_EXISTS) {
                    *data = (uint8_t*)&infos->toggle;
                    return REG_VAL_ALSTATUSCODE_NOERROR;
                }
                break;
			default:
				break;
        }
        return REG_VAL_ALSTATUSCODE_ESD_MAP_MAPPEDENTRYERR;
#endif

    } else if (obj->flags & ESS_OD_OBJECT_FLAGS_ARRAY) {
        const ESS_OD_OBJECT_INFOS_ARRAY* infos = (const ESS_OD_OBJECT_INFOS_ARRAY*)obj->infos;

        if (infos->data == NULL)
            return REG_VAL_ALSTATUSCODE_ESD_MAP_MAPPEDOBJERR;

        if (subIdx > infos->entriesCount)
            return REG_VAL_ALSTATUSCODE_ESD_MAP_MAPPEDENTRYMISSING;

        *data = &infos->data[(subIdx - 1) * infos->entriesDataStride];
        return REG_VAL_ALSTATUSCODE_NOERROR;
    }

    entry = FindEntry(obj, subIdx);
    if (entry == NULL)
        return REG_VAL_ALSTATUSCODE_ESD_MAP_MAPPEDENTRYMISSING;

#if (!CFG_ESS_COE_ALLOWTOMAPNULLENTRIES)
    if (entry->data == NULL)
        return REG_VAL_ALSTATUSCODE_ESD_MAP_MAPPEDENTRYERR;
#endif

    *data = entry->data;

    return REG_VAL_ALSTATUSCODE_NOERROR;
}

#endif /* CFG_ESS_SUPPORT_COE */



ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODCreate(ESS_HANDLE h, ESS_OD_FLAGS flags)
{
#if (CFG_ESS_SUPPORT_COE)
    ESC_DATA* ed = (ESC_DATA*)h;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->currState != ESC_STATE_INIT)
        return ESS_RESULT_WRONG_ECAT_STATE;

#if (!CFG_ESS_OD_ALLOW_HANDLE_SM_TYPE)
    if (flags & ESS_OD_FLAGS_HANDLE_SM_TYPES) {
        ESS_DBG(("ERROR: ESS_OD_FLAGS_HANDLE_SM_TYPES given, but CFG_ESS_OD_ALLOW_HANDLE_SM_TYPE not defined!"));
        return ESS_RESULT_INVALID_ARG;
    }
#endif

    if (ed->dict.objects)
        return ESS_RESULT_ALREADY_ACTIVE;

    /* Make sure obj. capacity is > obj count */
    if (VerifyDictCapacity(&ed->dict))
        return ESS_RESULT_NO_MEM;

    /* Handle flags: */
    ed->dict.flags = flags;

#if (CFG_ESS_OD_ALLOW_HANDLE_SM_TYPE)
    if (flags & ESS_OD_FLAGS_HANDLE_SM_TYPES) {
        /* Create object 0x1c00 => type of every SM */
        ESS_RESULT res;
        int cnt = 0;
        int i;

        for (i = 0; ((i < ed->config->smConfigCount) && (i < CFG_ESS_MAX_SM_COUNT)); ++i) {
            const uint8_t smType = ed->config->smConfigs[i].contrByte & REG_MASK_SMCONTROL_TYPE;

            ed->smObj0x1c00Data[i] = (smType == SM_TYPE_MBXOUT) ? 1 : (smType == SM_TYPE_MBXIN) ? 2 : (smType == SM_TYPE_OUTPUTS) ? 3 : (smType == SM_TYPE_INPUTS) ? 4 : 0xff;
            cnt++;
        }

        res = essODAddArrayObject(h, 0x1c00, "Sync Manager Communication Type",
				  (uint8_t)cnt, COE_DATATYPE_USINT, 8, COE_ACCESS_R, &ed->smObj0x1c00Data[0], 1);
        if (res != ESS_RESULT_SUCCESS) {
            ESS_DBG(("ERROR: essODAddArrayObject() for \"SM Types\" failed with %s!", essFormatResult(res)));
            return res;
        }
    }
#endif

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)flags;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODDelete(ESS_HANDLE h, uint16_t reserved)
{
#if (CFG_ESS_SUPPORT_COE)
    ESC_DATA* const ed = (ESC_DATA*)h;
    uint16_t i;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (reserved)
        return ESS_RESULT_INVALID_ARG;

    if (ed->currState != ESC_STATE_INIT)
        return ESS_RESULT_WRONG_ECAT_STATE;

    if (ed->dict.objects == NULL)
        return ESS_RESULT_SUCCESS;

    for (i = 0; i < ed->dict.objectsCount; ++i) {
        ESS_OD_OBJECT* const obj = ed->dict.objects[i];
        uint16_t j;

        if (obj->flags & ESS_OD_OBJECT_FLAGS_ALLOCEDINFOS)
            ESS_FREE(obj->infos);

        for (j = 0; j < obj->entriesCount; ++j) {
            ESS_OD_ENTRY* const entry = obj->entries[j];

            ESS_FREE(entry);
        }
        ESS_FREE(obj->entries);

        ESS_FREE(obj);
    }

    ESS_FREE(ed->dict.objects);

    ESS_MEMZERO(&ed->dict, sizeof(ed->dict));

    return ESS_RESULT_SUCCESS;

#else
    (void)h;
    (void)reserved;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODObjectAdd(ESS_HANDLE h, uint16_t idx, ESS_OD_OBJECT_FLAGS flags, const ESS_OD_OBJECT_INFOS* infos)
{
#if (CFG_ESS_SUPPORT_COE)
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->dict.objects == NULL)
        return ESS_RESULT_NO_DICT;

    if (FindObj(&ed->dict, idx) != NULL)
        return ESS_RESULT_OBJ_EXISTS;

    if ((ed->currState != ESC_STATE_INIT) && (ed->currState != ESC_STATE_PREOP))
        return ESS_RESULT_WRONG_ECAT_STATE;

    /* Make sure obj. capacity is > obj count */
    if (VerifyDictCapacity(&ed->dict))
        return ESS_RESULT_NO_MEM;

    /* Create object and set its members: */
    ESS_ALLOC(obj, ESS_OD_OBJECT*, sizeof(*obj));
    if (obj == NULL) {
        ESS_DBG(("ERROR: alloc for %lu byte failed!", (unsigned long)sizeof(*obj)));
        return ESS_RESULT_NO_MEM;
    }

    obj->entries = NULL;
    obj->entriesCapacity = 0;
    obj->entriesCount = 0;
    obj->flags = flags;
    obj->index = idx;
    obj->infos = NULL;

    /* Add object to objects list: */
    InsertObject(&ed->dict, obj);

    /* Handle flags and special objects: */
    {
#       define CHECK_AND_ALLOC_INFOS(infosType)                                               \
			do {                                                                              \
			  if (infos) {                                                                    \
			    ESS_DBG(("ERROR: Obj with special flag AND infos given?!"));                  \
				return ESS_RESULT_INVALID_ARG;                                                \
			  }                                                                               \
			  ESS_ALLOC(info, infosType *, sizeof(*info));                                    \
			  if (info == NULL) {                                                             \
			    ESS_DBG(("ERROR: alloc for %lu byte failed!", (unsigned long)sizeof(*info))); \
				(void)essODObjectDelete(h, idx);                                              \
				return ESS_RESULT_NO_MEM;                                                     \
			  }                                                                               \
			} while (0)

        if (flags & ESS_OD_OBJECT_FLAGS_PDO) {
            ESS_OD_OBJECT_INFOS_PDO* info;
            CHECK_AND_ALLOC_INFOS(ESS_OD_OBJECT_INFOS_PDO);
            obj->infos = info;

            info->itemsUsed = 0;
            info->objName = "";

        } else if (flags & ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT) {
            ESS_OD_OBJECT_INFOS_PDOASSIGN* info;
            CHECK_AND_ALLOC_INFOS(ESS_OD_OBJECT_INFOS_PDOASSIGN);
            obj->infos = info;

            info->itemsUsed = 0;
            info->objName = "";

#if (CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS)
        } else if (flags & ESS_OD_OBJECT_FLAGS_PDO_PARAMETER) {
            ESS_OD_OBJECT_INFOS_PDOPARAM* info;
            CHECK_AND_ALLOC_INFOS(ESS_OD_OBJECT_INFOS_PDOPARAM);
            obj->infos        = info;
            info->exclData    = NULL;
            info->flags       = 0;
            info->exclDataLen = 0;
            info->maxSubIdx   = 0;
            info->state       = ESS_TRUE;
            info->control     = ESS_TRUE;
            info->toggle      = ESS_FALSE;
#endif

        } else if (flags & ESS_OD_OBJECT_FLAGS_ARRAY) {
            ESS_OD_OBJECT_INFOS_ARRAY* info;
            CHECK_AND_ALLOC_INFOS(ESS_OD_OBJECT_INFOS_ARRAY);
            obj->infos = info;
            /* Don't need to initialize info */

        } else {
            obj->infos = (void*)infos;
        }

#       undef CHECK_AND_ALLOC_INFOS
    }

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)idx;
    (void)flags;
    (void)infos;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODObjectDelete(ESS_HANDLE h, uint16_t idx)
{
#if (CFG_ESS_SUPPORT_COE)
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;
    uint16_t i;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->dict.objects == NULL)
        return ESS_RESULT_NO_DICT;

    if ((ed->currState != ESC_STATE_INIT) && (ed->currState != ESC_STATE_PREOP))
        return ESS_RESULT_WRONG_ECAT_STATE;

    obj = FindObj(&ed->dict, idx);
    if (obj == NULL)
        return ESS_RESULT_OBJ_NOT_FOUND;

    /* Remove item from the list: */
    for (i = 0; i < ed->dict.objectsCount; ++i) {
        if (obj == ed->dict.objects[i])
            break;
    }
    ed->dict.objectsCount--;
    for (; i < ed->dict.objectsCount; ++i)
        ed->dict.objects[i] = ed->dict.objects[i + 1];

    /* Then actually free it: */
    if (obj->flags & ESS_OD_OBJECT_FLAGS_ALLOCEDINFOS)
        ESS_FREE(obj->infos);
    for (i = 0; i < obj->entriesCount; ++i)
        ESS_FREE(obj->entries[i]);
    ESS_FREE(obj->entries);
    ESS_FREE(obj);

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)idx;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODEntryAdd(ESS_HANDLE h, uint16_t idx, uint8_t subIdx, uint32_t bitLen, volatile void* data, COE_ACCESS acc, ESS_OD_ENTRY_FLAGS flags, const ESS_OD_ENTRY_INFOS* infos)
{
#if (CFG_ESS_SUPPORT_COE)
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_ENTRY* entry;
    ESS_OD_OBJECT* obj;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->dict.objects == NULL)
        return ESS_RESULT_NO_DICT;

    if ((ed->currState != ESC_STATE_INIT) && (ed->currState != ESC_STATE_PREOP))
        return ESS_RESULT_WRONG_ECAT_STATE;

    obj = FindObj(&ed->dict, idx);
    if (obj == NULL)
        return ESS_RESULT_OBJ_NOT_FOUND;

    if (obj->flags & ESS_OD_OBJECT_FLAGS_PRIVATE_MASK) {
        ESS_DBG(("ERROR: Can't add entry to obj 0x%.4x because it's a \"special\" object!", idx));
        return ESS_RESULT_INVALID_OBJECT;
    }

    if (flags & ESS_OD_ENTRY_FLAGS_UPLOAD_TRIMMED) {
        if (data == NULL)
            return ESS_RESULT_INVALID_ARG;

        if ((infos) && ((infos->defaultValue) || (infos->minValue) || (infos->maxValue)))
            return ESS_RESULT_INVALID_ARG; /* If item shall be trimmed it can't contain min/max/def cause their sizes would have to be equal cause protocol does not contain separate lengths for these. */
    }

    /* If entry is writable and of string type, automatically set ALLOWPARTIALDOWNLOAD flag: */
    if ((acc & COE_ACCESS_W) && (infos) && ((infos->dataType == COE_DATATYPE_STRING) || (infos->dataType == COE_DATATYPE_WSTRING) || (infos->dataType == COE_DATATYPE_OCTETSTRING)))
        flags |= ESS_OD_ENTRY_FLAGS_ALLOWPARTIALDOWNLOAD;

    entry = FindEntry(obj, subIdx);
    if (entry != NULL)
        return ESS_RESULT_ENTRY_EXISTS;

    if (VerifyObjectCapacity(obj))
        return ESS_RESULT_NO_MEM;

    /* Create entry and set its members: */
    ESS_ALLOC(entry, ESS_OD_ENTRY*, sizeof(*entry));
    if (entry == NULL) {
        ESS_DBG(("ERROR: alloc for %lu byte failed!", (unsigned long)sizeof(*entry)));
        return ESS_RESULT_NO_MEM;
    }

    entry->access = acc;
    entry->bitLen = bitLen;
    entry->data = (uint8_t*)data;
    entry->flags = flags;
    entry->infos = (void*)infos;
    entry->subIdx = subIdx;

    InsertEntry(obj, entry);

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)idx;
    (void)subIdx;
    (void)bitLen;
    (void)data;
    (void)acc;
    (void)flags;
    (void)infos;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODEntryDelete(ESS_HANDLE h, uint16_t idx, uint8_t subIdx)
{
#if (CFG_ESS_SUPPORT_COE)
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_ENTRY* entry;
    ESS_OD_OBJECT* obj;
    uint16_t i;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->dict.objects == NULL)
        return ESS_RESULT_NO_DICT;

    if ((ed->currState != ESC_STATE_INIT) && (ed->currState != ESC_STATE_PREOP))
        return ESS_RESULT_WRONG_ECAT_STATE;

    obj = FindObj(&ed->dict, idx);
    if (obj == NULL)
        return ESS_RESULT_OBJ_NOT_FOUND;

    if (obj->flags) { /* this must be changed when public object flags exist */
        ESS_DBG(("ERROR: Can't remove entry from obj 0x%.4x because it's a \"special\" object!", idx));
        return ESS_RESULT_INVALID_OBJECT;
    }

    entry = FindEntry(obj, subIdx);
    if (entry == NULL)
        return ESS_RESULT_ENTRY_NOT_FOUND;

    /* Remove item from the list: */
    for (i = 0; i < obj->entriesCount; ++i) {
        if (entry == obj->entries[i])
            break;
    }
    obj->entriesCount--;
    for (; i < obj->entriesCount; ++i)
        obj->entries[i] = obj->entries[i + 1];

    /* Then actually free it: */
    ESS_FREE(entry);
    /* When here is more to do, don't forget to update the entry cleanup in essODObjectDelete() */

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)idx;
    (void)subIdx;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODAddArrayObject(ESS_HANDLE h, uint16_t idx, const char* name,
                                                       uint8_t entriesCount, COE_DATATYPE entriesDataType,
                                                       uint32_t entryBitLen, COE_ACCESS entriesAcc,
                                                       volatile void* entriesData, uint16_t entriesDataStride)
{
#if (CFG_ESS_SUPPORT_COE)
    ESS_OD_OBJECT_INFOS_ARRAY* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;
    ESS_RESULT res;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (FindObj(&ed->dict, idx) != NULL)
        return ESS_RESULT_OBJ_EXISTS;

    if (entriesCount == 0)
        return ESS_RESULT_INVALID_ARG;

    res = essODObjectAdd(h, idx, ESS_OD_OBJECT_FLAGS_ARRAY, NULL);
    if (res != ESS_RESULT_SUCCESS)
        return res;

    obj = FindObj(&ed->dict, idx);
	if (NULL == obj)		/* Make static code analysis happy */
		return(ESS_RESULT_OBJ_NOT_FOUND);
	else
		infos = (ESS_OD_OBJECT_INFOS_ARRAY*)obj->infos;

    infos->data = (uint8_t*)entriesData;
    infos->objName = name;
    infos->entryBitLen = entryBitLen;
    infos->entriesDataType = entriesDataType;
    infos->entriesDataStride = entriesDataStride;
    infos->entriesAcc = entriesAcc;
    infos->entriesCount = entriesCount;
    infos->maxSubIdx = entriesCount;

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)idx;
    (void)name;
    (void)entriesCount;
    (void)entriesDataType;
    (void)entryBitLen;
    (void)entriesAcc;
    (void)entriesData;
    (void)entriesDataStride;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODSetArrayMax(ESS_HANDLE h, uint16_t idx, uint8_t entriesMax)
{
#if (CFG_ESS_SUPPORT_COE)
    ESS_OD_OBJECT_INFOS_ARRAY* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if(0 == (ed->dict.flags & ESS_OD_SDO_INFO_MODE_ETG1020))
        return(ESS_RESULT_INVALID_OD_FLAGS);

    obj = FindObj(&ed->dict, idx);
    if (NULL == obj)
        return(ESS_RESULT_OBJ_NOT_FOUND);

    if (0 == (obj->flags & ESS_OD_OBJECT_FLAGS_ARRAY))
        return(ESS_RESULT_INVALID_OBJECT);

    infos = (ESS_OD_OBJECT_INFOS_ARRAY*)obj->infos;

    if(infos->entriesCount > entriesMax)
        return(ESS_RESULT_INVALID_ARG);

    infos->maxSubIdx = entriesMax;
    return ESS_RESULT_SUCCESS;

#else
    (void)h;
    (void)idx;
    (void)entriesMax;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

/*lint -e838 (Info: Previously assigned value to variable 'res' has not been used) */
ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODAddGenericObjects(ESS_HANDLE h, const uint32_t* deviceType, const uint32_t* vendorId, const uint32_t* productCode, const uint32_t* revNo, const uint32_t* serialNo, const char* devName, const char* verHardware, const char* verSoftware)
{
#if (CFG_ESS_SUPPORT_COE)
#   define CHECK_RES() do { if (res != ESS_RESULT_SUCCESS) { ESS_DBG(("ERROR: essODAddGenericObjects() failed near line %u", (unsigned)__LINE__)); return res; } } while (0)
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_RESULT res = ESS_RESULT_SUCCESS;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if ((ed->currState != ESC_STATE_INIT) && (ed->currState != ESC_STATE_PREOP))
        return ESS_RESULT_WRONG_ECAT_STATE;

    if ((deviceType == NULL) || (vendorId == NULL) || (productCode == NULL) || (revNo == NULL) || (serialNo == NULL))
        return ESS_RESULT_INVALID_ARG;

#if defined(_DEBUG) && (defined(ESS_PLATFORM_LINUX) || defined(ESS_PLATFORM_WINDOWS))
    {
        const uint8_t* p[9];
        int i;

        p[0] = (const uint8_t*)&ed; p[1] = (const uint8_t*)deviceType; p[2] = (const uint8_t*)vendorId; p[3] = (const uint8_t*)productCode; p[4] = (const uint8_t*)revNo; p[5] = (const uint8_t*)serialNo; p[6] = (const uint8_t*)devName; p[7] = (const uint8_t*)verHardware; p[8] = (const uint8_t*)verSoftware;

        for (i = 1; i < (int)(sizeof(p) / sizeof(p[0])); ++i) {
            const uint64_t dist = (unsigned)((p[0] > p[i]) ? p[0] - p[i] : p[i] - p[0]);

            if (dist < 1024) {
                /* False positives are also possible... This is just a hint to have a look at the arguments! */
                ESS_DBG(("Warning: Please check arguments - looks like some data is one the stack!"));
                break;
            }
        }
    }
#endif

    /* Object 0x1000: */
    {
        DECLARE_OBJECT_INFOS(objInfo, "Device Type", COE_DATATYPE_UDINT, COE_CODE_VARIABLE);
        DECLARE_ENTRY_INFOS(entryInfo0, "Device Type", COE_DATATYPE_UDINT);

        res = essODObjectAdd((ESS_HANDLE)ed, 0x1000, ESS_OD_OBJECT_FLAGS_NONE, objInfo);
        CHECK_RES();
        res = essODEntryAdd((ESS_HANDLE)ed, 0x1000, 0, 32, (void*)deviceType, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, entryInfo0);
        CHECK_RES();
    }

    /* Object 0x1018: */
    {
        DECLARE_OBJECT_INFOS(objInfo, "Identity", COE_DATATYPE_IDENTITY, COE_CODE_RECORD);
        DECLARE_ENTRY_INFOS(entryInfo0, ESS_SUB0_NAME, COE_DATATYPE_USINT);
        DECLARE_ENTRY_INFOS(entryInfo1, "Vendor Id", COE_DATATYPE_UDINT);
        DECLARE_ENTRY_INFOS(entryInfo2, "Product Code", COE_DATATYPE_UDINT);
        DECLARE_ENTRY_INFOS(entryInfo3, "Revision Number", COE_DATATYPE_UDINT);
        DECLARE_ENTRY_INFOS(entryInfo4, "Serial Number", COE_DATATYPE_UDINT);

        res = essODObjectAdd((ESS_HANDLE)ed, 0x1018, ESS_OD_OBJECT_FLAGS_NONE, objInfo);
        CHECK_RES();
        res = essODEntryAdd((ESS_HANDLE)ed, 0x1018, 0, 8, (void*)"\x04", COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, entryInfo0);
        CHECK_RES();
        res = essODEntryAdd((ESS_HANDLE)ed, 0x1018, 1, 32, (void*)vendorId, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, entryInfo1);
        CHECK_RES();
        res = essODEntryAdd((ESS_HANDLE)ed, 0x1018, 2, 32, (void*)productCode, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, entryInfo2);
        CHECK_RES();
        res = essODEntryAdd((ESS_HANDLE)ed, 0x1018, 3, 32, (void*)revNo, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, entryInfo3);
        CHECK_RES();
        res = essODEntryAdd((ESS_HANDLE)ed, 0x1018, 4, 32, (void*)serialNo, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, entryInfo4);
        CHECK_RES();
    }

    /* Object 0x1008: */
    if (devName) {
        DECLARE_OBJECT_INFOS(objInfo, "Device Name", COE_DATATYPE_STRING, COE_CODE_VARIABLE);
        DECLARE_ENTRY_INFOS(entryInfo0, "Device Name", COE_DATATYPE_STRING);

        res = essODObjectAdd((ESS_HANDLE)ed, 0x1008, ESS_OD_OBJECT_FLAGS_NONE, objInfo);
        CHECK_RES();
        res = essODEntryAdd((ESS_HANDLE)ed, 0x1008, 0, StrLen((const uint8_t*)devName) * 8, (void*)devName, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, entryInfo0);
        CHECK_RES();
    }

    /* Object 0x1009: */
    if (verHardware) {
        DECLARE_OBJECT_INFOS(objInfo, "Hardware Version", COE_DATATYPE_STRING, COE_CODE_VARIABLE);
        DECLARE_ENTRY_INFOS(entryInfo0, "Hardware Version", COE_DATATYPE_STRING);

        res = essODObjectAdd((ESS_HANDLE)ed, 0x1009, ESS_OD_OBJECT_FLAGS_NONE, objInfo);
        CHECK_RES();
        res = essODEntryAdd((ESS_HANDLE)ed, 0x1009, 0, StrLen((const uint8_t*)verHardware) * 8, (void*)verHardware, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, entryInfo0);
        CHECK_RES();
    }

    /* Object 0x100a: */
    if (verSoftware) {
        DECLARE_OBJECT_INFOS(objInfo, "Software Version", COE_DATATYPE_STRING, COE_CODE_VARIABLE);
        DECLARE_ENTRY_INFOS(entryInfo0, "Software Version", COE_DATATYPE_STRING);

        res = essODObjectAdd((ESS_HANDLE)ed, 0x100a, ESS_OD_OBJECT_FLAGS_NONE, objInfo);
        CHECK_RES();
        res = essODEntryAdd((ESS_HANDLE)ed, 0x100a, 0, StrLen((const uint8_t*)verSoftware) * 8, (void*)verSoftware, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, entryInfo0);
        CHECK_RES();
    }

    return res;
#   undef CHECK_RES

#else
    (void)h;
    (void)deviceType;
    (void)vendorId;
    (void)productCode;
    (void)revNo;
    (void)serialNo;
    (void)devName;
    (void)verHardware;
    (void)verSoftware;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODUpdatePDOConfiguration(ESS_HANDLE h, uint16_t pdo, const char* name, const ESS_PDO_ENTRY* entries, uint8_t entryCount, ESS_PDO_CFG_FLAGS flags)
{
#if (CFG_ESS_SUPPORT_COE)
    ESS_OD_OBJECT_INFOS_PDO* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;
    ESS_RESULT res;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    /* Dynamic PDO configuration support and block copy exclude each other */
    if((ESS_PDO_CFG_FLAGS_WRITABLE_IN_PREOP | ESS_PDO_CFG_FLAGS_BLOCK_COPY) == (flags & (ESS_PDO_CFG_FLAGS_WRITABLE_IN_PREOP | ESS_PDO_CFG_FLAGS_BLOCK_COPY)))
        return ESS_RESULT_INVALID_ARG;

# if (CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT < 255)
    if (entryCount > CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT)
        return ESS_RESULT_INVALID_ARG;
# endif

    obj = FindObj(&ed->dict, pdo);

    if (entries == NULL) { /* pdo object shall be deleted */
        if (obj == NULL)
            return ESS_RESULT_OBJ_NOT_FOUND;

        return essODObjectDelete(h, pdo);
    }

    if (obj == NULL) {
        res = essODObjectAdd(h, pdo, ESS_OD_OBJECT_FLAGS_PDO, NULL);
        if (res != ESS_RESULT_SUCCESS)
            return res;

        obj = FindObj(&ed->dict, pdo);
		if (NULL == obj)		/* Make static code analysis happy */
			return(ESS_RESULT_OBJ_NOT_FOUND);	
    }

    if ((obj->flags & ESS_OD_OBJECT_FLAGS_PDO) == 0) {
        ESS_DBG(("ERROR: Object 0x%.4x was not created by essODUpdatePDOConfiguration()!", pdo));
        return ESS_RESULT_INVALID_OBJECT;
    }

    infos = (ESS_OD_OBJECT_INFOS_PDO*)obj->infos;
    infos->itemsUsed = entryCount;
    while (entryCount--)
        infos->items[entryCount] = entries[entryCount];
    infos->objName = (name) ? name : "";
    infos->flags = flags;

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)pdo;
    (void)name;
    (void)entries;
    (void)entryCount;
    (void)writable;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODGetPDOConfiguration(ESS_HANDLE h, uint16_t pdo, uint8_t* numEntries, ESS_PDO_ENTRY* dest, uint8_t reserved)
{
#if (CFG_ESS_SUPPORT_COE)
    const ESS_OD_OBJECT_INFOS_PDO* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;

    (void)reserved; /* TODO: essODGetPDOConfiguration() shall use its reserved param as offset. (To acquire the configuration with multiple calls but less memory) Don't forget to update sample then too! */

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    obj = FindObj(&ed->dict, pdo);
    if (obj == NULL)
        return ESS_RESULT_OBJ_NOT_FOUND;

    infos = (const ESS_OD_OBJECT_INFOS_PDO*)obj->infos;

    if (dest) {
        uint8_t maxItems = MIN(*numEntries, infos->itemsUsed);
        int i;

        *numEntries = maxItems;
        for (i = 0; i < maxItems; ++i)
            dest[i] = infos->items[i];

    } else {
        *numEntries = infos->itemsUsed;
    }

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)pdo;
    (void)numEntries;
    (void)dest;
    (void)reserved;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODUpdatePDOAssignment(ESS_HANDLE h, ESS_SM smIdx, const uint16_t* pdo, uint8_t pdoCount, ESS_BOOL writable)
{
#if (CFG_ESS_SUPPORT_COE)
    const uint16_t objIdx = 0x1c10 + smIdx;
    ESS_OD_OBJECT_INFOS_PDOASSIGN* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;
    ESS_RESULT res;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (smIdx >= CFG_ESS_MAX_SM_COUNT)
        return ESS_RESULT_INVALID_SM;

#if (CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT < 255)
    if (pdoCount > CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT)
        return ESS_RESULT_INVALID_ARG;
#endif

    obj = FindObj(&ed->dict, objIdx);
    if (obj == NULL) {
        res = essODObjectAdd(h, objIdx, ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT, NULL);
        if (res != ESS_RESULT_SUCCESS)
            return res;
    } else {
        if ((obj->flags & ESS_OD_OBJECT_FLAGS_PDO_ASSIGNMENT) == 0) {
            ESS_DBG(("ERROR: Object 0x%.4x was not created by essODUpdatePDOAssignment()!", objIdx));
            return ESS_RESULT_INVALID_OBJECT;
        }
    }
    obj = FindObj(&ed->dict, objIdx);
	if (NULL == obj)		/* Make static code analysis happy */
		return(ESS_RESULT_OBJ_NOT_FOUND);

    infos = (ESS_OD_OBJECT_INFOS_PDOASSIGN*)obj->infos;

    infos->itemsUsed = pdoCount;
    while (pdoCount--)
        infos->items[pdoCount] = pdo[pdoCount];
    infos->objName = "PDO Assignment";
    infos->writableInPreOp = writable;

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)smIdx;
    (void)pdo;
    (void)pdoCount;
    (void)writable;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODGetPDOAssignment(ESS_HANDLE h, ESS_SM smIdx, uint8_t* numPDOs, uint16_t* dest)
{
#if (CFG_ESS_SUPPORT_COE)
    const uint16_t objIdx = 0x1c10 + smIdx;
    const ESS_OD_OBJECT_INFOS_PDOASSIGN* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    obj = FindObj(&ed->dict, objIdx);
    if (obj == NULL)
        return ESS_RESULT_OBJ_NOT_FOUND;

    infos = (const ESS_OD_OBJECT_INFOS_PDOASSIGN*)obj->infos;

    if (dest) {
        uint8_t maxItems = MIN(infos->itemsUsed, *numPDOs);
        int i;

        *numPDOs = maxItems;
        for (i = 0; i < maxItems; ++i)
            dest[i] = infos->items[i];

    } else {
        *numPDOs = infos->itemsUsed;
    }

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)smIdx;
    (void)numPDOs;
    (void)dest;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}


#if ((CFG_ESS_SUPPORT_COE) && (CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS))

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODPDOParamCreate(ESS_HANDLE h, uint16_t pdo, const char* name, ESS_OD_PDOPARAM_FLAGS flags)
{
    const uint16_t objIdx = pdo - 0x200;
    ESS_OD_OBJECT_INFOS_PDOPARAM* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;
    ESS_RESULT res;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    obj = FindObj(&ed->dict, objIdx);
    if (obj) {
        ESS_DBG(("ERROR: PDO Parameter object 0x%.4x already existed!", objIdx));
        return ESS_RESULT_OBJ_EXISTS;
    }

    if (flags == ESS_OD_PDOPARAM_FLAGS_NONE)
        return ESS_RESULT_INVALID_ARG;

    res = essODObjectAdd(h, objIdx, ESS_OD_OBJECT_FLAGS_PDO_PARAMETER, NULL);
    if (res != ESS_RESULT_SUCCESS)
        return res;

    obj = FindObj(&ed->dict, objIdx);
	if (NULL == obj)		/* Make static code analysis happy */
		return(ESS_RESULT_OBJ_NOT_FOUND);

    infos = (ESS_OD_OBJECT_INFOS_PDOPARAM*)obj->infos;

    infos->flags = flags;
    infos->name = name;
    for(infos->maxSubIdx = 5; flags; flags >>= 1)
        infos->maxSubIdx++; /* Calculate max index. Destroys 'flags' !! */

    return ESS_RESULT_SUCCESS;
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODPDOParamUpdateExclude(ESS_HANDLE h, uint16_t pdo, const void* data, uint16_t dataLen)
{
    const uint16_t objIdx = pdo - 0x200;
    ESS_OD_OBJECT_INFOS_PDOPARAM* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if ((dataLen % 2) || ((data == NULL) && (dataLen > 0)))
        return ESS_RESULT_INVALID_ARG;

    obj = FindObj(&ed->dict, objIdx);
    if (obj == NULL)
        return ESS_RESULT_OBJ_NOT_FOUND;

    infos = (ESS_OD_OBJECT_INFOS_PDOPARAM*)obj->infos;
    infos->exclData = data;
    infos->exclDataLen = dataLen;

#ifdef _DEBUG
    if ((infos->flags & ESS_OD_PDOPARAM_FLAGS_EXCLPDOS_EXISTS) == 0) {
        ESS_DBG(("WARNING: essODPDOParamUpdateExclude() used, but obj has according flag NOT set!"));
    }
#endif

    return ESS_RESULT_SUCCESS;
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODPDOParamGetState(ESS_HANDLE h, uint16_t pdo, volatile ESS_BOOL** state)
{
    const uint16_t objIdx = pdo - 0x200;
    ESS_OD_OBJECT_INFOS_PDOPARAM* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    obj = FindObj(&ed->dict, objIdx);
    if (obj == NULL)
        return ESS_RESULT_OBJ_NOT_FOUND;

    infos = (ESS_OD_OBJECT_INFOS_PDOPARAM*)obj->infos;
    *state = &infos->state;

#ifdef _DEBUG
    if ((infos->flags & ESS_OD_PDOPARAM_FLAGS_STATE_EXISTS) == 0) {
        ESS_DBG(("WARNING: essODPDOParamGetState() used, but obj has according flag NOT set!"));
    }
#endif

    return ESS_RESULT_SUCCESS;
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODPDOParamGetControl(ESS_HANDLE h, uint16_t pdo, volatile ESS_BOOL** control)
{
    const uint16_t objIdx = pdo - 0x200;
    ESS_OD_OBJECT_INFOS_PDOPARAM* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_OD_OBJECT* obj;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    obj = FindObj(&ed->dict, objIdx);
    if (obj == NULL)
        return ESS_RESULT_OBJ_NOT_FOUND;

    infos = (ESS_OD_OBJECT_INFOS_PDOPARAM*)obj->infos;
    *control = &infos->control;

#ifdef _DEBUG
    if ((infos->flags & ESS_OD_PDOPARAM_FLAGS_CONTROL_EXISTS) == 0) {
        ESS_DBG(("WARNING: essODPDOParamGetControl() used, but obj has according flag NOT set!"));
    }
#endif

    return ESS_RESULT_SUCCESS;
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODPDOParamGetToggle(ESS_HANDLE h, uint16_t pdo, volatile ESS_BOOL** toggle)
{
    const uint16_t objIdx = pdo - 0x200;
    ESS_OD_OBJECT_INFOS_PDOPARAM* infos;
    ESC_DATA* ed = (ESC_DATA*)h;
    const ESS_OD_OBJECT* obj;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    obj = FindObj(&ed->dict, objIdx);
    if (obj == NULL)
        return ESS_RESULT_OBJ_NOT_FOUND;

    infos = (ESS_OD_OBJECT_INFOS_PDOPARAM*)obj->infos;
    *toggle = &infos->toggle;

#ifdef _DEBUG
    if ((infos->flags & ESS_OD_PDOPARAM_FLAGS_TOGGLE_EXISTS) == 0) {
        ESS_DBG(("WARNING: essODPDOParamGetToggle() used, but obj has according flag NOT set!"));
    }
#endif

    return ESS_RESULT_SUCCESS;
}

#else

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODPDOParamCreate(ESS_HANDLE h, uint16_t pdo, const char* name, ESS_OD_PDOPARAM_FLAGS flags)
{
    (void)h;
    (void)pdo;
    (void)name;
    (void)flags;
    return ESS_RESULT_NOT_IMPLEMENTED;
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODPDOParamUpdateExclude(ESS_HANDLE h, uint16_t pdo, const void* data, uint16_t dataLen)
{
    (void)h;
    (void)pdo;
    (void)data;
    (void)dataLen;

    return ESS_RESULT_NOT_IMPLEMENTED;
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODPDOParamGetState(ESS_HANDLE h, uint16_t pdo, volatile ESS_BOOL** state)
{
    (void)h;
    (void)pdo;
    (void)state;

    return ESS_RESULT_NOT_IMPLEMENTED;
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODPDOParamGetControl(ESS_HANDLE h, uint16_t pdo, volatile ESS_BOOL** control)
{
    (void)h;
    (void)pdo;
    (void)control;

    return ESS_RESULT_NOT_IMPLEMENTED;
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essODPDOParamGetToggle(ESS_HANDLE h, uint16_t pdo, volatile ESS_BOOL** toggle)
{
    (void)h;
    (void)pdo;
    (void)toggle;

    return ESS_RESULT_NOT_IMPLEMENTED;
}

#endif /* !CFG_ESS_OD_ALLOW_HANDLE_PDO_PARAMS */
