/*
 * Copyright (C) 2011-2017 Hottinger Bruel & Kjaer Benelux B.V.
 * Schutweg 15a
 * 5145NP Waalwijk
 * The Netherlands
 * http://www.hbm.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 *(at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * HBM FPGA driver core implementation
 *
 */
/* enable printing debugging info */
/* #define DEBUG */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/cdev.h>
#include <linux/poll.h>
#include <linux/bug.h>
#include <linux/gfp.h>
#include <linux/kfifo.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/mutex.h>
#include <linux/version.h>
#include <linux/ioctl.h>

#include "hbm-fpga.h"
#include "hbm-fpga-driver.h"
#include "hbm-fpga-core.h"
#include "hbm-fpga-i2c.h"
#include "hbm-fpga-lcd.h"
#include "hbm-fpga-button.h"


#define MODNAME			"hbm-fpga"
#define DRIVER_VERSION		"1.6"

#define EVENT_QUEUE_SIZE	(128 *1024)	/* In bytes, must be a power of 2 */

/* Index numbers to tie fpga_variant array members to hbm_fpga_pci_tbl array members */
#define BACKPLANE_2x		0
#define BACKPLANE_3x		1
#define BACKPLANE_4x		2	
#define BACKPLANE_7x		3
#define BACKPLANE_17P		4
#define OPTIONCARRIER		5

/* Flags for variant properties */
#define FLAGS__IS_BKP		(1UL << 0)	/* Indicates this FPGA is a backplane FPGA */
#define FLAGS__IS_OCC		(1UL << 1)	/* Indicates this FPGA is an option carrier card FPGA */
#define FLAGS__HAS_BOOTMODE	(1UL << 2)	/* Indicates this FPGA has a boot mode */
#define FLAGS__HAS_SLOTID	(1UL << 3)	/* Indicates this FPGA has a slot ID */
#define FLAGS__HAS_LCD		(1UL << 4)	/* Indicates this FPGA has an LCD attached connected */
#define FLAGS__HAS_BUTTON	(1UL << 5)	/* Indicates this FPGA has a button connected */


struct event_t {
	unsigned int           int_timebase_mask;
	unsigned int           int_timebase_evt;
	unsigned int           int_cmdnow_mask;
	unsigned int           int_cmdnow_evt;
	unsigned int           int_msgq_mask;
	unsigned int           int_msgq_evt;
	unsigned int           int_ms_mask;
	unsigned int           int_ms_evt;
	unsigned int           int_clock_generator_mask;
	unsigned int           int_clock_generator_evt;
	unsigned int           int_ext_start_mask;
	unsigned int           int_ext_start_evt;
	unsigned int           int_ext_stop_mask;
	unsigned int           int_ext_stop_evt;
};


/*
 * Global data
 */
static struct class            *class;                                  /* Device class for all FPGAs */
static dev_t                   cdev_region;                             /* Region to allocate all FPGA character devices in */
static const char              bootmode_normal[]  = "normal";		/* Keyword to return for normal boot mode */
static const char              bootmode_minimum[] = "minimum";		/* Keyword to return for minimum boot mode */
static const char              type_backplane[]   = "backplane";	/* Keyword to return for backplane type FPGAs */
static const char              type_occ[]         = "occ";		/* Keyword to return for backplane type FPGAs */

static struct event_t          bkp_events = {
	.int_timebase_mask        = BKP_INT_TIMEBASE_MASK,
	.int_timebase_evt         = BKP_INT_TIMEBASE,
	.int_cmdnow_mask          = BKP_INT_CMDNOW_MASK,
	.int_cmdnow_evt           = BKP_INT_CMDNOW,
	.int_msgq_mask            = BKP_INT_MSGQ_MASK,
	.int_msgq_evt             = BKP_INT_MSGQ,
	.int_ms_mask              = BKP_INT_MS_MASK,
	.int_ms_evt               = BKP_INT_MS,
	.int_clock_generator_mask = BKP_INT_CLOCK_GENERATOR_MASK,
	.int_clock_generator_evt  = BKP_INT_CLOCK_GENERATOR,
	.int_ext_start_mask       = BKP_INT_EXT_START_MASK,
	.int_ext_start_evt        = BKP_INT_EXT_START,
	.int_ext_stop_mask        = BKP_INT_EXT_STOP_MASK,
	.int_ext_stop_evt         = BKP_INT_EXT_STOP
};

static struct event_t          occ_events = {
	.int_timebase_mask        = OCC_INT_TIMEBASE_MASK,
	.int_timebase_evt         = OCC_INT_TIMEBASE,
	.int_cmdnow_mask          = OCC_INT_CMDNOW_MASK,
	.int_cmdnow_evt           = OCC_INT_CMDNOW,
	.int_msgq_mask            = OCC_INT_MSGQ_MASK,
	.int_msgq_evt             = OCC_INT_MSGQ,
	.int_ms_mask              = OCC_INT_MS_MASK,
	.int_ms_evt               = OCC_INT_MS,
	.int_clock_generator_mask = OCC_INT_CLOCK_GENERATOR_MASK,
	.int_clock_generator_evt  = OCC_INT_CLOCK_GENERATOR,
	.int_ext_start_mask       = OCC_INT_EXT_START_MASK,
	.int_ext_start_evt        = OCC_INT_EXT_START,
	.int_ext_stop_mask        = OCC_INT_EXT_STOP_MASK,
	.int_ext_stop_evt         = OCC_INT_EXT_STOP
};

static struct i2c_bus_props_t  backplane_i2c[] = {
	/* Order by interrupt priority: First entry is checked and serviced first */
	{
		.regbase  = BKP_I2C_SYSTEM, /* offsetof(struct hbm_fpga_regs, FPGA_SYSTEM_MAP) + offsetof(FPGA_SYSTEM_MAP, I2CPRERL) */
		.irq_mask = BKP_INT_I2C_SYSTEM_MASK,
		.event    = BKP_INT_I2C_SYSTEM
	},
	{
		.regbase  = BKP_I2C_POWER_MANAGEMENT,
		.irq_mask = BKP_INT_I2C_SEQUENCER_MASK,
		.event    = BKP_INT_I2C_SEQUENCER
	},
	{
		.regbase  = BKP_I2C_KEYBOARD,
		.irq_mask = BKP_INT_I2C_KEYBOARD_MASK,
		.event    = BKP_INT_I2C_KEYBOARD
	},
	{
		.regbase  = BKP_I2C_POWER_CONTROL,
		.irq_mask = BKP_INT_I2C_POWERCTRL_MASK,
		.event    = BKP_INT_I2C_POWERCTRL
	},
	{} /* Sentinel */
};

static struct i2c_bus_props_t  gen2tb_backplane_i2c[] = {
	/* Order by interrupt priority: First entry is checked and serviced first */
	{
		.regbase  = GEN2TB_I2C_SYSTEM,
		.irq_mask = GEN2TB_INT_I2C_SYSTEM_MASK,
		.event    = GEN2TB_INT_I2C_SYSTEM
	},
	{
		.regbase  = GEN2TB_I2C_MASTER_SLAVE,
		.irq_mask = GEN2TB_INT_I2C_MASTER_SLAVE_MASK,
		.event    = GEN2TB_INT_I2C_MASTER_SLAVE
	},
	{} /* Sentinel */
};

static struct i2c_bus_props_t  gen4tb_backplane_i2c[] = {
	/* Order by interrupt priority: First entry is checked and serviced first */
	{
		.regbase  = GEN4TB_I2C_SYSTEM,
		.irq_mask = GEN4TB_INT_I2C_SYSTEM_MASK,
		.event    = GEN4TB_INT_I2C_SYSTEM
	},
	{
		.regbase  = GEN4TB_I2C_MASTER_SLAVE,
		.irq_mask = GEN4TB_INT_I2C_MASTER_SLAVE_MASK,
		.event    = GEN4TB_INT_I2C_MASTER_SLAVE
	},
	{} /* Sentinel */
};

static struct i2c_bus_props_t  occ_i2c[] = {
	/* Order by interrupt priority: First entry is checked and serviced first */
	{
		.regbase  = OCC_I2C_SYSTEM,
		.irq_mask = OCC_INT_I2C_SYSTEM_MASK,
		.event    = OCC_INT_I2C_SYSTEM
	},
	{
		.regbase  = OCC_I2C_POWER_MANAGEMENT,
		.irq_mask = OCC_INT_I2C_SEQUENCER_MASK,
		.event    = OCC_INT_I2C_SEQUENCER
	},
	{
		.regbase  = OCC_I2C_KEYBOARD,
		.irq_mask = OCC_INT_I2C_KEYBOARD_MASK,
		.event    = OCC_INT_I2C_KEYBOARD
	},
	{
		.regbase  = OCC_I2C_POWER_CONTROL,
		.irq_mask = OCC_INT_I2C_PCIBRIDGE_MASK,
		.event    = OCC_INT_I2C_PCIBRIDGE
	},
	{} /* Sentinel */
};

static struct variant_t {
	const char              *type;
	const char              *name;
	u32                     id;
	struct event_t          *events;
	struct i2c_bus_props_t  *i2c_bus_props;
	u32                     flags;
} fpga_variant[] = {
	[BACKPLANE_2x] = {
		.type          = type_backplane,
		.name          = "GEN2tB",
		.id            = 8,
		.events        = &bkp_events,
		.i2c_bus_props = gen2tb_backplane_i2c,
		.flags         = FLAGS__IS_BKP |
		                 FLAGS__HAS_BOOTMODE |
		                 FLAGS__HAS_LCD |
		                 FLAGS__HAS_BUTTON
	},
	[BACKPLANE_3x] = {
		.type          = type_backplane,
		.name          = "GEN3t",
		.id            = 5,
		.events        = &bkp_events,
		.i2c_bus_props = backplane_i2c,
		.flags         = FLAGS__IS_BKP |
		                 FLAGS__HAS_BOOTMODE
	},
	[BACKPLANE_4x] = {
		.type          = type_backplane,
		.name          = "GEN4tB",
		.id            = 9,
		.events        = &bkp_events,
		.i2c_bus_props = gen4tb_backplane_i2c,
		.flags         = FLAGS__IS_BKP |
		                 FLAGS__HAS_BOOTMODE |
		                 FLAGS__HAS_LCD |
		                 FLAGS__HAS_BUTTON
	},
	[BACKPLANE_7x] = {
		.type          = type_backplane,
		.name          = "GEN7tA",
		.id            = 6,
		.events        = &bkp_events,
		.i2c_bus_props = backplane_i2c,
		.flags         = FLAGS__IS_BKP |
		                 FLAGS__HAS_BOOTMODE
	},
	[BACKPLANE_17P] = {
		.type          = type_backplane,
		.name          = "GEN17tA",
		.id            = 7,
		.events        = &bkp_events,
		.i2c_bus_props = backplane_i2c,
		.flags         = FLAGS__IS_BKP |
		                 FLAGS__HAS_BOOTMODE
	},
	[OPTIONCARRIER] = {
		.type          = type_occ,
		.name          = "OCC",
		.events        = &occ_events,
		.i2c_bus_props = occ_i2c,
		.flags         = FLAGS__IS_OCC |
		                 FLAGS__HAS_SLOTID
	}
};

DECLARE_BITMAP(instance_map, INSTANCES);	/* Bit map keeping track of used instance/minor numbers */


/**
 * Sets the interrupt mask on the FPGA
 *
 * @param [in] fpga - The FPGA for which to set the mask
 * @param [in] enable_mask - Interrupts to enable
 * @param [in] disable_mask - Interrupts to disable
 */
void hbm_fpga_change_irq_mask(struct hbm_fpga_dev  *fpga,
                              unsigned int         enable_mask,
                              unsigned int         disable_mask)
{
	unsigned long  flags;

	/* This function is called from both syscalls and interrupt context, so lock and save irq */
	spin_lock_irqsave(&fpga->irq_spinlock, flags);

	fpga->irq_mask |= enable_mask;
	fpga->irq_mask &= ~disable_mask;

	fpga->regs->system.INTEN = fpga->irq_mask;

	spin_unlock_irqrestore(&fpga->irq_spinlock, flags);
}


/**
 * Handles HBM_FPGA_IOCTL_CHANGE_IRQ_MASK
 *
 * @param [in] fpga - reference to FPGA
 * @param [in] irqParams - ioctl parameters
 *
 */
static void ioctl_change_irq_mask(struct hbm_fpga_dev *fpga, irq_params* irqParams)
{
	BUG_ON(!fpga);
	BUG_ON(!irqParams);

	hbm_fpga_change_irq_mask(fpga, irqParams->enable_mask, irqParams->disable_mask);
}


/**
 * Initializes a new fifo
 *
 * @param [in] fpga - FPGA reference, for debug info mainly.
 * @param [in] size - Size of the fifo in bytes (must be a power of 2)
 *
 * @return Reference to new fifo event queue or NULL on failure
 */
static struct kfifo *create_new_fifo(struct hbm_fpga_dev *fpga, int size)
{
	struct kfifo  *fifo;

	if (!(fifo = kzalloc(sizeof(struct kfifo), GFP_KERNEL))) {
		dev_err(&fpga->pci_dev->dev, "create_new_fifo: kzalloc failed\n");
		return NULL;
	}

	if ((kfifo_alloc(fifo, size, GFP_KERNEL)) ||
	    IS_ERR(fifo)) {
		kfree(fifo);
		dev_err(&fpga->pci_dev->dev, "create_new_fifo: kfifo_alloc failed\n");
		return NULL;
	}

	return fifo;
}


static int kfifo_len_locked(struct hbm_fpga_dev *fpga)
{
	int            ret;
	unsigned long  flags;

	spin_lock_irqsave(&fpga->event_queue_spinlock, flags);
	ret = kfifo_len(fpga->event_queue);
	spin_unlock_irqrestore(&fpga->event_queue_spinlock, flags);

	return ret;
}


void hbm_fpga_queue_event(struct hbm_fpga_dev *fpga, irq_event *event)
{
	int  ret;

	/* Used from interrupt contect only, so no irq save required */
	spin_lock(&fpga->event_queue_spinlock);
	if (!fpga->event_queue) {
		/* Device node is not opened; Nobody cares */
		spin_unlock(&fpga->event_queue_spinlock);
		return;
	}

	ret = kfifo_in(fpga->event_queue, event, sizeof(irq_event));
	spin_unlock(&fpga->event_queue_spinlock);

	if (ret < 0) {
		dev_err(&fpga->pci_dev->dev, "Error %d queueing event %d\n", ret, event->type);
		return;
	}

	/* Wake up the read() */
	if (ret)
		wake_up_interruptible(&fpga->wait_queue);

	if (ret != sizeof(irq_event))
		dev_err(&fpga->pci_dev->dev, "Error queueing event %d: kfifo full\n", event->type);
}


/**
 * Releases a previously created fifo
 *
 * @param [in] fifo - Fifo to be released
 */
static void release_fifo(struct kfifo* fifo)
{
	if (!fifo)
		return;

	kfifo_free(fifo);
	kfree(fifo);
}


static long chrdev_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
	struct inode         *i = f->f_dentry->d_inode;
#else
	struct inode         *i = f->f_inode;
#endif
	long                 ret = 0;
	struct hbm_fpga_dev  *fpga = NULL;
	irq_params           irqParams;

	fpga = (struct hbm_fpga_dev *)container_of(i->i_cdev, struct hbm_fpga_dev, fpga_cdev);
	BUG_ON(!fpga);

	mutex_lock(&(fpga->driver_mutex));

	switch (cmd) {
	case HBM_FPGA_IOCTL_READ_PHYSICAL_ADDRESS:
		if (copy_to_user((void *)arg, (void *)&(fpga->fpga_mem_base), sizeof(unsigned long)))
			ret = -EFAULT;
		break;

	case HBM_FPGA_IOCTL_READ_MEMORY_SIZE:
		if (copy_to_user((void *)arg, (void *)&(fpga->fpga_mem_size), sizeof(unsigned int)))
			ret = -EFAULT;
		break;

	case HBM_FPGA_IOCTL_SET_TRACE_FLAGS:
		if (copy_from_user((void *)&(fpga->trace_flags), (void *)arg, sizeof(unsigned int)))
			ret = -EFAULT;
		break;

	case HBM_FPGA_IOCTL_I2C_INIT:
	case HBM_FPGA_IOCTL_I2C_RESET:
	case HBM_FPGA_IOCTL_I2C_MSG:
		/* Forward I2C IOCTLs to I2C */
		ret = i2c_ioctl(fpga, cmd, arg);
		break;

	case HBM_FPGA_IOCTL_CHANGE_IRQ_MASK:
		if (!copy_from_user((void *)&(irqParams), (void *)arg, sizeof(irq_params)))
			ioctl_change_irq_mask(fpga, &irqParams);
		else
			ret = -EFAULT;
		break;

	case HBM_FPGA_IOCTL_GET_PCI_BUS_NUMBER:
		if (copy_to_user((void *)arg, (void *)&(fpga->pci_dev->bus->number), sizeof(fpga->pci_dev->bus->number)))
			ret = -EFAULT;
		break;

	case HBM_FPGA_IOCTL_GET_PCI_SUBDEVICE_ID:
		if (copy_to_user((void *)arg, (void *)&(fpga->subdevice), sizeof(fpga->subdevice)))
			ret = -EFAULT;
		break;

	default:
		ret = -ENOTTY; /* POSIX: "inappropriate ioctl for device" */
		break;
	}

	mutex_unlock(&(fpga->driver_mutex));

	return ret;
}


static int chrdev_mmap(struct file *f, struct vm_area_struct *vma)
{
	struct hbm_fpga_dev  *fpga;
	unsigned long        phys_addr;
	unsigned long        phys_size;

	BUG_ON(!f);
	BUG_ON(!vma);

	fpga = (struct hbm_fpga_dev *)f->private_data;
	BUG_ON(!fpga);

	dev_dbg(&fpga->pci_dev->dev, "chrdev_mmap\n");

	phys_addr = fpga->fpga_mem_base;
	phys_size = fpga->fpga_mem_size;

	dev_info(&fpga->pci_dev->dev, "mmap phys 0x%.8lx...0x%.8lx\n",
		phys_addr, phys_addr + phys_size - 1);

	if (remap_pfn_range(vma, vma->vm_start, phys_addr >> PAGE_SHIFT, phys_size, vma->vm_page_prot)) {
		dev_err(&fpga->pci_dev->dev, "chrdev_mmap: failed\n");
		return -EAGAIN;
	}

	dev_info(&fpga->pci_dev->dev, "mmap virt 0x%.8lx...0x%.8lx (flags = 0x%.8lx)\n",
		(unsigned long)vma->vm_start, (unsigned long)vma->vm_start + phys_size - 1, vma->vm_flags);

	return 0;
}


static void cmdnow_isr(struct hbm_fpga_dev *fpga)
{
	irq_event  event;

	event.type = fpga->variant->events->int_cmdnow_evt;
	event.u.cmd_now.TCDR = fpga->regs->timebase.TCDR;

	hbm_fpga_queue_event(fpga, &event);
}


static void timebase_isr(struct hbm_fpga_dev *fpga)
{
	irq_event  event;

	event.type = fpga->variant->events->int_timebase_evt;
	event.u.timebase.TSR  = fpga->regs->timebase.TSR;
	event.u.timebase.TS0R = fpga->regs->timebase.TS0R;
	event.u.timebase.TSIR = fpga->regs->timebase.TSIR;
	event.u.timebase.TIDR = fpga->regs->timebase.TIDR;

	if (!(event.u.timebase.TSR & (TSR__1PPS_INTERRUPT | TSR__TIMEBASE_INTERRUPT))) {
		dev_err(&fpga->pci_dev->dev, "timebase_isr: spurious interrupt\n");
		return;
	}

	hbm_fpga_queue_event(fpga, &event);
}


static void master_slave_isr(struct hbm_fpga_dev *fpga)
{
	volatile uint32_t  MSMSGSTAT;
	volatile uint32_t  MSINTSTAT;
	irq_event          event;

	/* First read MSINTSTAT to clear the interrupt... */
	MSINTSTAT = fpga->regs->masterslave.MSINTSTAT;

	/* ... then read the MSMSGSTAT value. This sequence is important, otherwise we could miss level changes. */
	MSMSGSTAT = fpga->regs->masterslave.MSMSGSTAT;

	event.type = fpga->variant->events->int_ms_evt;
	event.u.master_slave.MSMSGSTAT = MSMSGSTAT;

	hbm_fpga_queue_event(fpga, &event);
}


static void clock_generator_isr(struct hbm_fpga_dev *fpga)
{
	irq_event  event;

	event.type = fpga->variant->events->int_clock_generator_evt;
	event.u.clock_generator.CGSR = fpga->regs->timebase.CGCR;

	hbm_fpga_queue_event(fpga, &event);
}


static void ext_start_stop_isr(struct hbm_fpga_dev *fpga, int stop)
{
	irq_event          event;

	hbm_fpga_change_irq_mask(fpga, 0, stop ? fpga->variant->events->int_ext_stop_mask : fpga->variant->events->int_ext_start_mask);

	event.type = stop?fpga->variant->events->int_ext_stop_evt:fpga->variant->events->int_ext_start_evt;
	/* This event type has no payload */

	hbm_fpga_queue_event(fpga, &event);

	/* Clear the interrupt */
	fpga->regs->system.EXTSTSTP |= stop ? EXTSTSTP__CLEAR_STOP_INTERRUPT : EXTSTSTP__CLEAR_START_INTERRUPT;
}


static void msgqueue_isr(struct hbm_fpga_dev *fpga)
{
	irq_event          event;

	event.type = fpga->variant->events->int_msgq_evt;

	/* Queue events for as long as data is available */
	while ((event.u.message_queue.frame_offset = fpga->regs->msgq.recvMsgQ) != RECV_FIFO_EMPTY_VALUE)
		hbm_fpga_queue_event(fpga, &event);
}


/* Reset FPGA MsgQ FIFOs */
static void reset_msg_queue_fifos(struct hbm_fpga_dev *fpga)
{
	BUG_ON(!fpga);
	BUG_ON(!fpga->regs);

	/* Reset FIFOs */
	fpga->regs->msgq.MSGCR = (1 << MSGCR_RESET_FIFO);
}


/**
 *	Initializes all structures for the Message Queue event source.
 *
 *	@param [in] fpga - FPGA reference
 *
 *	@return 0 on success, negative value on failure.
 *
*/
static void init_evt_src_msgqueue(struct hbm_fpga_dev *fpga)
{
	BUG_ON(!fpga->regs);

	/* Initialise FPGA FIFO's */
	reset_msg_queue_fifos(fpga);
}


/**
 *	Releases all structures for the Message Queue event source
 *
 *	@param [in] fpga - FPGA reference
 *
 *	@return 0 on success, negative value on failure.
 *
*/
static void release_evt_src_msgqueue(struct hbm_fpga_dev *fpga)
{
	hbm_fpga_change_irq_mask(fpga, 0, fpga->variant->events->int_msgq_mask);

	/* clear Msg Q FIFO's */
	reset_msg_queue_fifos(fpga);

	/* tear down the wait queue */

	/* Note: no need to destroy spin_lock, simply don't use it anymore! */
}


/**
 *	Initializes all structures for the CMDNOW event source.
 *
 *	@param [in] fpga - FPGA reference
 *
 *	@return 0 on success, negative value on failure.
 *
*/
static void init_evt_src_cmdnow(struct hbm_fpga_dev *fpga)
{
	hbm_fpga_change_irq_mask(fpga, fpga->variant->events->int_cmdnow_mask, 0);
}


/**
 *	Releases all structures for the CMDNOW event source
 *
 *	@param [in] fpga - FPGA reference
 *
 *	@return 0 on success, negative value on failure.
 *
*/
static void release_evt_src_cmdnow(struct hbm_fpga_dev *fpga)
{
	hbm_fpga_change_irq_mask(fpga, 0, fpga->variant->events->int_cmdnow_mask);
}


/* Initialize all structures for the Timebase character device implementation */
static void init_evt_src_timebase(struct hbm_fpga_dev *fpga)
{
/*
	NOTE: timestamp queue size = 512 entries
	for (i = 0; i < TIMESTAMP_HW_QUEUE_SIZE; i++) {
		timebase_struct.TSR  = fpga->regs->timebase.TSR;
		timebase_struct.TS0R = fpga->regs->timebase.TS0R;
		timebase_struct.TIDR = fpga->regs->timebase.TIDR;
	} HBM FPGA
*/
	hbm_fpga_change_irq_mask(fpga, fpga->variant->events->int_timebase_mask, 0);
}


static void release_evt_src_timebase(struct hbm_fpga_dev *fpga)
{
	hbm_fpga_change_irq_mask(fpga, 0, fpga->variant->events->int_timebase_mask);
}


/**
 *
 * Initializes the event sources.
 *
 * @param [in] fpga - The FPGA device for which to initialize
 * @return 0 on success, negative value on failure
 */
static int init_event_sources(struct hbm_fpga_dev *fpga)
{
	int  ret;

	/* Init the wait queue so we can sleep while waiting for messages */
	init_waitqueue_head(&fpga->wait_queue);

	/* Event sources */
	init_evt_src_msgqueue(fpga);
	init_evt_src_cmdnow(fpga);
	init_evt_src_timebase(fpga);

	if ((ret = i2c_init(fpga, fpga->variant->i2c_bus_props)))
		goto err_i2c;
	return 0;

err_i2c:
	release_evt_src_timebase(fpga);
	release_evt_src_cmdnow(fpga);
	release_evt_src_msgqueue(fpga);

	return ret;
}


/**
 *
 * Releases the event sources.
 *
 * @param [in] fpga - The FPGA device for which to release
 */
static void release_event_sources(struct hbm_fpga_dev *fpga)
{
	i2c_release(fpga);

	release_evt_src_timebase(fpga);
	release_evt_src_cmdnow(fpga);
	release_evt_src_msgqueue(fpga);
}


/* open() */
static int chrdev_open(struct inode *i, struct file *f)
{
	struct hbm_fpga_dev  *fpga = container_of(i->i_cdev, struct hbm_fpga_dev, fpga_cdev);
	unsigned long        flags;
	int                  ret;

	BUG_ON(!fpga);

	/* Only Blocking mode supported */
	if (f->f_flags & O_NONBLOCK) {
		dev_err(&fpga->pci_dev->dev, "chrdev_open: non-blocking mode not supported!\n");
		ret = -EINVAL;
		goto err_mode;
	}

	spin_lock_irqsave(&fpga->event_queue_spinlock, flags);
	if (fpga->event_queue) {
		dev_err(&fpga->pci_dev->dev, "chrdev_open: driver already opened by other application. Only one instance allowed\n");
		ret = -EMFILE;
		goto err_queue;
	}

	if (!(fpga->event_queue = create_new_fifo(fpga, EVENT_QUEUE_SIZE))) {
		dev_err(&fpga->pci_dev->dev, "chrdev_open: create_new_fifo failed\n");
		ret = -EFAULT;
		goto err_fifo;
	}
	spin_unlock_irqrestore(&fpga->event_queue_spinlock, flags);

	if ((ret = init_event_sources(fpga)))
		goto err_event_sources;

	/* Setup private data used when read/write/.. is called */
	f->private_data = fpga;

	return 0;

err_event_sources:
	spin_lock_irqsave(&fpga->event_queue_spinlock, flags);
	release_fifo(fpga->event_queue);
	fpga->event_queue = NULL;
err_fifo:
err_queue:
	spin_unlock_irqrestore(&fpga->event_queue_spinlock, flags);
err_mode:
	return ret;
}


static int chrdev_release(struct inode *i, struct file *f)
{
	struct hbm_fpga_dev  *fpga = container_of(i->i_cdev, struct hbm_fpga_dev, fpga_cdev);
	unsigned long        flags;

	BUG_ON(!fpga);

	release_event_sources(fpga);

	spin_lock_irqsave(&fpga->event_queue_spinlock, flags);
	release_fifo(fpga->event_queue);
	fpga->event_queue = NULL;
	spin_unlock_irqrestore(&fpga->event_queue_spinlock, flags);

	return 0;
}


/*
 * read frame offset of oldest(first) message in FIFO
 */
static ssize_t chrdev_read(struct file *f, char __user *buffer, size_t count, loff_t *offset)
{
	int                  ret;
	struct hbm_fpga_dev  *fpga;
	unsigned long        flags;
	unsigned int         copied;

	BUG_ON(!f);

	fpga = (struct hbm_fpga_dev *)f->private_data;
	BUG_ON(!fpga);

	/* We only support blocking reads */
	if (f->f_flags & O_NONBLOCK)
		return -EAGAIN;

	/* Check parameters, ignore offset */
	if (!buffer) {
		dev_err(&fpga->pci_dev->dev, "chrdev_read: Null pointer parsed\n");
		return -EFAULT;
	}
	if (count != sizeof(irq_event)) {
		dev_err(&fpga->pci_dev->dev, "chrdev_read: Attempt to read %d bytes (%d expected)\n", count, sizeof(irq_event));
		return -EFAULT;
	}

	/* Check if / block until there's enough to read */
	if ((ret = wait_event_interruptible(fpga->wait_queue, kfifo_len_locked(fpga) >= sizeof(irq_event)))) {
		dev_dbg(&fpga->pci_dev->dev, "chrdev_read: wait_event interrupted (%d)\n", ret);

		/* Interrupted; Restart syscall */
		return -ERESTARTSYS;
	}

	dev_dbg(&fpga->pci_dev->dev, "chrdev_read: Data ready\n");

	spin_lock_irqsave(&fpga->event_queue_spinlock, flags);
	ret = kfifo_to_user(fpga->event_queue, buffer, count, &copied);
	spin_unlock_irqrestore(&fpga->event_queue_spinlock, flags);

	if (ret) {
		dev_err(&fpga->pci_dev->dev, "chrdev_read: Error %d copying data\n", ret);
		return ret;
	}

	return copied;
}


/*
 *	Recylces message queue frameOffset, one frameOffset per call
 */
static ssize_t chrdev_write(struct file *f, const char __user *buffer, size_t count, loff_t *offset)
{
	int                  ret = -EFAULT;
	struct hbm_fpga_dev  *fpga;
	unsigned int         frameOffset;
	char                 frameOffsetIsOK = 1;

	BUG_ON(!f);

	fpga = (struct hbm_fpga_dev *)f->private_data;
	BUG_ON(!fpga);
	BUG_ON(!fpga->regs);

	if (!buffer || (count != sizeof(u32)))
		return -EFAULT;

	if (copy_from_user(&frameOffset, buffer, sizeof(u32))) {
		dev_err(&fpga->pci_dev->dev, "chrdev_write: error copying data from user space\n");
		return -EFAULT;
	}

	/* frameOffset must be 4 bytes aligned */
	if (frameOffset % sizeof(u32)) {
		dev_err(&fpga->pci_dev->dev, "chrdev_write: frameOffset not a multiple of msg size: frameOffset0x%x\n",
			(unsigned int)frameOffset);
		frameOffsetIsOK = 0;
	}

	if (frameOffsetIsOK)
		ret = sizeof(u32);
	else
		frameOffset = (u32)RECV_FIFO_EMPTY_VALUE;

	fpga->regs->msgq.recvMsgQ = (u32)frameOffset;

	return ret;
}


static struct file_operations hbm_fpga_chrdev_fops = {
	.read		= chrdev_read,
	.write		= chrdev_write,
	.unlocked_ioctl	= chrdev_ioctl,
	.mmap		= chrdev_mmap,
	.open		= chrdev_open,
	.release	= chrdev_release,
	.owner		= THIS_MODULE
};


/**
 * Sets up the device("/dev/<myCharDevice>")
 *
 * @param [in] fpga - indicates the FPGA
 * @return 0 on success
 */
static int setup_char_device(struct hbm_fpga_dev *fpga)
{
	int         ret = 0;

	/* Allocate instance_nr based minor number */
	fpga->fpga_dev = MKDEV(MAJOR(cdev_region), fpga->instance_nr * DEV_NODES + FPGA_DEV_NODE_OFFSET);

	dev_info(&fpga->pci_dev->dev, "FPGA cdev: %d:%d\n", MAJOR(fpga->fpga_dev), MINOR(fpga->fpga_dev));

	/* Initialize a character device for this device */
	cdev_init(&fpga->fpga_cdev, &hbm_fpga_chrdev_fops);
	fpga->fpga_cdev.owner = THIS_MODULE;

	/* Add the character device */
	if ((ret = cdev_add(&fpga->fpga_cdev, fpga->fpga_dev, 1))) {
		dev_err(&fpga->pci_dev->dev, "Error %d adding FPGA char dev\n", ret);
		goto err_add;
	}

	/* Create device based on the module name and the instance number */
	if (!(ret = (int)device_create(class, &fpga->pci_dev->dev, fpga->fpga_dev, NULL, FPGA_DEVICE_NAME_FORMAT, fpga->instance_nr))) {
		ret = -ENXIO;
		dev_err(&fpga->pci_dev->dev, "Error creating FPGA device\n");
		goto err_create;
	}

	return 0;

	device_destroy(class, fpga->fpga_dev);
err_create:
	cdev_del(&fpga->fpga_cdev);
err_add:
	return ret;
}


/**
 * Removes the device("/dev/<myCharDevice>")
 *
 * @param [in] fpga - indicates the FPGA
 * @return 0 on success
 */
static void remove_char_device(struct hbm_fpga_dev *fpga)
{
	device_destroy(class, fpga->fpga_dev);
	cdev_del(&fpga->fpga_cdev);
}


/**
 * ISR dispatcher.
 */
static irqreturn_t core_irq_handler(int irq, void *dev_id)
{
	irqreturn_t          ret = IRQ_NONE;
	unsigned int         irq_status;
	struct hbm_fpga_dev  *fpga = (struct hbm_fpga_dev *)dev_id;

	BUG_ON(!fpga);
	BUG_ON(!fpga->regs);

	/* Dispatch on the IRQ types (clears the CMDNOW interrupt) */
	while ((irq_status = fpga->regs->system.INTSR & fpga->regs->system.INTEN)) {
		if (fpga->trace_flags & FPGA_TRACE_IRQ)
			dev_info(&fpga->pci_dev->dev, "core_irq_handler: irq_status: 0x%.8x\n", irq_status);

		if (irq_status & fpga->variant->events->int_cmdnow_mask)
			cmdnow_isr(fpga);

		if (irq_status & fpga->variant->events->int_timebase_mask)
			/*
			 Timebase interrupt source is one of the following two:
			 - 1PPS
			 - Timestamp on Command event(Timetag)
			 - Top Dead Center event(Timetag)
			 - Command Now event(Timetag)
			 - External clock 0 counter rollover
			 - Timestamp counter rollover
			*/
			timebase_isr(fpga);

		if (irq_status & fpga->variant->events->int_msgq_mask)
			msgqueue_isr(fpga);

		if (irq_status & fpga->variant->events->int_ms_mask)
			master_slave_isr(fpga);

		if (irq_status & fpga->variant->events->int_clock_generator_mask)
			clock_generator_isr(fpga);

		if (irq_status & fpga->variant->events->int_ext_start_mask)
			ext_start_stop_isr(fpga, 0);

		if (irq_status & fpga->variant->events->int_ext_stop_mask)
			ext_start_stop_isr(fpga, 1);

		/* Process these last, because these are handled by other interrupt handlers */
		if (fpga->i2c)
			i2c_handle_isr(fpga, irq_status);

		if (fpga->button)
			button_handle_isr(fpga, irq_status);

		ret = IRQ_HANDLED;
	}

	return ret;
}


/*
 * SYSFS RELATED SECTION
 */

#ifdef EXPOSE_REGISTERS
static ssize_t bsr_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));

	return snprintf(buf, PAGE_SIZE, "0x%.8x\n", fpga->regs->system.BSR);
}
static ssize_t bsr_store(struct device *dev, struct device_attribute *attr,
                          const char *buf, size_t count)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));
	unsigned long        value;

	if (kstrtoul(buf, 0, &value)) {
		printk(KERN_ERR "Conversion error\n");
		return count;
	}

	printk(KERN_ERR "Setting BSR @0x%p to 0x%.8lx\n", &fpga->regs->system.BSR, value);
	fpga->regs->system.BSR = value;

	return count;
}
static DEVICE_ATTR(bsr, S_IWUSR | S_IRUGO, bsr_show, bsr_store);


static ssize_t scratch1_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));

	return snprintf(buf, PAGE_SIZE, "0x%.8x\n", fpga->regs->debug.SCRATCH1);
}
static ssize_t scratch1_store(struct device *dev, struct device_attribute *attr,
                          const char *buf, size_t count)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));
	unsigned long        value;

	if (kstrtoul(buf, 0, &value)) {
		printk(KERN_ERR "Conversion error\n");
		return count;
	}

	printk(KERN_ERR "Setting SCRATCH1 @0x%p to 0x%.8lx\n", &fpga->regs->debug.SCRATCH1, value);
	fpga->regs->debug.SCRATCH1 = value;

	return count;
}
static DEVICE_ATTR(scratch1, S_IWUSR | S_IRUGO, scratch1_show, scratch1_store);


static ssize_t scratch2_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));

	return snprintf(buf, PAGE_SIZE, "0x%.8x\n", fpga->regs->debug.SCRATCH2);
}
static ssize_t scratch2_store(struct device *dev, struct device_attribute *attr,
                          const char *buf, size_t count)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));
	unsigned long        value;

	if (kstrtoul(buf, 0, &value)) {
		printk(KERN_ERR "Conversion error\n");
		return count;
	}

	printk(KERN_ERR "Setting SCRATCH2 @0x%p to 0x%.8lx\n", &fpga->regs->debug.SCRATCH2, value);
	fpga->regs->debug.SCRATCH2 = value;

	return count;
}
static DEVICE_ATTR(scratch2, S_IWUSR | S_IRUGO, scratch2_show, scratch2_store);
#endif /* EXPOSE_REGISTERS */

static ssize_t type_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));

	return snprintf(buf, PAGE_SIZE, fpga->variant->type);
}
static DEVICE_ATTR(type, S_IRUGO, type_show, NULL);


static ssize_t name_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));

	return snprintf(buf, PAGE_SIZE, fpga->variant->name);
}
static DEVICE_ATTR(name, S_IRUGO, name_show, NULL);


static ssize_t version_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));

	return snprintf(buf, PAGE_SIZE, "%d.%.2d.%.5d", FPGAVER__MAJOR(fpga->regs->system.FPGAVER),
							FPGAVER__MINOR(fpga->regs->system.FPGAVER),
							FPGAVER__BUILD(fpga->regs->system.FPGAVER));
}
static DEVICE_ATTR(version, S_IRUGO, version_show, NULL);


static ssize_t bootmode_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));

	return snprintf(buf, PAGE_SIZE, ((fpga->regs->system.BSR & BSR_BOOTMODE_MASK) == BSR_BOOTMODE_NORMAL)?bootmode_normal:bootmode_minimum);
}
static DEVICE_ATTR(bootmode, S_IRUGO, bootmode_show, NULL);


static ssize_t slotid_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct hbm_fpga_dev  *fpga = pci_get_drvdata(to_pci_dev(dev));

	return snprintf(buf, PAGE_SIZE, "%d", (fpga->regs->system.BSR & BSR_SLOT_NUMBER_MASK) >> BSR_SLOT_NUMBER_SHIFT);
}
static DEVICE_ATTR(slotid, S_IRUGO, slotid_show, NULL);


/*
 * PCI DRIVER RELATED SECTION
 */


#ifdef CONFIG_PM
static int hbm_fpga_pci_suspend(struct pci_dev *pdev, pm_message_t state)
{
	struct hbm_fpga_dev  *dev = pci_get_drvdata(pdev);

	pci_save_state(pdev);
	pci_set_power_state(pdev, PCI_D3hot);

	return 0; /* Returning anything else will prevent the kernel to go into suspend */
}

static int hbm_fpga_pci_resume(struct pci_dev *pdev)
{
	struct hbm_fpga_dev  *dev = pci_get_drvdata(pdev);

	pci_set_power_state(pdev, PCI_D0);
	pci_restore_state(pdev);

	return 0;
}
#endif /* CONFIG_PM */


/**
 * Sets up the FPGA
 *
 * @param [inout] pdev - reference to the FPGA hardware
 * @param [inout] fpga - the driver's logical representation of the hardware.
 *
 * @return 0 on success, negative value otherwise
 */
static int setup_pci_device(struct pci_dev *pdev, struct hbm_fpga_dev *fpga)
{
	int                     ret;
	unsigned long           flags;
	volatile unsigned long  intsr;

	/* Retrieve the mapped region first valid address */
	if (!(fpga->fpga_mem_base = pci_resource_start(pdev, 0)))
		dev_warn(&fpga->pci_dev->dev, "Warning: Region mapped to address 0\n");
	/* Retrieve the mapped region size */
	if (!(fpga->fpga_mem_size = pci_resource_len(pdev, 0))) {
		dev_err(&fpga->pci_dev->dev, "Error: Region is not mapped (size is 0)\n");
		ret = -ENOMEM;
		goto err_size;
	}
	/* Retrieve the mapped region resource flags */
	if (!(flags = pci_resource_flags(pdev, 0))) {
		dev_err(&fpga->pci_dev->dev, "Error: Region flags are 0\n");
		ret = -ENOMEM;
		goto err_resource;
	}
	/* Check if this region is mapped to memory space (as in: not to I/O space) */
	if (!(flags & IORESOURCE_MEM)) {
		dev_err(&fpga->pci_dev->dev, "Error: Region is not memory mapped\n");
		ret = -ENOMEM;
		goto err_flags;
	}

	dev_info(&fpga->pci_dev->dev, "phys mapped: 0x%.8lx...0x%.8lx (%lu bytes), flags = 0x%.8lx\n",
		fpga->fpga_mem_base, fpga->fpga_mem_base + fpga->fpga_mem_size - 1, (unsigned long)fpga->fpga_mem_size, flags);

	if (!(fpga->regs = ioremap(fpga->fpga_mem_base, fpga->fpga_mem_size))) {
		dev_err(&pdev->dev, "Error remapping region\n");
		ret = -ENOMEM;
		goto err_remap;
	}
	dev_info(&pdev->dev, "virt mapped: 0x%p\n", fpga->regs);

	/* Disable IRQ generation in FPGA, mask all interrupts */
	fpga->regs->system.INTEN = 0x0;

	/* Clear any pending IRQ's */
	intsr = fpga->regs->system.INTSR;

	/* Clear trace flags */
	fpga->trace_flags = FPGA_TRACE_NONE;

	if ((ret = pci_enable_msi(pdev))) {
		dev_err(&pdev->dev, "Error %d enabling MSI\n", ret);
		goto err_msi;
	}

	/* Setup the IRQ */
	/* PATCH GR */
	if ((ret = request_irq(pdev->irq, core_irq_handler, IRQF_SHARED, MODNAME, fpga))) {
		dev_err(&pdev->dev, "Error %d requesting irq = %d\n", ret, pdev->irq);
		goto err_irq;
	}

	spin_lock_init(&fpga->irq_spinlock);
	fpga->irq = pdev->irq; /* The kernel already knows the irq number assigned */
	fpga->irq_mask = 0;

	spin_lock_init(&fpga->event_queue_spinlock);

	return 0;

	free_irq(fpga->irq, fpga);
err_irq:
	pci_disable_msi(pdev);
err_msi:
	iounmap(fpga->regs);
err_remap:
err_flags:
err_resource:
err_size:
	return ret;
}


static void remove_pci_device(struct pci_dev *pdev, struct hbm_fpga_dev *fpga)
{
	free_irq(fpga->irq, fpga);
	pci_disable_msi(pdev);
	iounmap(fpga->regs);
}


/**
 * Called by the OS when probing for this driver's hardware
 * Will call setup_pci_device()
 *
 * @param [inout] pdev - reference to the FPGA hardware
 * @param [inout] ent - not used here.
 *
 * @return 0 on success, negative value otherwise.
 */
static int hbm_fpga_pci_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int                         ret;
	int                         instance_nr;
	const struct pci_device_id  *id_table_entry;
	struct hbm_fpga_dev         *fpga;

	BUG_ON(!pdev);
	BUG_ON(!pdev->driver);
	BUG_ON(!pdev->driver->id_table);

	/* Find a free instance number for this FPGA */
	do {
		instance_nr = find_first_zero_bit(instance_map, INSTANCES);
		if (instance_nr >= INSTANCES) {
			goto err_inst;
		}
	} while (test_and_set_bit(instance_nr, instance_map));

	dev_info(&pdev->dev, "Found HBM FPGA nr. %d\n", instance_nr);

	/* Allocate memory for driver instance data */
	if (!(fpga = kzalloc(sizeof(struct hbm_fpga_dev), GFP_KERNEL))) {
		dev_err(&pdev->dev, "Error allocating "MODNAME" instance\n");
		ret = -ENOMEM;
		goto err_mem;
	}
	fpga->instance_nr = instance_nr;

	/* Store driver instance data in pci_dev and pci_dev in driver instance for later reference */
	pci_set_drvdata(pdev, fpga);
	fpga->pci_dev = pdev;

	/* Find out which pci_device_id triggered this probe */
	if (!(id_table_entry = pci_match_id(pdev->driver->id_table, pdev))) {
		dev_err(&pdev->dev, "Error identifying\n");
		ret = -ENODEV;
		goto err_id;
	}
	fpga->variant = (struct variant_t *)id_table_entry->driver_data;
	fpga->subdevice = id_table_entry->subdevice;
	dev_info(&pdev->dev, "Identified as %s (id = %d)\n", fpga->variant->name, fpga->variant->id);

	/* Enable the device */
	if ((ret = pci_enable_device(pdev))) {
		dev_err(&pdev->dev, "Error %d enabling device\n", ret);
		goto err_enable;
	}

	/* Claim the PCI memory regions */
	if ((ret = pci_request_regions(pdev, MODNAME))) {
		dev_err(&pdev->dev, "Error %d claiming pci regions\n", ret);
		goto err_request;
	}

	/* Setup the PCI device resources */
	if ((ret = setup_pci_device(pdev, fpga)))
		/* Error reported in function already */
		goto err_pci_setup;

	/* Enable DMA by setting the bus master bit in the PCI_COMMAND register */
	pci_set_master(pdev);

	/* Setup a character device to access the FPGA */
	if ((ret = setup_char_device(fpga)))
		/* Error reported in function already */
		goto err_char_setup;

	/* Setup the LCD (which in turn will setup a character device itself */
	if ((fpga->variant->flags & FLAGS__HAS_LCD) &&
	    ((ret = lcd_setup(class, &cdev_region, fpga))))
		/* Error reported in function already */
		goto err_lcd_setup;

	/* Setup the button */
	if ((fpga->variant->flags & FLAGS__HAS_BUTTON) &&
	    ((ret = button_setup(fpga))))
		/* Error reported in function already */
		goto err_button_setup;

	mutex_init(&fpga->driver_mutex);

	dev_info(&pdev->dev, "Version: %d.%.2d.%.5d\n", FPGAVER__MAJOR(fpga->regs->system.FPGAVER),
							FPGAVER__MINOR(fpga->regs->system.FPGAVER),
							FPGAVER__BUILD(fpga->regs->system.FPGAVER));
	if (fpga->variant->flags & FLAGS__HAS_LCD)
		lcd_splash(fpga);

#ifdef EXPOSE_REGISTERS
	/* Generic sysfs file 'bsr' */
	if ((ret = device_create_file(&pdev->dev, &dev_attr_bsr))) {
		dev_err(&pdev->dev, "Error %d creating sysfs file 'bsr'\n", ret);
		goto err_sysfs_bsr;
	}

	/* Generic sysfs file 'scratch1' */
	if ((ret = device_create_file(&pdev->dev, &dev_attr_scratch1))) {
		dev_err(&pdev->dev, "Error %d creating sysfs file 'scratch1'\n", ret);
		goto err_sysfs_scratch1;
	}

	/* Generic sysfs file 'scratch2' */
	if ((ret = device_create_file(&pdev->dev, &dev_attr_scratch2))) {
		dev_err(&pdev->dev, "Error %d creating sysfs file 'scratch2'\n", ret);
		goto err_sysfs_scratch2;
	}
#endif /* EXPOSE_REGISTERS */

	/* Generic sysfs file 'type' */
	if ((ret = device_create_file(&pdev->dev, &dev_attr_type))) {
		dev_err(&pdev->dev, "Error %d creating sysfs file 'type'\n", ret);
		goto err_sysfs_type;
	}

	/* Generic sysfs file 'name' */
	if ((ret = device_create_file(&pdev->dev, &dev_attr_name))) {
		dev_err(&pdev->dev, "Error %d creating sysfs file 'name'\n", ret);
		goto err_sysfs_name;
	}

	/* Generic sysfs file 'version' */
	if ((ret = device_create_file(&pdev->dev, &dev_attr_version))) {
		dev_err(&pdev->dev, "Error %d creating sysfs file 'version'\n", ret);
		goto err_sysfs_version;
	}

	/* Variant specific sysfs file 'bootmode' */
	if ((fpga->variant->flags & FLAGS__HAS_BOOTMODE) &&
	    (ret = device_create_file(&pdev->dev, &dev_attr_bootmode))) {
		dev_err(&pdev->dev, "Error %d creating sysfs file 'bootmode'\n", ret);
		goto err_sysfs_bootmode;
	}

	/* Variant specific sysfs file 'slotid' */
	if ((fpga->variant->flags & FLAGS__HAS_SLOTID) &&
	    (ret = device_create_file(&pdev->dev, &dev_attr_slotid))) {
		dev_err(&pdev->dev, "Error %d creating sysfs file 'slotid'\n", ret);
		goto err_sysfs_slotid;
	}

	return 0;

	if (fpga->variant->flags & FLAGS__HAS_SLOTID)
		device_remove_file(&pdev->dev, &dev_attr_slotid);
err_sysfs_slotid:
	if (fpga->variant->flags & FLAGS__HAS_BOOTMODE)
		device_remove_file(&pdev->dev, &dev_attr_bootmode);
err_sysfs_bootmode:
	device_remove_file(&pdev->dev, &dev_attr_version);
err_sysfs_version:
	device_remove_file(&pdev->dev, &dev_attr_name);
err_sysfs_name:
	device_remove_file(&pdev->dev, &dev_attr_type);
err_sysfs_type:
#ifdef EXPOSE_REGISTERS
	device_remove_file(&pdev->dev, &dev_attr_scratch2);
err_sysfs_scratch2:
	device_remove_file(&pdev->dev, &dev_attr_scratch1);
err_sysfs_scratch1:
	device_remove_file(&pdev->dev, &dev_attr_bsr);
err_sysfs_bsr:
#endif /* EXPOSE_REGISTERS */
	if (fpga->variant->flags & FLAGS__HAS_BUTTON)
		button_remove(fpga);
err_button_setup:
	if (fpga->variant->flags & FLAGS__HAS_LCD)
		lcd_remove(class, &cdev_region, fpga);
err_lcd_setup:
	remove_char_device(fpga);
err_char_setup:
	pci_clear_master(pdev);
	remove_pci_device(pdev, fpga);
err_pci_setup:
	pci_release_regions(pdev);
err_request:
	pci_disable_device(pdev);
err_enable:
err_id:
	pci_set_drvdata(pdev, NULL);
	kfree(fpga);
err_mem:
	clear_bit(instance_nr, instance_map);
err_inst:
	return ret;
}


/**
 * Called by the OS when uninstalling the driver
 *
 */
static void hbm_fpga_pci_remove(struct pci_dev *pdev)
{
	struct hbm_fpga_dev  *fpga;
	int                  instance_nr;

	BUG_ON(!pdev);

	fpga = pci_get_drvdata(pdev);
	BUG_ON(!fpga);

	/* Cache the instance number */
	instance_nr = fpga->instance_nr;

	/* First disable irq */
	disable_irq(fpga->irq);

	if (fpga->variant->flags & FLAGS__HAS_SLOTID)
		device_remove_file(&pdev->dev, &dev_attr_slotid);
	if (fpga->variant->flags & FLAGS__HAS_BOOTMODE)
		device_remove_file(&pdev->dev, &dev_attr_bootmode);
	device_remove_file(&pdev->dev, &dev_attr_version);
	device_remove_file(&pdev->dev, &dev_attr_name);
	device_remove_file(&pdev->dev, &dev_attr_type);
#ifdef EXPOSE_REGISTERS
	device_remove_file(&pdev->dev, &dev_attr_scratch2);
	device_remove_file(&pdev->dev, &dev_attr_scratch1);
	device_remove_file(&pdev->dev, &dev_attr_bsr);
#endif /* EXPOSE_REGISTERS */
	if (fpga->variant->flags & FLAGS__HAS_BUTTON)
		button_remove(fpga);
	if (fpga->variant->flags & FLAGS__HAS_LCD)
		lcd_remove(class, &cdev_region, fpga);
	remove_char_device(fpga);
	pci_clear_master(pdev);
	remove_pci_device(pdev, fpga);
	pci_release_regions(pdev);
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
	kfree(fpga);
	clear_bit(instance_nr, instance_map);

	dev_info(&pdev->dev, "Removed HBM FPGA nr. %d\n", instance_nr);
}


/*

 ------------ OS driver plumbing --------------

*/

MODULE_AUTHOR("Copyright(C) 2011-2017 Hottinger Bruel & Kjaer Benelux B.V.");
MODULE_DESCRIPTION("HBM FPGA driver");
MODULE_LICENSE("GPL");


static const struct pci_device_id hbm_fpga_pci_tbl[] =
{
	{ 	.vendor      = PCI_VENDOR_ID_XILINX,
		.device      = PCI_DEVICE_ID_HBM,
		.subvendor   = PCI_SUBVENDOR_ID_XILINX,
		.subdevice   = PCI_SUBDEVICE_ID_HBM_GEN2x,
		.driver_data = (unsigned long)&fpga_variant[BACKPLANE_2x]
	},
	{ 	.vendor      = PCI_VENDOR_ID_XILINX,
		.device      = PCI_DEVICE_ID_HBM,
		.subvendor   = PCI_SUBVENDOR_ID_XILINX,
		.subdevice   = PCI_SUBDEVICE_ID_HBM_GEN3x,
		.driver_data = (unsigned long)&fpga_variant[BACKPLANE_3x]
	},
	{	.vendor      = PCI_VENDOR_ID_XILINX,
		.device      = PCI_DEVICE_ID_HBM,
		.subvendor   = PCI_SUBVENDOR_ID_XILINX,
		.subdevice   = PCI_SUBDEVICE_ID_HBM_GEN3xB,
		.driver_data = (unsigned long)&fpga_variant[BACKPLANE_3x]
	},
	{ 	.vendor      = PCI_VENDOR_ID_XILINX,
		.device      = PCI_DEVICE_ID_HBM,
		.subvendor   = PCI_SUBVENDOR_ID_XILINX,
		.subdevice   = PCI_SUBDEVICE_ID_HBM_GEN4x,
		.driver_data = (unsigned long)&fpga_variant[BACKPLANE_4x]
	},
	{ 	.vendor      = PCI_VENDOR_ID_XILINX,
		.device      = PCI_DEVICE_ID_HBM,
		.subvendor   = PCI_SUBVENDOR_ID_XILINX,
		.subdevice   = PCI_SUBDEVICE_ID_HBM_GEN7x,
		.driver_data = (unsigned long)&fpga_variant[BACKPLANE_7x]
	},
	{ 	.vendor      = PCI_VENDOR_ID_XILINX,
		.device      = PCI_DEVICE_ID_HBM,
		.subvendor   = PCI_SUBVENDOR_ID_XILINX,
		.subdevice   = PCI_SUBDEVICE_ID_HBM_GEN7xB,
		.driver_data = (unsigned long)&fpga_variant[BACKPLANE_7x]
	},
	{ 	.vendor      = PCI_VENDOR_ID_XILINX,
		.device      = PCI_DEVICE_ID_HBM,
		.subvendor   = PCI_SUBVENDOR_ID_XILINX,
		.subdevice   = PCI_SUBDEVICE_ID_HBM_GEN17P,
		.driver_data = (unsigned long)&fpga_variant[BACKPLANE_17P]
	},
	{ 	.vendor      = PCI_VENDOR_ID_XILINX,
		.device      = PCI_DEVICE_ID_HBM,
		.subvendor   = PCI_SUBVENDOR_ID_XILINX,
		.subdevice   = PCI_SUBDEVICE_ID_HBM_GEN17PB,
		.driver_data = (unsigned long)&fpga_variant[BACKPLANE_17P]
	},
	{ 	.vendor      = PCI_VENDOR_ID_XILINX,
		.device      = PCI_DEVICE_ID_HBM,
		.subvendor   = PCI_SUBVENDOR_ID_HBM,
		.subdevice   = PCI_SUBDEVICE_ID_HBM_OCC,
		.driver_data = (unsigned long)&fpga_variant[OPTIONCARRIER]
	},
	{ 	.vendor      = PCI_VENDOR_ID_XILINX,
		.device      = PCI_DEVICE_ID_HBM,
		.subvendor   = PCI_SUBVENDOR_ID_HBM,
		.subdevice   = PCI_SUBDEVICE_ID_HBM_OCCB,
		.driver_data = (unsigned long)&fpga_variant[OPTIONCARRIER]
	},
	{} /* Sentinel */
};
MODULE_DEVICE_TABLE(pci, hbm_fpga_pci_tbl);


static struct pci_driver hbm_fpga_pci_driver = {
	.name		= MODNAME,
	.id_table	= hbm_fpga_pci_tbl,
	.probe		= hbm_fpga_pci_probe,
	.remove		= hbm_fpga_pci_remove,
#ifdef CONFIG_PM
	.suspend	= hbm_fpga_pci_suspend,
	.resume		= hbm_fpga_pci_resume,
#endif /* CONFIG_PM */
};


/**
 * Called when this driver is loaded
 */
static int __init hbm_fpga_core_init_module(void)
{
	int  ret;

	/* When compiled as a module, this is printed whether or not devices are found in probe */
	printk(KERN_INFO MODNAME " driver v"DRIVER_VERSION" ("__DATE__" "__TIME__")\n");

	/* Create device class */
	class = class_create(THIS_MODULE, MODNAME);
	if (IS_ERR(class)) {
		ret = PTR_ERR(class);
		printk(KERN_ERR MODNAME "Error %d allocating device class\n", ret);
		goto err_class;
	}

	/* Dynamically allocate a major device ID with a range of minor device IDs */
	if ((ret = alloc_chrdev_region(&cdev_region, 0, DEV_NODES, MODNAME))) {
		printk(KERN_ERR MODNAME "Error %d allocating major/minor cdev range\n", ret);
		goto err_chrdev_region;
	}

	/* Register this driver in the PCI subsystem, allowing it to probe us with all supported devices found */
	if ((ret = pci_register_driver(&hbm_fpga_pci_driver))) {
		printk(KERN_ERR MODNAME "Error %d registering driver\n", ret);
		goto err_register;
	}

	return 0;

	pci_unregister_driver(&hbm_fpga_pci_driver);
err_register:
	unregister_chrdev_region(cdev_region, DEV_NODES);
err_chrdev_region:
	class_destroy(class);
err_class:
	return ret;
}


/**
 * Called when this driver is unloaded
 */
static void __exit hbm_fpga_core_cleanup_module(void)
{
	pci_unregister_driver(&hbm_fpga_pci_driver);
	unregister_chrdev_region(cdev_region, DEV_NODES);
	class_destroy(class);
}


module_init(hbm_fpga_core_init_module);
module_exit(hbm_fpga_core_cleanup_module);

