/*
 * EtherCAT Slave Controller (ESC) UIO driver for esd PCI/PCIe boards
 *
 * Copyright (c) 2010 - 2021
 * Matthias Fuchs <mf@esd.eu>               (out-of-date)
 * Thomas Koerper <thomas.koerper@esd.eu>   (out-of-date)
 * Manuel Koeppen <manuel.koeppen@esd.eu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/version.h>
#include <linux/device.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/uio_driver.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <linux/sysfs.h>

#ifdef CONFIG_X86
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0)
#include <asm/set_memory.h>
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
#include <asm/cacheflush.h>
#endif
#endif

#define DRV_NAME			"pexesc_uio"
#define DRV_VERSION			"2.2.4"

#define ESD_DRIVER_INFOSTR		"esd PCIe EtherCAT Slave Driver, uio, Version " DRV_VERSION

#define PCI_SUBDEVICE_ID_PEXESC		0x1100
#define PCI_SUBDEVICE_ID_XMCESC		0x0703

#ifndef PCI_DEVICE_ID_PLX_8111
#define PCI_DEVICE_ID_PLX_8111		0x8111
#endif
#ifndef PCI_DEVICE_ID_PLX_9056
#define PCI_DEVICE_ID_PLX_9056		0x9056
#endif

#define UIOMAP_PCIE1100_ESC		0
#define UIOMAP_PCIE1100_9056		1
#define UIOMAP_PCIE1100_8111		2

#define UIOMAP_XMCFPGA_UTIL		0
#define UIOMAP_XMCFPGA_ESC		1
#define UIOMAP_XMCFPGA_DMA		2

#define XMCFPGA_UTIL_OF_SYSID		0x0000
#define XMCFPGA_UTIL_OF_LEDS		0x0020
#define XMCFPGA_UTIL_OF_FLASH		0x0800
#define XMCFPGA_UTIL_OF_PCIEP		0x4000
#define XMCFPGA_ESC_OF_SYNC0		0x10000
#define XMCFPGA_ESC_OF_SYNC1		0x10001

#define XMCFPGA_IRQEN_IRQ_DEF		0x00000000u /* Def (to avoid read) */

#define XMCFPGA_IRQ_PDI			0x00000001u
#define XMCFPGA_IRQ_SYNC0		0x00000002u
#define XMCFPGA_IRQ_SYNC1		0x00000004u
#define XMCFPGA_IRQ_TIMER		0x00000008u
#define XMCFPGA_IRQ_FLASH		0x00000010u
#define XMCFPGA_IRQ_I2C			0x00000020u
#define XMCFPGA_IRQ_JTAGUART		0x00000040u
#define XMCFPGA_IRQ_PCIE		0x00000080u
#define XMCFPGA_IRQ_NIOSGEN		0x00000100u
#define XMCFPGA_IRQ_MAILBOX0		0x00010000u
#define XMCFPGA_IRQ_MAILBOX1		0x00020000u
#define XMCFPGA_IRQ_MAILBOX2		0x00040000u
#define XMCFPGA_IRQ_MAILBOX3		0x00080000u
#define XMCFPGA_IRQ_MAILBOX4		0x00100000u
#define XMCFPGA_IRQ_MAILBOX5		0x00200000u
#define XMCFPGA_IRQ_MAILBOX6		0x00400000u
#define XMCFPGA_IRQ_MAILBOX7		0x00800000u

#define XMCFPGA_IRQS			(XMCFPGA_IRQ_PDI)
//#define XMCFPGA_IRQS			(XMCFPGA_IRQ_PDI|XMCFPGA_IRQ_SYNC0|XMCFPGA_IRQ_SYNC1)

#define XMCFPGA_IRQEN_IRQ_ON		(XMCFPGA_IRQEN_IRQ_DEF | XMCFPGA_IRQS)
#define XMCFPGA_IRQEN_IRQ_OFF		XMCFPGA_IRQEN_IRQ_DEF

/* PLX9056 registers */
#define PCIE1100_PLX_LMISC2		0x0f
#define PCIE1100_PLX_RDYTIMEOUTEN	0x01
#define PCIE1100_PLX_RDYTIMEOUTSEL	0x02
#define PCIE1100_PLX_INTCSR		0x68 /* Interrupt Control/Status */
#define PCIE1100_PLX_PCI_INT_EN		(1 << 8)
#define PCIE1100_PLX_LINTI		(1 << 11)
#define PCIE1100_PLX_LINTI_STATUS	(1 << 15)
#define PCIE1100_LINTI_ENACT		(PCIE1100_PLX_PCI_INT_EN | \
						PCIE1100_PLX_LINTI | \
						PCIE1100_PLX_LINTI_STATUS)
#define PCIE1100_PLX_CNTRL		0x6c      /* Control / Softw. Reset */
#define PCIE1100_PLX_PCI_RCR		(1 << 29) /* Read Config Registers */
#define PCIE1100_PLX_PCI_RESET		(1 << 30) /* PCI Adapt. Softw. Reset */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
#define __devinit
#define __devexit
#define __devinitdata
#endif

enum CARD_TYPE {
	CARD_TYPE_UNKNOWN = 0,
	CARD_TYPE_ECSPCIe1100,
	CARD_TYPE_ECSXMCFPGA,
};

#define ESD_CARD_TYPE_TO_STR(ct) (\
			((ct) == CARD_TYPE_ECSPCIe1100) ? "ECS-PCIe/1100" : \
			((ct) == CARD_TYPE_ECSXMCFPGA) ? "ECS-xMC/FPGA" : \
			"Unknown Card!")


struct card_info {
	int pic_ids_idx;
	enum CARD_TYPE card_type;
	const char* name;
};

static struct pci_device_id pexesc_pci_ids[] = {
	{
		.vendor =	PCI_VENDOR_ID_PLX,
		.device =	PCI_DEVICE_ID_PLX_9056,
		.subvendor =	PCI_VENDOR_ID_ESDGMBH,
		.subdevice =	PCI_SUBDEVICE_ID_PEXESC,
	},
	{
		.vendor =	PCI_VENDOR_ID_ESDGMBH,
		.device =	PCI_SUBDEVICE_ID_XMCESC,
		.subvendor =	PCI_VENDOR_ID_ESDGMBH,
		.subdevice =	PCI_SUBDEVICE_ID_XMCESC,
	},
	{ 0, }
};

static const struct card_info card_infos[] = {
	{
		.pic_ids_idx = 0,
		.card_type = CARD_TYPE_ECSPCIe1100,
		.name = "esd ECS-PCIe/1100",
	},
	{
		.pic_ids_idx = 1,
		.card_type = CARD_TYPE_ECSXMCFPGA,
		.name = "esd ECS-xMC/FPGA",
	},
};

struct uio_info_priv {
	enum CARD_TYPE card_type;
	struct pci_dev *bridge;
	bool have_msi;
	phys_addr_t     dma_addr;
	resource_size_t dma_size;
	void           *dma_virt;
};

static bool nomsi = false;
module_param(nomsi, bool, 0);
MODULE_PARM_DESC(nomsi, "Prevent use of MSI");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
void __iomem *pci_ioremap_bar(struct pci_dev *pdev, int bar)
{
	/*
	 * Make sure the BAR is actually a memory resource, not an IO resource
	 */
	if (!(pci_resource_flags(pdev, bar) & IORESOURCE_MEM)) {
		WARN_ON(1);
		return NULL;
	}
	return ioremap_nocache(pci_resource_start(pdev, bar),
			       pci_resource_len(pdev, bar));
}
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28) */

static int map_uio_phys(struct pci_dev *dev, int bar, struct uio_mem *dest,
			const char* name)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,30)
	dest->name = name;
#endif
	dest->addr = pci_resource_start(dev, bar);
	if (!dest->addr)
		return -1;
	dest->internal_addr = pci_ioremap_bar(dev, bar);
	if (!dest->internal_addr)
		return -1;

	dest->size = pci_resource_len(dev, bar);
	dest->memtype = UIO_MEM_PHYS;

	return 0;
}

static ssize_t pexesc_cardtype_show(struct device *d,
				    struct device_attribute *attr, char *buf)
{
	struct pci_dev *dev = to_pci_dev(d);
	struct uio_info *info = pci_get_drvdata(dev);
	struct uio_info_priv *priv = info->priv;

	return scnprintf(buf, PAGE_SIZE, "%u\n", priv->card_type);
}
static DEVICE_ATTR(pexesc_cardtype, S_IRUGO, pexesc_cardtype_show, NULL);

static ssize_t pexesc_buildinfo_show(struct device *d,
				     struct device_attribute *attr, char *buf)
{
	struct pci_dev *dev = to_pci_dev(d);
	struct uio_info *info = pci_get_drvdata(dev);
	struct uio_info_priv *priv = info->priv;

	return scnprintf(buf, PAGE_SIZE, "%s, %s\n", ESD_DRIVER_INFOSTR,
			 ESD_CARD_TYPE_TO_STR(priv->card_type));
}
static DEVICE_ATTR(pexesc_buildinfo, S_IRUGO, pexesc_buildinfo_show, NULL);

static irqreturn_t pcie1100_interrupt(int irq, struct uio_info *info)
{
	void __iomem *intscr = info->mem[UIOMAP_PCIE1100_9056].internal_addr +
				PCIE1100_PLX_INTCSR;
	u32 status = ioread32(intscr);

	if ((status & PCIE1100_LINTI_ENACT) != PCIE1100_LINTI_ENACT)
		return IRQ_NONE;

	/* Disable interrupt */
	iowrite32(status & ~PCIE1100_PLX_PCI_INT_EN, intscr);
	return IRQ_HANDLED;
}

static irqreturn_t xmcfpga_interrupt(int irq, struct uio_info *info)
{
	void __iomem *pciep = info->mem[UIOMAP_XMCFPGA_UTIL].internal_addr +
				XMCFPGA_UTIL_OF_PCIEP;
	const u32 st = ioread32(pciep + 0x0040);

	if ((st & XMCFPGA_IRQS) == 0u)
		return IRQ_NONE;
	if(!((struct uio_info_priv *)info->priv)->have_msi) {
		iowrite32(XMCFPGA_IRQEN_IRQ_OFF, pciep + 0x0050); /* disable all irqs */
	}
	return IRQ_HANDLED;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,30)
static int pcie1100_irqcontrol(struct uio_info *info, s32 irq_on)
{
	u32 val;
	void __iomem *intscr = info->mem[UIOMAP_PCIE1100_9056].internal_addr +
				PCIE1100_PLX_INTCSR;

	/*
	 * Enable/disable interrupts from bridge and Local_1
	 */
	val = ioread32(intscr);
	if (irq_on)
		val |= PCIE1100_PLX_LINTI | PCIE1100_PLX_PCI_INT_EN;
	else
		val &= ~(PCIE1100_PLX_LINTI | PCIE1100_PLX_PCI_INT_EN);

	iowrite32(val, intscr);

	return 0;
}

static int xmcfpga_irqcontrol(struct uio_info *info, s32 irq_on)
{
	void __iomem *pciep = info->mem[UIOMAP_XMCFPGA_UTIL].internal_addr +
		XMCFPGA_UTIL_OF_PCIEP;

	iowrite32((irq_on) ? XMCFPGA_IRQEN_IRQ_ON : XMCFPGA_IRQEN_IRQ_OFF,
		pciep + 0x0050);
	return 0;
}
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,30) */

static void pcie1100_reset(struct uio_info *info)
{
	void __iomem *conf = info->mem[UIOMAP_PCIE1100_9056].internal_addr;
	u32 cntrl;

	/* issue a local bus reset */
	cntrl = ioread32(conf + PCIE1100_PLX_CNTRL);
	cntrl |= PCIE1100_PLX_PCI_RESET;
	iowrite32(cntrl, conf + PCIE1100_PLX_CNTRL);
	udelay(100);
	cntrl ^= PCIE1100_PLX_PCI_RESET;
	iowrite32(cntrl, conf + PCIE1100_PLX_CNTRL);

	/* reload local configuration from EEPROM */
	cntrl |= PCIE1100_PLX_PCI_RCR;
	iowrite32(cntrl, conf + PCIE1100_PLX_CNTRL);

	/*
	 * There is no safe way to poll for the end
	 * of reconfiguration process...
	 */
	mdelay(100);

	cntrl ^= PCIE1100_PLX_PCI_RCR;
	iowrite32(cntrl, conf + PCIE1100_PLX_CNTRL);

	/* configure 9056 for localbus timeout (1024 cylces) */
	iowrite8(PCIE1100_PLX_RDYTIMEOUTEN | PCIE1100_PLX_RDYTIMEOUTSEL,
		 conf + PCIE1100_PLX_LMISC2);
}

#if 0
static void xmcfpga_reset(struct uio_info *info)
{
	struct uio_info_priv *priv = info->priv;

	(void)priv;
	/* TODO: do we need an xmcfpga_reset()? */
}
#endif

static int card_probe_esc(struct pci_dev *dev, void __iomem *esc)
{
	u8 type;
	u8 fmmu;
	u8 sm;
	u8 ram;
	int i;

	if (!esc)
		return -1;

	i = 10;
	do {
		type = ioread8(esc);
		fmmu = ioread8(esc + 4);
		sm = ioread8(esc + 5);
		ram = ioread8(esc + 6);
		if (type != 0xff) {
			dev_info(&dev->dev,
				 "ESC Type 0x%.2x, %u kB, %u SMs, %u FMMUs\n",
				 type, ram, sm, fmmu);
			return 0;
		}

		mdelay(10);
	} while (i--);

	/* Might return err here, but as stack also probes... */
	dev_warn(&dev->dev, "failed to probe ESC!\n");
	return 0;
}

/* card_finish() is called on error! */
static int card_init(struct pci_dev *dev)
{
	struct uio_info *info = pci_get_drvdata(dev);
	struct uio_info_priv *priv = info->priv;
	void __iomem *esc;

	switch (priv->card_type) {
	case CARD_TYPE_ECSPCIe1100:
	{
		/* Get 8111 with our 9056 on secondary bus to map its BAR0: */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
		priv->bridge = pci_get_domain_bus_and_slot(pci_domain_nr(dev->bus), dev->bus->number - 1, PCI_DEVFN(0, 0));
#else
		priv->bridge = pci_get_bus_and_slot(dev->bus->number - 1, PCI_DEVFN(0, 0));
#endif
		if (!priv->bridge) {
			dev_err(&dev->dev, "Can't get bridge!\n");
			return -1;
		}

		if ((priv->bridge->vendor != PCI_VENDOR_ID_PLX) ||
			(priv->bridge->device != PCI_DEVICE_ID_PLX_8111)) {
			dev_err(&dev->dev,
				"Unexpected bridge! (0x%.4x/0x%.4x)\n",
				priv->bridge->vendor, priv->bridge->device);
			return -1;
		}

		if (pci_request_regions(priv->bridge, "pexesc")) {
			dev_err(&dev->dev,
				"pci_request_regions() for bridge failed!\n");
			priv->bridge = NULL;
			return -1;
		}

		if (map_uio_phys(priv->bridge, 0,
				 &info->mem[UIOMAP_PCIE1100_8111], "8111")) {
			dev_err(&dev->dev,
				"map_uio_phys() for bridge failed!\n");
			return -1;
		}

		if (map_uio_phys(dev, 2,
				 &info->mem[UIOMAP_PCIE1100_ESC], "esc")) {
			dev_err(&dev->dev,
				"map_uio_phys() for BAR%u failed!\n", 2);
			return -1;
		}

		if (map_uio_phys(dev, 0,
				&info->mem[UIOMAP_PCIE1100_9056], "9056")) {
			dev_err(&dev->dev,
				"map_uio_phys() for BAR%u failed!\n", 0);
			return -1;

		}

		pcie1100_reset(info);

		info->handler = pcie1100_interrupt;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,30)
		info->irqcontrol = pcie1100_irqcontrol;
#endif

		esc = info->mem[UIOMAP_PCIE1100_ESC].internal_addr;
		break;
	} /* case CARD_TYPE_ECSPCIe1100 */

	case CARD_TYPE_ECSXMCFPGA:
	{
		if (map_uio_phys(dev, 0,
				 &info->mem[UIOMAP_XMCFPGA_UTIL], "util")) {
			dev_err(&dev->dev,
				"map_uio_phys() for BAR%u failed!\n", 0);
			return -1;
		}
		if (map_uio_phys(dev, 1,
				 &info->mem[UIOMAP_XMCFPGA_ESC], "esc")) {
			dev_err(&dev->dev,
				"map_uio_phys() for BAR%u failed!\n", 1);
			return -1;
		}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,30)
			info->mem[UIOMAP_XMCFPGA_DMA].name = "dma";
#endif
			priv->dma_size = 0x40000;
#if 0
			priv->dma_virt=dma_alloc_coherent(&dev->dev, priv->dma_size, &priv->dma_addr, GFP_KERNEL|GFP_DMA);
			if (!priv->dma_virt) {
				dev_err(&dev->dev,"dma_alloc_coherent() failed!\n");
				return -1;
			}
#else
			priv->dma_virt=pci_alloc_consistent(dev,  priv->dma_size, &priv->dma_addr);
			if (!priv->dma_virt) {
				dev_err(&dev->dev,"pci_alloc_consistent() failed!\n");
				return -1;
			}
#endif
#if defined CONFIG_X86 && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25))
			set_memory_uc((unsigned long)priv->dma_virt, priv->dma_size >> PAGE_SHIFT);
#endif
			info->mem[UIOMAP_XMCFPGA_DMA].memtype       = UIO_MEM_PHYS;
			info->mem[UIOMAP_XMCFPGA_DMA].size          = priv->dma_size / 2;
			info->mem[UIOMAP_XMCFPGA_DMA].internal_addr = priv->dma_virt;
			info->mem[UIOMAP_XMCFPGA_DMA].addr          = priv->dma_addr;
			{
				resource_size_t align=priv->dma_addr & ((priv->dma_size / 2)-1);
				if(align) { /* not aligned */
					align = (priv->dma_size / 2) - align;
					info->mem[UIOMAP_XMCFPGA_DMA].addr += align;
					info->mem[UIOMAP_XMCFPGA_DMA].internal_addr += align;
				}
			}
			{
				void __iomem *pciep = info->mem[UIOMAP_XMCFPGA_UTIL].internal_addr + XMCFPGA_UTIL_OF_PCIEP;
				u64 m = info->mem[UIOMAP_XMCFPGA_DMA].addr;
				/* Bits0-1 are special and should be 0 here */
				/* m = m & ~3ull; */				/* this is taken care of by the alignment already */
				if(m & 0xffffffff00000000ull) {
					m |= 1ull; /* 64 bit translation */
				}
				iowrite32((u32)m, pciep + 0x1000);          /* LSB */
				iowrite32((u32)(m>>32u), pciep + 0x1004);   /* MSB */
			}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12))
		pci_set_master(dev);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,1)
		if (nomsi) {
			dev_info(&dev->dev,"MSI not enabled because of nomsi option!\n");
		} else if (pci_enable_msi(dev)) {
			dev_info(&dev->dev,"pci_enable_msi failed!\n");
		} else {
			priv->have_msi = true;
		}
#endif
		info->handler = xmcfpga_interrupt;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,30)
		info->irqcontrol = xmcfpga_irqcontrol;
#endif

		esc = info->mem[UIOMAP_XMCFPGA_ESC].internal_addr;
		break;
	} /* case CARD_TYPE_ECSXMCFPGA */

	default:
		return 0;
	}

	return card_probe_esc(dev, esc);
}

static void card_finish(struct pci_dev *dev)
{
	struct uio_info *info = pci_get_drvdata(dev);
	struct uio_info_priv *priv = info->priv;

	switch (priv->card_type) {
	case CARD_TYPE_ECSPCIe1100:
	{
		if (priv->bridge) {
			pci_release_regions(priv->bridge);
			priv->bridge = NULL;
		}
		if (info->mem[UIOMAP_PCIE1100_ESC].internal_addr)
			iounmap(info->mem[UIOMAP_PCIE1100_ESC].internal_addr);
		if (info->mem[UIOMAP_PCIE1100_9056].internal_addr)
			iounmap(info->mem[UIOMAP_PCIE1100_9056].internal_addr);
		if (info->mem[UIOMAP_PCIE1100_8111].internal_addr)
			iounmap(info->mem[UIOMAP_PCIE1100_8111].internal_addr);
		break;
	} /* case CARD_TYPE_ECSPCIe1100 */

	case CARD_TYPE_ECSXMCFPGA:
	{
		if (info->mem[UIOMAP_XMCFPGA_UTIL].internal_addr)
			iounmap(info->mem[UIOMAP_XMCFPGA_UTIL].internal_addr);
		if (info->mem[UIOMAP_XMCFPGA_ESC].internal_addr)
			iounmap(info->mem[UIOMAP_XMCFPGA_ESC].internal_addr);
		if (priv->dma_virt) {
#if defined CONFIG_X86 && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25))
			set_memory_wb((unsigned long)priv->dma_virt, priv->dma_size >> PAGE_SHIFT);
#endif
#if 0
			dma_free_coherent(&dev->dev,
#else
			pci_free_consistent(dev,
#endif
				priv->dma_size,
				priv->dma_virt,
				priv->dma_addr);
		}
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12))
		pci_clear_master(dev);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,1)
		if (priv->have_msi) {
			pci_disable_msi(dev);
		}
#endif
		break;
	} /* case CARD_TYPE_ECSXMCFPGA */

	default:
		break;
	}
}

static enum CARD_TYPE card_detect(struct pci_dev *dev,
				  const struct pci_device_id *id)
{
	int i = (int)(sizeof(card_infos) / sizeof(card_infos[0]));

	while (i--) {
		const struct pci_device_id * const p =
				&pexesc_pci_ids[card_infos[i].pic_ids_idx];

		if ((id->vendor == p->vendor) && (id->device == p->device) &&
			(id->subvendor == p->subvendor) &&
			(id->subdevice == p->subdevice)) {

			dev_info(&dev->dev, "detected %s\n",
				 card_infos[i].name);
			return card_infos[i].card_type;
		}
	}

	dev_warn(&dev->dev, "unknown card! (0x%.4x/0x%.4x, 0x%.4x/0x%.4x)\n",
		 id->vendor, id->device,
		 id->subvendor, id->subdevice);
	return CARD_TYPE_UNKNOWN;
}

static int __devinit pexesc_pci_probe(struct pci_dev *dev,
				      const struct pci_device_id *id)
{
	struct uio_info_priv *priv;
	struct uio_info *info;

	info = kzalloc(sizeof(*info) + sizeof(*priv), GFP_KERNEL);
	if (!info)
		return -ENOMEM;

	priv = (struct uio_info_priv *)(info + 1);
	info->priv = priv;

	if (pci_enable_device(dev))
		goto out_free;

	if (pci_request_regions(dev, "pexesc"))
		goto out_disable_dev;

	priv->card_type = card_detect(dev, id);
	priv->have_msi = false;
	info->name = "pexesc";
	info->version = DRV_VERSION;
	pci_set_drvdata(dev, info);

	if (card_init(dev))
		goto out_finish;

	info->irq = dev->irq;
	info->irq_flags = IRQF_SHARED;

	if (device_create_file(&dev->dev, &dev_attr_pexesc_cardtype)) {
		dev_err(&dev->dev, "device_create_file() failed!\n");
		goto out_finish;
	}
	if (device_create_file(&dev->dev, &dev_attr_pexesc_buildinfo)) {
		dev_err(&dev->dev, "device_create_file() failed!\n");
		goto out_remattr_cardtype;
	}

	if (uio_register_device(&dev->dev, info)) {
		dev_err(&dev->dev, "uio_register_device() failed!\n");
		goto out_remattr;
	}

	return 0;

out_remattr:
	device_remove_file(&dev->dev, &dev_attr_pexesc_buildinfo);
out_remattr_cardtype:
	device_remove_file(&dev->dev, &dev_attr_pexesc_cardtype);
out_finish:
	card_finish(dev);
	pci_release_regions(dev);
out_disable_dev:
	pci_disable_device(dev);
out_free:
	kfree(info);

	return -ENODEV;
}

static void __devexit pexesc_pci_remove(struct pci_dev *dev)
{
	struct uio_info *info = pci_get_drvdata(dev);

	device_remove_file(&dev->dev, &dev_attr_pexesc_cardtype);
	device_remove_file(&dev->dev, &dev_attr_pexesc_buildinfo);

	uio_unregister_device(info);
	card_finish(dev);
	pci_release_regions(dev);
	pci_disable_device(dev);

	pci_set_drvdata(dev, NULL);
	kfree(info);
}

static struct pci_driver pexesc_pci_driver = {
	.name = DRV_NAME,
	.id_table = pexesc_pci_ids,
	.probe = pexesc_pci_probe,
	.remove = pexesc_pci_remove,
};

static int __init pexesc_init(void)
{
	pr_info(ESD_DRIVER_INFOSTR);
	return pci_register_driver(&pexesc_pci_driver);
}
module_init(pexesc_init);

static void __exit pexesc_exit(void)
{
	pci_unregister_driver(&pexesc_pci_driver);
}
module_exit(pexesc_exit);

MODULE_LICENSE("GPL v2");
MODULE_VERSION(DRV_VERSION);
MODULE_AUTHOR("Manuel Koeppen <support@esd.eu>");

