/*
*       Copyright (c) 2011-2022 by esd electronics gmbh
*
*  This software is copyrighted by and is the sole property of
*  esd gmbh.  All rights, title, ownership, or other interests
*  in the software remain the property of esd gmbh. This
*  software may only be used in accordance with the corresponding
*  license agreement.  Any unauthorized use, duplication, transmission,
*  distribution, or disclosure of this software is expressly forbidden.
*
*  This Copyright notice may not be removed or modified without prior
*  written consent of esd gmbh.
*
*  esd gmbh, reserves the right to modify this software without notice.
*
*  esd electronics gmbh                   Tel. +49-511-37298-0
*  Vahrenwalder Str 207                   Fax. +49-511-37298-68
*  30165 Hannover                         http://www.esd.eu
*  Germany                                sales@esd.eu
*
*************************************************************************
*
* Description:
* =============================================================================
* This module contains the main stack code.
*
*/

#include "../include/essSystem.h"
#include "../include/essPrivate.h"
#include "../hal/essHAL.h"
#include "../include/ess.h"

    /* Exceptions from the esd PC-lint/FlexeLint policy for this file */
/*lint -e697  [Warn: Quasi-booleans should be equality-compared only with 0] */
/*lint -e717  [Info: do ... while(0)]                                        */
/*lint -e750  [Info: local macro not referenced]                             */
/*lint -e801  [Info: Use of goto is deprecated]                              */

                                                    /* 1 for DBGs for every write access to ESC memory */
#define DEBUG_ESCWRITES                             0

                                                    /* 1 to show SM changes */
#define DEBUG_SYNCMANAGERS                          0

                                                    /* 1 to get DBGs when send mbx is written/read-out, etc. */
#define DEBUG_MAILBOX_TRANSFER                      0

                                                    /* 1 to have a debug output for each (that's many!) interrupt */
#define DEBUG_INTERRUPTS                            0

                                                    /* 1 to print every AL Event Request (0x0220) change */
#define DEBUG_AL_EVENT                              0

                                                    /* 1 for debug output with duration to handle AL Event Request */
#define DEBUG_AL_EVENT_DURATION                     0

                                                    /* 1 for debug output for CoE stuff, such as SDO Requests, etc. */
#define DEBUG_COE                                   0

                                                    /* 1 for debug output when updating a served LED */
#define DEBUG_LEDSERVE                              0

                                                    /* 1 for debug output for events related to distributed clocks */
#define DEBUG_DC                                    0

                                                    /* 1 for debug output for each EEPROM access etc. */
#define DEBUG_EEPROM_EMULATION                      0

                                                    /* 1 to enable minor debug outputs for ED_DO_EVENT() usage */
#define DEBUG_EVENTS                                0




#define DELAY(milliSeconds)                         do { const ESS_TIMESTAMP _tsStart = essHALGetTime(); while ((essHALGetTime() - _tsStart) < (milliSeconds)) ; } while (0)
#define DO_INOUT_ACTIV_CB(ped, inpts, starting)     do { ESS_CBDATA_INOUTPUTS_ACTIVATE _cb; _cb.hDev = (ESS_HANDLE)(ped); _cb.inputs = (inpts); _cb.started = (starting); ED_DO_CALLBACK(ped, cbInOutputsActivate, &_cb); } while (0)

#if (CFG_ESS_SERVE_ERR_LED)
# define SET_ERROR_LED(ped, ledState)               LED_SetErrLED((ped), (ledState))
#else
# define SET_ERROR_LED(unused1, unused2)            do { } while (0)
#endif

#if (CFG_ESS_SERVE_RUN_LED)
# define SET_RUN_LED(ped, ledState)                 LED_SetRunLED((ped), (ledState))
#else
# define SET_RUN_LED(unused1, unused2)              do { } while (0)
#endif

#if (CFG_ESC_HAVE_POINTER)
# if (CFG_ESC_HAVE_READ_FUNC)
#  define ESC_READ(ped, ad, pDst, len) do { essHALReadESCMem((ped)->devIdx, ad, pDst, len); } while (0)
# else
#  define ESC_READ(ped, ad, pDst, len)                          \
    do {                                                        \
      if ((len) > 0) {                                          \
        void* _src = (uint8_t*)((ped)->escPointer) + (ad);      \
        void* _dst = (pDst);                                    \
        uint16_t _len = (len);                                  \
        if (((uintptr_t)_src & 1) && ((uintptr_t)_dst & 1)) {   \
          *((uint8_t*)_dst) = *((volatile uint8_t*)_src);       \
          _len--;                                               \
          _dst = (uint8_t*)_dst + 1;                            \
          _src = (uint8_t*)_src + 1;                            \
        }                                                       \
        if (((uintptr_t)_src & 1) || ((uintptr_t)_dst & 1)) {   \
          while (_len > 0) {                                    \
            *((uint8_t*)_dst) = *((volatile uint8_t*)_src);     \
            _len--;                                             \
            _dst = (uint8_t*)_dst + 1;                          \
            _src = (uint8_t*)_src + 1;                          \
          }                                                     \
        } else {                                                \
          while (_len > 1) {                                    \
            *((uint16_t*)_dst) = *((volatile uint16_t*)_src);   \
            _len -= 2;                                          \
            _dst = (uint16_t*)_dst + 1;                         \
            _src = (uint16_t*)_src + 1;                         \
        }                                                       \
        if (_len > 0)                                           \
          *((uint8_t*)_dst) = *((volatile uint8_t*)_src);       \
        }                                                       \
      }                                                         \
    } while (0)
# endif

# if (CFG_ESC_HAVE_WRITE_FUNC)
#  define ESC_WRITE(ped, ad, pSrc, len)              do { essHALWriteESCMem((ped)->devIdx, ad, pSrc, len); ESS_DBG_ESCWRITEX(ad, pSrc, len); } while (0)
# else
#  define ESC_WRITE(ped, ad, pSrc, len)                              \
    do {                                                            \
      if ((len) > 0) {                                              \
        const void* _src = (pSrc);                                  \
        void* _dst = (uint8_t*)((ped)->escPointer) + (ad);          \
        uint16_t _len = (len);                                      \
        if (((uintptr_t)_src & 1) && ((uintptr_t)_dst & 1)) {       \
          *((volatile uint8_t*)_dst) = *((const uint8_t*)_src);     \
          _len--;                                                   \
          _dst = (uint8_t*)_dst + 1;                                \
          _src = (const uint8_t*)_src + 1;                          \
        }                                                           \
        if (((uintptr_t)_src & 1) || ((uintptr_t)_dst & 1)) {       \
          while (_len > 0) {                                        \
            *((volatile uint8_t*)_dst) = *((const uint8_t*)_src);   \
            _len--;                                                 \
            _dst = (uint8_t*)_dst + 1;                              \
            _src = (const uint8_t*)_src + 1;                        \
          }                                                         \
        } else {                                                    \
          while (_len > 1) {                                        \
            *((volatile uint16_t*)_dst) = *((const uint16_t*)_src); \
            _len -= 2;                                              \
            _dst = (uint16_t*)_dst + 1;                             \
            _src = (const uint16_t*)_src + 1;                       \
          }                                                         \
          if (_len > 0)                                             \
            *((volatile uint8_t*)_dst) = *((const uint8_t*)_src);   \
        }                                                           \
      }                                                             \
      ESS_DBG_ESCWRITEX(ad, pSrc, len);                             \
    } while (0)
# endif

# define ESC_READ8(ped, ad, dest)                   do { (dest) = *((volatile uint8_t*)((ped)->escPointer) + (ad)); } while (0)
# define ESC_WRITE8(ped, ad, val)                   do { *((volatile uint8_t*)((ped)->escPointer) + (ad)) = (uint8_t)(val); ESS_DBG_ESCWRITE("0x%.2x", ad, val); } while (0)
# define ESC_READ16_NOSWAP(ped, ad, dest)           do { ESS_ASSERT(((ad) & 1) == 0, ("%s (0x%.4x)", "Unaligned access!", (ad))); (dest) = *((volatile uint16_t*)((ped)->escPointer) + ((ad) >> 1)); } while (0)
# define ESC_READ32_NOSWAP(ped, ad, dest)           do { ESS_ASSERT(((ad) & 3) == 0, ("%s (0x%.4x)", "Unaligned access!", (ad))); (dest) = *((volatile uint32_t*)((ped)->escPointer) + ((ad) >> 2)); } while (0)
# define ESC_WRITE16_NOSWAP(ped, ad, val)           do { ESS_ASSERT(((ad) & 1) == 0, ("%s (0x%.4x)", "Unaligned access!", (ad))); *((volatile uint16_t*)((ped)->escPointer) + ((ad) >> 1)) = (val); } while (0)
# define ESC_WRITE32_NOSWAP(ped, ad, val)           do { ESS_ASSERT(((ad) & 3) == 0, ("%s (0x%.4x)", "Unaligned access!", (ad))); *((volatile uint32_t*)((ped)->escPointer) + ((ad) >> 2)) = (val); } while (0)

#else /* Don't have ESC pointer: */
# define ESC_READ(ped, ad, pDst, len)               do { essHALReadESCMem((ped)->devIdx, ad, pDst, len); } while (0)
# define ESC_WRITE(ped, ad, pSrc, len)              do { essHALWriteESCMem((ped)->devIdx, ad, pSrc, len); ESS_DBG_ESCWRITEX(ad, pSrc, len); } while (0)

# define ESC_READ8(ped, ad, dest)                   do { uint8_t _temp = 0; essHALReadESCMem((ped)->devIdx, ad, &_temp, 1); (dest) = _temp; } while (0)
# define ESC_WRITE8(ped, ad, val)                   do { uint8_t _temp = (val); essHALWriteESCMem((ped)->devIdx, ad, &_temp, 1); ESS_DBG_ESCWRITE("0x%.2x", ad, val); } while (0)

# define ESC_READ16_NOSWAP(ped, ad, dest)           do { uint16_t _temp = 0; essHALReadESCMem((ped)->devIdx, ad, &_temp, 2); (dest) = _temp; } while (0)
# define ESC_READ32_NOSWAP(ped, ad, dest)           do { uint32_t _temp = 0; essHALReadESCMem((ped)->devIdx, ad, &_temp, 4); (dest) = _temp; } while (0)
# define ESC_WRITE16_NOSWAP(ped, ad, val)           do { uint16_t _temp = (val); essHALWriteESCMem((ped)->devIdx, ad, &_temp, 2); } while (0)
# define ESC_WRITE32_NOSWAP(ped, ad, val)           do { uint32_t _temp = (val); essHALWriteESCMem((ped)->devIdx, ad, &_temp, 4); } while (0)
#endif

#if (CFG_BIG_ENDIAN)
# define ESC_READ16(ped, ad, dest)                  do { ESC_READ16_NOSWAP(ped, ad, dest); (dest) = SWAP16(dest); } while (0)
# define ESC_READ32(ped, ad, dest)                  do { ESC_READ32_NOSWAP(ped, ad, dest); (dest) = SWAP32(dest); } while (0)
# define ESC_WRITE16(ped, ad, val)                  do { ESC_WRITE16_NOSWAP(ped, ad, SWAP16(val)); } while (0)
# define ESC_WRITE32(ped, ad, val)                  do { ESC_WRITE32_NOSWAP(ped, ad, SWAP32(val)); } while (0)
#else
# define ESC_READ16(ped, ad, dest)                  do { ESC_READ16_NOSWAP(ped, ad, dest); } while (0)
# define ESC_READ32(ped, ad, dest)                  do { ESC_READ32_NOSWAP(ped, ad, dest); } while (0)
# define ESC_WRITE16(ped, ad, val)                  do { ESC_WRITE16_NOSWAP(ped, ad, val); ESS_DBG_ESCWRITE("0x%.4x", ad, val); } while (0)
# define ESC_WRITE32(ped, ad, val)                  do { ESC_WRITE32_NOSWAP(ped, ad, val); ESS_DBG_ESCWRITE("0x%.8x", ad, val); } while (0)
#endif

#if (DEBUG_ESCWRITES)
# define ESS_DBG_ESCWRITEX(ad, data, len)           do { ESS_DBG(("~~~ Wrote %u bytes to ESC address 0x%.4x:", (unsigned int)(len), (unsigned short)(ad))); { int _i; for (_i = 0; _i < (len); ) { ESS_DBG((" %.2x", ((uint8_t*)data)[_i])); if ((++_i % 16) == 0) ESS_DBG(("\n")); } ESS_DBG(("\n")); }  } while (0)
# define ESS_DBG_ESCWRITE(fmt, ad, val)             ESS_DBG(("~~~ Wrote " fmt " to ESC register 0x%.4x ~~~", (unsigned int)(val), (unsigned short)(ad)))
#else
# define ESS_DBG_ESCWRITEX(ad, data, len)
# define ESS_DBG_ESCWRITE(fmt, ad, val)
#endif

#if (DEBUG_SYNCMANAGERS)
# define ESS_DBG_SYNCMANAGERS(x)                    ESS_DBG(x)
#else
# define ESS_DBG_SYNCMANAGERS(x)                    do { } while (0)
#endif

#if (DEBUG_MAILBOX_TRANSFER)
# define ESS_DBG_MAILBOX_TRANSFER(x)                ESS_DBG(x)
#else
# define ESS_DBG_MAILBOX_TRANSFER(x)                do { } while (0)
#endif

#if 0 /* Only a light version of DEBUG_MAILBOX_TRANSFER */
# define ESS_DBG_MBX_MINIMAL(x)                     do { ESS_PRINTF("[MBX_DBG] "); ESS_PRINTF x; ESS_PRINTF("\n"); } while (0)
#else
# define ESS_DBG_MBX_MINIMAL(x)                     do { } while (0)
#endif

#if (DEBUG_INTERRUPTS)
# define ESS_DBG_INTERRUPTS(x)                      ESS_DBG(x)
#else
# define ESS_DBG_INTERRUPTS(x)                      do { } while (0)
#endif

#if (DEBUG_LEDSERVE)
# define ESS_DBG_LEDSERVE(x)                        ESS_DBG(x)
#else
# define ESS_DBG_LEDSERVE(x)                        do { } while (0)
#endif

#if (DEBUG_DC)
# define ESS_DBG_DC(x)                              ESS_DBG(x)
#else
# define ESS_DBG_DC(x)                              do { } while (0)
#endif

#if (DEBUG_EEPROM_EMULATION)
# define ESS_DBG_EEE(x)                             ESS_DBG(x)
#else
# define ESS_DBG_EEE(x)                             do { } while (0)
#endif

#if (DEBUG_EVENTS)
# define ESS_DBG_EVENTS(x)                          ESS_DBG(x)
#else
# define ESS_DBG_EVENTS(x)                          do { } while (0)
#endif


/* Macros used for special "code injection" - shouldn't be needed! */
#ifndef ESS_DO_MAINLOOP_READALEVENT
# define ESS_DO_MAINLOOP_READALEVENT()
#endif
#ifndef ESS_DO_SYNCEDINPUTS
# define ESS_DO_SYNCEDINPUTS()
#endif




/* Context infos for each ESC: */
static ESC_DATA escData[CFG_ESS_MAX_DEVICES];



/* Forward declarations: */
static void DoStateChangeCallback(ESC_DATA* ed);





#if (CFG_ESS_SERVE_RUN_LED || CFG_ESS_SERVE_ERR_LED || CFG_ESS_SERVE_CUSTOM_LED)

/* Called every 50 ms */
static void HandleLEDs(void)
{
#   define USED_LED_STATES 8
    static const struct { uint8_t states; uint8_t numStatesAnd50Flag; } \
        table[USED_LED_STATES] = {
                /* off: */
                { 0x00, 1 | 0x80 }, /* 1 state: 0b0 (in interval 50) */
                /* flash 1: */
                { 0x01, 6 },        /* 6 states: 0b000001 (in interval 200) */
                /* flash 2: */
                { 0x05, 8 },        /* 8 states: 0b00000101 (in interval 200) */
                /* flash 3: */
                { 0x15, 10 },       /* 10 states: 0b0000010101 (in interval 200) */
                /* flash 4: */
                { 0x55, 12 },       /* 12 states: 0b000001010101 (in interval 200) */
                /* blink: */
                { 0x01, 2 },        /* 2 states: 0b01 (in interval 200) */
                /* flicker: */
                { 0x01, 2 | 0x80 }, /* 2 states: 0b01 (in interval 50) */
                /* on: */
                { 0x01, 1 | 0x80 }  /* 1 state: 0b1 (in interval 50) */
        };
    static uint8_t currPositions[USED_LED_STATES]; /* bit position in table[].states [0..numStates-1] */
    static const uint8_t ledStateToTableIdx[16] = { 0, 1, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 5, 6, 7 };
    static int interv50Count;
    const int incrInterv200 = ((interv50Count % 4) == 0);
    ESC_DATA* ed = &escData[0];
    int i;

    /*
     * With CFG_ESS_MAX_DEVICES=2, CFG_ESS_COMBINED_RUNERR_LED=1 and ESC_LED_TYPE_NUMBER_OF_LEDS=4
     * all this takes averagely 400 cycles on 720 MHz Cortex-A8
     */

    /* Set LED on/off according to current state: */
    for (i = 0; i < CFG_ESS_MAX_DEVICES; ++i, ed++) {
        ESC_LED_TYPE j = ESC_LED_TYPE_NUMBER_OF_LEDS;
#if (CFG_ESS_COMBINED_RUNERR_LED)
        const int runLEDOverridden = (ed->ledStates[ESC_LED_TYPE_ERR].state != ESC_LED_STATE_OFF);
#endif

        while (j--) {
            ESS_LED_STATEINFOS* const si = &ed->ledStates[j];
            const int t = ledStateToTableIdx[si->state];
            ESS_BOOL onOff = ((1U << currPositions[t]) & table[t].states) != 0;

#if (CFG_ESS_COMBINED_RUNERR_LED)
            if ((j == ESC_LED_TYPE_RUN) && (runLEDOverridden))
                onOff = 0;
#endif
            if (si->lastWritten != onOff) {
                (void)essHALSetLEDState(ed->devIdx, j, onOff);
                si->lastWritten = onOff;
            }
        }
    }

    /* Increment states: */
    for (i = 0; i < USED_LED_STATES; ++i) {
        if ((table[i].numStatesAnd50Flag & 0x80) || (incrInterv200)) {
            const int max = (table[i].numStatesAnd50Flag & 0x7f) - 1;
            currPositions[i] = (currPositions[i] < max) ? currPositions[i] + 1 : 0;
        }
    }

    interv50Count++;
#   undef USED_LED_STATES
}

#endif /* (CFG_ESS_SERVE_ERR_LED) || (CFG_ESS_SERVE_RUN_LED) */

#if (CFG_ESS_SERVE_ERR_LED)

/* Updates ed->ledStates[ESC_LED_TYPE_ERR], no change when already signaling an error */
static void LED_SetErrLED(ESC_DATA* ed, ESC_LED_STATE st)
{
    if (st == ESC_LED_STATE_OFF) {
        ESS_DBG_LEDSERVE(("LEDs: Setting LED ERR to OFF"));

        ed->ledStates[ESC_LED_TYPE_ERR].state = ESC_LED_STATE_OFF;
        return;
    }

    if (ed->ledStates[ESC_LED_TYPE_ERR].state != ESC_LED_STATE_OFF) {
        ESS_DBG_LEDSERVE(("LEDs: NOT setting LED ERR to %s, because it's already %s", ESC_LED_STATE_TO_STRING(st), ESC_LED_STATE_TO_STRING(ed->ledStates[ESC_LED_TYPE_ERR].state)));
        return;
    }

    ESS_DBG_LEDSERVE(("LEDs: Setting LED ERR to %s", ESC_LED_STATE_TO_STRING(st)));
    ed->ledStates[ESC_LED_TYPE_ERR].state = st;
}

#endif /* CFG_ESS_SERVE_ERR_LED */

#if (CFG_ESS_SERVE_RUN_LED)

/* Updates ed->ledStates[ESC_LED_TYPE_RUN]... */
static void LED_SetRunLED(ESC_DATA* ed, ESC_LED_STATE st)
{
    ESS_DBG_LEDSERVE(("LEDs: Setting LED RUN to %s", ESC_LED_STATE_TO_STRING(st)));

    ed->ledStates[ESC_LED_TYPE_RUN].state = st;
}

#endif /* CFG_ESS_SERVE_RUN_LED */



#if (CFG_ESS_EEPROM_EMULATION)

/* Returns ESS_FALSE on success */
static ESS_BOOL EEPROMLoad(ESC_DATA* ed, ESS_BOOL isReload)
{
    uint8_t eepromConfigArea[16];
    ESS_BOOL crcValid;

    if (essEEPROMRead(ed, 0, sizeof(eepromConfigArea), eepromConfigArea))
        return ESS_TRUE;

    if ((eepromConfigArea[14] == 0xa4) && (eepromConfigArea[15] == 0x88)) { /* EEPROM checksum test disabled*/
        crcValid = ESS_TRUE;
        ESS_DBG_EEE(("EEE: checksum DISABLED"));

    } else {
        uint16_t crc = 0xff;
        uint8_t i, j;

        for (i = 0; i < 14; i++) {
            crc ^= eepromConfigArea[i];
            for (j = 0; j < 8; j++)
                crc = (crc & 0x80) ? (crc << 1) ^ 0x07 : crc << 1;
        }

        crcValid = (LO_BYTE(crc) == eepromConfigArea[14]);
        ESS_DBG_EEE(("EEE: checksum %s", (crcValid) ? "OK" : "WRONG!"));
    }

    if (crcValid) {
        const uint16_t pdiCtrl = MAKE_WORD(eepromConfigArea[0], eepromConfigArea[1]);
        const uint16_t pdiConfig = MAKE_WORD(eepromConfigArea[2], eepromConfigArea[3]);
        const uint16_t syncLen = MAKE_WORD(eepromConfigArea[4], eepromConfigArea[5]);
        const uint16_t statAlias = MAKE_WORD(eepromConfigArea[8], eepromConfigArea[9]);

        ESS_DBG_EEE(("EEE: loaded pdiCtrl = 0x%.4x, pdiConfig = 0x%.4x, syncLen = 0x%.4x, alias = 0x%.4x", pdiCtrl, pdiConfig, syncLen, statAlias));

        if (!isReload) {
            uint16_t dlStatus;

            ESC_WRITE16(ed, ESC_REG_PDICONFIGURATION, pdiConfig);
            ESC_WRITE16(ed, ESC_REG_SYNCPULSELEN, syncLen);

            ESC_READ16(ed, ESC_REG_DLSTATUS, dlStatus);
            dlStatus = (pdiCtrl & REG_MASK_PDICONTROL_ENHLINKDETECTION) ? dlStatus | REG_MASK_DLSTATUS_ENHLINKDETECTION : dlStatus & ~REG_MASK_DLSTATUS_ENHLINKDETECTION;
            ESC_WRITE16(ed, ESC_REG_DLSTATUS, dlStatus);
        }

        ESC_WRITE16(ed, ESC_REG_PDICONTROL, pdiCtrl);
        ESC_WRITE16(ed, ESC_REG_STATIONALIAS, statAlias);

        /* Storing some reloaded data as if we were the IP Core: */
        ESC_WRITE16(ed, ESC_REG_EEPROMDATA, statAlias);
        ESC_WRITE16(ed, ESC_REG_EEPROMDATA + 2, ((pdiCtrl >> 9) & 0x01) | (((pdiCtrl >> 12) & 0x0f) << 1));
        ESC_WRITE32_NOSWAP(ed, ESC_REG_EEPROMDATAHI, 0);

        return ESS_FALSE;
    }

    return ESS_TRUE;
}

#endif /* CFG_ESS_EEPROM_EMULATION */



/* Reads MBoxOut SM buffer to given mb (Receive from master) */
static void MBX_ReadPacket(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    uint16_t addr = ed->cbDataSM[ed->smIdxMBoxOut].smPhysAddr;

    ED_DO_EVENT3(ed, ESS_EVENT_READMBOXOUTSTART, ed->cbDataSM[ed->smIdxMBoxOut].smPhysAddr, ed->cbDataSM[ed->smIdxMBoxOut].smLen, ed->smIdxMBoxOut);

    /* Header: */
    ESC_READ16(ed, addr, mb->length);
    ESC_READ16(ed, addr + 2, mb->address);
    ESC_READ8(ed, addr + 4, mb->channelAndPrio);
    ESC_READ8(ed, addr + 5, mb->typeAndCounter);
    addr += MBX_HEADERLEN;

    /*
    ESS_DBG(("len = %u, addr = 0x%.4x, channelAndPrio = 0x%.2x, typeAndCounter = 0x%.2x", mb->length, mb->address, mb->channelAndPrio, mb->typeAndCounter));
    */

    /* Data: */
    if (mb->length > sizeof(mb->data)) {
        ESS_DBG(("WARNING: MBX Data truncated!"));
        mb->length = sizeof(mb->data);
    }

    ESC_READ(ed, addr, mb->data, mb->length);

    /* Make sure last byte is also read, to unlock the mailbox again: */
    if (mb->length < ed->mbxOutDataLen) {
        uint8_t foo;
        ESC_READ8(ed, ed->cbDataSM[ed->smIdxMBoxOut].smPhysAddr + ed->cbDataSM[ed->smIdxMBoxOut].smLen - 1, foo);
        if (foo) {}
    }

    ED_DO_EVENT3(ed, ESS_EVENT_READMBOXOUTEND, ed->cbDataSM[ed->smIdxMBoxOut].smPhysAddr, ed->cbDataSM[ed->smIdxMBoxOut].smLen, ed->smIdxMBoxOut);
}

/* Writes given mb to MBoxIn SM buffer (Send to master). Repeats last packet when mb is NULL */
static void MBX_WritePacket(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    uint16_t addr = ed->cbDataSM[ed->smIdxMBoxIn].smPhysAddr;

    if (mb == NULL) { /* Repeat last packet: */
        mb = &ed->mbxLastSentPacket;
        ed->mbxCnt = MBX_HEADER_GET_COUNTER(mb->typeAndCounter);
    } else {
        ed->mbxCnt = 1 + (ed->mbxCnt % 7);
        mb->typeAndCounter = MBX_HEADER_MAKE_TYPEANDCOUNTER(MBX_HEADER_GET_TYPE(mb->typeAndCounter), ed->mbxCnt);
    }

    ED_DO_EVENT3(ed, ESS_EVENT_WRITEMBOXINSTART, addr, mb->length, ed->smIdxMBoxIn);

    /* Header: */
    ESC_WRITE16(ed, addr, mb->length);
    ESC_WRITE16(ed, addr + 2, mb->address);
    ESC_WRITE8(ed, addr + 4, mb->channelAndPrio);
    ESC_WRITE8(ed, addr + 5, mb->typeAndCounter);
    addr += MBX_HEADERLEN;

    ed->mbxSendFull = ESS_TRUE;

    /* Store a copy: */
#if defined(_DEBUG)
    if ((size_t)(MBX_HEADERLEN + mb->length) > sizeof(*mb)) {
        ESS_DBG(("ERROR: mbx packet with %u data bytes too large for mbx buffer of %u bytes!", mb->length, (unsigned int)sizeof(*mb)));
        mb->length = sizeof(*mb) - MBX_HEADERLEN;
    }
#endif
    ESS_MEMCOPY(&ed->mbxLastStoredPacket, mb, MBX_HEADERLEN + mb->length);

    /* Then write data: */
    ESC_WRITE(ed, addr, mb->data, mb->length);

    /* Make sure last byte is also written, to unlock the mailbox again: */
    if (mb->length < ed->mbxInDataLen)
        ESC_WRITE8(ed, ed->cbDataSM[ed->smIdxMBoxIn].smPhysAddr + ed->cbDataSM[ed->smIdxMBoxIn].smLen - 1, 0);

    ESS_DBG_MAILBOX_TRANSFER(("MBX: Wrote %smbx packet to send mbx, cnt = %u (cnt might be wrong)", (mb != &ed->mbxLastStoredPacket) ? "" : "(REPETITION!)", MBX_HEADER_GET_COUNTER(mb->typeAndCounter)));
    ESS_DBG_MBX_MINIMAL(("--> Put cnt %d (= mbxLastStoredPacket now%s)", MBX_HEADER_GET_COUNTER(mb->typeAndCounter), (mb != &ed->mbxLastStoredPacket) ? "" : ", was REPETITION"));

    ED_DO_EVENT3(ed, ESS_EVENT_WRITEMBOXINEND, ed->cbDataSM[ed->smIdxMBoxIn].smPhysAddr, ed->cbDataSM[ed->smIdxMBoxIn].smLen, ed->smIdxMBoxIn);
}

/* Turns the given mb packet into an error reply and sends it to the master */
static void MBX_WriteErrorReply(ESC_DATA* ed, ESS_MBX_PACKET* mb, uint16_t errCode)
{
    /* This error reply is defined in ETG.1000.4, but usually masters do NOT support it, so protocol specific error is preferable! */

    mb->length = 4;
    mb->typeAndCounter = 0; /* type shall be 0 and counter is set by MBX_WritePacket() */
    mb->data[0] = 1; /* mailbox error cmd */
    mb->data[1] = 0;
    mb->data[2] = LO_BYTE(errCode);
    mb->data[3] = HI_BYTE(errCode);

    MBX_WritePacket(ed, mb);
}


#if (CFG_ESS_SUPPORT_COE)

# if (DEBUG_COE)
#  define ESS_DBG_COE(x)                            ESS_DBG(x)
# else
#  define ESS_DBG_COE(x)
# endif

# if (CFG_ESS_COE_ALLOWTOMAPNULLENTRIES)
#  define CONTINUE_IF_MI_DATA_NULL()                if (mi->data == NULL) continue
# else
#  define CONTINUE_IF_MI_DATA_NULL()
# endif


/* Don't use in conjunction with ESC memory! */
INLINE static void MemCopyExtSrc(uint8_t* dest, const uint8_t* src, uint32_t bitLen, uint8_t srcBit)
{
    if (((bitLen % 8) == 0) && (srcBit == 0)) {
        ESS_MEMCOPY(dest, src, bitLen >> 3);

    } else {
        uint32_t srcBitPos = srcBit;

        /* Copy complete PDO. */
        if (srcBit & ESS_FLAG_OPT_COPY) {
            bitLen += ((srcBit >> 4) & 0x07) * 256;
            ESS_MEMCOPY(dest, src, bitLen);
            return;
        }

        /* Might access one more byte than necessary in src array */

        while (bitLen >= 8) {
            *dest++ = ((src[srcBitPos >> 3] >> (srcBitPos % 8)) | ((src[(srcBitPos >> 3) + 1] << (8 - (srcBitPos) % 8)) & 0xFF));
            srcBitPos += 8;
            bitLen -= 8;
        }

        if (bitLen > 0) {
            uint8_t byteRead;

            byteRead = ((src[srcBitPos >> 3] >> (srcBitPos % 8)) | ((src[(srcBitPos >> 3) + 1] << (8 - (srcBitPos) % 8)) & 0xFF));
            byteRead &= ((1U << bitLen) - 1);

            *dest &= ~((1U << bitLen) - 1);
            *dest |= byteRead;
        }
    }
}

/* Don't use in conjunction with ESC memory! */
INLINE static void MemCopyExtDest(uint8_t* dest, const uint8_t* src, uint32_t bitLen, uint8_t destBit)
{
    if (((bitLen % 8) == 0) && (destBit == 0)) {
        ESS_MEMCOPY(dest, src, bitLen >> 3);

    } else {
        uint32_t srcBitPos = 0;

        /* Copy complete PDO */
        if (destBit & ESS_FLAG_OPT_COPY) {
            bitLen += ((destBit >> 4) & 0x07) * 256;
            ESS_MEMCOPY(dest, src, bitLen);
            return;
        }

        /* First bits: */
        if (destBit) {
            uint8_t bl = (uint8_t)(MIN((uint32_t)(8 - destBit), bitLen));

            *dest &= ~MAKE_BIT_MASK(destBit, bl);
            *dest |= (((src[0] & ((1U << bl) - 1)) << destBit) & 0xFF);

            dest++;
            srcBitPos += bl;
        }

        /* Full bytes in between: */
        while ((bitLen - srcBitPos) >= 8) { /* Could do word/dword copies here, too... but usually this is only called for very short entries => so: not needed! */
            *dest = (uint8_t)BYTE_FROM_BYTEARR(src, srcBitPos);

            dest++;
            srcBitPos += 8;
        }

        /* Last bits: */
        if (srcBitPos < bitLen) {
            uint8_t bl = (uint8_t)(bitLen - srcBitPos);

            *dest &= ~((1U << (bl)) - 1);
            *dest |= (uint8_t)BYTE_FROM_BYTEARR(src, srcBitPos) & ((1U << bl) - 1);
        }
    }
}


/* Turns the given mb packet into an sdo error request and sends it to the master */
static void CoE_SendSDOAbort(ESC_DATA* ed, ESS_MBX_PACKET* mb, uint16_t objIdx, uint8_t subIdx, uint32_t abortCode)
{
    ESS_DBG_COE(("INFO: Writing CoE abort 0x%.8x, obj = 0x%.4x, subIdx = 0x%.2x", abortCode, objIdx, subIdx));

    mb->address = 0;
    mb->channelAndPrio = 0xc0; /* Highest pri */
    mb->typeAndCounter = MBX_TYPE_COE; /* counter is set by MBX_WritePacket() */
    mb->length = 10;

    mb->data[0] = 0x00; /* CoE Header: Service = "SDO Request" */
    mb->data[1] = 0x20; /* ... */

    mb->data[2] = 0x80; /* = 'Abort transfer request' (0x04) at upper 3 bits */
    mb->data[3] = LO_BYTE(objIdx);
    mb->data[4] = HI_BYTE(objIdx);
    mb->data[5] = subIdx;
    mb->data[6] = LO_BYTE(LO_WORD(abortCode));
    mb->data[7] = HI_BYTE(LO_WORD(abortCode));
    mb->data[8] = LO_BYTE(HI_WORD(abortCode));
    mb->data[9] = HI_BYTE(HI_WORD(abortCode));

    MBX_WritePacket(ed, mb);
}

/* Copies all input entries to the SM buffer they're mapped to  */
static void CoE_SyncInputMappedEntries(ESC_DATA* ed, const ESS_CBDATA_SM* sm, uint8_t* smBuffer)
{
    int i = ed->smMappedEntries[sm->smIdx].entriesCount;

    while (i--) {
        const ESS_PDO_MAP_INFO* mi = &ed->smMappedEntries[sm->smIdx].entries[i];

        CONTINUE_IF_MI_DATA_NULL(); /* With (CFG_ESS_COE_ALLOWTOMAPNULLENTRIES) mi->data can be NULL */

        MemCopyExtDest(&smBuffer[mi->bufferOffset], mi->data, mi->bitLen, mi->addrBit);
    }

#if (CFG_HAL_NEEDS_EVENTS)
    ED_DO_EVENT3(ed, ESS_EVENT_NEEDSMADDRINPUTS, sm->smPhysAddr, sm->smLen, sm->smIdx);

    {
        const uint16_t currAddr = (uint16_t)ED_EVENT_RESULT(ed);

        if ((currAddr < 0x1000) || (currAddr == 0xffff)) {
            ESS_DBG_EVENTS(("ED_DO_EVENT: ERROR: ESS_EVENT_NEEDSMADDRINPUTS failed! (Returned 0x%.4x, devIdx = %d)", currAddr, ed->devIdx));
        } else {
            ESC_WRITE(ed, currAddr, smBuffer, sm->smLen);
            ED_DO_EVENT3(ed, ESS_EVENT_WROTEINPUTS, sm->smPhysAddr, sm->smLen, sm->smIdx);
        }
    }

#else
    ESC_WRITE(ed, sm->smPhysAddr, smBuffer, sm->smLen);
#endif

    ESS_DO_SYNCEDINPUTS();
}

static void CoE_SyncOutputMappedEntries(ESC_DATA* ed, const ESS_CBDATA_SM* sm, uint8_t* smBuffer)
{
    int i = ed->smMappedEntries[sm->smIdx].entriesCount;

    while (i--) {
        ESS_PDO_MAP_INFO* mi = &ed->smMappedEntries[sm->smIdx].entries[i];

        CONTINUE_IF_MI_DATA_NULL(); /* With (CFG_ESS_COE_ALLOWTOMAPNULLENTRIES) mi->data can be NULL */

        MemCopyExtSrc(mi->data, &smBuffer[mi->bufferOffset], mi->bitLen, mi->addrBit);
    }
}

static void CoE_CreateEmergency_SMAddrErr(ESS_COE_EMERGENCY* em, ESS_SM smIdx, ESS_BOOL whilePreOpSafeOp, uint16_t minAddr, uint16_t maxAddr)
{
    em->data[0] = (uint8_t)(1 | (smIdx << 2));
    em->data[1] = LO_BYTE(minAddr);
    em->data[2] = HI_BYTE(minAddr);
    em->data[3] = LO_BYTE(maxAddr);
    em->data[4] = HI_BYTE(maxAddr);
    em->errReg = 0x00;
    em->errCode = (whilePreOpSafeOp) ? 0xa000 : 0xa001;
}

static void CoE_CreateEmergency_SMLenErr(ESS_COE_EMERGENCY* em, ESS_SM smIdx, ESS_BOOL whilePreOpSafeOp, uint16_t minSize, uint16_t maxSize)
{
    em->data[0] = (uint8_t)(smIdx << 2);
    em->data[1] = LO_BYTE(minSize);
    em->data[2] = HI_BYTE(minSize);
    em->data[3] = LO_BYTE(maxSize);
    em->data[4] = HI_BYTE(maxSize);
    em->errReg = 0x00;
    em->errCode = (whilePreOpSafeOp) ? 0xa000 : 0xa001;
}

static void CoE_CreateEmergency_SMSettingsErr(ESS_COE_EMERGENCY* em, ESS_SM smIdx, ESS_BOOL whilePreOpSafeOp, uint8_t expBufferType, uint8_t expDirection, ESS_BOOL expALEvtEnable, ESS_BOOL expEnable)
{
    em->data[0] = (uint8_t)(2 | smIdx << 2);
    em->data[1] = (expBufferType & 3) | ((expDirection & 3) << 2);
    if(ESS_TRUE == expALEvtEnable) {
        em->data[1] |= 1U << 5;
    }
    em->data[2] = 0;
    em->data[3] = (expEnable != 0);
    em->data[4] = 0;
    em->errReg = 0x00;
    em->errCode = (whilePreOpSafeOp) ? 0xa000 : 0xa001;
}

static void CoE_SendQueuedEmergency(ESC_DATA* ed)
{
    struct {
        uint16_t length;
        uint16_t address;
        uint8_t channelAndPrio;
        uint8_t typeAndCounter;
        uint8_t data[10];
    } mb;

    mb.address = 0;
    mb.channelAndPrio = 0xc0; /* Highest pri */
    mb.typeAndCounter = MBX_TYPE_COE; /* counter is set by MBX_WritePacket() */
    mb.length = 10;

    mb.data[0] = 0x00; /* CoE Header: Service = "Emergency" */
    mb.data[1] = 0x10; /* ... */

    mb.data[2] = LO_BYTE(ed->coeEmergency.errCode);
    mb.data[3] = HI_BYTE(ed->coeEmergency.errCode);
    mb.data[4] = ed->coeEmergency.errReg;
    mb.data[5] = ed->coeEmergency.data[0];
    mb.data[6] = ed->coeEmergency.data[1];
    mb.data[7] = ed->coeEmergency.data[2];
    mb.data[8] = ed->coeEmergency.data[3];
    mb.data[9] = ed->coeEmergency.data[4];

    MBX_WritePacket(ed, (ESS_MBX_PACKET*)&mb);

    ed->coeEmergencyReady = ESS_FALSE;
}

static void CoE_FinishSegmentedTransfer(ESC_DATA* ed, ESS_BOOL upload, ESS_BOOL unexpected)
{
    if (upload) {
        ed->coeSDOUploadState.active = ESS_FALSE;
        (void)essOD_DoEntryEventCallback(ed, ed->coeSDOUploadState.objIdx, ed->coeSDOUploadState.subIdx, (unexpected) ? ESS_OD_ENTRY_CALLBACK_ABORTED_UPLOAD : ESS_OD_ENTRY_CALLBACK_COMPLETED_UPLOAD);

    } else {
        ed->coeSDODownloadState.active = ESS_FALSE;
        (void)essOD_DoEntryEventCallback(ed, ed->coeSDODownloadState.objIdx, ed->coeSDODownloadState.subIdx, (unexpected) ? ESS_OD_ENTRY_CALLBACK_ABORTED_DOWNLOAD : ESS_OD_ENTRY_CALLBACK_COMPLETED_DOWNLOAD);
    }
}

static uint16_t CoE_Handle_SDOAbort(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    (void)mb;

    if (ed->coeSDODownloadState.active)
        CoE_FinishSegmentedTransfer(ed, ESS_FALSE, ESS_TRUE);

    if (ed->coeSDOUploadState.active)
        CoE_FinishSegmentedTransfer(ed, ESS_TRUE, ESS_TRUE);

    return MBX_ERR_SUCCESS;
}


#if (CFG_ESS_COE_SUPPORT_SDO_INFO_SERVICE)

static void CoE_Handle_SegmentedListTransfer(ESC_DATA* ed)
{
    const int mbMaxDataLen = ed->mbxInDataLen - MBX_COE_HEADERLEN - MBX_COE_SDOINFO_HEADERLEN;
    ESS_MBX_PACKET* const mbxData = &ed->mbxTempPacket;
    int mbDataLenRemain, indicesToCopy;
    uint16_t fragmentsLeft = 0;
    uint8_t* pIndices;

    mbxData->address = 0;
    mbxData->channelAndPrio = 0;
    mbxData->typeAndCounter = MBX_HEADER_MAKE_TYPEANDCOUNTER(MBX_TYPE_COE, 0); /* counter is set by MBX_WritePacket() */
    mbxData->data[MBX_COE_HEADERPOS + 1] = 0x80; /* CoE service = SDO Information */
    mbxData->data[MBX_COE_SDOINFO_HEADERPOS] = MBX_COE_SDOINFO_OP_ODLISTRESP;

    mbDataLenRemain = mbMaxDataLen;
    pIndices = &mbxData->data[MBX_COE_SDOINFO_DATAPOS];
    mbxData->length = MBX_COE_SDOINFO_DATAPOS;

    if (ed->sdoInfoState.nextItem == 0) {
        /* Prepend two bytes ListType: */
        mbDataLenRemain -= 2;
        pIndices += 2;
        mbxData->length += 2;

        mbxData->data[MBX_COE_SDOINFO_DATAPOS] = LO_BYTE(ed->sdoInfoState.listType);
        mbxData->data[MBX_COE_SDOINFO_DATAPOS + 1] = HI_BYTE(ed->sdoInfoState.listType);
    }

    indicesToCopy = MIN(mbDataLenRemain / 2, ed->sdoInfoState.objCount - ed->sdoInfoState.nextItem);
    mbxData->length += (uint16_t)(indicesToCopy * 2);
    while (indicesToCopy--) {
        const uint16_t objIdx = essOD_SDOInfo_GetObjIndex(&ed->dict, ed->sdoInfoState.listType, ed->sdoInfoState.nextItem);

        *pIndices++ = LO_BYTE(objIdx);
        *pIndices++ = HI_BYTE(objIdx);
        ed->sdoInfoState.nextItem++;
    }

    if (mbMaxDataLen / 2) { /* Should never happen, but static code analyzer gave "division by zero" warning */
        fragmentsLeft = (uint16_t)((ed->sdoInfoState.objCount - ed->sdoInfoState.nextItem) / (mbMaxDataLen / 2)); /* = indicesLeft / indicesPerPacket */
        if(((ed->sdoInfoState.objCount - ed->sdoInfoState.nextItem) % (mbMaxDataLen / 2) != 0)) {
            fragmentsLeft++;
        }
    }
    if (fragmentsLeft)
        mbxData->data[MBX_COE_SDOINFO_HEADERPOS] |= 0x80;
    else
        ed->sdoInfoState.active = ESS_FALSE;

    mbxData->data[MBX_COE_SDOINFO_HEADERPOS + 2] = LO_BYTE(fragmentsLeft);
    mbxData->data[MBX_COE_SDOINFO_HEADERPOS + 3] = HI_BYTE(fragmentsLeft);

    MBX_WritePacket(ed, mbxData);
}

static uint16_t CoE_Handle_ListRequest(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    uint16_t listType = MAKE_WORD(mb->data[MBX_COE_SDOINFO_DATAPOS], mb->data[MBX_COE_SDOINFO_DATAPOS + 1]);
    uint16_t listCounts[5];

    essOD_SDOInfo_GetListCounts(&ed->dict, listCounts);

#if 0 /* Only used for debugging essOD_SDOInfo_GetListCounts() and essOD_SDOInfo_GetObjIndex() */
    {
        int iL;

        for (iL = MBX_COE_SDOINFO_ODLISTREQ_ALLOBJECTS; iL <= MBX_COE_SDOINFO_ODLISTREQ_STARTUPOBJECTS; ++iL) {
            const uint16_t objCount = listCounts[iL - 1];
            int iObj;

            ESS_DBG(("    %u objects in list %u:", objCount, iL));
            for (iObj = 0; iObj < objCount; ++iObj) {
                const uint16_t objIdx = essOD_SDOInfo_GetObjIndex(&ed->dict, iL, iObj);

                ESS_DBG(("        object%u = 0x%.4x", iObj, objIdx));
            }
        }
    }
#endif


    if (listType == MBX_COE_SDOINFO_ODLISTREQ_NUMBERS) {
        /* Number of objects in each of the 5 lists requested: */
        ESS_DBG_COE(("INFO: Requested number of CoE objects in OD lists: all = %u, Rx = %u, Tx = %u, backup = %u, startup = %u", listCounts[0], listCounts[1], listCounts[2], listCounts[3], listCounts[4]));

        mb->length = MBX_COE_SDOINFO_DATAPOS + 12;
        mb->data[MBX_COE_SDOINFO_HEADERPOS] = MBX_COE_SDOINFO_OP_ODLISTRESP;
        mb->data[(MBX_COE_SDOINFO_DATAPOS +  2)] = LO_BYTE(listCounts[0]);
        mb->data[(MBX_COE_SDOINFO_DATAPOS +  3)] = HI_BYTE(listCounts[0]);
        mb->data[(MBX_COE_SDOINFO_DATAPOS +  4)] = LO_BYTE(listCounts[1]);
        mb->data[(MBX_COE_SDOINFO_DATAPOS +  5)] = HI_BYTE(listCounts[1]);
        mb->data[(MBX_COE_SDOINFO_DATAPOS +  6)] = LO_BYTE(listCounts[2]);
        mb->data[(MBX_COE_SDOINFO_DATAPOS +  7)] = HI_BYTE(listCounts[2]);
        mb->data[(MBX_COE_SDOINFO_DATAPOS +  8)] = LO_BYTE(listCounts[3]);
        mb->data[(MBX_COE_SDOINFO_DATAPOS +  9)] = HI_BYTE(listCounts[3]);
        mb->data[(MBX_COE_SDOINFO_DATAPOS + 10)] = LO_BYTE(listCounts[4]);
        mb->data[(MBX_COE_SDOINFO_DATAPOS + 11)] = HI_BYTE(listCounts[4]);

        MBX_WritePacket(ed, mb);
        return MBX_ERR_SUCCESS;

    } else if ((listType >= MBX_COE_SDOINFO_ODLISTREQ_ALLOBJECTS) && (listType <= MBX_COE_SDOINFO_ODLISTREQ_STARTUPOBJECTS)) {
        /* Indices of objects in a specific list requested: */
        uint16_t objCount = listCounts[listType - 1];
        uint16_t dataPos = MBX_COE_SDOINFO_DATAPOS + 2;
        uint16_t i;

        mb->data[MBX_COE_SDOINFO_HEADERPOS] = MBX_COE_SDOINFO_OP_ODLISTRESP;
        mb->length = MBX_COE_SDOINFO_DATAPOS + 2 + objCount + objCount;

        if (mb->length > ed->mbxInDataLen) {
            ed->sdoInfoState.listType = listType;
            ed->sdoInfoState.objCount = objCount;
            ed->sdoInfoState.nextItem = 0;
            ed->sdoInfoState.active = ESS_TRUE;

            CoE_Handle_SegmentedListTransfer(ed);
            return MBX_ERR_SUCCESS;
        }

        for (i = 0; i < objCount; i++) {
            uint16_t objIdx = essOD_SDOInfo_GetObjIndex(&ed->dict, listType, i);
            mb->data[dataPos++] = LO_BYTE(objIdx);
            mb->data[dataPos++] = HI_BYTE(objIdx);
        }

        MBX_WritePacket(ed, mb);
        return MBX_ERR_SUCCESS;

    } else {
        ESS_DBG_COE(("WARNING: SDO Request for unknown listType: %u", listType));
        return MBX_ERR_SERVICENOTSUPPORTED;
    }
}

static uint16_t CoE_Handle_ObjectRequest(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    const uint16_t objIdx = MAKE_WORD(mb->data[MBX_COE_SDOINFO_DATAPOS], mb->data[MBX_COE_SDOINFO_DATAPOS + 1]);
    ESS_OD_OBJECT_INFOS infos;
    COE_DATATYPE dataType;
    uint8_t maxSubIdx;
    COE_ABORTCODE ab;

    ab = essOD_SDOInfo_GetObjInfos(&ed->dict, objIdx, &infos, &maxSubIdx, &dataType);
    if (ab != COE_ABORTCODE_NONE) {
        CoE_SendSDOAbort(ed, mb, objIdx, 0, ab);
        return MBX_ERR_SUCCESS;
    }

    mb->length = MBX_COE_SDOINFO_DATAPOS + 6;
    mb->data[MBX_COE_SDOINFO_HEADERPOS] = MBX_COE_SDOINFO_OP_OBJRESP;
    mb->data[MBX_COE_SDOINFO_DATAPOS + 2] = LO_BYTE(dataType);
    mb->data[MBX_COE_SDOINFO_DATAPOS + 3] = HI_BYTE(dataType);
    mb->data[MBX_COE_SDOINFO_DATAPOS + 4] = maxSubIdx;
    mb->data[MBX_COE_SDOINFO_DATAPOS + 5] = infos.code;

    if (infos.name) {
        while ((*infos.name != '\0') && (mb->length < ed->mbxInDataLen)) {
            mb->data[mb->length] = (uint8_t)(*infos.name);
            mb->length++;
            infos.name++;
        }
    }

    MBX_WritePacket(ed, mb);
    return MBX_ERR_SUCCESS;
}

static uint16_t CoE_Handle_EntryRequest(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    const uint16_t objIdx = MAKE_WORD(mb->data[MBX_COE_SDOINFO_DATAPOS], mb->data[MBX_COE_SDOINFO_DATAPOS + 1]);
    const COE_VALUEINFO reqValues = mb->data[MBX_COE_SDOINFO_DATAPOS + 3];
    const uint8_t subIdx = mb->data[MBX_COE_SDOINFO_DATAPOS + 2];
    ESS_OD_ENTRY_INFOS infos;
    uint32_t entryByteLen;
    COE_ACCESS acc;
    COE_ABORTCODE ab;
    uint16_t dataPos;
    uint32_t bitLen;

    ab = essOD_SDOInfo_GetEntryInfos(&ed->dict, objIdx, subIdx, &infos, &acc, &bitLen);
    if (ab != COE_ABORTCODE_NONE) {
        CoE_SendSDOAbort(ed, mb, objIdx, 0, ab);
        return MBX_ERR_SUCCESS;
    }
    entryByteLen = (bitLen + 7) >> 3;

    /* Determine total mb packet length: */
    mb->length = MBX_COE_SDOINFO_DATAPOS + 10;
    if ((reqValues & COE_VALUEINFO_UNITTYPE) && (infos.unitType != 0))
        mb->length += (uint16_t)4;
    if ((reqValues & COE_VALUEINFO_DEFAULTVALUE) && (infos.defaultValue != NULL))
        mb->length += LO_WORD(entryByteLen);
    if ((reqValues & COE_VALUEINFO_MINVALUE) && (infos.minValue != NULL))
        mb->length += LO_WORD(entryByteLen);
    if ((reqValues & COE_VALUEINFO_MAXVALUE) && (infos.maxValue != NULL))
        mb->length += LO_WORD(entryByteLen);
    if (mb->length > ed->mbxInDataLen) {
        ESS_DBG(("ERROR: Mailbox too small to return CoE entry 0x%.4x.%.2x (Req.: value infos: 0x%.2x) (%u > %u)", subIdx, objIdx, reqValues, mb->length, ed->mbxInDataLen));
        return MBX_ERR_NORESOURCES;
    }

    /* Write data: */
    mb->data[MBX_COE_SDOINFO_HEADERPOS] = MBX_COE_SDOINFO_OP_ENTRYRESP;
    dataPos = MBX_COE_SDOINFO_DATAPOS + 3;
    mb->data[dataPos++] = COE_VALUEINFO_NONE;
    mb->data[dataPos++] = LO_BYTE(infos.dataType);
    mb->data[dataPos++] = HI_BYTE(infos.dataType);
    if (bitLen >= 0xffff) {
        mb->data[dataPos++] = 0xff;
        mb->data[dataPos++] = 0xff;
    } else {
        mb->data[dataPos++] = LO_BYTE(bitLen);
        mb->data[dataPos++] = HI_BYTE(bitLen);
    }
    mb->data[dataPos++] = LO_BYTE(acc);
    mb->data[dataPos++] = HI_BYTE(acc);

    if ((reqValues & COE_VALUEINFO_UNITTYPE) && (infos.unitType != 0)) {
        mb->data[dataPos++] = LO_BYTE(LO_WORD(infos.unitType));
        mb->data[dataPos++] = HI_BYTE(LO_WORD(infos.unitType));
        mb->data[dataPos++] = LO_BYTE(HI_WORD(infos.unitType));
        mb->data[dataPos++] = HI_BYTE(HI_WORD(infos.unitType));
        mb->data[MBX_COE_SDOINFO_DATAPOS + 3] |= COE_VALUEINFO_UNITTYPE;
    }
    if ((reqValues & COE_VALUEINFO_DEFAULTVALUE) && (infos.defaultValue != NULL)) {
        ESS_MEMCOPY(&mb->data[dataPos], infos.defaultValue, entryByteLen);
        dataPos += LO_WORD(entryByteLen);
        mb->data[MBX_COE_SDOINFO_DATAPOS + 3] |= COE_VALUEINFO_DEFAULTVALUE;
    }
    if ((reqValues & COE_VALUEINFO_MINVALUE) && (infos.minValue != NULL)) {
        ESS_MEMCOPY(&mb->data[dataPos], infos.minValue, entryByteLen);
        dataPos += LO_WORD(entryByteLen);
        mb->data[MBX_COE_SDOINFO_DATAPOS + 3] |= COE_VALUEINFO_MINVALUE;
    }
    if ((reqValues & COE_VALUEINFO_MAXVALUE) && (infos.maxValue != NULL)) {
        ESS_MEMCOPY(&mb->data[dataPos], infos.maxValue, entryByteLen);
        mb->data[MBX_COE_SDOINFO_DATAPOS + 3] |= COE_VALUEINFO_MAXVALUE;
    }

    if (infos.name) {
        while ((*infos.name != '\0') && (mb->length < ed->mbxInDataLen)) {
            mb->data[mb->length] = (uint8_t)(*infos.name);
            mb->length++;
            infos.name++;
        }
    }

    MBX_WritePacket(ed, mb);
    return MBX_ERR_SUCCESS;
}

#endif


static uint16_t CoE_Handle_Download(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    const uint16_t objIdx = MAKE_WORD(mb->data[MBX_COE_SDOREQ_HEADERPOS + 1], mb->data[MBX_COE_SDOREQ_HEADERPOS + 2]);
    const ESS_BOOL complAccess = MAKE_ESS_BOOL(mb->data[MBX_COE_SDOREQ_HEADERPOS] & MBX_COE_SDOREQ_HEADER_MASK_COMPLACCESS);
    const ESS_BOOL expedited = MAKE_ESS_BOOL(mb->data[MBX_COE_SDOREQ_HEADERPOS] & MBX_COE_SDOREQ_HEADER_MASK_TRANSFTYPE);
    const uint8_t subIdx = mb->data[MBX_COE_SDOREQ_HEADERPOS + 3];
    const uint32_t totalDownloadLen = (expedited) ? (uint32_t)(4 - MBX_COE_SDOREQ_HEADER_GET_DATASETSIZE(mb->data[MBX_COE_SDOREQ_HEADERPOS])) : MAKE_DWORD(mb->data[MBX_COE_SDOREQ_HEADERPOS + 4], mb->data[MBX_COE_SDOREQ_HEADERPOS + 5], mb->data[MBX_COE_SDOREQ_HEADERPOS + 6], mb->data[MBX_COE_SDOREQ_HEADERPOS + 7]);
    const uint16_t chunkLen = (uint16_t)((expedited) ? totalDownloadLen : mb->length - 10U);
    ESS_OD_VIRTUAL_ENTRY entry;
    COE_ABORTCODE ab;

#if (DEBUG_COE)
    ESS_DBG(("Received SDO Download Request for %u bytes (%u total) for 0x%.4x.%.2x%s%s", chunkLen, totalDownloadLen, objIdx, subIdx, (complAccess) ? ", CompleteAccess" : "", (expedited) ? ", Expedited" : ""));
    ESS_MEM_DUMP("    => Data:", &mb->data[(expedited) ? 6 : 10], chunkLen);
#endif

    (void)essOD_DoEntryEventCallback(ed, objIdx, subIdx, ESS_OD_ENTRY_CALLBACK_REQUESTED_DOWNLOAD);

    if (ed->coeSDODownloadState.active)
        CoE_FinishSegmentedTransfer(ed, ESS_FALSE, ESS_TRUE);

    /* Complete access?: */
    if (complAccess) {
        if (subIdx > 1) {
            ESS_DBG_COE(("ERROR: Compl. obj requested, but SubIdx is %u!?", subIdx));
            CoE_SendSDOAbort(ed, mb, objIdx, subIdx, COE_ABORTCODE_P_INCOMP);
            return MBX_ERR_SUCCESS;
        }

#if (!CFG_ESS_COE_SUPPORT_COMPLETE_ACCESS)
        ESS_DBG_COE(("ERROR: Complete access not supported (obj 0x%.4x requested)", objIdx));
        CoE_SendSDOAbort(ed, mb, objIdx, subIdx, COE_ABORTCODE_ACCESS);
        return MBX_ERR_SUCCESS;
#endif
    }

    /* Size indicated? */
    if ((mb->data[MBX_COE_SDOREQ_HEADERPOS] & MBX_COE_SDOREQ_HEADER_MASK_SIZEIND) == 0) {
        ESS_DBG_COE(("ERROR: Missing size indicator!"));
        CoE_SendSDOAbort(ed, mb, objIdx, subIdx, COE_ABORTCODE_GENERAL);
        return MBX_ERR_SUCCESS;
    }

    /* Get entry and size: */
    ab = essOD_GetEntry(ed, &entry, objIdx, (complAccess) ? -1 - subIdx : subIdx);
    if (ab != COE_ABORTCODE_NONE) {
        CoE_SendSDOAbort(ed, mb, objIdx, subIdx, ab);
        return MBX_ERR_SUCCESS;
    }

    /* Set entry data: */
    ab = essOD_DoEntryDownload(ed, &entry, &mb->data[(expedited) ? 6 : 10], totalDownloadLen, 0, chunkLen, (chunkLen >= totalDownloadLen));
    if (ab != COE_ABORTCODE_NONE) {
        CoE_SendSDOAbort(ed, mb, objIdx, subIdx, ab);
        return MBX_ERR_SUCCESS;
    }

    /* Respond: */
    mb->data[MBX_COE_HEADERPOS + 1] |= 0x10; /* Turn CoE Service "SDO Request" into "SDO Response" */
    mb->data[MBX_COE_SDOINFO_HEADERPOS] = MBX_COE_SDORESP_HDR(ESS_FALSE, ESS_FALSE, 0, ESS_FALSE, MBX_COE_SDORESP_CMDSPECIFIER_DOWNLOAD);
    /* Reserved DWORD, according ETG.1000.6: */
    mb->data[6] = 0; mb->data[7] = 0; mb->data[8] = 0; mb->data[9] = 0;
    mb->length = 10;
    MBX_WritePacket(ed, mb);

    if (chunkLen < totalDownloadLen) {
        /* Start segmented transfer: */
        ed->coeSDODownloadState.active = ESS_TRUE;
        ed->coeSDODownloadState.bytesDone = chunkLen;
        ed->coeSDODownloadState.entry = entry;
        ed->coeSDODownloadState.lastPacket = essGetTime();
        ed->coeSDODownloadState.nextToggleBit = 0;
        ed->coeSDODownloadState.objIdx = objIdx;
        ed->coeSDODownloadState.subIdx = subIdx;
        ed->coeSDODownloadState.totalBytes = totalDownloadLen;
    } else {
        (void)essOD_DoEntryEventCallback(ed, objIdx, subIdx, ESS_OD_ENTRY_CALLBACK_COMPLETED_DOWNLOAD);
    }

    return MBX_ERR_SUCCESS;
}

static uint16_t CoE_Handle_DownloadSegment(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    const uint16_t chunkLen = (mb->length == 10) ? 7 - (uint16_t)MBX_COE_SDOREQ_HEADER_MASK_DOWNLSEGSIZE(mb->data[MBX_COE_SDOREQ_HEADERPOS]) : mb->length - 3;
    const ESS_BOOL toggleBit = MAKE_ESS_BOOL(mb->data[MBX_COE_SDOREQ_HEADERPOS] & MBX_COE_SDOREQ_HEADER_MASK_TOGGLEBIT);
    ESS_SDOTRANSFER_STATUS* si = &ed->coeSDODownloadState;
    COE_ABORTCODE ab;

    if (si->active == ESS_FALSE) {
        ESS_DBG_COE(("WARNING: Download was not set active previously!?"));
        CoE_SendSDOAbort(ed, mb, 0x0000, 0x00, COE_ABORTCODE_GENERAL);
        return MBX_ERR_SUCCESS;
    }

    if (toggleBit != (ESS_BOOL)si->nextToggleBit) {
        ESS_DBG_COE(("WARNING: Togglebit (%u) not as expected! (%u)", toggleBit, si->nextToggleBit));
        CoE_FinishSegmentedTransfer(ed, ESS_FALSE, ESS_TRUE);
        CoE_SendSDOAbort(ed, mb, si->objIdx, si->subIdx, COE_ABORTCODE_TOGGLE);
        return MBX_ERR_SUCCESS;
    }

    if ((si->bytesDone + chunkLen) > si->totalBytes) {
        ESS_DBG_COE(("ERROR: Too many data!"));
        CoE_FinishSegmentedTransfer(ed, ESS_FALSE, ESS_TRUE);
        CoE_SendSDOAbort(ed, mb, si->objIdx, si->subIdx, COE_ABORTCODE_LEN_TOO_HIGH);
        return MBX_ERR_SUCCESS;
    }

    ab = essOD_DoEntryDownload(ed, &si->entry, &mb->data[3], si->totalBytes, si->bytesDone, chunkLen, ((si->bytesDone + chunkLen) >= si->totalBytes));
    if (ab != COE_ABORTCODE_NONE) {
        CoE_SendSDOAbort(ed, mb, si->objIdx, si->subIdx, ab);
        CoE_FinishSegmentedTransfer(ed, ESS_FALSE, ESS_TRUE);
        return MBX_ERR_SUCCESS;
    }

    /* Send response: */
    mb->data[MBX_COE_HEADERPOS + 1] |= 0x10; /* Turn CoE Service "SDO Request" into "SDO Response" */
    mb->data[MBX_COE_SDOREQ_HEADERPOS] = MBX_COE_SDORESP_DOWNLSEG_HDR(toggleBit);
    /* Reserved BYTE[7], according ETG.1000.6: */
    mb->data[3] = 0; mb->data[4] = 0; mb->data[5] = 0; mb->data[6] = 0; mb->data[7] = 0; mb->data[8] = 0; mb->data[9] = 0;
    mb->length = 10;

    si->bytesDone += chunkLen;
    if (si->bytesDone >= si->totalBytes) {
        CoE_FinishSegmentedTransfer(ed, ESS_FALSE, ESS_FALSE);

    } else {
        si->nextToggleBit = !toggleBit;
        si->lastPacket = essGetTime();
    }

    MBX_WritePacket(ed, mb);
    return MBX_ERR_SUCCESS;
}

static uint16_t CoE_Handle_Upload(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    const uint16_t objIdx = MAKE_WORD(mb->data[MBX_COE_SDOREQ_HEADERPOS + 1], mb->data[MBX_COE_SDOREQ_HEADERPOS + 2]);
    const ESS_BOOL complAccess = MAKE_ESS_BOOL(mb->data[MBX_COE_SDOREQ_HEADERPOS] & MBX_COE_SDOREQ_HEADER_MASK_COMPLACCESS);
    const uint8_t subIdx = mb->data[MBX_COE_SDOREQ_HEADERPOS + 3];
    ESS_OD_VIRTUAL_ENTRY entry;
    uint32_t entryByteLen;
    ESS_BOOL expedited;
    uint16_t chunkLen;
    COE_ABORTCODE ab;

    ESS_DBG_COE(("INFO: Received SDO Upload Request for 0x%.4x.%.2x%s", objIdx, subIdx, (complAccess) ? ", CompleteAccess" : ""));
    (void)essOD_DoEntryEventCallback(ed, objIdx, subIdx, ESS_OD_ENTRY_CALLBACK_REQUESTED_UPLOAD);

    if (ed->coeSDOUploadState.active)
        CoE_FinishSegmentedTransfer(ed, ESS_TRUE, ESS_TRUE);

    /* Complete access?: */
    if (complAccess) {
        if (subIdx > 1) {
            ESS_DBG_COE(("ERROR: Compl. obj requested, but SubIdx is %u!?", subIdx));
            CoE_SendSDOAbort(ed, mb, objIdx, subIdx, COE_ABORTCODE_P_INCOMP);
            return MBX_ERR_SUCCESS;
        }

#if (!CFG_ESS_COE_SUPPORT_COMPLETE_ACCESS)
        ESS_DBG_COE(("ERROR: Complete access not supported (obj 0x%.4x requested)", objIdx));
        CoE_SendSDOAbort(ed, mb, objIdx, subIdx, COE_ABORTCODE_ACCESS);
        return MBX_ERR_SUCCESS;
#endif
    }

    /* Get entry and its length: */
    ab = essOD_GetEntry(ed, &entry, objIdx, (complAccess) ? -1 - subIdx : subIdx);
    if (ab != COE_ABORTCODE_NONE) {
        CoE_SendSDOAbort(ed, mb, objIdx, subIdx, ab);
        return MBX_ERR_SUCCESS;
    }
	
    entryByteLen = essOD_GetEntrySize(ed, &entry, ESS_TRUE);

	/*
	 * Handle objects which indicate a dynamic size for upload.
	 * If no handler is attached use the configured (download) size.
	 */
    if(0 == entryByteLen) {
        entryByteLen = (uint32_t)essOD_DoEntryEventCallback(ed, objIdx, subIdx, ESS_OD_ENTRY_CALLBACK_GET_UPLOAD_SIZE);
        if(0 == entryByteLen) {
            entryByteLen = essOD_GetEntrySize(ed, &entry, ESS_FALSE);
        }
    }

    expedited = (entryByteLen <= 4);
    chunkLen = (uint16_t)MIN((uint32_t)ed->mbxInDataLen - 10, entryByteLen);

    /* Get entry data: */
    ab = essOD_DoEntryUpload(ed, &entry, &mb->data[(expedited) ? 6 : 10], 0, chunkLen, (chunkLen >= entryByteLen));
    if (ab != COE_ABORTCODE_NONE) {
        CoE_SendSDOAbort(ed, mb, objIdx, subIdx, ab);
        return MBX_ERR_SUCCESS;
    }

    /* Send Reply: */
    mb->data[MBX_COE_HEADERPOS + 1] |= 0x10; /* Turn CoE Service "SDO Request" into "SDO Response" */
    if (expedited) {
        mb->length = 10;
        mb->data[MBX_COE_SDOINFO_HEADERPOS] = MBX_COE_SDORESP_HDR(ESS_TRUE, ESS_TRUE, 4 - entryByteLen, ESS_FALSE, MBX_COE_SDOREQ_CMDSPECIFIER_UPLOAD);
        MBX_WritePacket(ed, mb);
        (void)essOD_DoEntryEventCallback(ed, objIdx, subIdx, ESS_OD_ENTRY_CALLBACK_COMPLETED_UPLOAD);

    } else {
        mb->length = 10 + chunkLen;
        mb->data[MBX_COE_SDOINFO_HEADERPOS] = MBX_COE_SDORESP_HDR(ESS_TRUE, ESS_FALSE, 0, ESS_FALSE, MBX_COE_SDOREQ_CMDSPECIFIER_UPLOAD);
        mb->data[MBX_COE_SDOINFO_HEADERPOS + 4] = LO_BYTE(LO_WORD(entryByteLen));
        mb->data[MBX_COE_SDOINFO_HEADERPOS + 5] = HI_BYTE(LO_WORD(entryByteLen));
        mb->data[MBX_COE_SDOINFO_HEADERPOS + 6] = LO_BYTE(HI_WORD(entryByteLen));
        mb->data[MBX_COE_SDOINFO_HEADERPOS + 7] = HI_BYTE(HI_WORD(entryByteLen));

        MBX_WritePacket(ed, mb);

        if (chunkLen < entryByteLen) {
            /* Start segmented transfer: */
            ed->coeSDOUploadState.active = ESS_TRUE;
            ed->coeSDOUploadState.bytesDone = chunkLen;
            ed->coeSDOUploadState.entry = entry;
            ed->coeSDOUploadState.lastPacket = essGetTime();
            ed->coeSDOUploadState.nextToggleBit = 0;
            ed->coeSDOUploadState.objIdx = objIdx;
            ed->coeSDOUploadState.subIdx = subIdx;
            ed->coeSDOUploadState.totalBytes = entryByteLen;
        } else {
            (void)essOD_DoEntryEventCallback(ed, objIdx, subIdx, ESS_OD_ENTRY_CALLBACK_COMPLETED_UPLOAD);
        }
    }

    return MBX_ERR_SUCCESS;
}

static uint16_t CoE_Handle_UploadSegment(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    const ESS_BOOL toggleBit = MAKE_ESS_BOOL(mb->data[MBX_COE_SDOREQ_HEADERPOS] & MBX_COE_SDOREQ_HEADER_MASK_TOGGLEBIT);
    ESS_SDOTRANSFER_STATUS* si = &ed->coeSDOUploadState;
    uint16_t chunkLen;
    COE_ABORTCODE ab;
    uint8_t sds;

    if (si->active == ESS_FALSE) {
        ESS_DBG_COE(("WARNING: Upload was not set active previously!?"));
        CoE_SendSDOAbort(ed, mb, 0x0000, 0x00, COE_ABORTCODE_GENERAL);
        return MBX_ERR_SUCCESS;
    }

    if (toggleBit != (ESS_BOOL)si->nextToggleBit) {
        ESS_DBG_COE(("WARNING: Togglebit not as expected!"));
        CoE_FinishSegmentedTransfer(ed, ESS_TRUE, ESS_TRUE);
        CoE_SendSDOAbort(ed, mb, si->objIdx, si->subIdx, COE_ABORTCODE_TOGGLE);
        return MBX_ERR_SUCCESS;
    }

    chunkLen = (uint16_t)MIN((int32_t)(si->totalBytes - si->bytesDone), ed->mbxInDataLen - 3);
    mb->data[MBX_COE_HEADERPOS + 1] |= 0x10; /* Turn CoE Service "SDO Request" into "SDO Response" */
    mb->length = 3 + chunkLen;
    if (mb->length < 10) {
        mb->length = 10;
        sds = 7 - (uint8_t)chunkLen;
    } else {
        sds = 0;
    }

    /* Get entry data: */
    ab = essOD_DoEntryUpload(ed, &si->entry, &mb->data[3], si->bytesDone, chunkLen, ((si->bytesDone + chunkLen) >= si->totalBytes));
    if (ab != COE_ABORTCODE_NONE) {
        CoE_SendSDOAbort(ed, mb, si->objIdx, si->subIdx, ab);
        CoE_FinishSegmentedTransfer(ed, ESS_TRUE, ESS_TRUE);
        return MBX_ERR_SUCCESS;
    }

    si->bytesDone += chunkLen;
    if (si->bytesDone >= si->totalBytes) {
        mb->data[MBX_COE_SDOINFO_HEADERPOS] = MBX_COE_SDORESP_UPLSEG_HDR(ESS_TRUE, sds, toggleBit);
        CoE_FinishSegmentedTransfer(ed, ESS_TRUE, ESS_FALSE);

    } else {
        mb->data[MBX_COE_SDOINFO_HEADERPOS] = MBX_COE_SDORESP_UPLSEG_HDR(ESS_FALSE, sds, toggleBit);
        si->lastPacket = essGetTime();
        si->nextToggleBit = !toggleBit;
    }

    MBX_WritePacket(ed, mb);
    return MBX_ERR_SUCCESS;
}

static ESS_BOOL CoE_DoPDODoMapping_VerifySMBufferSize(ESC_DATA* ed, ESS_SM sm)
{
#if (CFG_ESS_SMCOPYINUSERBUFFER)
    const ESS_SM_USERBUFFER* const ub = &ed->config->smBufferCopies[sm];

    ed->smBuffer[sm] = (uint8_t*)ub->data;

    return (ed->cbDataSM[sm].smLen > ub->capacity);
#else
    ESS_DBG(("Ensuring smBuffer len for SM%u: %u (current: %u)", sm, ed->cbDataSM[sm].smLen, ed->smBufferAllocSize[sm]));

    if (ed->smBufferAllocSize[sm] < ed->cbDataSM[sm].smLen) {
        uint16_t allocSize = MAX(ed->cbDataSM[sm].smLen, CFG_ESS_MIN_SM_ALLOC);
        uint8_t* newBuf;

        ESS_REALLOC(newBuf, uint8_t*, ed->smBuffer[sm], allocSize);
        if (newBuf) {
            ed->smBuffer[sm] = newBuf;
            ed->smBufferAllocSize[sm] = allocSize;
        } else {
            ESS_DBG(("ERROR: failed to alloc %u bytes for SM%u buffer copy!", allocSize, sm));
            return ESS_TRUE;
        }
    }

    return ESS_FALSE;
#endif
}

static ESS_BOOL CoE_DoPDODoMapping_VerifySMMappedEntriesSize(ESC_DATA* ed, ESS_SM sm)
{
    if (ed->smMappedEntries[sm].entriesCount >= ed->smMappedEntries[sm].entriesCapacity) {
        ESS_PDO_MAP_INFO* pNew;

        ESS_REALLOC(pNew, ESS_PDO_MAP_INFO*, ed->smMappedEntries[sm].entries, sizeof(*pNew) * (ed->smMappedEntries[sm].entriesCapacity + CFG_ESS_COE_MAPPED_LIST_INCREMENT));
        if (pNew) {
            ed->smMappedEntries[sm].entries = pNew;
            ed->smMappedEntries[sm].entriesCapacity += CFG_ESS_COE_MAPPED_LIST_INCREMENT;
        } else {
            ESS_DBG(("ERROR: failed to realloc smMappedEntries[%u]!", sm));
            return ESS_TRUE;
        }
    }

    return ESS_FALSE;
}

/*
 * Iterate list of entries mapped into this PDO and check if the following
 * requirements for an optimized copy operation are met:
 *   - The physical start address is byte aligned
 *   - All entries have a byte granularity
 *   - The destination address of each entry is consecutive.
 * Returns PDO size in bit or 0 in case of any error.
 */
static uint16_t CoE_PDOSize(const uint32_t physBitStart, const ESC_DATA* ed, const ESS_OD_OBJECT_INFOS_PDO* pdoInfos)
{
	int iEntry;
	uint32_t physBitPos = physBitStart;
	uint16_t errCode;
	uint8_t *p1, *p2;

	for (iEntry = 0, p1 = NULL; iEntry < pdoInfos->itemsUsed; iEntry++) {
		const ESS_PDO_ENTRY* pdoEntry = &pdoInfos->items[iEntry];
        if(0x0 == pdoEntry->objIdx) {
            /* Dummy mapping entry */
            if(p1 != NULL) {
                p1 += (pdoEntry->len >> 3);
            }
        } else {
            /* Existing mapping entry */
		    errCode = essOD_GetEntryData(ed, pdoEntry->objIdx, pdoEntry->subIdx, &p2);
            if(errCode != 0) {
                ESS_DBG_COE(("WARNING: Object (0x%04x/0x%02x) lookup error %u.",
                    pdoEntry->objIdx, pdoEntry->subIdx, errCode));
                return(0);
            } else if((p1 != NULL) && (p1 != p2)) {
			    ESS_DBG_COE(("WARNING: Object (0x%04x/0x%02x) address nonconsecutive.",
				    pdoEntry->objIdx, pdoEntry->subIdx));
			    return(0);
		    }
		    p1 = p2 + (pdoEntry->len >> 3);
        }
        physBitPos += pdoEntry->len;
        if((physBitPos & 0x7) != 0) {
            ESS_DBG_COE(("WARNING: Object (0x%04x/0x%02x) failed optimized copy granularity requirement.",
                pdoEntry->objIdx, pdoEntry->subIdx));
            return(0);
        }
	} /* for each PDO entry */

	return((uint16_t)(physBitPos - physBitStart));
}

/*
 * Called when changing state to SafeOp. Returns true when error was found (err code was written to ed->cbDataStateRequest.. then) */
static uint16_t CoE_DoPDODoMapping(ESC_DATA* ed)
{
    ESS_SM iSM;

    /* Find PDOs assigned to each SM, objects in there, and then do the mapping (i.e. calc phys addr of the entries): */
    for (iSM = 0; iSM < ed->smCount; iSM++) {
        ESS_SM_MAPPINGS *pSmMapEntries = &ed->smMappedEntries[iSM];
        const ESS_OD_OBJECT_INFOS_PDOASSIGN* pdoAssignInfos;
        uint16_t errCode; /* REG_VAL_ALSTATUSCODE_... */
        uint32_t physBitPos;
        int iPDO;

        pSmMapEntries->entriesCount = 0;  /* Clear mapping */
        physBitPos = (uint32_t)8 * ed->cbDataSM[iSM].smPhysAddr;

        /* SM disabled, or wrong mailbox type ? continue! */
        if ((!(ed->cbDataSM[iSM].smActivate & REG_MASK_SMACTIVATE_SMENABLE)) || ((ed->cbDataSM[iSM].smType != SM_TYPE_OUTPUTS) && (ed->cbDataSM[iSM].smType != SM_TYPE_INPUTS)))
            continue;

        /* Get assignment object, 0x1c12/0x1c13, etc. */
        errCode = essOD_GetPDOAssignInfos(ed, iSM, &pdoAssignInfos);
        if (errCode != REG_VAL_ALSTATUSCODE_NOERROR)
            return errCode;

        ESS_DBG_COE(("INFO: %u PDOs assigned to SM%u:", pdoAssignInfos->itemsUsed, iSM));

        /* Loop through all PDOs (0x1600/0x1a00 etc.) and handle their entries: */
        for (iPDO = 0; iPDO < pdoAssignInfos->itemsUsed; ++iPDO) {
            ESS_PDO_MAP_INFO* mapInfo;
            const ESS_OD_OBJECT_INFOS_PDO* pdoInfos;
            int iEntry;

            ESS_DBG_COE(("  PDO 0x%.4x:", pdoAssignInfos->items[iPDO]));

            errCode = essOD_GetPDOInfos(ed, pdoAssignInfos->items[iPDO], &pdoInfos);
            if (errCode != REG_VAL_ALSTATUSCODE_NOERROR)
                return errCode;

            for (mapInfo = NULL, iEntry = 0; iEntry < pdoInfos->itemsUsed; ++iEntry) {
                const ESS_PDO_ENTRY* pdoEntry = &pdoInfos->items[iEntry];
                ESS_BOOL isMapExtend = ESS_IS_EXTEND_OBJECT(pdoEntry) ? ESS_TRUE : ESS_FALSE;

                ESS_DBG_COE(("    Obj. 0x%.4x.%.2x with %d bit is within that PDO", pdoEntry->objIdx, pdoEntry->subIdx, pdoEntry->len));

                if ((pdoEntry->objIdx != 0) || isMapExtend) {
                    if ((physBitPos + pdoEntry->len) > (uint32_t)((ed->cbDataSM[iSM].smPhysAddr + ed->cbDataSM[iSM].smLen) * 8)) {
                        ESS_DBG(("    Mapping entry 0x%.4x.%.2x to address 0x%.4x.%u would exceed SM len of %u byte!", pdoEntry->objIdx, pdoEntry->subIdx, physBitPos >> 3, physBitPos % 8, ed->cbDataSM[iSM].smLen));
                        return REG_VAL_ALSTATUSCODE_INVALIDSMCFG;
                    }

                    /* Add infos about mapped the entry to the ed->smMappedEntries array: */
                    if (CoE_DoPDODoMapping_VerifySMMappedEntriesSize(ed, iSM))
                        return REG_VAL_ALSTATUSCODE_NOMEMORY;

                    /* &ed->smMappedEntries[iSM].entries[ed->smMappedEntries[iSM].entriesCount] */
                    mapInfo = &pSmMapEntries->entries[pSmMapEntries->entriesCount];

                    if(isMapExtend) {
                        ESS_PDO_MAP_INFO* mapInfoPrev = &pSmMapEntries->entries[pSmMapEntries->entriesCount - 1];
                        mapInfo->data = mapInfoPrev->data + (mapInfoPrev->bitLen / 8);
                    } else {
                        errCode = essOD_GetEntryData(ed, pdoEntry->objIdx, pdoEntry->subIdx, &mapInfo->data);
                        if (errCode != REG_VAL_ALSTATUSCODE_NOERROR)
                            return errCode;
                    }

                    /* CHECKME: PDO Mapping: also verify mapped len and actual entry len? */

                    mapInfo->bufferOffset = (uint16_t)(physBitPos >> 3) - ed->cbDataSM[iSM].smPhysAddr;
                    mapInfo->addrBit = physBitPos % 8;
                    mapInfo->bitLen = pdoEntry->len;
                    ed->smMappedEntries[iSM].entriesCount++;

                    ESS_DBG_COE(("    Mapped it to 0x%.4x.%u (data @%p)", ed->cbDataSM[iSM].smPhysAddr + mapInfo->bufferOffset, mapInfo->addrBit, mapInfo->data));
                }

                /*
                 * For the optimized copy create just one large map entry for the complete
                 * PDO instead one mapping entry for each object mapped into this PDO
                 */
                
                if(pdoInfos->flags & ESS_PDO_CFG_FLAGS_BLOCK_COPY) {
                    uint16_t pdoBitSize = CoE_PDOSize(physBitPos, ed, pdoInfos);
                    if(0 == pdoBitSize)
                        return(REG_VAL_ALSTATUSCODE_ESD_MAP_MAPPEDENTRYERR);
                    ESS_DBG_COE(("    This is the start of an optimized PDO copy with %u bits", pdoBitSize));
                    ESS_SET_OPTIMZED_COPY_SIZE(mapInfo, pdoBitSize);
                    physBitPos += pdoBitSize;
                    break;
                }
                
                physBitPos += pdoEntry->len;

            } /* for each PDO entry */

        } /* for each PDO */

        /* Mapping less is OK. (Needed for Beckhoff's tools, they round SM sizes up to even len) */
        if ((((physBitPos + 7) >> 3) - ed->cbDataSM[iSM].smPhysAddr) > ed->cbDataSM[iSM].smLen) {
            ESS_DBG(("    Mapped %u byte to SM%u, but SM's len is only %u byte!", ((physBitPos + 7) >> 3) - ed->cbDataSM[iSM].smPhysAddr, iSM, ed->cbDataSM[iSM].smLen));
            return REG_VAL_ALSTATUSCODE_INVALIDSMCFG;
        }

        /* Allocate buffer for our SM buffer copy: */
        if (CoE_DoPDODoMapping_VerifySMBufferSize(ed, iSM))
            return REG_VAL_ALSTATUSCODE_NOMEMORY;

    } /* for each sm */

    return REG_VAL_ALSTATUSCODE_NOERROR;
}

/* Called by SM_HandleMBoxOutInterrupt(), just calls the CoE_Handle_ functions */
static uint16_t MBX_HandleCoE(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    const uint16_t coeHdr = MAKE_WORD(mb->data[0], mb->data[1]);
    const uint8_t coeService = MBX_COE_GET_SERVICE(coeHdr);

    if (mb->length < MBX_COE_HEADERLEN) {
        ESS_DBG_COE(("WARNING: Received CoE packet with invalid len (%u)", mb->length));
        return MBX_ERR_LENTOOSHORT;
    }

    switch (coeService) {
        case MBX_COE_SERVICE_SDOINFO:
# if (CFG_ESS_COE_SUPPORT_SDO_INFO_SERVICE)
            if (mb->length < (MBX_COE_SDOINFO_HEADERLEN + MBX_COE_SDOINFO_HEADERPOS)) {
                ESS_DBG_COE(("WARNING: Received CoE SDOInfo packet with invalid len (%u)", mb->length));
                return MBX_ERR_LENINVALID;
            }

            if (mb->data[MBX_COE_SDOINFO_HEADERPOS] & MBX_COE_SDOINFO_HEADER_MASK_INCOMPLETE) {
                ESS_DBG_COE(("WARNING: Fragmented SDO Info service not supported"));
                return MBX_ERR_NORESOURCES;
            }

            /*lint -e527 [Unreachable code] */
            switch (mb->data[2] & MBX_COE_SDOINFO_HEADER_MASK_OPCODE) {
                case MBX_COE_SDOINFO_OP_ODLISTREQ:
                    return CoE_Handle_ListRequest(ed, mb);

                case MBX_COE_SDOINFO_OP_OBJREQ:
                    return CoE_Handle_ObjectRequest(ed, mb);

                case MBX_COE_SDOINFO_OP_ENTRYREQ:
                    return CoE_Handle_EntryRequest(ed, mb);

                default:
                    ESS_DBG_COE(("WARNING: CoE SDOInfo OpCode %u not supported", mb->data[2] & MBX_COE_SDOINFO_HEADER_MASK_OPCODE));
                    return MBX_ERR_SERVICENOTSUPPORTED;
            }

#else
            return MBX_ERR_SERVICENOTSUPPORTED;
#endif

            break;  /* Never reached. Suppress Eclipse static code analysis warning */

        case MBX_COE_SERVICE_SDOREQUEST:
            if (mb->length < (MBX_COE_SDOREQ_HEADERLEN + MBX_COE_SDOREQ_HEADERPOS)) {
                ESS_DBG_COE(("WARNING: Received CoE SDORequest packet with invalid len (%u)", mb->length));
                return MBX_ERR_LENINVALID;
            }

            /*lint -e527 [Unreachable code] */
            switch (MBX_COE_SDOREQ_HEADER_GET_CMDSPEC(mb->data[MBX_COE_SDOREQ_HEADERPOS])) {
                case MBX_COE_SDOREQ_CMDSPECIFIER_DOWNLOAD:
                    return CoE_Handle_Download(ed, mb);

                case MBX_COE_SDOREQ_CMDSPECIFIER_DOWNLOADSEGMENT:
                    return CoE_Handle_DownloadSegment(ed, mb);

                case MBX_COE_SDOREQ_CMDSPECIFIER_UPLOAD:
                    return CoE_Handle_Upload(ed, mb);

                case MBX_COE_SDOREQ_CMDSPECIFIER_UPLOADSEGMENT:
                    return CoE_Handle_UploadSegment(ed, mb);

                case MBX_COE_SDOREQ_CMDSPECIFIER_ABORT:
                    return CoE_Handle_SDOAbort(ed, mb);

                default:
                    {
                        const uint16_t objIdx = MAKE_WORD(mb->data[MBX_COE_SDOREQ_HEADERPOS + 1], mb->data[MBX_COE_SDOREQ_HEADERPOS + 2]);
                        const uint8_t subIdx = mb->data[MBX_COE_SDOREQ_HEADERPOS + 3];

                        ESS_DBG_COE(("WARNING: Unknown cmdSpecifier: %u", MBX_COE_SDOREQ_HEADER_GET_CMDSPEC(mb->data[MBX_COE_SDOREQ_HEADERPOS])));
                        CoE_SendSDOAbort(ed, mb, objIdx, subIdx, COE_ABORTCODE_CCS_SCS);
                    }
                    return MBX_ERR_SUCCESS;
            }

            break;  /* Never reached. Suppress Eclipse static code analysis warning */

        case 0: /* To please the CTT: */
            return MBX_ERR_HEADERINVALID;

        default:
            ESS_DBG_COE(("WARNING: Unsupported CoE service: %u", coeService));
            return MBX_ERR_SERVICENOTSUPPORTED;
    }
}

#endif /* CFG_ESS_SUPPORT_COE */




#if (CFG_ESS_SUPPORT_EOE)

/* Called by MBX_HandleEoE() */
static Mbx_Err MBX_HandleEoE_SetIPParams(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    ESS_CBDATA_EOE_SETIPPARAM cbData;
    uint16_t currPos = 8;
    char dnsName[33];

    /*
    Code is more complex than necessary, cause initial idea was to have "currPos" incremented only
    when item is included...
    */

    if (EOE_HDR_GET_IPPARAM_MACINCLUDED(mb)) {
        cbData.macAddress = &mb->data[currPos];
    } else {
        cbData.macAddress = NULL;
    }
    currPos += 6;
    if (currPos > mb->length) {
        ESS_DBG(("ERROR: EoE: not enough data in set ip param packet!"));
        return MBX_ERR_LENTOOSHORT;
    }

    if (EOE_HDR_GET_IPPARAM_IPADDRINCLUDED(mb)) {
        cbData.ipAddress = &mb->data[currPos];
    } else {
        cbData.ipAddress = NULL;
    }
    currPos += 4;
    if (currPos > mb->length) {
        ESS_DBG(("ERROR: EoE: not enough data in set ip param packet!"));
        return MBX_ERR_LENTOOSHORT;
    }

    if (EOE_HDR_GET_IPPARAM_SUBNMASKINCLUDED(mb)) {
        cbData.subnetMask = &mb->data[currPos];
    } else {
        cbData.subnetMask = NULL;
    }
    currPos += 4;
    if (currPos > mb->length) {
        ESS_DBG(("ERROR: EoE: not enough data in set ip param packet!"));
        return MBX_ERR_LENTOOSHORT;
    }

    if (EOE_HDR_GET_IPPARAM_DEFGWINCLUDED(mb)) {
        cbData.defaultGateway = &mb->data[currPos];
    } else {
        cbData.defaultGateway = NULL;
    }
    currPos += 4;
    if (currPos > mb->length) {
        ESS_DBG(("ERROR: EoE: not enough data in set ip param packet!"));
        return MBX_ERR_LENTOOSHORT;
    }

    if (EOE_HDR_GET_IPPARAM_DNSSERVERINCLUDED(mb)) {
        cbData.dnsServerIP = &mb->data[currPos];
    } else {
        cbData.dnsServerIP = NULL;
    }
    currPos += 4;
    if (currPos > mb->length) {
        ESS_DBG(("ERROR: EoE: not enough data in set ip param packet!"));
        return MBX_ERR_LENTOOSHORT;
    }

    if (EOE_HDR_GET_IPPARAM_DNSNAMEINCLUDED(mb)) {
        if ((currPos + 32) > mb->length) {
            ESS_DBG(("ERROR: EoE: not enough data in set ip param packet!"));
            return MBX_ERR_LENTOOSHORT;
        }
        ESS_MEMCOPY(dnsName, &mb->data[currPos], 32);
        dnsName[32] = '\0';
        cbData.dnsName = dnsName;
    } else {
        cbData.dnsName = NULL;
    }

    cbData.result = EOE_RESULTCODE_UNSPECIFIEDERR;
    cbData.hDev = (ESS_HANDLE)ed;

    if (ED_CALLBACK(ed, cbEoESetIPParam)) {
        ED_DO_CALLBACK(ed, cbEoESetIPParam, &cbData);
    }

    mb->length = 4;
    mb->data[0] = EOE_FRAMETYPE_SETIPPARAMRRESPONSE; /* Frametype = 3, Port = 0 */
    mb->data[1] = 1; /* Last fragment = 1, Time appended/requested = 0, reserved = 0 */
    mb->data[2] = LO_BYTE(cbData.result);
    mb->data[3] = HI_BYTE(cbData.result);

    MBX_WritePacket(ed, mb);
    return MBX_ERR_SUCCESS;
}

/* Called by MBX_HandleEoE() */
static Mbx_Err MBX_HandleEoE_SetAddrFilter(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    ESS_CBDATA_EOE_SETADDRFILTER cbData;

    cbData.hDev = (ESS_HANDLE)ed;
    cbData.filterBroadcasts = MAKE_ESS_BOOL(EOE_HDR_GET_SETFILTER_INHIBITBROADCAST(mb));
    cbData.macAddrCount = EOE_HDR_GET_SETFILTER_MACFILTERCNT(mb);
    cbData.macAddrFilterCount = EOE_HDR_GET_SETFILTER_MACFILTERMASKCNT(mb);
    cbData.macAddresses = &mb->data[6];
    cbData.macAddrFilter = &mb->data[6 + cbData.macAddrCount * 6];

    cbData.result = EOE_RESULTCODE_UNSPECIFIEDERR;
    if (ED_CALLBACK(ed, cbEoESetAddrFilter)) {
        ED_DO_CALLBACK(ed, cbEoESetAddrFilter, &cbData);
    }

    mb->length = 4;
    mb->data[0] = EOE_FRAMETYPE_SETADDRFILTERRESPONSE; /* Frametype = 5, Port = 0 */
    mb->data[1] = 1; /* Last fragment = 1, Time appended/requested = 0, reserved = 0 */
    mb->data[2] = LO_BYTE(cbData.result);
    mb->data[3] = HI_BYTE(cbData.result);

    MBX_WritePacket(ed, mb);
    return MBX_ERR_SUCCESS;
}

/* Called by MBX_HandleEoE() */
static Mbx_Err MBX_HandleEoE_Fragment(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    uint8_t fragmNo = EOE_HDR_GET_FRAGMENTNUMBER(mb);
    ESS_BOOL lastFragment = MAKE_ESS_BOOL(EOE_HDR_GET_LASTFRAGMENT(mb));
    uint16_t fragmLen;

    if (fragmNo > 0x2f) {
        ESS_DBG(("WARNING: EoE: invalid fragment number! (%u)", fragmNo));
        return MBX_ERR_HEADERINVALID;
    }

    if (fragmNo == 0) {
        ed->eoeReceive.lastFragmNo = 0;
        ed->eoeReceive.timeAppended = MAKE_ESS_BOOL(EOE_HDR_GET_TIMEAPPENDED(mb));
        ed->eoeReceive.currDataLen = 0;
        ed->eoeReceive.frameNo = EOE_HDR_GET_FRAMENUMBER(mb);

    } else {
        if ((ed->eoeReceive.lastFragmNo + 1) != fragmNo) {
            ESS_DBG(("WARNING: EoE: Last fragment was %u, but this is %u!?", ed->eoeReceive.lastFragmNo, fragmNo));
            return 0xffff; /* arbitrary... */
        }

        ed->eoeReceive.lastFragmNo = fragmNo;
    }

    fragmLen = (lastFragment) ? ((ed->eoeReceive.timeAppended) ? mb->length - EOE_HEADERLEN - 4 : mb->length - EOE_HEADERLEN) : mb->length - EOE_HEADERLEN;
    if ((ed->eoeReceive.currDataLen + fragmLen) > sizeof(ed->eoeReceive.data)) {
        ESS_DBG(("ERROR: EoE fragment: too many data!"));
        return MBX_ERR_LENINVALID;
    }
    ESS_MEMCOPY(&ed->eoeReceive.data[ed->eoeReceive.currDataLen], &mb->data[EOE_HEADERLEN], fragmLen);
    ed->eoeReceive.currDataLen += fragmLen;

    /* ESS_DBG(("INFO: EoE: ======= Fragment%u%s: len = %u, %s = %u, Rcvd len now = %u", fragmNo, (lastFragment) ? " (LAST)" : "", fragmLen, (fragmNo) ? "fragment offset" : "total chunks len", EOE_HDR_GET_COMPLETESIZE(mb) << 5, ed->eoeReceive.currDataLen, ed->eoeReceive.currDataLen)); */

    if (lastFragment) {
        ESS_CBDATA_EOE_FRAME cbData;

        cbData.hDev = (ESS_HANDLE)ed;
        cbData.data = ed->eoeReceive.data;
        cbData.frameNo = ed->eoeReceive.frameNo;

        cbData.dataLen = ed->eoeReceive.currDataLen;
        /* CHECKME: enable timestamps for received EoE frames? (Might already work, but could not be tested) */
        /* "reserved2" => "timeStampValid" */
        cbData.reserved2 = ed->eoeReceive.timeAppended;
        /* "reserved1" => "timeStamp" */
        cbData.reserved1 = (cbData.reserved2) ? EOE_HDR_GET_TIMESTAMP(mb) : 0;

#if (CFG_ESS_SERVE_STATISTICS)
        ed->config->stats->eoeReceived++;
#endif

        if (ED_CALLBACK(ed, cbEoEFrame)) {
            ED_DO_CALLBACK(ed, cbEoEFrame, &cbData);
        }
    }

    return MBX_ERR_SUCCESS;
}

/* Called by MainLoop() when there is an active EOE send transfer and our send mbx is ready */
static void MBX_HandleEoEFramesToSend(ESC_DATA* ed)
{
    const uint16_t maxMbxDataLen = ed->cbDataSM[ed->smIdxMBoxIn].smLen - MBX_HEADERLEN - EOE_HEADERLEN;
    uint16_t maxMbxDataChunks = maxMbxDataLen >> 5; /* 32 byte per chunk */
    uint16_t dataChunks = (unsigned)((ed->eoeSend.dataLen - ed->eoeSend.dataSent) + 31) >> 5;
    uint16_t offsBufSize = (ed->eoeSend.fragmNo == 0) ? ((ed->eoeSend.dataLen + 31) >> 5) : (ed->eoeSend.dataSent >> 5);
    uint16_t fragNoOffsFrameNo = (ed->eoeSend.fragmNo & 0x3f) | ((offsBufSize & 0x3f) << 6) | ((ed->eoeSend.frameNo & 0x0f) << 12);
    ESS_MBX_PACKET* const mbxData = &ed->mbxTempPacket;
    ESS_BOOL lastFragment;

    mbxData->address = 0;
    mbxData->channelAndPrio = 0;
    mbxData->typeAndCounter = MBX_HEADER_MAKE_TYPEANDCOUNTER(MBX_TYPE_EOE, 0); /* counter is set by MBX_WritePacket() */

    if (dataChunks <= maxMbxDataChunks) {
        lastFragment = ESS_TRUE;
    } else {
        lastFragment = ESS_FALSE;
        dataChunks = maxMbxDataChunks;
    }

    mbxData->data[0] = EOE_FRAMETYPE_FRAGMENTREQUEST; /* Port = 0 */
    mbxData->data[1] = (uint8_t)lastFragment;
    mbxData->data[2] = LO_BYTE(fragNoOffsFrameNo);
    mbxData->data[3] = HI_BYTE(fragNoOffsFrameNo);
    mbxData->length = ((ed->eoeSend.dataSent >= ed->eoeSend.dataLen) ? 0 : ((lastFragment) ? (ed->eoeSend.dataLen - ed->eoeSend.dataSent) : (dataChunks << 5)) & 0xFFFF);

    ESS_MEMCOPY(&mbxData->data[EOE_HEADERLEN], &ed->eoeSend.data[ed->eoeSend.dataSent], mbxData->length);
    ed->eoeSend.dataSent += mbxData->length;
    mbxData->length += EOE_HEADERLEN;

    ed->eoeSend.fragmNo++;
    if (lastFragment) {
        ed->eoeSend.active = ESS_FALSE;
#if (CFG_ESS_SERVE_STATISTICS)
        ed->config->stats->eoeSent++;
#endif
    }

    MBX_WritePacket(ed, mbxData);
}

/* Called when an EoE packet is received */
static Mbx_Err MBX_HandleEoE(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    if (mb->length < EOE_HEADERLEN) {
        ESS_DBG(("WARNING: Received EoE packet with invalid len (%u)", mb->length));
        return MBX_ERR_LENTOOSHORT;
    }

    if (EOE_HDR_GET_PORT(mb) != 0) {
        ESS_DBG(("WARNING: Received EoE packet for specific port: not supported! (port must be 0)"));
        return MBX_ERR_PROTOCOLNOTSUPPORTED;
    }

    switch (EOE_HDR_GET_FRAMETYPE(mb)) {
        case EOE_FRAMETYPE_SETIPPARAMREQUEST:
            if (mb->length < EOE_HEADERLEN + 4) {
                ESS_DBG(("WARNING: Received EoE packet (set IP param) with invalid len (%u)", mb->length));
                return MBX_ERR_LENTOOSHORT;
            }
            return MBX_HandleEoE_SetIPParams(ed, mb);

        case EOE_FRAMETYPE_SETADDRFILTERREQUEST:
            if (mb->length <= EOE_HEADERLEN) {
                ESS_DBG(("WARNING: Received EoE packet (set addr filter) with invalid len (%u)", mb->length));
                return MBX_ERR_LENTOOSHORT;
            }
            return MBX_HandleEoE_SetAddrFilter(ed, mb);

        case EOE_FRAMETYPE_FRAGMENTREQUEST:
            if (mb->length <= EOE_HEADERLEN) {
                ESS_DBG(("WARNING: Received EoE packet (fragment rcvd) with invalid len (%u)", mb->length));
                return MBX_ERR_LENTOOSHORT;
            }
            return MBX_HandleEoE_Fragment(ed, mb);

        case EOE_FRAMETYPE_INITRESPONSE:
            return MBX_ERR_PROTOCOLNOTSUPPORTED;

        case EOE_FRAMETYPE_GETIPPARAMREQUEST:
            /* Protocol enhancement, not yet needed: */
            return MBX_ERR_PROTOCOLNOTSUPPORTED;

        case EOE_FRAMETYPE_GETADDRFILTERREQUEST:
            /* Protocol enhancement, not yet needed: */
            return MBX_ERR_PROTOCOLNOTSUPPORTED;

        default:
            ESS_DBG(("WARNING: EoE: unknown frame type: %u", EOE_HDR_GET_FRAMETYPE(mb)));
            return MBX_ERR_PROTOCOLNOTSUPPORTED;
    }
}

#endif /* (CFG_ESS_SUPPORT_EOE) */




#if (CFG_ESS_SUPPORT_FOE)

static void FoE_DoCloseCallback(ESC_DATA* ed, ESS_BOOL isUpload, ESS_BOOL unexpected)
{
    ESS_CBDATA_FOE_CLOSE cbClose;

    cbClose.appHandle = ed->foeState.appHandle;
    cbClose.hDev = (ESS_HANDLE)ed;
    cbClose.isUpload = isUpload;
    cbClose.unexpected = unexpected;

    ED_DO_CALLBACK(ed, cbFoEClose, &cbClose);

    ed->foeState.operation = 0;
}

/*
 * Turns the given mb packet into an OPCODE_ERROR reply and sends it to the master
 * Any BUSY reply error code is turned into FOE_ERRCODE_UNDEFINED.
 */
static void MBX_WriteFoEErrorReply(ESC_DATA* ed, ESS_MBX_PACKET* mb,
                                   uint32_t errCode, const char* errText)
{
    /*lint -e845 [The [left/right] argument ... is certain to be 0] */
    if ((errCode & 0xffffff00) == FOE_ERRCODE_BUSY(0)) {
        errCode = FOE_ERRCODE_UNDEFINED;
    }

    mb->data[0] = FOE_OPCODE_ERROR;
    mb->data[1] = 0;

    mb->data[2] = LO_BYTE(LO_WORD(errCode));
    mb->data[3] = HI_BYTE(LO_WORD(errCode));
    mb->data[4] = LO_BYTE(HI_WORD(errCode));
    mb->data[5] = HI_BYTE(HI_WORD(errCode));

    mb->length = 6;
    if (errText) {
        while ((*errText) && (mb->length < ed->mbxInDataLen)) {
            mb->data[mb->length] = (uint8_t)(*errText);
            mb->length++;
            errText++;
        }
    }

    MBX_WritePacket(ed, mb);
}

/* Turns the given mb packet into a OPCODE_BUSY reply and sends it to the master */
static void MBX_WriteFoEBusyReply(ESC_DATA* ed, ESS_MBX_PACKET* mb, uint8_t percentDone, const char* busyTxt)
{
    mb->data[0] = FOE_OPCODE_BUSY;
    mb->data[1] = 0;

    mb->data[2] = percentDone;
    mb->data[3] = 0;
    mb->data[4] = 0;
    mb->data[5] = 0;

    mb->length = 6;
    if (busyTxt) {
        while ((*busyTxt) && (mb->length < ed->mbxInDataLen)) {
            mb->data[mb->length] = (uint8_t)(*busyTxt);
            mb->length++;
            busyTxt++;
        }
    }

    MBX_WritePacket(ed, mb);

    ed->foeState.lastPacket = essHALGetTime();
}

/*
 * Pass FoE upload request to the application.
 *
 * As long as the application returns FOE_ERRCODE_NONE the data segments are
 * uploaded to the EtherCAT master. Less data as the indicated maximum net
 * mailbox size is interpreted as the last segment. 
 *
 * The application may also return no data and stall or fail the FoE upload.
 */
static void FoE_RequestAppData(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    uint16_t maxDataLen = ed->cbDataSM[ed->smIdxMBoxIn].smLen - MBX_HEADERLEN - FOE_HEADERLEN - FOE_PACKETNOLEN;
    ESS_CBDATA_FOE_DATA cbData;

    /* Request upload data from application */
    cbData.appHandle = ed->foeState.appHandle;
    cbData.destSrc = &mb->data[FOE_HEADERLEN + FOE_PACKETNOLEN];
    cbData.errCode = FOE_ERRCODE_UNDEFINED;
    cbData.errorText = NULL;
    cbData.hDev = (ESS_HANDLE)ed;
    cbData.isUpload = ESS_TRUE;
    cbData.len = maxDataLen;
    cbData.maxDataLen = maxDataLen;
    cbData.offset = ed->foeState.packetNo * cbData.len;

    ED_DO_CALLBACK(ed, cbFoEData, &cbData);

    /* Handle application terminates or stalls the transfer */
    if (cbData.errCode != FOE_ERRCODE_NONE) {
        /*lint -e845 [The [left/right] argument ... is certain to be 0] */
        if ((cbData.errCode & 0xffffff00) == FOE_ERRCODE_BUSY(0)) {
            MBX_WriteFoEBusyReply(ed, mb, cbData.errCode & 0xff, cbData.errorText);
        } else {
            MBX_WriteFoEErrorReply(ed, mb, cbData.errCode, cbData.errorText);
            FoE_DoCloseCallback(ed, ESS_TRUE, ESS_TRUE);
            return;
        }
    }

    /*
     * Configure data phase and return if application requests to stall the transfer
     * without incrementing the packetNo.
     */
    ed->foeState.operation = FOE_OPCODE_READ;
    if(FOE_OPCODE_NONE == ed->foeState.operation) {
        ed->foeState.operation  = FOE_OPCODE_READ;
    }
    if ((cbData.errCode & 0xffffff00) == FOE_ERRCODE_BUSY(0)) {
        return;
    }

    /* Update internal FoE state */
    ed->foeState.packetNo++;
    ed->foeState.lastPacket = essHALGetTime();

    /* Upload application data */
    mb->data[0] = FOE_OPCODE_DATA;
    mb->data[2] = LO_BYTE(LO_WORD(ed->foeState.packetNo));
    mb->data[3] = HI_BYTE(LO_WORD(ed->foeState.packetNo));
    mb->data[4] = LO_BYTE(HI_WORD(ed->foeState.packetNo));
    mb->data[5] = HI_BYTE(HI_WORD(ed->foeState.packetNo));
    mb->length = cbData.len + FOE_HEADERLEN + FOE_PACKETNOLEN;
    MBX_WritePacket(ed, mb);

    /* Application returned less than maxDataLen bytes -> Upload completed */
    if (cbData.len < maxDataLen) {
        ed->foeState.uploadDone = ESS_TRUE;
        FoE_DoCloseCallback(ed, ESS_TRUE, ESS_FALSE);
    }

    return;
}

/* Called when a FoE packet is received */
static Mbx_Err MBX_HandleFoE(ESC_DATA* ed, ESS_MBX_PACKET* mb)
{
    if ((ED_CALLBACK(ed, cbFoEOpen) == NULL) || (ED_CALLBACK(ed, cbFoEClose) == NULL) || (ED_CALLBACK(ed, cbFoEData) == NULL))
        return MBX_ERR_PROTOCOLNOTSUPPORTED;

    if (mb->length < (FOE_HEADERLEN + 4)) {
        ESS_DBG(("WARNING: Received FoE packet with invalid len (%u)", mb->length));
        return MBX_ERR_LENTOOSHORT;
    }

    switch (mb->data[0]) {
        case FOE_OPCODE_READ:
            /* Master initiates FoE upload */
            {
                char fileName[CFG_ESS_MAX_MBX_LEN - MBX_HEADERLEN];
                ESS_CBDATA_FOE_OPEN cbOpen;

                if (ed->foeState.operation) {
                    ESS_DBG(("WARNING: Received FoE Read request while an FoE transfer seems to be still active!"));
                    FoE_DoCloseCallback(ed, (ed->foeState.operation == FOE_OPCODE_READ), ESS_TRUE);
                }

                /* Ask app to open file: */
                cbOpen.appHandle = NULL;
                cbOpen.errCode = FOE_ERRCODE_UNDEFINED;
                cbOpen.errorText = NULL;
                cbOpen.fileName = fileName;
                cbOpen.fileNameLen = mb->length - FOE_HEADERLEN - FOE_PACKETNOLEN;
                ESS_MEMCOPY(fileName, &mb->data[6], cbOpen.fileNameLen);
                fileName[cbOpen.fileNameLen] = '\0';
                cbOpen.hDev = (ESS_HANDLE)ed;
                cbOpen.isUpload = ESS_TRUE;
                cbOpen.maxDataLen = ed->cbDataSM[ed->smIdxMBoxIn].smLen - MBX_HEADERLEN - FOE_HEADERLEN - FOE_PACKETNOLEN;
                cbOpen.password = MAKE_DWORD(mb->data[2], mb->data[3], mb->data[4], mb->data[5]);

                /* Pass open request to application and evaluate result to terminate or proceed  */
                ED_DO_CALLBACK(ed, cbFoEOpen, &cbOpen);
                if (cbOpen.errCode != FOE_ERRCODE_NONE) {
                    MBX_WriteFoEErrorReply(ed, mb, cbOpen.errCode, cbOpen.errorText);
                    return MBX_ERR_SUCCESS;
                } else {
                    ed->foeState.operation  = FOE_OPCODE_NONE;
                    ed->foeState.appHandle  = cbOpen.appHandle;
                    ed->foeState.uploadDone = ESS_FALSE;
                    ed->foeState.packetNo   = 0;
                }

                /* Request upload of 1st data segment from application */
                FoE_RequestAppData(ed, mb);
            }
            return MBX_ERR_SUCCESS;

        case FOE_OPCODE_ACK:
            /* ACK for data chunk we sent to master. Sending next chunk: */
            {
                const uint32_t ackPackNo = MAKE_DWORD(mb->data[2], mb->data[3], mb->data[4], mb->data[5]);

                if ((ed->foeState.uploadDone) && (ed->foeState.operation == 0)) {
                    /* Upload was finished, but size was multiple of maxDataLen. We can ignore this additional ack */
                    ed->foeState.uploadDone = ESS_FALSE;
                    return MBX_ERR_SUCCESS;
                }

                if (ed->foeState.operation != FOE_OPCODE_READ) {
                    ESS_DBG(("WARNING: Received FoE ACK in invalid FoE transfer state (%u)", ed->foeState.operation));
                    return MBX_ERR_SYNTAX;
                }

                if (ackPackNo != ed->foeState.packetNo) {
                    ESS_DBG(("WARNING: Received FoE ACK for packet %u where %u was expected", ackPackNo, ed->foeState.packetNo));
                    MBX_WriteFoEErrorReply(ed, mb, FOE_ERRCODE_WRONGPACKETNO, NULL);
                    FoE_DoCloseCallback(ed, ESS_TRUE, ESS_TRUE);
                    return MBX_ERR_SUCCESS;
                }

                /* Request upload of next data segment from application */
                FoE_RequestAppData(ed, mb);
            }

            return MBX_ERR_SUCCESS;

        case FOE_OPCODE_WRITE:
            /* Master wants to write a file to us (FoE download): */
            {
                char fileName[CFG_ESS_MAX_MBX_LEN - MBX_HEADERLEN];
                ESS_CBDATA_FOE_OPEN cbOpen;

                if (ed->foeState.operation) {
                    ESS_DBG(("WARNING: Received FoE Read request while an FoE transfer seems to be still active!"));
                    FoE_DoCloseCallback(ed, (ed->foeState.operation == FOE_OPCODE_READ), ESS_TRUE);
                }

                /* Ask app to open file: */
                cbOpen.appHandle = NULL;
                cbOpen.errCode = FOE_ERRCODE_UNDEFINED;
                cbOpen.errorText = NULL;
                cbOpen.fileName = fileName;
                cbOpen.fileNameLen = mb->length - FOE_HEADERLEN - FOE_PACKETNOLEN;
                ESS_MEMCOPY(fileName, &mb->data[6], cbOpen.fileNameLen);
                fileName[cbOpen.fileNameLen] = '\0';
                cbOpen.hDev = (ESS_HANDLE)ed;
                cbOpen.isUpload = ESS_FALSE;
                cbOpen.maxDataLen = ed->cbDataSM[ed->smIdxMBoxOut].smLen - MBX_HEADERLEN - FOE_HEADERLEN - FOE_PACKETNOLEN;
                cbOpen.password = MAKE_DWORD(mb->data[2], mb->data[3], mb->data[4], mb->data[5]);

                ED_DO_CALLBACK(ed, cbFoEOpen, &cbOpen);

                /* Handle application terminates the transfer */
                if (cbOpen.errCode != FOE_ERRCODE_NONE) {
                    MBX_WriteFoEErrorReply(ed, mb, cbOpen.errCode, cbOpen.errorText);
                    return MBX_ERR_SUCCESS;
                }

                /* App accepted, send ACK to master: */
                ed->foeState.appHandle = cbOpen.appHandle;
                ed->foeState.packetNo = 0;
                ed->foeState.lastPacket = essHALGetTime();
                ed->foeState.operation = FOE_OPCODE_WRITE;

                mb->data[0] = FOE_OPCODE_ACK;
                mb->data[2] = LO_BYTE(LO_WORD(ed->foeState.packetNo));
                mb->data[3] = HI_BYTE(LO_WORD(ed->foeState.packetNo));
                mb->data[4] = LO_BYTE(HI_WORD(ed->foeState.packetNo));
                mb->data[5] = HI_BYTE(HI_WORD(ed->foeState.packetNo));
                mb->length = FOE_HEADERLEN + FOE_PACKETNOLEN;
                MBX_WritePacket(ed, mb);
            }
            return MBX_ERR_SUCCESS;

        case FOE_OPCODE_DATA:
            /* Data chunk for file write, answering with FOE_OPCODE_ACK: */
            {
                const uint16_t maxDataLen = ed->cbDataSM[ed->smIdxMBoxOut].smLen - MBX_HEADERLEN - FOE_HEADERLEN - FOE_PACKETNOLEN;
                const uint32_t ackPackNo = MAKE_DWORD(mb->data[2], mb->data[3], mb->data[4], mb->data[5]);
                ESS_CBDATA_FOE_DATA cbData;

                if (ed->foeState.operation != FOE_OPCODE_WRITE) {
                    ESS_DBG(("WARNING: Received FoE DATA whithout an active FoE write transfer"));
                    return MBX_ERR_SYNTAX;
                }

                if (ackPackNo == 0) {
                    ESS_DBG(("WARNING: Received FoE DATA with ack for packet 0"));
                    MBX_WriteFoEErrorReply(ed, mb, FOE_ERRCODE_WRONGPACKETNO, NULL);
                    return MBX_ERR_SUCCESS;
                }

                if (ackPackNo != ed->foeState.packetNo + 1) {
                    ESS_DBG(("WARNING: Received FoE data for packet %u where %u was expected", ackPackNo, ed->foeState.packetNo + 1));
                    MBX_WriteFoEErrorReply(ed, mb, FOE_ERRCODE_WRONGPACKETNO, NULL);
                    FoE_DoCloseCallback(ed, ESS_FALSE, ESS_TRUE);
                    return MBX_ERR_SUCCESS;
                }

                /* Give received data to app: */
                cbData.appHandle = ed->foeState.appHandle;
                cbData.destSrc = &mb->data[FOE_HEADERLEN + FOE_PACKETNOLEN];
                cbData.errCode = FOE_ERRCODE_UNDEFINED;
                cbData.errorText = NULL;
                cbData.hDev = (ESS_HANDLE)ed;
                cbData.isUpload = ESS_FALSE;
                cbData.len = mb->length - FOE_HEADERLEN - FOE_PACKETNOLEN;
                cbData.maxDataLen = maxDataLen;
                cbData.offset = (ackPackNo - 1) * maxDataLen;

                ED_DO_CALLBACK(ed, cbFoEData, &cbData);

                /* Handle application terminates or stalls the transfer */
                if (cbData.errCode != FOE_ERRCODE_NONE) {
                    if ((cbData.errCode & 0xffffff00) == FOE_ERRCODE_BUSY(0)) {
                        MBX_WriteFoEBusyReply(ed, mb, cbData.errCode & 0xff, cbData.errorText);
                    } else {
                        MBX_WriteFoEErrorReply(ed, mb, cbData.errCode, cbData.errorText);
                        FoE_DoCloseCallback(ed, ESS_FALSE, ESS_TRUE);
                    }
                    return MBX_ERR_SUCCESS;
                }

                /* OK, request next packet from master by sending ACK: */
                ed->foeState.packetNo = ackPackNo;
                ed->foeState.lastPacket = essHALGetTime();

                mb->data[0] = FOE_OPCODE_ACK;
                mb->data[2] = LO_BYTE(LO_WORD(ed->foeState.packetNo));
                mb->data[3] = HI_BYTE(LO_WORD(ed->foeState.packetNo));
                mb->data[4] = LO_BYTE(HI_WORD(ed->foeState.packetNo));
                mb->data[5] = HI_BYTE(HI_WORD(ed->foeState.packetNo));
                mb->length = FOE_HEADERLEN + FOE_PACKETNOLEN;
                MBX_WritePacket(ed, mb);

                if (cbData.len < maxDataLen) {
                    FoE_DoCloseCallback(ed, ESS_FALSE, ESS_FALSE);
                }
            }
            return MBX_ERR_SUCCESS;

        case FOE_OPCODE_ERROR:
            ESS_DBG(("INFO: Received MBX_FOE_OPCODE_ERROR packet"));
            if (ed->foeState.operation) {
                ESS_DBG(("Aborting FoE transfer due to error packet"));
                FoE_DoCloseCallback(ed, (ed->foeState.operation == FOE_OPCODE_READ), ESS_TRUE);
            }
            return MBX_ERR_SUCCESS;

        case FOE_OPCODE_BUSY:
            /* According to current ETG.1020 (V1.0.0) this is not allowed (only other direction) */
            ESS_DBG(("ERROR: Received MBX_FOE_OPCODE_BUSY packet!"));
            return MBX_ERR_SERVICENOTSUPPORTED;

        default:
            ESS_DBG(("ERROR: Unsupported FoE opcode: %u", mb->data[0]));
            return MBX_ERR_SERVICENOTSUPPORTED;
    }
}

#endif /* (CFG_ESS_SUPPORT_FOE) */




#if (CFG_ESS_SUPPORT_VOE)

/* Called when a VoE packet is received */
static Mbx_Err MBX_HandleVoE(ESC_DATA* ed, ESS_MBX_PACKET* mb, ESS_BOOL fragment)
{
    ESS_CBDATA_VOE cbData;

    if (ED_CALLBACK(ed, cbVoE) == NULL)
        return MBX_ERR_PROTOCOLNOTSUPPORTED;

    if (mb->length < MBX_VOE_HEADERLEN)
        return MBX_ERR_LENTOOSHORT;

    cbData.hDev = (ESS_HANDLE)ed;
    cbData.mb = mb;
    cbData.mbMaxLen = ed->mbxInDataLen;
    cbData.result = MBX_ERR_PROTOCOLNOTSUPPORTED;
    cbData.fragment = fragment;

    ED_DO_CALLBACK(ed, cbVoE, &cbData);
    if (cbData.result == MBX_ERR_SUCCESS) {
        MBX_WritePacket(ed, mb);
        ed->voeMoreFragments = cbData.fragment;
    }

    return cbData.result;
}

#endif /* CFG_ESS_SUPPORT_VOE */



#if (CFG_ESS_SUPPORT_SOE)

/* Called when an SoE packet is received */
static Mbx_Err MBX_HandleSoE(ESC_DATA* ed, ESS_MBX_PACKET* mb, ESS_BOOL fragment)
{
    ESS_CBDATA_SOE cbData;

    if (ED_CALLBACK(ed, cbSoE) == NULL)
        return MBX_ERR_PROTOCOLNOTSUPPORTED;

    if (mb->length < MBX_SOE_HEADERLEN)
        return MBX_ERR_LENTOOSHORT;

    cbData.hDev = (ESS_HANDLE)ed;
    cbData.mb = mb;
    cbData.mbMaxLen = ed->mbxInDataLen;
    cbData.result = MBX_ERR_PROTOCOLNOTSUPPORTED;
    cbData.fragment = fragment;

    ED_DO_CALLBACK(ed, cbSoE, &cbData);
    if (cbData.result == MBX_ERR_SUCCESS) {
        MBX_WritePacket(ed, mb);
        ed->soeMoreFragments = cbData.fragment;
    }

    return cbData.result;
}

#endif /* CFG_ESS_SUPPORT_SOE */



#if (CFG_ESS_SUPPORT_AOE)

/* Called when an AoE packet is received */
static Mbx_Err MBX_HandleAoE(ESC_DATA* ed, ESS_MBX_PACKET* mb, ESS_BOOL fragment)
{
    ESS_CBDATA_AOE cbData;

    if (ED_CALLBACK(ed, cbAoE) == NULL)
        return MBX_ERR_PROTOCOLNOTSUPPORTED;

    if (mb->length < 32)
        return MBX_ERR_LENTOOSHORT;

    cbData.hDev = (ESS_HANDLE)ed;
    cbData.mb = mb;
    cbData.mbMaxLen = ed->mbxInDataLen;
    cbData.result = MBX_ERR_PROTOCOLNOTSUPPORTED;
    cbData.fragment = fragment;

    ED_DO_CALLBACK(ed, cbAoE, &cbData);
    if (cbData.result == MBX_ERR_SUCCESS) {
        MBX_WritePacket(ed, mb);
        ed->aoeMoreFragments = cbData.fragment;
    }

    return cbData.result;
}

#endif /* CFG_ESS_SUPPORT_AOE */





/*
 * Checks mailbox input/output SMs (Start/Size/Control).
 *
 * For a state change into BOOTSTRAP the special bootstrap mailbox configuration
 *  is used (if defined by the application) and state changes to INIT always
 *  succeed.
 * 
 * RETURN:
 *   Returns ESS_TRUE in case of an error, otherwise ESS_FALSE.
 *
 * REMARK:
 *   In case of an error AL status code is written to ed->cbDataStateRequest and
 *   a forced state change to INIT is requested.
 */
static ESS_BOOL SM_VerifyMailboxSMs(ESC_DATA* ed, ESC_STATE requestedState)
{
    uint16_t errCode = REG_VAL_ALSTATUSCODE_NOERROR;
    int i, j;

    /* No input/output mailbox SM configuration check for transition to INIT */
    if(ESC_STATE_INIT == requestedState) {
        return ESS_FALSE;
    }

    for (i = 0; i < ed->config->smConfigCount; i++) {
        const ESS_SM_CONFIGURATION* smc = &ed->config->smConfigs[i];
        const ESS_CBDATA_SM* sm;
        uint8_t  mbxType = (smc->contrByte & REG_MASK_SMCONTROL_TYPE);

        if ((mbxType != SM_TYPE_MBXOUT) && (mbxType != SM_TYPE_MBXIN))
            continue;

        if (i >= ed->smCount) {
            ESS_DBG(("ERROR: Not enough SMs (SM%i missing)", i));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMCFG;
            break;
        }
        sm = &ed->cbDataSM[i];
        if(requestedState == ESC_STATE_BOOT) {
            const ESS_SM_CONFIGURATION* smcb = &ed->config->smConfigs[ed->config->smConfigCount];
            for(j = 0; j < ed->config->smConfigCountBootstrap; j++, smcb++) {
                if(mbxType == (smcb->contrByte & REG_MASK_SMCONTROL_TYPE)) {
                    smc = smcb;
                    break;
                }
            }
        }

        if ((sm->smActivate & REG_MASK_SMACTIVATE_SMENABLE) == 0) {
            ESS_DBG(("ERROR: SM%i (%s) is not enabled, but it should be!", i,
                     SM_TYPE_TO_STRING(smc->contrByte & SM_TYPE_MASK)));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDMBXCFGINPREOP;
            break;
        }
        if ((smc->contrByte & REG_MASK_SMCONTROL_TYPE) != (sm->smContrByte & REG_MASK_SMCONTROL_TYPE)) {
            ESS_DBG(("ERROR: Control byte of SM%i (%s?) not as expected (0x%.2x != 0x%.2x)", i,
                     SM_TYPE_TO_STRING(smc->contrByte & SM_TYPE_MASK), sm->smContrByte, smc->contrByte));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDMBXCFGINPREOP;
            break;
        }
        if ((smc->startAddr) && (sm->smPhysAddr != smc->startAddr)) {
            ESS_DBG(("ERROR: SM%i (%s) physAddr (0x%.4x) differs from configured one (0x%.4x)", i,
                      SM_TYPE_TO_STRING(sm->smType & SM_TYPE_MASK), sm->smPhysAddr, smc->startAddr));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDMBXCFGINPREOP;
            break;
        }
        if ((smc->maxSize > 0) && (sm->smLen > smc->maxSize)) {
            ESS_DBG(("ERROR: SM%i (%s) len > configured maxSize (%u > %u)", i,
                     SM_TYPE_TO_STRING(sm->smType & SM_TYPE_MASK), sm->smLen, smc->maxSize));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDMBXCFGINPREOP;
            break;
        }
        if (sm->smLen < smc->minSize) {
            ESS_DBG(("ERROR: SM%i (%s) len < configured minSize (%u < %u)", i,
                     SM_TYPE_TO_STRING(sm->smType & SM_TYPE_MASK), sm->smLen, smc->minSize));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDMBXCFGINPREOP;
            break;
        }
    }

    if(REG_VAL_ALSTATUSCODE_NOERROR == errCode) {
        ed->mbxOutDataLen = ed->cbDataSM[ed->smIdxMBoxOut].smLen - MBX_HEADERLEN;
        ed->mbxInDataLen = ed->cbDataSM[ed->smIdxMBoxIn].smLen - MBX_HEADERLEN;
        return ESS_FALSE;
    } else {
        if((REG_VAL_ALSTATUSCODE_INVALIDMBXCFGINPREOP == errCode) && (ESC_STATE_BOOT == requestedState)) {
            errCode = REG_VAL_ALSTATUSCODE_INVALIDMBXCFGINBOOT;
        }
        ed->cbDataStateRequest.statusCode = errCode;
        ed->cbDataStateRequest.newState   = ESC_STATE_INIT;
        return ESS_TRUE;
    }
}

/*
 * Check PD output SMs (Start/Size/Control).
 * 
 * RETURN:
 *   Returns ESS_TRUE in case of an error, otherwise ESS_FALSE.
 *
 * REMARK:
 *   In case of an error AL status code is written to ed->cbDataStateRequest and
 *   a forced state change to the current state is requested.
 *
 *   If CoE is supported a CoE EMCY is generated.
 */
static ESS_BOOL SM_VerifyOutputSMs(ESC_DATA* ed)
{
    uint16_t errCode = REG_VAL_ALSTATUSCODE_NOERROR;
    ESS_SM i;

    for (i = 0; i < ed->config->smConfigCount; i++) {
        const ESS_SM_CONFIGURATION* smc = &ed->config->smConfigs[i];
        const ESS_CBDATA_SM* sm;
        ESS_BOOL mustBeEnabled;

        if ((smc->contrByte & REG_MASK_SMCONTROL_TYPE) != SM_TYPE_OUTPUTS)
            continue;

        if (i >= ed->smCount) {
            ESS_DBG(("ERROR: Not enough SMs (SM%u missing)", i));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMCFG;
            break;
        }
        sm = &ed->cbDataSM[i];
        mustBeEnabled = (sm->smLen > 0);

        if (MAKE_ESS_BOOL(sm->smActivate & REG_MASK_SMACTIVATE_SMENABLE) != mustBeEnabled) {
            ESS_DBG(("ERROR: SM%u (%s) is %senabled, but it should %sbe!", i,
                     SM_TYPE_TO_STRING(smc->contrByte & SM_TYPE_MASK),
                     (sm->smActivate & REG_MASK_SMACTIVATE_SMENABLE) ? "" : "not ", (mustBeEnabled) ? "" : "not "));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG;
#if (CFG_ESS_SUPPORT_COE)
            CoE_CreateEmergency_SMSettingsErr(&ed->coeEmergency, i, ESS_FALSE,
                                              smc->contrByte & REG_MASK_SMCONTROL_OPMODE,
                                              (smc->contrByte & REG_MASK_SMCONTROL_DIRECTION) >> 2,
                                              (smc->contrByte & REG_MASK_SMCONTROL_PDIINT) != 0, mustBeEnabled);
            ed->coeEmergencyReady = ESS_TRUE;
#endif
            break;
        }

        if ((smc->contrByte & REG_MASK_SMCONTROL_TYPE) != (sm->smContrByte & REG_MASK_SMCONTROL_TYPE)) {
            if ((sm->smLen == 0) && (sm->smContrByte == 0)) { /* Allow disabling unused SM: */
                ESS_DBG(("WARNING: Control byte of SM%u (%s?) not as expected (0x%.2x != 0x%.2x)", i,
                         SM_TYPE_TO_STRING(smc->contrByte & SM_TYPE_MASK), sm->smContrByte, smc->contrByte));

            } else {
                ESS_DBG(("ERROR: Control byte of SM%u (%s?) not as expected (0x%.2x != 0x%.2x)", i,
                         SM_TYPE_TO_STRING(smc->contrByte & SM_TYPE_MASK), sm->smContrByte, smc->contrByte));
                errCode = REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG;
#if (CFG_ESS_SUPPORT_COE)
                CoE_CreateEmergency_SMSettingsErr(&ed->coeEmergency, i, ESS_FALSE,
                                                  smc->contrByte & REG_MASK_SMCONTROL_OPMODE,
                                                  (smc->contrByte & REG_MASK_SMCONTROL_DIRECTION) >> 2,
                                                  (smc->contrByte & REG_MASK_SMCONTROL_PDIINT) != 0, mustBeEnabled);
                ed->coeEmergencyReady = ESS_TRUE;
#endif
                break;
            }
        }

        if ((smc->startAddr) && (sm->smPhysAddr != smc->startAddr)) {
            ESS_DBG(("ERROR: SM%u (%s) physAddr (0x%.4x) differs from configured one (0x%.4x)", i,
                     SM_TYPE_TO_STRING(sm->smType & SM_TYPE_MASK), sm->smPhysAddr, smc->startAddr));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG;
#if (CFG_ESS_SUPPORT_COE)
            CoE_CreateEmergency_SMAddrErr(&ed->coeEmergency, i, ESS_FALSE, smc->startAddr, smc->startAddr);
            ed->coeEmergencyReady = ESS_TRUE;
#endif
            break;
        }

        if ((smc->maxSize > 0) && (sm->smLen > smc->maxSize)) {
            ESS_DBG(("ERROR: SM%u (%s) len > configured maxSize (%u > %u)", i,
                SM_TYPE_TO_STRING(sm->smType & SM_TYPE_MASK), sm->smLen, smc->maxSize));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG;
#if (CFG_ESS_SUPPORT_COE)
            CoE_CreateEmergency_SMLenErr(&ed->coeEmergency, i, ESS_FALSE,
                                         ed->config->smConfigs[i].minSize, ed->config->smConfigs[i].maxSize);
            ed->coeEmergencyReady = ESS_TRUE;
#endif
            break;
        }

        if (sm->smLen < smc->minSize) {
            ESS_DBG(("ERROR: SM%u (%s) len < configured minSize (%u < %u)", i,
                     SM_TYPE_TO_STRING(sm->smType & SM_TYPE_MASK), sm->smLen, smc->minSize));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG;
#if (CFG_ESS_SUPPORT_COE)
            CoE_CreateEmergency_SMLenErr(&ed->coeEmergency, i, ESS_FALSE,
                                         ed->config->smConfigs[i].minSize, ed->config->smConfigs[i].maxSize);
            ed->coeEmergencyReady = ESS_TRUE;
#endif
            break;
        }

    } /* Foreach SM */

    if (REG_VAL_ALSTATUSCODE_NOERROR == errCode) {
        return ESS_FALSE;
    } else {
        ESS_DBG(("SM_VerifyOutputsSMs: FOUND ERROR %04x\n", errCode));
        ed->cbDataStateRequest.statusCode = errCode;
        ed->cbDataStateRequest.newState = ed->currState;
        return ESS_TRUE;
    }
}

/*
 * Check PD input SMs (Start/Size/Control).
 * 
 * RETURN:
 *   Returns ESS_TRUE in case of an error, otherwise ESS_FALSE.
 *
 * REMARK:
 *   In case of an error AL status code is written to ed->cbDataStateRequest and
 *   a forced state change to the current state is requested.
 *
 *   If CoE is supported a CoE EMCY is generated.
 */
static ESS_BOOL SM_VerifyInputSMs(ESC_DATA* ed)
{
    uint16_t errCode = REG_VAL_ALSTATUSCODE_NOERROR;
    uint8_t i;

    for (i = 0; i < ed->config->smConfigCount; i++) {
        const ESS_SM_CONFIGURATION* smc = &ed->config->smConfigs[i];
        const ESS_CBDATA_SM* sm;
        ESS_BOOL mustBeEnabled;

        if ((smc->contrByte & REG_MASK_SMCONTROL_TYPE) != SM_TYPE_INPUTS)
            continue;

        if (i >= ed->smCount) {
            ESS_DBG(("ERROR: Not enough SMs (SM%u missing)", i));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMCFG;
            break;
        }
        sm = &ed->cbDataSM[i];
        mustBeEnabled = (sm->smLen > 0);

        if (MAKE_ESS_BOOL(sm->smActivate & REG_MASK_SMACTIVATE_SMENABLE) != mustBeEnabled) {
            ESS_DBG(("ERROR: SM%u (%s) is %senabled, but it should %sbe!", i,
                     SM_TYPE_TO_STRING(smc->contrByte & SM_TYPE_MASK),
                     (sm->smActivate & REG_MASK_SMACTIVATE_SMENABLE) ? "" : "not ", (mustBeEnabled) ? "" : "not "));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMINCFG;
#if (CFG_ESS_SUPPORT_COE)
            CoE_CreateEmergency_SMSettingsErr(&ed->coeEmergency, i, ESS_FALSE,
                                              smc->contrByte & REG_MASK_SMCONTROL_OPMODE,
                                              (smc->contrByte & REG_MASK_SMCONTROL_DIRECTION) >> 2,
                                              (smc->contrByte & REG_MASK_SMCONTROL_PDIINT) != 0, mustBeEnabled);
            ed->coeEmergencyReady = ESS_TRUE;
#endif
            break;
        }
        if ((smc->contrByte & REG_MASK_SMCONTROL_TYPE) != (sm->smContrByte & REG_MASK_SMCONTROL_TYPE)) {
            if ((sm->smLen == 0) && (sm->smContrByte == 0)) { /* Allow disabling unused SM: */
                ESS_DBG(("WARNING: Control byte of SM%u (%s?) not as expected (0x%.2x != 0x%.2x)", i,
                          SM_TYPE_TO_STRING(smc->contrByte & SM_TYPE_MASK),
                          sm->smContrByte, smc->contrByte));

            } else {
                ESS_DBG(("ERROR: Control byte of SM%u (%s?) not as expected (0x%.2x != 0x%.2x)", i,
                         SM_TYPE_TO_STRING(smc->contrByte & SM_TYPE_MASK), sm->smContrByte, smc->contrByte));
                errCode = REG_VAL_ALSTATUSCODE_INVALIDSMINCFG;
#if (CFG_ESS_SUPPORT_COE)
                CoE_CreateEmergency_SMSettingsErr(&ed->coeEmergency, i, ESS_TRUE,
                                                  smc->contrByte & REG_MASK_SMCONTROL_OPMODE,
                                                  (smc->contrByte & REG_MASK_SMCONTROL_DIRECTION) >> 2,
                                                  (smc->contrByte & REG_MASK_SMCONTROL_PDIINT) != 0, mustBeEnabled);
                ed->coeEmergencyReady = ESS_TRUE;
#endif
                break;
            }
        }
        if ((smc->startAddr) && (sm->smPhysAddr != smc->startAddr)) {
            ESS_DBG(("ERROR: SM%u (%s) physAddr (0x%.4x) differs from configured one (0x%.4x)", i,
                SM_TYPE_TO_STRING(smc->contrByte & SM_TYPE_MASK), sm->smPhysAddr, smc->startAddr));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMINCFG;
#if (CFG_ESS_SUPPORT_COE)
            CoE_CreateEmergency_SMAddrErr(&ed->coeEmergency, i, ESS_TRUE, smc->startAddr, smc->startAddr);
            ed->coeEmergencyReady = ESS_TRUE;
#endif
            break;
        }
        if ((smc->maxSize > 0) && (sm->smLen > smc->maxSize)) {
            ESS_DBG(("ERROR: SM%u (%s) len > configured maxSize (%u > %u)", i,
                     SM_TYPE_TO_STRING(sm->smType & SM_TYPE_MASK), sm->smLen, smc->maxSize));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMINCFG;
#if (CFG_ESS_SUPPORT_COE)
            CoE_CreateEmergency_SMLenErr(&ed->coeEmergency, i, ESS_TRUE,
                                         ed->config->smConfigs[i].minSize, ed->config->smConfigs[i].maxSize);
            ed->coeEmergencyReady = ESS_TRUE;
#endif
            break;
        }
        if (sm->smLen < smc->minSize) {
            ESS_DBG(("ERROR: SM%u (%s) len < configured minSize (%u < %u)", i,
                      SM_TYPE_TO_STRING(sm->smType & SM_TYPE_MASK), sm->smLen, smc->minSize));
            errCode = REG_VAL_ALSTATUSCODE_INVALIDSMINCFG;
#if (CFG_ESS_SUPPORT_COE)
            CoE_CreateEmergency_SMLenErr(&ed->coeEmergency, i, ESS_TRUE,
                                         ed->config->smConfigs[i].minSize, ed->config->smConfigs[i].maxSize);
            ed->coeEmergencyReady = ESS_TRUE;
#endif
            break;
        }

    }

    if (REG_VAL_ALSTATUSCODE_NOERROR == errCode) {
        return ESS_FALSE;
    } else {
        ESS_DBG(("SM_VerifyInputsSMs: FOUND ERROR %04x\n", errCode));
        ed->cbDataStateRequest.statusCode = errCode;
        ed->cbDataStateRequest.newState = ed->currState;
        return ESS_TRUE;
    }
}

/* Returns 0 when no error was found */
static uint16_t SM_VerifyNoOverlapping(ESC_DATA* ed, ESC_STATE requestedState)
{
    int i;

    for (i = 0; i < ed->smCount; i++) {
        int j;

        if ((ed->cbDataSM[i].smActivate & REG_MASK_SMACTIVATE_SMENABLE) == 0)
            continue;

        /* Ignore inputs and outputs when going back to init: */
        if ((requestedState == ESC_STATE_INIT) && ((ed->cbDataSM[i].smType == SM_TYPE_INPUTS) || (ed->cbDataSM[i].smType == SM_TYPE_OUTPUTS)))
            continue;
        /* Ignore inputs and outputs when going to PreOp: */
        if ((requestedState == ESC_STATE_PREOP) && ((ed->cbDataSM[i].smType == SM_TYPE_INPUTS) || (ed->cbDataSM[i].smType == SM_TYPE_OUTPUTS)))
            continue;

        for (j = i + 1; j < ed->smCount; j++) {
            if ((ed->cbDataSM[j].smActivate & REG_MASK_SMACTIVATE_SMENABLE) == 0)
                continue;

            /* Ignore inputs and outputs when going back to init: */
            if ((requestedState == ESC_STATE_INIT) && ((ed->cbDataSM[j].smType == SM_TYPE_INPUTS) || (ed->cbDataSM[j].smType == SM_TYPE_OUTPUTS)))
                continue;
            /* Ignore inputs and outputs when going to PreOp: */
            if ((requestedState == ESC_STATE_PREOP) && ((ed->cbDataSM[j].smType == SM_TYPE_INPUTS) || (ed->cbDataSM[j].smType == SM_TYPE_OUTPUTS)))
                continue;

            if (((ed->cbDataSM[i].smPhysAddr >= ed->cbDataSM[j].smPhysAddr) && (ed->cbDataSM[i].smPhysAddr < (ed->cbDataSM[j].smPhysAddr + ed->cbDataSM[j].smLen))) || ((ed->cbDataSM[j].smPhysAddr >= ed->cbDataSM[i].smPhysAddr) && (ed->cbDataSM[j].smPhysAddr < (ed->cbDataSM[i].smPhysAddr + ed->cbDataSM[i].smLen)))) {
                ESS_DBG(("ERROR: SM%i overlaps with SM%i!", i, j));
                return MAKE_WORD(i, j);
            }
        }
    }

    return 0;
}

/* Enables a SM (SM PDI control bit0) */
static void SM_Enable(ESC_DATA* ed, ESS_SM smIdx)
{
    uint8_t currSMPDIContr;

    ESC_READ8(ed, ESC_REG_SMPDICONTROL(smIdx), currSMPDIContr);
    ESS_DBG_SYNCMANAGERS(("SMs: Enabling SM%u...%s", smIdx, (currSMPDIContr & 1) ? "" : " (was already)"));

    if (currSMPDIContr & 1) { /* Disabled bit */
        currSMPDIContr &= ~1;
        ESC_WRITE8(ed, ESC_REG_SMPDICONTROL(smIdx), currSMPDIContr);
        ED_DO_EVENT2(ed, ESS_EVENT_WROTEPDICONTROL, smIdx, currSMPDIContr);
    }

    ED_DO_EVENT3(ed, ESS_EVENT_SMENABLED, smIdx, ed->cbDataSM[smIdx].smPhysAddr, ed->cbDataSM[smIdx].smLen);
}

/* Disables a SM (SM PDI control bit0) */
static void SM_Disable(ESC_DATA* ed, ESS_SM smIdx)
{
    uint8_t currSMPDIContr;

    ESC_READ8(ed, ESC_REG_SMPDICONTROL(smIdx), currSMPDIContr);
    ESS_DBG_SYNCMANAGERS(("SMs: Disabling SM%u...%s", smIdx, (currSMPDIContr & 1) ? " (was already)" : ""));

    if (!(currSMPDIContr & 1)) { /* Disabled bit */
        currSMPDIContr |= 1;
        ESC_WRITE8(ed, ESC_REG_SMPDICONTROL(smIdx), currSMPDIContr);
        ED_DO_EVENT2(ed, ESS_EVENT_WROTEPDICONTROL, smIdx, currSMPDIContr);
    }

    ED_DO_EVENT3(ed, ESS_EVENT_SMDISABLED, smIdx, ed->cbDataSM[smIdx].smPhysAddr, ed->cbDataSM[smIdx].smLen);
}

/* Updates ed->smCount, ed->cbDataSM[] and ed->smIdxMBoxOut etc. */
INLINE static void SM_UpdateAll(ESC_DATA* ed)
{
    uint8_t i;

    ed->smIdxMBoxOut = ESS_SM_NONE;
    ed->smIdxMBoxIn = ESS_SM_NONE;

    for (i = 0; i < ed->smCount; i++) {
        uint32_t smCfg[2];

        ED_DO_EVENT1(ed, ESS_EVENT_READSMACTIVATE, i);
        /* Remember: reading smActivate clears SM activate bit in AL_EVENT_REQUEST */
        ESC_READ32_NOSWAP(ed, ESC_REG_SMCONTROL(i), smCfg[1]);
        ESC_READ32(ed, ESC_REG_SMPHYSADDR(i), smCfg[0]);

        /*lint -e419 [Apparent data overrun for function] */
	ESS_MEMCOPY(&ed->cbDataSM[i].smPhysAddr, smCfg, sizeof(smCfg));
#if CFG_BIG_ENDIAN
        SWAP_WORD(ed->cbDataSM[i].smPhysAddr, ed->cbDataSM[i].smLen);
#endif
        ed->cbDataSM[i].smIdx  = i;     /* Restore SM index overwritten above */

        ed->cbDataSM[i].smType = ed->cbDataSM[i].smContrByte & REG_MASK_SMCONTROL_TYPE;

        if ((ed->cbDataSM[i].smLen > CFG_ESS_MAX_MBX_LEN) && ((ed->cbDataSM[i].smType == SM_TYPE_MBXOUT) || (ed->cbDataSM[i].smType == SM_TYPE_MBXIN))) {
            ESS_DBG(("ERROR: Mailbox too large! (%u), max supported len is %i", ed->cbDataSM[i].smLen, CFG_ESS_MAX_MBX_LEN));
            ed->cbDataSM[i].smLen = CFG_ESS_MAX_MBX_LEN;
        }

        ED_DO_EVENT3(ed, ESS_EVENT_SMUPDATED, i, ed->cbDataSM[i].smPhysAddr, ed->cbDataSM[i].smLen);

        if (ed->cbDataSM[i].smType == SM_TYPE_MBXOUT) {
            ed->smIdxMBoxOut = i;

        } else if (ed->cbDataSM[i].smType == SM_TYPE_MBXIN) {
            ed->smIdxMBoxIn = i;

            if (ed->mbxSendFull) {
                ESS_DBG_MAILBOX_TRANSFER(("MBX: Send mbx full while new settings rcvd, assuming empty now (SM CTRL says %s)", (ed->cbDataSM[i].smContrByte & REG_MASK_SMSTATUS_MBXFULL) ? "FULL" : "EMPTY"));
                ESS_DBG_MBX_MINIMAL(("<~~ SMs changed with pkt waitin for readout by master. Set mbox to empty! (SM CTRL: was %s)...", (ed->cbDataSM[i].smContrByte & REG_MASK_SMSTATUS_MBXFULL) ? "FULL" : "empty"));
                /* Assuming packet was read out by master is not possible! */
                ed->mbxSendFull = ESS_FALSE;
            }
            if ((ed->cbDataSM[i].smLen > 0) && (ed->cbDataSM[i].smPhysAddr >= 0x1000)) { /* these checks just to be sure... (SMs will be verified in state transitions) */
                /* Write first byte to lock the send mailbox: */
                ED_DO_EVENT3(ed, ESS_EVENT_WRITEMBOXINSTART, ed->cbDataSM[i].smPhysAddr, ed->cbDataSM[i].smLen, i);
                ESC_WRITE8(ed, ed->cbDataSM[i].smPhysAddr, (uint8_t)ed->mbxLastStoredPacket.length);
            }
        }

    } /* for each sm */
}

/*
 * Safe update of all SMs with a retry of the update if the SMs are changed
 * during the read access to the ESC.
 */
static void SM_UpdateAllSafe(ESC_DATA* ed)
{
    int totalTries = 16;
    volatile uint16_t alEvent;

    do {
        SM_UpdateAll(ed); /* Reads 0x0806 etc. which clears the SMCHANGED bit in AL Event Req */

        ESC_READ16(ed, ESC_REG_ALEVENTREQ, alEvent);

        ESS_DBG_SYNCMANAGERS(("SMs: Updated SMs. %s", (alEvent & REG_MASK_ALEVENT_SMCHANGED) ? \
            "SM changed during update, starting over..." : "Finished."));

    } while ((alEvent & REG_MASK_ALEVENT_SMCHANGED) && (totalTries--));

    if (totalTries < 0) { /* Workaround for a special ESC only... */
        ESC_READ16(ed, ESC_REG_ALEVENTREQ, alEvent);
        ESC_WRITE16(ed, ESC_REG_ALEVENTREQ, alEvent & ~REG_MASK_ALEVENT_SMCHANGED);
        ESC_READ16(ed, ESC_REG_ALEVENTREQ, alEvent);
        ESS_DBG(("WARNING: SMCHANGED bit in 0x0220%s stuck!", (alEvent & REG_MASK_ALEVENT_SMCHANGED) ? "" : " WAS"));
    }

#if DEBUG_SYNCMANAGERS
    {
        int i;

        for (i = 0; i < ed->smCount; i++) {
            if ((ed->cbDataSM[i].smActivate & REG_MASK_SMACTIVATE_SMENABLE) || (i < ed->config->smConfigCount))
                ESS_DBG(("INFO: SM%u: %s, %u byte @0x%.4x, activate = 0x%.2x, control = 0x%.2x", i,
                         SM_TYPE_TO_STRING(ed->cbDataSM[i].smType & SM_TYPE_MASK), ed->cbDataSM[i].smLen,
                         ed->cbDataSM[i].smPhysAddr, ed->cbDataSM[i].smActivate, ed->cbDataSM[i].smContrByte));
        }
    }
#endif

    if ((ed->smIdxMBoxOut == ESS_SM_NONE) || (ed->smIdxMBoxIn == ESS_SM_NONE)) {
        ed->mbxActive = ESS_FALSE;
    }
}


/* Called from main loop when AL EVENT signals an SM change event */
static void SM_HandleChange(ESC_DATA* ed, ESC_STATE reqState)
{
#if defined(_DEBUG)
    if(reqState == ed->currState) {
        ESS_DBG(("INFO: SM changed event (%u SMs, curr state = %s, curr statusCode = 0x%.4x)",
            ed->smCount, ESC_STATE_TO_STRING(ed->currState), ed->cbDataStateRequest.statusCode));
    }
#endif

    SM_UpdateAllSafe(ed);

    if ((ed->cbDataStateRequest.statusCode == REG_VAL_ALSTATUSCODE_NOERROR) && (ed->currState != ESC_STATE_INIT)) {
        uint16_t overlapCheck;

        /* Check Mailbox SM configuration: */
        if (SM_VerifyMailboxSMs(ed, reqState)) {
            ESS_DBG(("WARNING: Invalid mailbox SM! going back to %s with AL status code (0x%.4x)...",
                    ESC_STATE_TO_STRING(ed->cbDataStateRequest.newState), ed->cbDataStateRequest.statusCode));
            DoStateChangeCallback(ed);
            return;
        }

        /* No input/output SM configuration check for transition to PREOP */
        if(ESC_STATE_PREOP == reqState) {
            return;
        }

        if ((ed->currState == ESC_STATE_SAFEOP) || (ed->currState == ESC_STATE_OP)) {
            /* Check Inputs SM configuration: */
            if (SM_VerifyInputSMs(ed)) {
                ESS_DBG(("WARNING: Invalid input SMs! going back to PreOp..."));
                ed->cbDataStateRequest.newState = ESC_STATE_PREOP;
                ed->cbDataStateRequest.statusCode = REG_VAL_ALSTATUSCODE_INVALIDSMINCFG;
                DoStateChangeCallback(ed);
                return;
            }

            /* Check Outputs SM configuration: */
            if (SM_VerifyOutputSMs(ed)) {
                ESS_DBG(("WARNING: Invalid output SMs! going back to PreOp..."));
                ed->cbDataStateRequest.newState = ESC_STATE_PREOP;
                ed->cbDataStateRequest.statusCode = REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG;
                DoStateChangeCallback(ed);
                return;
            }
        }

        overlapCheck = SM_VerifyNoOverlapping(ed, ed->currState);
        if (overlapCheck) {
            int CTTKnowsItBetter = 0;

            /*
             * CTT expects us detecting the wrong of two overlapped SMs. As ess is very flexible and has no restrictions on
             * how to use the SMs it's basically not possible! ETG says that the former accepted settings are to be used
             * as base, i.e. when more SMs are activated in PS/SO these new SMs must be one of the wrong one when one overlaps.
             *
             * This code achieves that. (Works only with recommended SM usage, i.e. Mbx SMs < PD SMs)
             */
            if ((ed->currState == ESC_STATE_SAFEOP) || (ed->currState == ESC_STATE_OP)) {
                const uint8_t smL = LO_BYTE(overlapCheck);
                const uint8_t smH = HI_BYTE(overlapCheck);

                if ((ed->cbDataSM[smL].smType == SM_TYPE_MBXIN) || (ed->cbDataSM[smL].smType == SM_TYPE_MBXOUT)) {
                    if (ed->cbDataSM[smH].smType == SM_TYPE_INPUTS) {
                        CTTKnowsItBetter = 1;
                        ed->cbDataStateRequest.newState = ESC_STATE_PREOP;
                        ed->cbDataStateRequest.statusCode = REG_VAL_ALSTATUSCODE_INVALIDSMINCFG;

                    } else if (ed->cbDataSM[smH].smType == SM_TYPE_OUTPUTS) {
                        CTTKnowsItBetter = 1;
                        ed->cbDataStateRequest.newState = ESC_STATE_PREOP;
                        ed->cbDataStateRequest.statusCode = REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG;
                    }
                }
            }

            if (!CTTKnowsItBetter) {
                ed->cbDataStateRequest.newState = ESC_STATE_INIT;
                ed->cbDataStateRequest.statusCode = REG_VAL_ALSTATUSCODE_INVALIDSMCFG;
            }

            ESS_DBG(("WARNING: SM overlapping detected!"));
            DoStateChangeCallback(ed);
        }
    }
}

/* Only called when send mailbox is empty */
static void SM_HandleMBoxOutInterrupt(ESC_DATA* ed)
{
    ESS_MBX_PACKET* mbxData = &ed->mbxTempPacket;
    uint16_t res = MBX_ERR_SUCCESS;
    uint8_t type;
    uint8_t cnt;

    MBX_ReadPacket(ed, mbxData);
    ed->mbxOutInterruptPending = ESS_FALSE;

    type = MBX_HEADER_GET_TYPE(mbxData->typeAndCounter);
    cnt = MBX_HEADER_GET_COUNTER(mbxData->typeAndCounter);

    ESS_DBG_MAILBOX_TRANSFER(("MBX: Received packet from master with cnt = %d (prev was %d)", cnt, ed->mbxCntMaster));
    ESS_DBG_MBX_MINIMAL(("<== Read cnt %d that master wrote to us (prev was %d)", cnt, ed->mbxCntMaster));

    if ((cnt != 0) && (ed->mbxCntMaster != 0)) {
        if (cnt == ed->mbxCntMaster) {
            ESS_DBG_MAILBOX_TRANSFER(("MBX: Master's mailbox service counter indicates a repeat request, ignoring that packet..."));
            /* Repeat bit in "SM activate" is used to actually handle the repeat request */
            return;
        } else if (cnt != (1 + (ed->mbxCntMaster % 7))) {
            ESS_DBG_MAILBOX_TRANSFER(("MBX: Unexpected mbx counter received => aborting segmented CoE transfers"));
#if (CFG_ESS_SUPPORT_COE)
            if (ed->coeSDODownloadState.active)
                CoE_FinishSegmentedTransfer(ed, ESS_FALSE, ESS_TRUE);

            if (ed->coeSDOUploadState.active)
                CoE_FinishSegmentedTransfer(ed, ESS_TRUE, ESS_TRUE);
#endif
        }
    }
    ed->mbxCntMaster = cnt;

    if ((MBX_HEADER_GET_CHANNEL(mbxData->channelAndPrio) != 0) && ((type == MBX_TYPE_EOE) || (type == MBX_TYPE_COE) || (type != MBX_TYPE_FOE))) {
        ESS_DBG(("WARNING: Won't handle MBox service of type %u, cause channel is 0x%.2x (pri = 0x%.2x)", type, MBX_HEADER_GET_CHANNEL(mbxData->channelAndPrio), MBX_HEADER_GET_PRIORITY(mbxData->channelAndPrio)));
        MBX_WriteErrorReply(ed, mbxData, MBX_ERR_CHANNELINVALID);
        return;
    }

    switch (type) {
        case MBX_TYPE_ERROR:
            ESS_DBG(("WARNING: Received mbx packet: Error (unsupported)"));
            res = MBX_ERR_PROTOCOLNOTSUPPORTED;
            break;

        case MBX_TYPE_EOE:
#if (CFG_ESS_SUPPORT_EOE)
            res = MBX_HandleEoE(ed, mbxData);
#else
            res = MBX_ERR_PROTOCOLNOTSUPPORTED;
#endif
            break;

        case MBX_TYPE_COE:
#if (CFG_ESS_SUPPORT_COE)
            res = MBX_HandleCoE(ed, mbxData);
#else
            res = MBX_ERR_PROTOCOLNOTSUPPORTED;
#endif
            break;

        case MBX_TYPE_FOE:
#if (CFG_ESS_SUPPORT_FOE)
            res = MBX_HandleFoE(ed, mbxData);
#else
            res = MBX_ERR_PROTOCOLNOTSUPPORTED;
#endif
            break;

        case MBX_TYPE_VOE:
#if (CFG_ESS_SUPPORT_VOE)
            res = MBX_HandleVoE(ed, mbxData, ESS_FALSE);
#else
            res = MBX_ERR_PROTOCOLNOTSUPPORTED;
#endif
            break;

        case MBX_TYPE_SOE:
#if (CFG_ESS_SUPPORT_SOE)
            res = MBX_HandleSoE(ed, mbxData, ESS_FALSE);
#else
            res = MBX_ERR_PROTOCOLNOTSUPPORTED;
#endif
            break;

        case MBX_TYPE_AOE:
#if (CFG_ESS_SUPPORT_AOE)
            res = MBX_HandleAoE(ed, mbxData, ESS_FALSE);
#else
            res = MBX_ERR_PROTOCOLNOTSUPPORTED;
#endif
            break;

        default:
            ESS_DBG(("WARNING: Received mbx packet: UNKNOWN"));
            res = MBX_ERR_PROTOCOLNOTSUPPORTED;
            break;
    }

    if ((res != MBX_ERR_SUCCESS) && (res != MBX_ERR_ESS_NOREPLY)) {
        ESS_DBG(("WARNING: Failed to handle MBox service %s (%u), res = %u", MBX_TYPE_TO_STRING(type), type, res));
        MBX_WriteErrorReply(ed, mbxData, res);
    }
}

/* Called when master read our mailbox packet */
static void SM_HandleMBoxInInterrupt(ESC_DATA* ed, ESS_SM smIdx)
{
    ESS_DBG_MAILBOX_TRANSFER(("MBX: packet was taken out of our send buffer, cnt = %u (cnt might be wrong)", MBX_HEADER_GET_COUNTER(ed->mbxLastStoredPacket.typeAndCounter)));
    ESS_DBG_MBX_MINIMAL(("<-- Master read cnt %d (mbxLastStoredPacket, copied to mbxLastSentPacket now)", MBX_HEADER_GET_COUNTER(ed->mbxLastStoredPacket.typeAndCounter)));

    ed->mbxSendFull = ESS_FALSE;

#if defined(_DEBUG)
    if ((size_t)(MBX_HEADERLEN + ed->mbxLastStoredPacket.length) > sizeof(ed->mbxLastSentPacket)) {
        ESS_DBG(("ERROR: mbx packet with %u data bytes too large for mbx buffer of %u bytes!", ed->mbxLastStoredPacket.length, (unsigned int)sizeof(ed->mbxLastSentPacket)));
        ed->mbxLastStoredPacket.length = sizeof(ed->mbxLastSentPacket) - MBX_HEADERLEN;
    }
#endif
    ESS_MEMCOPY(&ed->mbxLastSentPacket, &ed->mbxLastStoredPacket, MBX_HEADERLEN + ed->mbxLastStoredPacket.length);

    /* Write first byte to lock (and clear the interrupt of) the mailbox again: */
    ED_DO_EVENT3(ed, ESS_EVENT_WRITEMBOXINSTART, ed->cbDataSM[smIdx].smPhysAddr, ed->cbDataSM[smIdx].smLen, smIdx);
    ESC_WRITE8(ed, ed->cbDataSM[smIdx].smPhysAddr, 0);

#if (CFG_ESS_SUPPORT_SOE)
    if (ed->soeMoreFragments) {
        ESS_MBX_PACKET* mbxData = &ed->mbxLastSentPacket;
        Mbx_Err res;

        ed->soeMoreFragments = ESS_FALSE;
        res = MBX_HandleSoE(ed, mbxData, ESS_TRUE);
        if ((res != MBX_ERR_SUCCESS) && (res != MBX_ERR_ESS_NOREPLY))
            MBX_WriteErrorReply(ed, mbxData, res);

        return;
    }
#endif

#if (CFG_ESS_SUPPORT_VOE)
    if (ed->voeMoreFragments) {
        ESS_MBX_PACKET* mbxData = &ed->mbxLastSentPacket;
        Mbx_Err res;

        ed->voeMoreFragments = ESS_FALSE;
        res = MBX_HandleVoE(ed, mbxData, ESS_TRUE);
        if ((res != MBX_ERR_SUCCESS) && (res != MBX_ERR_ESS_NOREPLY))
            MBX_WriteErrorReply(ed, mbxData, res);

        return;
    }
#endif

#if (CFG_ESS_SUPPORT_AOE)
    if (ed->aoeMoreFragments) {
        ESS_MBX_PACKET* mbxData = &ed->mbxLastSentPacket;
        Mbx_Err res;

        ed->aoeMoreFragments = ESS_FALSE;
        res = MBX_HandleAoE(ed, mbxData, ESS_TRUE);
        if ((res != MBX_ERR_SUCCESS) && (res != MBX_ERR_ESS_NOREPLY))
            MBX_WriteErrorReply(ed, mbxData, res);

        return;
    }
#endif
}

/* Called by MainLoop() for SMs configured for outputs */
static void SM_HandleOutputsInterrupt(ESC_DATA* ed, ESS_SM smIdx)
{
#if (CFG_ESS_SUPPORT_COE)
    const ESS_CBDATA_SM* sm = &ed->cbDataSM[smIdx];
    uint8_t* smBuffer = ed->smBuffer[smIdx];

# if (CFG_HAL_NEEDS_EVENTS)
    ED_DO_EVENT3(ed, ESS_EVENT_NEEDSMADDROUTPUTS, sm->smPhysAddr, sm->smLen, sm->smIdx);

    {
        const uint16_t currAddr = (uint16_t)ED_EVENT_RESULT(ed);

        if ((currAddr < 0x1000) || (currAddr == 0xffff)) {
            ESS_DBG_EVENTS(("ED_DO_EVENT: ERROR: ESS_EVENT_NEEDSMADDROUTPUTS failed! (Returned 0x%.4x, devIdx = %d)", currAddr, ed->devIdx));
        } else {
            ESC_READ(ed, currAddr, smBuffer, sm->smLen);
            ED_DO_EVENT3(ed, ESS_EVENT_READOUTPUTS, sm->smPhysAddr, sm->smLen, sm->smIdx);
        }
    }

# else
    ESC_READ(ed, sm->smPhysAddr, smBuffer, sm->smLen);
# endif

    CoE_SyncOutputMappedEntries(ed, sm, smBuffer);
#endif /* CFG_ESS_SUPPORT_COE */

    ed->cbDataOutputsUpdated.sm |= (1U << smIdx);
}

/* Called by MainLoop() for SMs configured for inputs */
static void SM_HandleInputsInterrupt(ESC_DATA* ed, ESS_SM smIdx)
{
    ed->cbDataInputsUpdated.sm |= (1U << smIdx);
}

/* Enables the mailbox SMs (if any) and sets ed->mbxActive to true */
static void MBX_Start(ESC_DATA* ed)
{
    ESS_DBG_MAILBOX_TRANSFER(("MBX: Enabling MBX SMs..."));

    if ((ed->smIdxMBoxOut == ESS_SM_NONE) || (ed->smIdxMBoxIn == ESS_SM_NONE)) {
        ESS_DBG_MAILBOX_TRANSFER(("MBX: Failed!"));
        return;
    }

    SM_Enable(ed, ed->smIdxMBoxOut);
    SM_Enable(ed, ed->smIdxMBoxIn);
    ed->mbxActive = ESS_TRUE;

    ed->mbxCntMaster = 0;
    ed->mbxCnt = 0;
    ed->mbxRepeatBitMaster = ESS_FALSE;
}

/* Disables the mailbox SMs and sets ed->mbxActive to false */
static void MBX_Stop(ESC_DATA* ed)
{
    ESS_DBG_MAILBOX_TRANSFER(("MBX: Disabling MBX SMs... (MBoxOut = SM%u, MBoxIn = SM%u)", ed->smIdxMBoxOut, ed->smIdxMBoxIn));

    if (ed->smIdxMBoxIn != ESS_SM_NONE) {
        uint8_t currSMPDIContr;

        ESC_READ8(ed, ESC_REG_SMPDICONTROL(ed->smIdxMBoxIn), currSMPDIContr);
        currSMPDIContr &= 0xfd; /* Clear mbx repeat ack */
        currSMPDIContr |= 1; /* Disable */
        ESC_WRITE8(ed, ESC_REG_SMPDICONTROL(ed->smIdxMBoxIn), currSMPDIContr);
        ED_DO_EVENT2(ed, ESS_EVENT_WROTEPDICONTROL, ed->smIdxMBoxIn, currSMPDIContr);
    }

    if (ed->smIdxMBoxOut != ESS_SM_NONE) {
        SM_Disable(ed, ed->smIdxMBoxOut);
    }

#if (CFG_ESS_SUPPORT_COE)
    /* When there's a queued emergency, discard it... */
    if (ed->coeEmergencyReady) {
        ESS_DBG(("WARNING: Discarding CoE emergency with errCode 0x%.4x", ed->coeEmergency.errCode));
        ed->coeEmergencyReady = ESS_FALSE;
    }

# if (CFG_ESS_COE_SUPPORT_SDO_INFO_SERVICE)
    if (ed->sdoInfoState.active) {
        ESS_DBG(("INFO: Segmented SDO Info transfer interrupted!"));
        ed->sdoInfoState.active = ESS_FALSE;
    }
# endif
#endif

    ed->mbxActive = ESS_FALSE;
}

/* Enables input SMs, updates their buffers with entry data and sets ed->inputsActive to true  */
static void Inputs_Start(ESC_DATA* ed)
{
    ESS_SM smIdx;

    ed->inputsActive = ESS_FALSE;
    for (smIdx = 0; smIdx < ed->smCount; smIdx++) {
        ESS_CBDATA_SM* sm = &ed->cbDataSM[smIdx];

        if ((sm->smType == SM_TYPE_INPUTS) && (sm->smPhysAddr) && (sm->smLen)) {
#if (CFG_ESS_SUPPORT_COE)
            if (ed->smMappedEntries[smIdx].entriesCount > 0) {

                CoE_SyncInputMappedEntries(ed, sm, ed->smBuffer[smIdx]);
                ed->inputsActive = ESS_TRUE;
            }
#else
            ed->inputsActive = ESS_TRUE;
#endif
        }
    }

    if (ed->inputsActive) {
        ESS_DBG(("INFO: Have inputs!"));
        DO_INOUT_ACTIV_CB(ed, ESS_TRUE, ESS_TRUE);
    } else {
        ESS_DBG(("INFO: No inputs."));
    }
}

/* Disables input SMs and sets ed->inputsActive to false */
static void Inputs_Stop(ESC_DATA* ed)
{
    ESS_SM smIdx;

    for (smIdx = 0; smIdx < ed->smCount; smIdx++)
        if (ed->cbDataSM[smIdx].smType == SM_TYPE_INPUTS)
            SM_Disable(ed, smIdx);

    ed->inputsActive = ESS_FALSE;

    DO_INOUT_ACTIV_CB(ed, ESS_TRUE, ESS_FALSE);
}

/* Enables output SMs and sets ed->outputsActive to true */
static void Outputs_Start(ESC_DATA* ed)
{
    ESS_SM smIdx;

    ed->outputsActive = ESS_FALSE;
    for (smIdx = 0; smIdx < ed->smCount; smIdx++) {
        ESS_CBDATA_SM* sm = &ed->cbDataSM[smIdx];

        if ((sm->smType == SM_TYPE_OUTPUTS) && (sm->smPhysAddr) && (sm->smLen)) {
#if (CFG_ESS_SUPPORT_COE)
            if (ed->smMappedEntries[smIdx].entriesCount > 0)
#endif
            {
                SM_Enable(ed, (ESS_SM)smIdx);
                ed->outputsActive = ESS_TRUE;
            }
        }
    }

    if (ed->outputsActive) {
        ESS_DBG(("INFO: Have outputs!"));
        DO_INOUT_ACTIV_CB(ed, ESS_FALSE, ESS_TRUE);
    } else {
        ESS_DBG(("INFO: No outputs."));
    }
}

/* Set ed->outputsActive to false and disable SMs if this is the result of error */
static void Outputs_Stop(ESC_DATA* ed, uint16_t errCode)
{
    ESS_SM smIdx;

    if(errCode != REG_VAL_ALSTATUSCODE_NOERROR) {
        for (smIdx = 0; smIdx < ed->smCount; smIdx++)
            if (ed->cbDataSM[smIdx].smType == SM_TYPE_OUTPUTS)
                SM_Disable(ed, smIdx);
    }

    ed->outputsActive = ESS_FALSE;

    DO_INOUT_ACTIV_CB(ed, ESS_FALSE, ESS_FALSE);
}

/* Called after a new state was successfully set (either by stack or application). Needs ed->cbDataStateRequest.transition to be set correctly! */
static void StateChanged(ESC_DATA* ed)
{
    switch (ed->cbDataStateRequest.transition) {

        case ESC_TRANSITION_II:
            /* This transition also occurs when essOpen() is called */
            Outputs_Stop(ed, REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG);
            Inputs_Stop(ed);
            MBX_Stop(ed);
            SET_RUN_LED(ed, ESC_LED_STATE_OFF);
            break;

        case ESC_TRANSITION_IB:
            MBX_Start(ed);
            SET_RUN_LED(ed, ESC_LED_STATE_FLICKER);
            break;

        case ESC_TRANSITION_BB:
            break;

        case ESC_TRANSITION_IP:
            MBX_Start(ed);
            SET_RUN_LED(ed, ESC_LED_STATE_BLINK);
            break;

        case ESC_TRANSITION_PS:
            {   /* Enable ALL SMs: (conformance test verifies this) */
                ESS_SM smIdx;

                for (smIdx = 0; smIdx < ed->smCount; smIdx++)
                    SM_Enable(ed, (ESS_SM)smIdx);
            }
            Inputs_Start(ed);
            SET_RUN_LED(ed, ESC_LED_STATE_FLASH_1);
            break;

        case ESC_TRANSITION_SO:
            Outputs_Start(ed);
            SET_RUN_LED(ed, ESC_LED_STATE_ON);
            break;

        case ESC_TRANSITION_OS:
            Outputs_Stop(ed, ed->cbDataStateRequest.statusCode);
            SET_RUN_LED(ed, ESC_LED_STATE_FLASH_1);
            break;

        case ESC_TRANSITION_OP:
            Outputs_Stop(ed, REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG);
            Inputs_Stop(ed);
            SET_RUN_LED(ed, ESC_LED_STATE_BLINK);
            break;

        case ESC_TRANSITION_SS:
            break;

        case ESC_TRANSITION_SP:
        case ESC_TRANSITION_PP:
            Outputs_Stop(ed, REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG);
            Inputs_Stop(ed);
            SET_RUN_LED(ed, ESC_LED_STATE_BLINK);
            break;

        case ESC_TRANSITION_OI:
        case ESC_TRANSITION_SI:
        case ESC_TRANSITION_PI:
        case ESC_TRANSITION_BI:
            Outputs_Stop(ed, REG_VAL_ALSTATUSCODE_INVALIDSMOUTCFG);
            Inputs_Stop(ed);
            MBX_Stop(ed);
            SET_RUN_LED(ed, ESC_LED_STATE_OFF);
            break;

        case ESC_TRANSITION_OO:
            break;

        default:
            ESS_DBG(("WARNING: Unknown transition: 0x%.4x (Doing nothing...)", ed->cbDataStateRequest.transition));
            break;
    }

    if (ed->cbDataStateRequest.statusCode != REG_VAL_ALSTATUSCODE_NOERROR)
        SET_ERROR_LED(ed, ESC_LED_STATE_BLINK);

    /* When transition is done, further calls should show state only: */
    ed->cbDataStateRequest.transition = MAKE_WORD(ed->currState, ed->currState);
 }

/* Does app callback and, when app did not tell us to try again, calls StateChanged() and actually writes new state to ALSTATUS */
static void DoStateChangeCallback(ESC_DATA* ed)
{
    REG_VAL_ALSTATUSCODE statusCode; 
    uint16_t status;

    ed->cbDataStateRequest.tryAgain = ESS_FALSE;
    ED_DO_CALLBACK(ed, cbStateRequest, &ed->cbDataStateRequest);
    if (ed->cbDataStateRequest.tryAgain)
        return;

    ed->cbDataStateRequest.transition = MAKE_WORD(ed->currState, ed->cbDataStateRequest.newState);
    ed->currState = ed->cbDataStateRequest.newState;
    StateChanged(ed);

    statusCode = ed->cbDataStateRequest.statusCode;
    status = (statusCode != REG_VAL_ALSTATUSCODE_NOERROR) ? (ed->cbDataStateRequest.newState | REG_MASK_ALSTATUS_ERRIND) : ed->cbDataStateRequest.newState;

#if (CFG_ESS_EXPLICIT_DEVICE_ID)
    if(ed->cbDataStateRequest.devIdReq && (REG_VAL_ALSTATUSCODE_NOERROR == statusCode)) {
        status |= REG_MASK_ALSTATUS_LOADDEVID;
        statusCode = ed->cbDataStateRequest.devId;
    }
#endif

    ESC_WRITE16(ed, ESC_REG_ALSTATUSCODE, statusCode);
    ESC_WRITE16(ed, ESC_REG_ALSTATUS, status);
    ED_DO_EVENT(ed, ESS_EVENT_WROTEALSTATUS);
}

/* Just makes sure newState is <= current state etc. */  
INLINE static void VerifyAppStateChange(ESC_DATA* ed, ESC_STATE* newState)
{
    if ((ed->currState == ESC_STATE_BOOT) && (*newState != ESC_STATE_INIT) && (*newState != ESC_STATE_BOOT)) {
        /* In Bootstrap allow only Bootstrap and Init: */
        ESS_DBG(("WARNING: Changing from Bootstrap to %s is not allowed, will remain in Bootstrap", ESC_STATE_TO_STRING(*newState)));
        *newState = ed->currState;

    } else if ((*newState == ESC_STATE_BOOT) && (ed->currState != ESC_STATE_BOOT)) {
        /* To Bootstrap is only possible when already in Bootstrap: */
        ESS_DBG(("WARNING: Changing to Bootstrap from %s is not allowed, will remain in current state", ESC_STATE_TO_STRING(ed->currState)));
        *newState = ed->currState;

    } else if (*newState > ed->currState) {
        /* In all other states just make sure newState is not higher than current state: */
        ESS_DBG(("WARNING: Changing state from %s to %s is not allowed, will remain in current state", ESC_STATE_TO_STRING(ed->currState), ESC_STATE_TO_STRING(*newState)));
        *newState = ed->currState;
    }
}

/* Called from main loop after AL Control was written by the master and before we're calling the application callback. To look for something that would prohibit the state change. Stores results in ed->cbDataStateRequest */
static void CheckStateChange(ESC_DATA* ed, ESC_STATE requestedState)
{
    ESS_DBG(("INFO: Requested state change: %s (0x%.2x) --> %s (0x%.2x), err ack = %s, curr. status code = %.4x ***",
            ESC_STATE_TO_STRING(ed->currState), ed->currState, ESC_STATE_TO_STRING(requestedState),
            requestedState, (ed->cbDataStateRequest.errAck) ? "YES" : "NO", ed->cbDataStateRequest.statusCode));

    if (ed->cbDataStateRequest.statusCode != REG_VAL_ALSTATUSCODE_NOERROR) {
        if (ed->cbDataStateRequest.errAck) {
            ESS_DBG(("    => Err Ack cleared status code 0x%.4x...", ed->cbDataStateRequest.statusCode));
            SET_ERROR_LED(ed, ESC_LED_STATE_OFF);
            ed->cbDataStateRequest.statusCode = REG_VAL_ALSTATUSCODE_NOERROR;
        } else if (requestedState != ESC_STATE_INIT) {
            /* Deny state change when err ack not set: (but to init state is always allowed) */
            ESS_DBG(("    => Denied!: err ack is not set"));
            ed->cbDataStateRequest.newState = ed->currState;
            return;
        }
    }

    if (ESC_TRANSITION_INVALID(ed->cbDataStateRequest.transition)) {
        ESS_DBG(("    => Denied!: Transition from %s (0x%.2x) is not allowed", ESC_STATE_TO_STRING(ed->currState), ed->currState));
        ed->cbDataStateRequest.statusCode = REG_VAL_ALSTATUSCODE_INVALIDALCONTROL;
        ed->cbDataStateRequest.newState = (ed->currState == ESC_STATE_OP) ? ESC_STATE_SAFEOP : ed->currState;
        return;
    }

    if ((requestedState == ESC_STATE_BOOT) && ((ed->config->flags & ESS_CONFIG_FLAGS_APPSUPP_BOOTSTRAP) == 0)) {
        ed->cbDataStateRequest.statusCode = REG_VAL_ALSTATUSCODE_BOOTNOTSUPP;
        ed->cbDataStateRequest.newState = ed->currState;
        return;
    }

    /* Check Mailbox SM configuration: */
    if (((requestedState == ESC_STATE_PREOP) || (requestedState == ESC_STATE_BOOT)) &&
         (SM_VerifyMailboxSMs(ed, requestedState))) {
        return;
    }

    /* Check Process data SM configuration: */
    if (((requestedState == ESC_STATE_SAFEOP) || (requestedState == ESC_STATE_OP))) {
        if (SM_VerifyInputSMs(ed))
            return;

        if (SM_VerifyOutputSMs(ed))
            return;

#if (CFG_ESS_SUPPORT_COE)
        if(requestedState == ESC_STATE_SAFEOP) {
            uint16_t mappResult = CoE_DoPDODoMapping(ed);

            if (mappResult != REG_VAL_ALSTATUSCODE_NOERROR) {
                ed->cbDataStateRequest.statusCode = mappResult;
                ed->cbDataStateRequest.newState = ed->currState;
                return;
            }
        }
#endif
    }

    /* Check all activated SMs for overlapping: */
    if ((requestedState != ESC_STATE_INIT) && (SM_VerifyNoOverlapping(ed, requestedState))) {
        ed->cbDataStateRequest.statusCode = REG_VAL_ALSTATUSCODE_INVALIDSMCFG;
        ed->cbDataStateRequest.newState = ed->currState;
        ESS_DBG(("WARNING: SM overlapping detected!"));
        return;
    }

    /* nothing found: new state is accepted and no error will be indicated: (unless app decides other state/error) */
    ed->cbDataStateRequest.statusCode = REG_VAL_ALSTATUSCODE_NOERROR;
    ed->cbDataStateRequest.newState = requestedState;

    ESS_DBG(("    => OK, may change to %s (application's called to confirm)", ESC_STATE_TO_STRING(requestedState)));
}

#if (CFG_ESS_DC_CALLBACK)

void Sync0_Event(void* arg)
{
    ESC_DATA* ed = (ESC_DATA*)arg;
    ESS_CBDATA_DC cbd;
    
    ESS_DBG_INTERRUPTS(("INTERRUPT: SYNC0 Event"));
    ESS_DBG_DC(("DC: SYNC0 Event"));

    cbd.hDev = (ESS_HANDLE)ed;
    cbd.evtType = ESS_DC_EVENT_SYNC0;
    ED_DO_CALLBACK(ed, cbDCEvent, &cbd);

    /*lint -e{438, 550} [Assigned value unused, Symbol not accessed] */
    if(ed->flags & ESS_FLAG_SYNC_ACK_MODE) {
        uint8_t dummy _ESS_UNUSED_;

        ESC_READ8(ed, ESC_REG_SYNC0STATUS, dummy);
        (void)dummy;
    }
}

void Sync1_Event(void* arg)
{
    ESC_DATA* ed = (ESC_DATA*)arg;
    ESS_CBDATA_DC cbd;

    ESS_DBG_INTERRUPTS(("INTERRUPT: SYNC1 Event"));
    ESS_DBG_DC(("DC: SYNC1 Event"));

    cbd.hDev = (ESS_HANDLE)ed;
    cbd.evtType = ESS_DC_EVENT_SYNC1;
    ED_DO_CALLBACK(ed, cbDCEvent, &cbd);

    /*lint -e{438, 550} [Assigned value unused, Symbol not accessed] */
    if(ed->flags & ESS_FLAG_SYNC_ACK_MODE) {
        uint8_t dummy _ESS_UNUSED_;

        ESC_READ8(ed, ESC_REG_SYNC1STATUS, dummy);
        (void)dummy;
    }
}

static void Latch_Event(ESC_DATA* ed)
{
    ESS_CBDATA_DC cbd;

    ESS_DBG_INTERRUPTS(("INTERRUPT: DC LATCH Event"));
    ESS_DBG_DC(("DC: Latch Event"));

    cbd.hDev = (ESS_HANDLE)ed;
    cbd.evtType = ESS_DC_EVENT_LATCH;
    ED_DO_CALLBACK(ed, cbDCEvent, &cbd);
}
#endif

/* Called either by HAL_TimerCb() or by HAL's ISR callback */
static void MainLoop(ESC_DATA* ed)
{
    uint32_t alEvent;
#if DEBUG_AL_EVENT_DURATION
    int32_t  start, stop, measure = 0;

    ESC_READ32(ed, ESC_REG_DCSYSTEMTIME, start);
#endif

#if (CFG_ESS_SERVE_STATISTICS)
    ed->config->stats->totalCalls++;
#endif

    /* Read AL EVENT register to check for pending events: */
    ESC_READ32(ed, ESC_REG_ALEVENTREQ, alEvent);
    ESS_DO_MAINLOOP_READALEVENT();


#if (DEBUG_AL_EVENT)
    {
        static uint32_t lastAlEvent[CFG_ESS_MAX_DEVICES];

        if (alEvent != lastAlEvent[ed->devIdx]) {
            /* This _AL_EVENT_TO_STRING() is incomplete! just as little helper... */
#           define _AL_EVENT_TO_STRING(alev) (((alev) == 0x01) ? " = Control Event" : ((alev) == 0x10) ? " = SM Activate changed" : ((alev) == 0x20) ? " = EEPROM Event" : ((alev) == 0x0100) ? " = SM0 Event (usually M->S Mbx)" : ((alev) == 0x0200) ? " = SM1 Event (usually S->M Mbx)" : ((alev) == 0x0400) ? " = SM2 Event (usually Outputs)" : ((alev) == 0x0800) ? " = SM3 Event (usually Inputs)" : ((alev) == 0x0c00) ? " = SM2+3 Event (usually Out- and Inputs)" : ((alev) == 0x11) ? " = Control Event + SM Activate changed" : "")
            ESS_DBG(("DEBUG: AL Event Req. (Reg. 0x0220) for dev %d changed from 0x%.8x to 0x%.8x%s", ed->devIdx, lastAlEvent[ed->devIdx], alEvent, _AL_EVENT_TO_STRING(alEvent)));
            lastAlEvent[ed->devIdx] = alEvent;
#           undef _AL_EVENT_TO_STRING
        }
    }
#endif


#if (CFG_ESS_EEPROM_EMULATION)
    /* Handle pending EEPROM commands: */
    if ((alEvent & REG_MASK_ALEVENT_EEPROMEVENT) && (ED_CALLBACK(ed, cbEEPROMEmulation))) {
        uint16_t cmdReg;

        ESC_READ16(ed, ESC_REG_EEPROMCTRLSTATUS, cmdReg);
        ESS_DBG_INTERRUPTS(("INTERRUPT: EEPROM cmd pending (0x0502 = 0x%.4x%s)", cmdReg, (cmdReg & REG_MASK_EEPROMCTRLSTATUS_BUSY) ? "" : " [not yet ready]"));

        if (cmdReg & REG_MASK_EEPROMCTRLSTATUS_BUSY) {
            uint32_t addrReg;

            switch (cmdReg & REG_MASK_EEPROMCTRLSTATUS_CMDREG) {
                /* Cmd Idle: Clear error bits */
                case REG_MASK_EEPROMCTRLSTATUS_CMDREG_IDLE:
                    cmdReg &= ~REG_MASK_EEPROMCTRLSTATUS_ERRBITS;
                    break;

                /* Cmd Read, forward to essEEPROMRead() */
                case REG_MASK_EEPROMCTRLSTATUS_CMDREG_READ:
                    cmdReg &= ~REG_MASK_EEPROMCTRLSTATUS_ERRBITS;

                    ESC_READ32(ed, ESC_REG_EEPROMADDRESS, addrReg);
                    addrReg <<= 1;

                    if (cmdReg & REG_MASK_EEPROMCTRLSTATUS_READ8BYTES) {
                        uint32_t data0, data1;

                        if ((essEEPROMRead((ESS_HANDLE)ed, addrReg, 4, &data0) == ESS_RESULT_SUCCESS) && (essEEPROMRead((ESS_HANDLE)ed, addrReg + 4, 4, &data1) == ESS_RESULT_SUCCESS)) {
                            ESC_WRITE32_NOSWAP(ed, ESC_REG_EEPROMDATA, data0);
                            ESC_WRITE32_NOSWAP(ed, ESC_REG_EEPROMDATAHI, data1);
                        } else {
                            cmdReg |= REG_MASK_EEPROMCTRLSTATUS_ACKCMDERR;
                        }

                    } else {
                        uint32_t data;

                        if (essEEPROMRead((ESS_HANDLE)ed, addrReg, 4, &data) == ESS_RESULT_SUCCESS) {
                            ESC_WRITE32_NOSWAP(ed, ESC_REG_EEPROMDATA, data);
                        } else {
                            cmdReg |= REG_MASK_EEPROMCTRLSTATUS_ACKCMDERR;
                        }
                    }
                    break;

                /* Cmd Write: forward to essEEPROMWrite() */
                case REG_MASK_EEPROMCTRLSTATUS_CMDREG_WRITE:
                    {
                        uint16_t data;

                        cmdReg &= ~REG_MASK_EEPROMCTRLSTATUS_ERRBITS;

                        ESC_READ32(ed, ESC_REG_EEPROMADDRESS, addrReg);
                        addrReg <<= 1;

                        ESC_READ16_NOSWAP(ed, ESC_REG_EEPROMDATA, data);

                        if (essEEPROMWrite((ESS_HANDLE)ed, addrReg, sizeof(data), &data) != ESS_RESULT_SUCCESS)
                            cmdReg |= REG_MASK_EEPROMCTRLSTATUS_ACKCMDERR;
                    }
                    break;

                /* Cmd Reload: Reload some registers from EEPROM */
                case REG_MASK_EEPROMCTRLSTATUS_CMDREG_RELOAD:
                    cmdReg &= ~REG_MASK_EEPROMCTRLSTATUS_ERRBITS;

                    if (EEPROMLoad(ed, ESS_TRUE)) {
                        ESS_DBG(("WARNING: EEPROM (Emulation) load failed!"));
                        cmdReg |= REG_MASK_EEPROMCTRLSTATUS_CRCERR;
                    }
                    break;

                default:
                    ESS_DBG_EEE(("EEE: Received unknown command! reg = 0x%.4x", cmdReg));
                    cmdReg |= REG_MASK_EEPROMCTRLSTATUS_ACKCMDERR;
                    break;
            }

            ESC_WRITE16(ed, ESC_REG_EEPROMCTRLSTATUS, cmdReg);
            ED_DO_EVENT(ed, ESS_EVENT_WROTEEEPROMCTRL);

        } /* Busy bit set*/

    } /* AL Event signals EEPROM cmd pending */

#endif /* CFG_ESS_EEPROM_EMULATION */


#if (CFG_ESS_DC_CALLBACK)
    /* CHECKME: count SYNC/LATCH events in stats? */

    /* Handle DC Latch events: */
    if (alEvent & REG_MASK_ALEVENT_DCLATCHEVENT) {
        Latch_Event(ed);
    }

# if (0 == CFG_HAL_USE_SYNC0_IRQ)
    /* Handle SYNC0 events: */
    if (alEvent & REG_MASK_ALEVENT_SYNC0EVENT) {
        Sync0_Event(ed);
    }
# endif


# if (0 == CFG_HAL_USE_SYNC1_IRQ)
    /* Handle SYNC1 events: */
    if (alEvent & REG_MASK_ALEVENT_SYNC1EVENT) {
        Sync1_Event(ed);
    }
# endif
#endif


    /* Handle Process data watchdog expired events: */
#if 0
    /* Disabled, cause we want the app to handle PD/SM watchdog itself */
    if (alEvent & REG_MASK_ALEVENT_PDWDEXPIRED) {
        uint16_t pdWdExpired;

        ESS_DBG_INTERRUPTS(("INTERRUPT: PD Watchdog expired"));

        /* This resets the (alEvent & REG_MASK_ALEVENT_PDWDEXPIRED) bit */
        ESC_READ16(ed, ESC_REG_WATCHDOG_STATUSPROCESSDATA, pdWdExpired);

        if ((pdWdExpired & 1) == 0) {
            ESS_DBG(("WARNING: Processdata watchdog expired!"));
            /* Might also change state here... but this is better decided by app! */
        }
    }
#endif


    /* Handle SM changed events: */
    if (alEvent & REG_MASK_ALEVENT_SMCHANGED) {
        if (alEvent & REG_MASK_ALEVENT_CONTROLEVENT) {
            ESS_DBG_INTERRUPTS(("INTERRUPT: SM Changed (Ignored, as CONTROLEVENT is also set)"));
            ESS_DBG_SYNCMANAGERS(("SMs: SMChanged interrupt (Ignored, as CONTROLEVENT is also set)"));

        } else {
            ESS_DBG_INTERRUPTS(("INTERRUPT: SM Changed"));
            ESS_DBG_SYNCMANAGERS(("SMs: SMChanged interrupt"));

            SM_HandleChange(ed, ed->currState);
            ed->mbxOutInterruptPending = ESS_FALSE;

            /* Check mailbox repeat request bit: */
            if ((ed->mbxActive) && (MAKE_ESS_BOOL(ed->cbDataSM[ed->smIdxMBoxIn].smActivate & REG_MASK_SMACTIVATE_REPEATREQ) != ed->mbxRepeatBitMaster)) {
                const uint32_t mboxInInterruptBit = (uint32_t)0x100 << ed->smIdxMBoxIn;
                uint8_t currSMPDIContr, smStatus;

                ESS_DBG_MAILBOX_TRANSFER(("MBX: SM%u signaled master repeat request (toggle bit at 0x%.4x.1 changed to %u, send mbx was %s)", ed->smIdxMBoxIn, ESC_REG_SMACTIVATE(ed->smIdxMBoxIn), MAKE_ESS_BOOL(ed->cbDataSM[ed->smIdxMBoxIn].smActivate & REG_MASK_SMACTIVATE_REPEATREQ), (ed->mbxSendFull) ? "FULL" : "EMPTY"));
                ESS_DBG_MBX_MINIMAL(("<-- Master toggled repeat bit! (our sendMBx: %s, lastStored cnt: %u, lastSend cnt: %u)", (ed->mbxSendFull) ? "FULL" : "EMPTY", MBX_HEADER_GET_COUNTER(ed->mbxLastStoredPacket.typeAndCounter), MBX_HEADER_GET_COUNTER(ed->mbxLastSentPacket.typeAndCounter)));

                ed->mbxRepeatBitMaster = !ed->mbxRepeatBitMaster;

                /* Make sure we know what was last read: */
                if (alEvent & mboxInInterruptBit) {
                    SM_HandleMBoxInInterrupt(ed, ed->smIdxMBoxIn);
                    alEvent &= ~mboxInInterruptBit;
                }

                /* Make sure mailbox is writable: */
                ESC_READ8(ed, ESC_REG_SMSTATUS(ed->smIdxMBoxIn), smStatus);
                if ((ed->mbxSendFull) || (smStatus & REG_MASK_SMSTATUS_MBXFULL)) {
                    ESS_DBG_MAILBOX_TRANSFER(("MBX: trying to make locked mbx writable again..."));
                    ESS_DBG_MBX_MINIMAL(("--- trying to make locked mbx writable again..."));

                    /* Make sure we can write our repeat into the SM buffer even if we've already written something and it's locked: */
                    SM_Disable(ed, ed->smIdxMBoxIn);
                    do {
                        /* We can't be sure if mbx is writable by just testing this bit without disabling the SM: */
                        ESC_READ8(ed, ESC_REG_SMSTATUS(ed->smIdxMBoxIn), smStatus);
                    } while (smStatus & REG_MASK_SMSTATUS_MBXFULL);
                    SM_Enable(ed, ed->smIdxMBoxIn);
                }

                /* Then actually repeat it: */
                MBX_WritePacket(ed, NULL);

                /* No race condition here: master starts to read only after he sees we changed our toggle bit */

                /* And tell the master about it, by toggling repeat ack bit (usually 0x080f): */
                ESC_READ8(ed, ESC_REG_SMPDICONTROL(ed->smIdxMBoxIn), currSMPDIContr);
                currSMPDIContr = (ed->mbxRepeatBitMaster) ? (currSMPDIContr | 0x02) : (currSMPDIContr & 0xfd);
                ESC_WRITE8(ed, ESC_REG_SMPDICONTROL(ed->smIdxMBoxIn), currSMPDIContr);
                ED_DO_EVENT2(ed, ESS_EVENT_WROTEPDICONTROL, ed->smIdxMBoxIn, currSMPDIContr);

                ESS_DBG_MAILBOX_TRANSFER(("MBX: Toggle bit (0x%.4x.1) was %s (cause master %s toggle bit at 0x%.4x.1)", ESC_REG_SMPDICONTROL(ed->smIdxMBoxIn), MAKE_ESS_BOOL(currSMPDIContr & 2) ? "SET" : "CLEARED", (ed->mbxRepeatBitMaster) ? "SET" : "CLEARED", ESC_REG_SMACTIVATE(ed->smIdxMBoxIn)));
                ESS_DBG_MBX_MINIMAL(("--> Toggle bit %s (cause master %s its toggle bit)", (currSMPDIContr & 2) ? "SET" : "CLEARED", (ed->mbxRepeatBitMaster) ? "SET" : "CLEARED"));
            } /* Check mailbox repeat request bit */

        } /* REG_MASK_ALEVENT_CONTROLEVENT not set*/

    } /* AL Event signals SM Changed */


    /* Handle AL CONTROL / State request events: */
    if (alEvent & REG_MASK_ALEVENT_CONTROLEVENT) {
        uint16_t  alControl;
        ESC_STATE reqState;

        /* This _AL_CONTROLEVENT_TO_STRING() is incomplete! just as little helper... */
#       define _AL_CONTROLEVENT_TO_STRING(alc)  (((alc) == 0x01) ? " = Init" : ((alc) == 0x02) ? " = PreOp" : ((alc) == 0x03) ? " = BootStrap" : ((alc) == 0x04) ? " = SafeOp" : ((alc) == 0x08) ? " = Op" : ((alc) == 0x11) ? " = Init+ErrAck" : ((alc) == 0x12) ? " = PreOp+ErrAck" : ((alc) == 0x13) ? " = BootStrap+ErrAck" : ((alc) == 0x14) ? " = SafeOp+ErrAck" : ((alc) == 0x18) ? " = Op+ErrAck" : "")
        ESC_READ16(ed, ESC_REG_ALCONTROL, alControl);
        reqState = alControl & REG_MASK_ALCONTROL_STATE;
        ESS_DBG_INTERRUPTS(("INTERRUPT: AL Control event, 0x0120 = 0x%.4x%s", alControl, _AL_CONTROLEVENT_TO_STRING(alControl)));
        ED_DO_EVENT(ed, ESS_EVENT_READALCONTROL);
#       undef _AL_CONTROLEVENT_TO_STRING

        if (alEvent & REG_MASK_ALEVENT_SMCHANGED)
            SM_UpdateAllSafe(ed); /* Ensure latest SM settings are used when changing state... */

        ed->cbDataStateRequest.transition = MAKE_WORD(ed->currState, reqState);
        ed->cbDataStateRequest.errAck = ((alControl & REG_MASK_ALCONTROL_ERRIND) || (ESC_STATE_INIT == reqState));
#if (CFG_ESS_EXPLICIT_DEVICE_ID)
        if(((alControl & REG_MASK_ALCONTROL_LOADDEVID) != 0) && (ed->currState != ESC_STATE_BOOT) && (reqState != ESC_STATE_BOOT)) {
            ed->cbDataStateRequest.devIdReq = ESS_TRUE;
        } else {
            ed->cbDataStateRequest.devIdReq = ESS_FALSE;
        }
#endif

        CheckStateChange(ed, reqState);

        /* Do Callback and set state: */
        DoStateChangeCallback(ed);
#if DEBUG_AL_EVENT_DURATION
        measure = 1;
#endif

    } else if (ed->cbDataStateRequest.tryAgain) {
        /* Try again: */
        DoStateChangeCallback(ed);
    }


    /* Handle SM interrupts: */
    if (alEvent & REG_MASK_ALEVENT_SMINTERRUPT) {
        uint8_t i;

        for (i = 0; i < ed->smCount; i++) {
            if (alEvent & ((uint32_t)0x0100 << i)) { /* 0x0100 to look in REG_MASK_ALEVENT_SMINTERRUPT */
                ESS_DBG_INTERRUPTS(("INTERRUPT: SM%u", i));

#if (CFG_ESS_SERVE_STATISTICS)
                ed->config->stats->smInterrupts[i]++;
#endif

#if (CFG_ESS_SM_CALLBACK)
                if (ED_DO_CALLBACK(ed, cbSMInterrupt, &ed->cbDataSM[i])) /* allow application to override our handling. Not very useful, because app must replace almost whole stack then... */
                    continue;
#endif

                /* Handle "Outputs" interrupts (Master has written PD for us) */
                if (ed->cbDataSM[i].smType == SM_TYPE_OUTPUTS) {
                    SM_HandleOutputsInterrupt(ed, i); /* With CFG_ESS_SUPPORT_COE it always clears the interrupt by reading the SM buffer */

                /* Handle "Inputs" interrupts (Master has read our PD) */
                } else if (ed->cbDataSM[i].smType == SM_TYPE_INPUTS) {
                    if (ed->inputsActive)
                        SM_HandleInputsInterrupt(ed, i); /* Does NOT clear the interrupt by writing the buffer, this is triggered only by essODSyncInputMappedEntries() */

                /* Handle "MBoxOut" interrupts (Master has written a mbx packet for us): */
                } else if (ed->cbDataSM[i].smType == SM_TYPE_MBXOUT) {
                    if (ed->mbxSendFull) { /* can't handle that packet now, cause we couldn't send answer packet*/
                        if (ed->mbxOutInterruptPending == ESS_FALSE) {
                            ESS_DBG_MAILBOX_TRANSFER(("MBX: pending packet from master..."));
                            ESS_DBG_MBX_MINIMAL(("<== packet from master pending..."));
                            ed->mbxOutInterruptPending = ESS_TRUE;
                        }
                    } else {
                        SM_HandleMBoxOutInterrupt(ed);
                    }

                /* Handle "MBoxIn" interrupts (Master has read our mbx packet): */
                } else if (ed->cbDataSM[i].smType == SM_TYPE_MBXIN) {
                    SM_HandleMBoxInInterrupt(ed, i);
                }

            } /* SM interrupt */

        } /* for each SM */

    } /* Any SM interrupt */


    if (ed->mbxActive) {
#if (CFG_ESS_SUPPORT_EOE)
        /* Queued EoE frame? send it now: */
        if ((ed->eoeSend.active) && (ed->mbxSendFull == ESS_FALSE))
            MBX_HandleEoEFramesToSend(ed);
#endif

#if (CFG_ESS_SUPPORT_COE)
# if (CFG_ESS_COE_SUPPORT_SDO_INFO_SERVICE)
        /* Segmented SDO list request reply active? continue: */
        if ((ed->sdoInfoState.active) && (ed->mbxSendFull == ESS_FALSE))
            CoE_Handle_SegmentedListTransfer(ed);
# endif

        /* Queued CoE emergency? send it now: */
        if ((ed->coeEmergencyReady) && (ed->mbxSendFull == ESS_FALSE))
            CoE_SendQueuedEmergency(ed);
#endif
    }


    /* Inform the application of outputs update: */
    if ((ed->cbDataOutputsUpdated.sm) && (ED_CALLBACK(ed, cbOutputsUpdated))) {
        /* AM335x: We need to forward the actual SM buffer address to the application, too. (Only reqd when app is not using our object dictionary, etc.) */
        if (ed->cbDataOutputsUpdated.sm) {
            if(ESS_FALSE == ed->outputsActive) {
                ed->cbDataOutputsUpdated.sm |= ESS_SM_EVENT_FLAG_SAFE_OUTPUTS;  /* Say: Local output handler disabled -> Set outputs to "safe state" */
            }
            ED_DO_CALLBACK(ed, cbOutputsUpdated, &ed->cbDataOutputsUpdated);
        }
        ed->cbDataOutputsUpdated.sm = 0;
    }


    /* Inform the application of inputs update: */
    if ((ed->cbDataInputsUpdated.sm) && (ED_CALLBACK(ed, cbInputsUpdated))) {
        /* AM335x: We need to forward the actual SM buffer address to the application, too. (Only reqd when app is not using our object dictionary, etc.) */
        ED_DO_CALLBACK(ed, cbInputsUpdated, &ed->cbDataInputsUpdated);
        ed->cbDataInputsUpdated.sm = 0;
    }


#if (CFG_ESS_SUPPORT_COE)
    if ((ed->smUpdateInputs) && (ed->inputsActive) && (ed->coeSDODownloadState.active == ESS_FALSE) && (ed->coeSDOUploadState.active == ESS_FALSE)) {
        ESS_SM smIdx;

        for (smIdx = 0; smIdx < ed->smCount; smIdx++) {
            ESS_CBDATA_SM* sm = &ed->cbDataSM[smIdx];

            if ((sm->smType != SM_TYPE_INPUTS) || (sm->smPhysAddr == 0x0000) || (ed->smBuffer[smIdx] == NULL))
                continue;
# if (CFG_ESS_SMCOPYINUSERBUFFER)
            if (ed->config->smBufferCopies[smIdx].capacity < sm->smLen)
                continue;
# else
            if (ed->smBufferAllocSize[smIdx] < sm->smLen)
                continue;
# endif

            CoE_SyncInputMappedEntries(ed, sm, ed->smBuffer[smIdx]);
        }

        ed->smUpdateInputs = ESS_FALSE;
    }
#endif

#if (CFG_ESS_SERVE_RUN_LED || CFG_ESS_SERVE_ERR_LED || CFG_ESS_SERVE_CUSTOM_LED || CFG_ESS_SUPPORT_COE || CFG_ESS_SUPPORT_FOE)
    {
        const ESS_TIMESTAMP now = essHALGetTime();

# if (CFG_ESS_SUPPORT_COE)
        /* Handle CoE Timeouts: */
        if ((ed->coeSDODownloadState.active) && ((now - ed->coeSDODownloadState.lastPacket) > CFG_ESS_COE_SDO_SEGMENT_TIMEOUT)) {
            ESS_DBG(("WARNING: Aborting SDO download due to timeout..."));
            CoE_FinishSegmentedTransfer(ed, ESS_FALSE, ESS_TRUE);
        }

        if ((ed->coeSDOUploadState.active) && ((now - ed->coeSDOUploadState.lastPacket) > CFG_ESS_COE_SDO_SEGMENT_TIMEOUT)) {
            ESS_DBG(("WARNING: Aborting SDO upload due to timeout..."));
            CoE_FinishSegmentedTransfer(ed, ESS_TRUE, ESS_TRUE);
        }
# endif


# if (CFG_ESS_SUPPORT_FOE)
        /* Handle FoE Timeouts: */
        if ((ed->foeState.operation) && ((now - ed->foeState.lastPacket) > CFG_ESS_FOE_TRANSFER_TIMEOUT)) {
            ESS_DBG(("WARNING: Aborting FoE transfer due to timeout"));
            FoE_DoCloseCallback(ed, (ed->foeState.operation == FOE_OPCODE_READ), ESS_TRUE);
        }
# endif

# if (CFG_ESS_SERVE_RUN_LED || CFG_ESS_SERVE_ERR_LED || CFG_ESS_SERVE_CUSTOM_LED)
        {
            /* CHECKME: LED handling for all devices by only HandleLEDs() leads to minor race condition if devices handled by separate threads... */
            static ESS_TIMESTAMP nextUpdate;

            if (ESS_TIMESTAMP_PASSED(nextUpdate, now)) {
                nextUpdate = now + 50;
                HandleLEDs();
            }
        }
# endif
    }
#endif

#if DEBUG_AL_EVENT_DURATION
    if(measure != 0) {
        ESC_READ32(ed, ESC_REG_DCSYSTEMTIME, stop);
        ESS_PRINTF("Time: %d\n", (uint32_t)(stop - start));
    }
#endif

}

/* Called cyclically by HAL. (HAL makes sure this is not interrupted by ISR, etc.) */
static void HAL_TimerCb(void* ctx)
{
    ESC_DATA* ed = (ESC_DATA*)ctx;
    ESS_CBDATA_CYCLIC cbData;

    cbData.hDev = (ESS_HANDLE)ed;
    cbData.essTime = essHALGetTime();
    ED_DO_CALLBACK(ed, cbCyclic, &cbData);

#if (CFG_ESS_SERVE_STATISTICS)
    ed->config->stats->cyclicCalls++;
#endif
    MainLoop(ed);
}


/* Just calls essHALStart() */
ESS_EXPORT ESS_RESULT ESS_CALLTYPE essStart(ESS_HANDLE h)
{
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_HAL_CALLBACK* isrCallback;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->config == NULL)
        return ESS_RESULT_INVALID_CONFIG;

    isrCallback = (ed->config->flags & ESS_CONFIG_FLAGS_USE_ISR) ? (ESS_HAL_CALLBACK*)MainLoop : NULL;
    return essHALStart(ed->devIdx, isrCallback, HAL_TimerCb, ed->config->timerInterval, ed);
}

/* Just calls essHALStop() */
ESS_EXPORT ESS_RESULT ESS_CALLTYPE essStop(ESS_HANDLE h)
{
    ESC_DATA* ed = (ESC_DATA*)h;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    return essHALStop(ed->devIdx);
}

/* Called by application when it wants to signal an error outside a state transition */
ESS_EXPORT ESS_RESULT ESS_CALLTYPE essIndicateError(ESS_HANDLE h, ESC_STATE newState, REG_VAL_ALSTATUSCODE errCode)
{
    ESC_DATA* ed = (ESC_DATA*)h;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->config == NULL)
        return ESS_RESULT_INVALID_CONFIG;

    ESS_DBG(("INFO: App wants newState %s (0x%.4x) %s", ESC_STATE_TO_STRING(newState), newState, (errCode == REG_VAL_ALSTATUSCODE_NOERROR) ? "[No Error]" : "[Error indicated]"));
    VerifyAppStateChange(ed, &newState); /* Just makes sure newState is <= current state etc. */

    ed->cbDataStateRequest.transition = MAKE_WORD(ed->currState, ed->currState);
    ed->cbDataStateRequest.statusCode = errCode;
    ed->cbDataStateRequest.newState = newState;

    DoStateChangeCallback(ed);

#if 0
    if ((ed->outputsActive) && (errCode != REG_VAL_ALSTATUSCODE_NOERROR))
        Outputs_Stop(ed); /* Needed for transition to SafeOp */
#endif

    return ESS_RESULT_SUCCESS;
}

/* Queues a CoE emergency that will be sent by main loop when mailbox is empty */
ESS_EXPORT ESS_RESULT ESS_CALLTYPE essCoESendEmergency(ESS_HANDLE h, const ESS_COE_EMERGENCY* data)
{
#if (CFG_ESS_SUPPORT_COE)
    ESC_DATA* ed = (ESC_DATA*)h;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->currState == ESC_STATE_INIT)
        return ESS_RESULT_WRONG_ECAT_STATE;

    if ((ed->smIdxMBoxIn == ESS_SM_NONE) || (ed->smIdxMBoxOut == ESS_SM_NONE))
        return ESS_RESULT_INVALID_CONFIG;

    if (ed->coeEmergencyReady)
        return ESS_RESULT_TRY_AGAIN;

    ed->coeEmergency = *data;
    ed->coeEmergencyReady = ESS_TRUE;

    return ESS_RESULT_SUCCESS;
#else
    (void)h;
    (void)data;
    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essSyncInputs(ESS_HANDLE h, ESS_ISYN_FLAGS flags)
{
#if (CFG_ESS_SUPPORT_COE)
    ESC_DATA* ed = (ESC_DATA*)h;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->config == NULL)
        return ESS_RESULT_INVALID_CONFIG;

    if (flags & 0x0000ffff) { /* Bits0..15 indicate immediate update of that SM is wanted */
        const int maxIdx = MIN(16, CFG_ESS_MAX_SM_COUNT);
        int updated = 0;
        int smIdx;

        if ((ed->currState != ESC_STATE_SAFEOP) && (ed->currState != ESC_STATE_OP))
            return ESS_RESULT_WRONG_ECAT_STATE;

        for (smIdx = 0; smIdx < maxIdx; ++smIdx) {
            if ((1U << smIdx) & flags) {
                const ESS_CBDATA_SM* const sm = &ed->cbDataSM[smIdx];

                if (sm->smType != SM_TYPE_INPUTS) {
                    ESS_DBG(("SM%i is not SM_TYPE_INPUTS!", smIdx));
                    continue;
                }

                /* Sanity checks required: Even while successfully in SafeOp/Op our SMs might be changed/invalidated. */
                if ((sm->smPhysAddr == 0x0000) || (ed->smBuffer[smIdx] == NULL))
                    continue;
# if (CFG_ESS_SMCOPYINUSERBUFFER)
                if (ed->config->smBufferCopies[smIdx].capacity < sm->smLen)
                    continue;
# else
                if (ed->smBufferAllocSize[smIdx] < sm->smLen)
                    continue;
# endif
                CoE_SyncInputMappedEntries(ed, sm, ed->smBuffer[smIdx]);
                updated++;
            }
        }
        return (updated) ? ESS_RESULT_SUCCESS : ESS_RESULT_UNKNOWN;

    } else {
        ed->smUpdateInputs = ESS_TRUE;
        return ESS_RESULT_SUCCESS;
    }
#else
    (void)h;
    (void)flags;
    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

/* Returns ESS_VERSION with date built */
ESS_EXPORT const char* ESS_CALLTYPE essGetVersion(void)
{
    /*lint -e506 [Constant value Boolean]                           */
    /*lint -e774 [Boolean within if always evaluates to true/false] */
    if (sizeof(void*) == 8)
        return "ess Version " ESS_VERSION ", built at " __DATE__ " " __TIME__ " (64 bit)";
    else
        return "ess Version " ESS_VERSION ", built at " __DATE__ " " __TIME__;
}

ESS_EXPORT ESS_TIMESTAMP ESS_CALLTYPE essGetTime(void)
{
    return essHALGetTime();
}

/* Returns err its define (ESS_FORMAT_SHORT) */
ESS_EXPORT const char* ESS_CALLTYPE essFormatResult(ESS_RESULT err)
{
    switch (err) {
        case ESS_RESULT_SUCCESS:
            return "ESS_RESULT_SUCCESS";
        case ESS_RESULT_UNKNOWN:
            return "ESS_RESULT_UNKNOWN";
        case ESS_RESULT_INVALID_INDEX:
            return "ESS_RESULT_INVALID_INDEX";
        case ESS_RESULT_NOT_READY:
            return "ESS_RESULT_NOT_READY";
        case ESS_RESULT_ALREADY_ACTIVE:
            return "ESS_RESULT_ALREADY_ACTIVE";
        case ESS_RESULT_CANT_OPEN:
            return "ESS_RESULT_CANT_OPEN";
        case ESS_RESULT_TRY_AGAIN:
            return "ESS_RESULT_TRY_AGAIN";
        case ESS_RESULT_MISSING_CALLBACK:
            return "ESS_RESULT_MISSING_CALLBACK";
        case ESS_RESULT_INVALID_HANDLE:
            return "ESS_RESULT_INVALID_HANDLE";
        case ESS_RESULT_INVALID_LIST:
            return "ESS_RESULT_INVALID_LIST";
        case ESS_RESULT_OBJ_NOT_FOUND:
            return "ESS_RESULT_OBJ_NOT_FOUND";
        case ESS_RESULT_ENTRY_NOT_FOUND:
            return "ESS_RESULT_ENTRY_NOT_FOUND";
        case ESS_RESULT_NO_DICT:
            return "ESS_RESULT_NO_DICT";
        case ESS_RESULT_NO_MEM:
            return "ESS_RESULT_NO_MEM";
        case ESS_RESULT_OBJ_EXISTS:
            return "ESS_RESULT_OBJ_EXISTS";
        case ESS_RESULT_ENTRY_EXISTS:
            return "ESS_RESULT_ENTRY_EXISTS";
        case ESS_RESULT_WRONG_ECAT_STATE:
            return "ESS_RESULT_WRONG_ECAT_STATE";
        case ESS_RESULT_INVALID_CONFIG:
            return "ESS_RESULT_INVALID_CONFIG";
        case ESS_RESULT_INVALID_OD_FLAGS:
            return "ESS_RESULT_INVALID_OD_FLAGS";
        case ESS_RESULT_INVALID_SUBINDEX:
            return "ESS_RESULT_INVALID_SUBINDEX";
        case ESS_RESULT_INVALID_SM:
            return "ESS_RESULT_INVALID_SM";
        case ESS_RESULT_INVALID_ARG:
            return "ESS_RESULT_INVALID_ARG";
        case ESS_RESULT_NOT_IMPLEMENTED:
            return "ESS_RESULT_NOT_IMPLEMENTED";
        case ESS_RESULT_INVALID_SIZE:
            return "ESS_RESULT_INVALID_SIZE";
        case ESS_RESULT_BUSY:
            return "ESS_RESULT_BUSY";
        case ESS_RESULT_INVALID_PDI_CONTROL:
            return "ESS_RESULT_INVALID_PDI_CONTROL";
        case ESS_RESULT_ABI_INCOMPATIBLE:
            return "ESS_RESULT_ABI_INCOMPATIBLE";
        case ESS_RESULT_INVALID_TIMER_CONFIG:
            return "ESS_RESULT_INVALID_TIMER_CONFIG";
        case ESS_RESULT_MISSING_ACK:
            return "ESS_RESULT_MISSING_ACK";
        case ESS_RESULT_INVALID_OBJECT:
            return "ESS_RESULT_INVALID_OBJECT";
        case ESS_RESULT_INSUFFICIENT_BUFFER:
            return "ESS_RESULT_INSUFFICIENT_BUFFER";
        case ESS_RESULT_INVALID_PDI_CONFIGURATION:
            return "ESS_RESULT_INVALID_PDI_CONFIGURATION";
        case ESS_RESULT_ERROR_EEPROM_CONTROL:
            return "ESS_RESULT_ERROR_EEPROM_CONTROL";
        default:
            return "[Undefined error code]";
    }
}

/* Just extracted from essOpen() */
static ESS_RESULT _essOpen_CheckConfig(const ESS_CONFIGURATION *config, ESC_DATA *ed)
{
    if (config->essABIVersion != 7) { /* don't use macro instead of number ...to be sure you know what you do when ABI changes. */
        ESS_DBG(("ERROR: App uses ABI version %u, this library is version %i!",
                 config->essABIVersion, 7));
        return ESS_RESULT_ABI_INCOMPATIBLE;
    }

    /* Handle difference between number of configured and available SMs */
    if (ed->smCount > CFG_ESS_MAX_SM_COUNT) {
        ESS_DBG(("WARNING: SM configuration mismatch (%u > %d), adapt CFG_ESS_MAX_SM_COUNT!",
                 ed->smCount, CFG_ESS_MAX_SM_COUNT));
        ed->smCount = CFG_ESS_MAX_SM_COUNT;
    } else if (ed->smCount < CFG_ESS_MAX_SM_COUNT) {
        ESS_DBG(("ERROR: SM configuration mismatch (%u < %d), adapt CFG_ESS_MAX_SM_COUNT!",
                 ed->smCount, CFG_ESS_MAX_SM_COUNT));
        return ESS_RESULT_INVALID_SM;
    }

#if (CFG_ESS_SMCOPYINUSERBUFFER)
    if (config->smBufferCopies == NULL) {
        ESS_DBG(("ERROR: %s", "smBufferCopies is NULL but CFG_ESS_SMCOPYINUSERBUFFER set!?"));
        return ESS_RESULT_INVALID_CONFIG;
    }
#endif


#if 1 /* Verify callbacks: (always active, just as code-folding helper...) */
    if (config->cb == NULL) {
        ESS_DBG(("ERROR: %s", "callbacks pointer is NULL!"));
        return ESS_RESULT_MISSING_CALLBACK;
    }

    if (config->cb->cbCyclic == NULL) {
        ESS_DBG(("ERROR: %s", "cbCyclic is missing!"));
        return ESS_RESULT_MISSING_CALLBACK;
    }

    if (config->cb->cbStateRequest == NULL) {
        ESS_DBG(("ERROR: %s", "cbStateRequest is missing!"));
        return ESS_RESULT_MISSING_CALLBACK;
    }

# if (CFG_ESS_SM_CALLBACK)
    if (config->cb->cbSMInterrupt == NULL) {
        ESS_DBG(("ERROR: %s", "cbSMInterrupt is missing!"));
        return ESS_RESULT_MISSING_CALLBACK;
    }
# else
    if (config->cb->cbSMInterrupt)
        ESS_DBG(("INFO: CFG_ESS_SM_CALLBACK is not set, but callback is set... is that intended?"));
# endif

    if (config->cb->cbInOutputsActivate == NULL) {
        ESS_DBG(("ERROR: %s", "cbInOutputsActivate is missing!"));
        return ESS_RESULT_MISSING_CALLBACK;
    }

    /* cbOutputsUpdated is always optional */

    /* cbInputsUpdated is always optional */

# if (!CFG_ESS_SUPPORT_COE)
    if ((config->cb->cbCoEReadWrite) || (config->cb->cbCoEEvent))
        ESS_DBG(("INFO: CoE support is disabled but CoE callbacks are set... is that intended?"));
# endif

# if (!CFG_ESS_SUPPORT_FOE)
    if ((config->cb->cbFoEClose) || (config->cb->cbFoEData) || (config->cb->cbFoEOpen))
        ESS_DBG(("INFO: FoE support is disabled but FoE callbacks are set... is that intended?"));
# endif

# if (!CFG_ESS_SUPPORT_EOE)
    if ((config->cb->cbEoEFrame) || (config->cb->cbEoESetAddrFilter) || (config->cb->cbEoESetIPParam))
        ESS_DBG(("INFO: EoE support is disabled but EoE callbacks are set... is that intended?"));
# endif

    /* cbVoE is always optional */
    /* cbSoE is always optional */
    /* cbAoE is always optional */

# if (CFG_ESS_EEPROM_EMULATION)
#  if (CFG_ESS_MAX_DEVICES == 1)
    if (config->cb->cbEEPROMEmulation == NULL) {
        ESS_DBG(("ERROR: cbEEPROMEmulation is missing!"));
        return ESS_RESULT_MISSING_CALLBACK;
    }
#  endif
# else
    if (config->cb->cbEEPROMEmulation)
        ESS_DBG(("INFO: CFG_ESS_EEPROM_EMULATION is not set, but callback is set... is that intended?"));
# endif

# if (CFG_ESS_DC_CALLBACK)
    if (config->cb->cbDCEvent == NULL) {
        ESS_DBG(("ERROR: %s", "CFG_ESS_DC_CALLBACK set but cbDCEvent is missing!"));
        return ESS_RESULT_MISSING_CALLBACK;
    }
# else
    if (config->cb->cbDCEvent)
        ESS_DBG(("INFO: CFG_ESS_DC_CALLBACK is not set, but callback is set... is that intended?"));
# endif

#endif /* Verify callbacks */


#if 1 /* Verify SM configuration: (minor checks only) */
    {
        int mbxInCount = 0, mbxOutCount = 0;
        int i;

        /* Bootstrap mailbox configurations can only be 0 or 2 */
        if((config->smConfigCountBootstrap & ~0x02) != 0) {
            ESS_DBG(("ERROR: Invalid value %d for bootstrap SM configuration", config->smConfigCountBootstrap));
            return ESS_RESULT_INVALID_CONFIG;
        }

        for (i = 0; i < config->smConfigCount + config->smConfigCountBootstrap; ++i) {
            const ESS_SM_CONFIGURATION* const smc = &config->smConfigs[i];

            if ((smc->maxSize) && (smc->defSize > smc->maxSize)) {
                ESS_DBG(("ERROR: Configured SM%d defSize > maxSize! (%u > %u)", i, smc->defSize, smc->maxSize));
                return ESS_RESULT_INVALID_CONFIG;
            }

            if (((smc->contrByte & REG_MASK_SMCONTROL_TYPE) == SM_TYPE_MBXIN) || ((smc->contrByte & REG_MASK_SMCONTROL_TYPE) == SM_TYPE_MBXOUT)) {
                if (smc->maxSize > CFG_ESS_MAX_MBX_LEN) {
                    ESS_DBG(("ERROR: Configured SM%d len exceeds CFG_ESS_MAX_MBX_LEN value! (%u > %i)",
                             i, smc->maxSize, CFG_ESS_MAX_MBX_LEN));
                    return ESS_RESULT_INVALID_CONFIG;
                }

                mbxInCount += (int)((smc->contrByte & REG_MASK_SMCONTROL_TYPE) == SM_TYPE_MBXIN);
                mbxOutCount += (int)((smc->contrByte & REG_MASK_SMCONTROL_TYPE) == SM_TYPE_MBXOUT);
                if((i >= config->smConfigCount) && (0 == (smc->contrByte & REG_MASK_SMCONTROL_BOOTSTRAP))) {
                    ESS_DBG(("WARNING: Last two SM entries are reserved for bootstrap configuration"));
                    return ESS_RESULT_INVALID_CONFIG;
                }
            }
        }

        if ((mbxInCount != mbxOutCount) || (mbxInCount != (config->smConfigCountBootstrap ? 2 : 1))) {
            ESS_DBG(("WARNING: SM configuration seems invalid! (%d MBXOUT, %d MBXIN)", mbxOutCount, mbxInCount));
            return ESS_RESULT_INVALID_CONFIG;
        }
    }
#endif /* Verify SM configuration */


#if (CFG_ESS_SERVE_STATISTICS)
    if (config->stats == NULL) {
        ESS_DBG(("ERROR: stats pointer not set, but CFG_ESS_SERVE_STATISTICS is 1!"));
        return ESS_RESULT_INVALID_ARG;
    }
#else
    if (config->stats)
        ESS_DBG(("INFO: stats pointer set, but CFG_ESS_SERVE_STATISTICS is 0... is that intended?"));
#endif

    return ESS_RESULT_SUCCESS;
}

/* Opens an ESC and initializes everything to run main loop */
ESS_EXPORT ESS_RESULT ESS_CALLTYPE essOpen(ESS_DEVICE_INDEX devIdx, const ESS_CONFIGURATION* config, ESS_HANDLE* resHandle)
{
    ESS_RESULT res;
    ESC_DATA*  ed;
    int        i;

#if 1 /* Verify environment/arguments: (always active, just as code-folding helper...) */
    /*lint -e506 [Constant value Boolean]                           */
    /*lint -e774 [Boolean within if always evaluates to true/false] */
    if ((sizeof(uintptr_t) < sizeof(void*)) || (sizeof(uint32_t) != 4) || (sizeof(uint16_t) != 2) || (sizeof(uint8_t) != 1)
# if (CFG_HAVE_STD_INT_TYPES == 0)
        || (sizeof(uint64_t) != 8) || (sizeof(int64_t) != 8) /* 64 bit typedefs not used by stack itself, so CFG_HAVE_STD_INT_TYPES can be used to disable sizeof checks for them too */
# endif
        ) {
        ESS_DBG(("ERROR: a sizeof() check failed! (please verify type definitions)"));
        return ESS_RESULT_UNKNOWN;
    }

    /* ESS_DBG(("DEBUG: sizeof(escData):  %lu * %lu = %lu byte", (unsigned long)(sizeof(escData) / sizeof(escData[0])), (unsigned long)sizeof(escData[0]), (unsigned long)sizeof(escData))); */

    if (devIdx >= CFG_ESS_MAX_DEVICES) {
        ESS_DBG(("ERROR: %s", "devIdx >= CFG_ESS_MAX_DEVICES!"));
        return ESS_RESULT_INVALID_INDEX;
    }

    if (resHandle == NULL) {
        ESS_DBG(("ERROR: %s", "resHandle is NULL!"));
        return ESS_RESULT_INVALID_HANDLE;
    }

    if (escData[devIdx].magic == 0x07bb0b04) {
        ESS_DBG(("ERROR: %s", "(In)valid magic!"));
        return ESS_RESULT_ALREADY_ACTIVE;
    }

    res = essHALOpen(devIdx);
    if (res != ESS_RESULT_SUCCESS) {
        if (config != NULL) /* Avoid error output while parsing device infos only */
            ESS_DBG(("ERROR: essHALOpen(%u) returned %s", devIdx, essFormatResult(res)));
        return res;
    }
#endif /* Verify environment/arguments: (always active, just as code-folding helper...) */



#if 1 /* Init ed and map ESC: (always active, just as code-folding helper...) */
    ed = &escData[devIdx];

    ed->devIdx = devIdx;
    ed->cbDataStateRequest.hDev = (ESS_HANDLE)ed;
    ed->cbDataInputsUpdated.hDev = (ESS_HANDLE)ed;
    ed->cbDataOutputsUpdated.hDev = (ESS_HANDLE)ed;
    for (i = 0; i < CFG_ESS_MAX_SM_COUNT; ++i) {
        ed->cbDataSM[i].hDev  = (ESS_HANDLE)ed;
        ed->cbDataSM[i].smIdx = (uint8_t)i;
    }
    ed->smIdxMBoxOut = ESS_SM_NONE;
    ed->smIdxMBoxIn = ESS_SM_NONE;
# if (CFG_ESS_SERVE_RUN_LED || CFG_ESS_SERVE_ERR_LED || CFG_ESS_SERVE_CUSTOM_LED)
    for (i = 0; i < ESC_LED_TYPE_NUMBER_OF_LEDS; ++i)
        ed->ledStates[i].lastWritten = -1;
# endif

# if (CFG_ESC_HAVE_POINTER)
    res = essHALMapESC(devIdx, &ed->escPointer);
    if (res != ESS_RESULT_SUCCESS) {
        ESS_DBG(("ERROR: essHALMapESC() failed!"));
        return res;
    }

    if (ed->escPointer == NULL) {
        ESS_DBG(("ERROR: essHALMapESC() returned NULL pointer!?"));
        res = ESS_RESULT_UNKNOWN;
        goto unmap_closehal_and_return_res;
    }
# endif

    ed->config = config;
    ed->magic = (uint32_t)0x07bb0b04;
    *resHandle = (ESS_HANDLE)ed;
# if defined(_DEBUG)
    ed->tIdOnOpen = _ESS_GETTHREADID(); /* Might be used for sanity checks in debug builds, but usually _ESS_GETTHREADID() is just defined to 0 */
    if (ed->tIdOnOpen != NULL) {
        if (config != NULL) /* Avoid output while parsing device infos only */
            ESS_DBG(("Thread Id: %p", ed->tIdOnOpen));
    }
# endif

#endif /* Init ed and map ESC: (always active, just as code-folding helper...) */


    if (config == NULL) {
        /* Device is only opened to use essIoctl(): */
        return ESS_RESULT_SUCCESS;
    }

    ESC_READ8(ed, ESC_REG_SMCOUNT, ed->smCount);

    res = _essOpen_CheckConfig(config, ed);
    if (res != ESS_RESULT_SUCCESS)
        goto unmap_closehal_and_return_res;


    /* ESS_DBG(("INFO: config OK, testing/initializing ESC...")); */
#if (CFG_ESS_COPY_CALLBACKS)
    ed->callbacks = *config->cb;
#endif
    ed->currState = ESC_STATE_INIT;
    ed->cbDataStateRequest.transition = ESC_TRANSITION_II;
    ed->cbDataStateRequest.devIdReq   = ESS_FALSE;
    StateChanged(ed);

    /* A sync pulse length of 0 defines SYNC acknowledge mode */
    {
        uint16_t   syncPulseLength;

        ESC_READ16(ed, ESC_REG_SYNCPULSELEN, syncPulseLength);
        if(0 == syncPulseLength ) {
            ed->flags |= ESS_FLAG_SYNC_ACK_MODE;
            ESS_DBG(("INFO: SYNC acknowledge mode configured"));
        }
    }

    /* Make a local copy of the Sync/Latch PDI configuration register 0x151 */
    ESC_READ8(ed, (ESC_REG_PDICONFIGURATION + 1), ed->escRegPdiCfg);

    {
        /* Default AL event mask to all non reserved (31..24 and 7) bits. */
        uint32_t   alEventMask = 0x00FFFF7F;

        /*
         * Make sure SYNC0/1 via PDI indication is only used if the SYNC IRQ is
         * not handled directly and the PDI is configured to acknowledge mode
         */
#if (CFG_HAL_USE_SYNC0_IRQ)
        alEventMask &= ~(REG_MASK_ALEVENT_SYNC0EVENT);
#else
        if((ed->escRegPdiCfg & (REG_MASK_PDICFG_SYNC0_OUTPUT | REG_MASK_PDICFG_SYNC0_MAPPED)) ==
                               (REG_MASK_PDICFG_SYNC0_OUTPUT | REG_MASK_PDICFG_SYNC0_MAPPED)) {
            if (0 == (ed->flags & ESS_FLAG_SYNC_ACK_MODE)) {
                ESS_DBG(("ERROR: Use SYNC0 via PDI IRQ only in acknowledge mode!"));
                res = ESS_RESULT_INVALID_PDI_CONFIGURATION;
                goto unmap_closehal_and_return_res;
            } 
        }
#endif

#if (CFG_HAL_USE_SYNC1_IRQ)
        alEventMask &= ~(REG_MASK_ALEVENT_SYNC1EVENT);
#else
        if((ed->escRegPdiCfg & (REG_MASK_PDICFG_SYNC1_OUTPUT | REG_MASK_PDICFG_SYNC1_MAPPED)) ==
                               (REG_MASK_PDICFG_SYNC1_OUTPUT | REG_MASK_PDICFG_SYNC1_MAPPED)) {
                if (0 == (ed->flags & ESS_FLAG_SYNC_ACK_MODE)) {
                    ESS_DBG(("ERROR: Use SYNC1 via PDI IRQ only in acknowledge mode!"));
                    res = ESS_RESULT_INVALID_PDI_CONFIGURATION;
                    goto unmap_closehal_and_return_res;
                } 
        }
#endif

        ESC_WRITE32(ed, ESC_REG_ALEVENTMASK, alEventMask);
    }
    ESC_WRITE16(ed, ESC_REG_ALSTATUSCODE, REG_VAL_ALSTATUSCODE_NOERROR);
    ESC_WRITE16(ed, ESC_REG_ALSTATUS, ESC_STATE_INIT);
    ED_DO_EVENT(ed, ESS_EVENT_WROTEALSTATUS);

    /* Perform some minor PDI control (0x0140:0x0141) checks: */
    {
        uint16_t pdiCtrl;

        ESC_READ16(ed, ESC_REG_PDICONTROL, pdiCtrl);

        /* Make sure PDI is at least enabled: */
        if (LO_WORD(pdiCtrl) == 0) {
            ESS_DBG(("ERROR: PDI is disabled!"));
            res = ESS_RESULT_INVALID_PDI_CONTROL;
            goto unmap_closehal_and_return_res;
        }

        /* Make sure ESC is not in device emulation mode: */
        if (pdiCtrl & REG_MASK_PDICONTROL_DEVEMULATION) {
            ESS_DBG(("ERROR: PDI control has DEV_EMULATION bit set!"));
            res = ESS_RESULT_INVALID_PDI_CONTROL;
            goto unmap_closehal_and_return_res;
        }

        /* ESS_DBG(("INFO: ESC%u is ready! (pdi control: 0x%.4x), starting...", ed->devIdx, pdiCtrl)); */
    }

    /* Perform sanity checks for ESC supported EEPROM mode (I2C or Emulation) */
    {
        uint16_t eepromCtrl;

        ESC_READ16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
        if (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_EMULATION) {
            ESS_DBG(("INFO: ESC EEPROM emulation mode configured"));
#if (!CFG_ESS_EEPROM_EMULATION)
            ESS_DBG(("ERROR: Stack does not support EEPROM emulation !"));
            res = ESS_RESULT_ERROR_EEPROM_CONTROL;
            goto unmap_closehal_and_return_res;
#endif
        }

#if (CFG_ESS_EEPROM_EMULATION)
       if ((ED_CALLBACK(ed, cbEEPROMEmulation)) && (EEPROMLoad(ed, ESS_FALSE))) {
          eepromCtrl |= REG_MASK_EEPROMCTRLSTATUS_CRCERR;
          ESC_WRITE16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
       }
#endif
    }

    return ESS_RESULT_SUCCESS;

unmap_closehal_and_return_res:
#if (CFG_ESC_HAVE_POINTER)
    (void)essHALUnmapESC(devIdx);
#endif
    (void)essHALClose(devIdx);

    ed->magic = 0;
    *resHandle = NULL;

    return res;
}

/* Closes the ESC (finishes the stack for that ESC) and invalidates the handle (h) */
ESS_EXPORT ESS_RESULT ESS_CALLTYPE essClose(ESS_HANDLE h)
{
    ESC_DATA* ed = (ESC_DATA*)h;
    ESS_RESULT res;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->currState != ESC_STATE_INIT) {
        ESC_WRITE16(ed, ESC_REG_ALSTATUSCODE, REG_VAL_ALSTATUSCODE_UNSPECIFIEDERROR);
        ESC_WRITE16(ed, ESC_REG_ALSTATUS, ESC_STATE_INIT);
        ed->currState = ESC_STATE_INIT;
    }

    (void)essHALStop(ed->devIdx); /* If HAL was successfully started it's likely that this was already called, so this does/must not hurt there. */

#if (CFG_ESC_HAVE_POINTER)
    (void)essHALUnmapESC(ed->devIdx);
#endif

    (void)essODDelete(h, 0);

#if (CFG_ESS_SUPPORT_COE)
    {
        int i;

        for (i = 0; i < CFG_ESS_MAX_SM_COUNT; ++i) {
#if (!CFG_ESS_SMCOPYINUSERBUFFER)
            ESS_FREE(ed->smBuffer[i]);
#endif
            ESS_FREE(ed->smMappedEntries[i].entries);
        }
    }
#endif

    res = essHALClose(ed->devIdx);
    if (res != ESS_RESULT_SUCCESS)
        ESS_DBG(("essHALClose(%u) returned %s", ed->devIdx, essFormatResult(res)));

    ESS_MEMZERO(ed, sizeof(*ed));

    return ESS_RESULT_SUCCESS;
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essEEPROMRead(ESS_HANDLE h, uint32_t addr, uint32_t len, void* dest)
{
#if (CFG_ESS_EEPROM_EMULATION)
    ESS_CBDATA_EEPROM_EMULATION cbData;
    ESC_DATA* ed = (ESC_DATA*)h;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->config == NULL)
        return ESS_RESULT_INVALID_CONFIG;

    if ((len == 0) || (dest == NULL)) {
        ESS_DBG_EEE(("EEE: Invalid len or dest!"));
        return ESS_RESULT_INVALID_ARG;
    }

    cbData.hDev = h;
    cbData.destSrc = (uint8_t*)dest;
    cbData.offset = addr;
    cbData.len = len;
    cbData.result = 0xff;
    cbData.isRead = ESS_TRUE;

    ED_DO_CALLBACK(ed, cbEEPROMEmulation, &cbData);

#if (DEBUG_EEPROM_EMULATION)
    /* Seeing all read accesses twice? Then this slave supports 8-byte-reads but master takes only 4 to support all slaves... so, don't care! */
    if (cbData.result) {
        ESS_DBG_EEE(("EEE: read failed for %u bytes from addr 0x%.4x (word 0x%.2x)", len, addr, addr >> 1));
    } else {
        if (len >= 4)
            ESS_DBG_EEE(("EEE: read OK for %u bytes from addr 0x%.4x (word 0x%.2x): bytes (hex) = %.2x %.2x %.2x %.2x%s", len, addr, addr >> 1, cbData.destSrc[0], cbData.destSrc[1], cbData.destSrc[2], cbData.destSrc[3], (len > 4) ? "..." : ""));
        else
            ESS_DBG_EEE(("EEE: read OK for %u bytes from addr 0x%.4x (word 0x%.2x)", len, addr, addr >> 1));
    }
#endif

    return (cbData.result == 0) ? ESS_RESULT_SUCCESS : ESS_RESULT_UNKNOWN;


#else

    ESC_DATA* ed = (ESC_DATA*)h;
    uint16_t eepromCtrl;
    uint8_t eepromConfig;
    uint8_t eepromAccess;
    ESS_TIMESTAMP tsStart;
    uint8_t ackRetries = 0;
    ESS_RESULT res = ESS_RESULT_SUCCESS;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if ((addr & 1) || (len == 0) || (dest == NULL))
        return ESS_RESULT_INVALID_ARG;

    /* Make sure we're allowed to have control: */
    ESC_READ8(ed, ESC_REG_EEPROMCONFIGURATION, eepromConfig);
    if (((eepromConfig & REG_MASK_EEPROMCONFIGURATION_PDICONTROL) == 0) || (eepromConfig & REG_MASK_EEPROMCONFIGURATION_FORCEECAT)) {
        ESS_DBG(("WARNING: ECAT has EEPROM access..."));
        return ESS_RESULT_CANT_OPEN;
    }

    /* Tell ECAT we have control: */
    ESC_READ8(ed, ESC_REG_EEPROMPDIACCESSSTATE, eepromAccess);
    eepromAccess |= REG_MASK_EEPROMPDIACCESSSTATE_PDIACCESS;
    ESC_WRITE8(ed, ESC_REG_EEPROMPDIACCESSSTATE, eepromAccess);

    /* Make sure EEPROM was loaded successfully: */
    ESC_READ16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
    if ((eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_LOADSTATE) || (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_CRCERR)) {
        ESS_DBG(("WARNING: EEPROM was not loaded correctly, Reg. 0x%.4x = 0x%.4x", ESC_REG_EEPROMCTRLSTATUS, eepromCtrl));
        res = ESS_RESULT_NOT_READY;
        goto finished;
    }

    /* Wait until EEPROM interface is not busy anymore: */
    tsStart = essHALGetTime();
    while (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_BUSY) {
        if ((essHALGetTime() - tsStart) > CFG_ESS_MAX_EEPROM_WAIT) {
            res = ESS_RESULT_BUSY;
            goto finished;
        }
        ESC_READ16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
    }

    do {
        uint16_t copyLen;

        /* Make sure error bits are cleared: */
        eepromCtrl &= ~REG_MASK_EEPROMCTRLSTATUS_CMDREG;
        if (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_ERRBITS)
            ESC_WRITE16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);

        /* Write address and command: */
        ESC_WRITE32(ed, ESC_REG_EEPROMADDRESS, addr >> 1);
        eepromCtrl |= (1 << 8); /* 1 = read cmd */
        ESC_WRITE16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
        ED_DO_EVENT(ed, ESS_EVENT_WROTEEEPROMCTRL);

        /* Wait until EEPROM interface is ready: */
        tsStart = essHALGetTime();
        ESC_READ16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
        while (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_BUSY) {
            if ((essHALGetTime() - tsStart) > CFG_ESS_MAX_EEPROM_WAIT){
                res = ESS_RESULT_BUSY;
                goto finished;
            }
            ESC_READ16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
        }

        /* Check ACK: */
        if (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_ACKCMDERR) {
            if (ackRetries++ > 10) {
                res = ESS_RESULT_MISSING_ACK;
                goto finished;
            }
            DELAY(CFG_ESS_EEPROM_ACK_WAIT);
            /* ESS_DBG(("INFO: EEPROM read access returned ACK error, retrying cmd...")); */
            continue;
        }
        ackRetries = 0;

        /* Get data: */
        copyLen = (uint16_t)(MIN((eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_READ8BYTES) ? (uint32_t)8 : (uint32_t)4, len));
        ESC_READ(ed, ESC_REG_EEPROMDATA, dest, copyLen);

        addr += copyLen;
        len -= copyLen;
        dest = (uint8_t*)dest + copyLen;

    } while (len > 0);

finished:
    /* Tell ECAT we have don't want control anymore: */
    eepromAccess &= ~REG_MASK_EEPROMPDIACCESSSTATE_PDIACCESS;
    ESC_WRITE8(ed, ESC_REG_EEPROMPDIACCESSSTATE, eepromAccess);

    return res;

#endif /* !CFG_ESS_EEPROM_EMULATION */
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essEEPROMWrite(ESS_HANDLE h, uint32_t addr, uint32_t len, const void* src)
{
#if (CFG_ESS_EEPROM_EMULATION)
    ESS_CBDATA_EEPROM_EMULATION cbData;
    ESC_DATA* ed = (ESC_DATA*)h;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if (ed->config == NULL)
        return ESS_RESULT_INVALID_CONFIG;

    if ((len == 0) || (src == NULL)) {
        ESS_DBG_EEE(("EEE: Invalid len or src!"));
        return ESS_RESULT_INVALID_ARG;
    }

    cbData.hDev = h;
    cbData.destSrc = (uint8_t*)src;
    cbData.offset = addr;
    cbData.len = len;
    cbData.result = 0xff;
    cbData.isRead = ESS_FALSE;

    ED_DO_CALLBACK(ed, cbEEPROMEmulation, &cbData);

#if (DEBUG_EEPROM_EMULATION)
    if (cbData.result) {
        ESS_DBG_EEE(("EEE: write failed for %u bytes to addr 0x%.4x (word 0x%.2x)", len, addr, addr >> 1));

    } else {
        if (len >= 2)
            ESS_DBG_EEE(("EEE: write OK for %u bytes to addr 0x%.4x (word 0x%.2x): bytes (hex) = %.2x %.2x%s", len, addr, addr >> 1, cbData.destSrc[0], cbData.destSrc[1], (len > 2) ? "..." : ""));
        else
            ESS_DBG_EEE(("EEE: write OK for %u bytes to addr 0x%.4x (word 0x%.2x)", len, addr, addr >> 1));
    }

#endif

    return (cbData.result == 0) ? ESS_RESULT_SUCCESS : ESS_RESULT_UNKNOWN;

#else

    ESC_DATA* ed = (ESC_DATA*)h;
    uint16_t eepromCtrl;
    uint8_t eepromConfig;
    uint8_t eepromAccess;
    ESS_TIMESTAMP tsStart;
    uint8_t ackRetries = 0;
    ESS_RESULT res = ESS_RESULT_SUCCESS;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if ((addr & 1) || (len & 1) || (len == 0) || (src == NULL))
        return ESS_RESULT_INVALID_ARG;

    /* Make sure we're allowed to have control: */
    ESC_READ8(ed, ESC_REG_EEPROMCONFIGURATION, eepromConfig);
    if (((eepromConfig & REG_MASK_EEPROMCONFIGURATION_PDICONTROL) == 0) || (eepromConfig & REG_MASK_EEPROMCONFIGURATION_FORCEECAT)) {
        ESS_DBG(("WARNING: ECAT has EEPROM access..."));
        return ESS_RESULT_CANT_OPEN;
    }

    /* Tell ECAT we have control: */
    ESC_READ8(ed, ESC_REG_EEPROMPDIACCESSSTATE, eepromAccess);
    eepromAccess |= REG_MASK_EEPROMPDIACCESSSTATE_PDIACCESS;
    ESC_WRITE8(ed, ESC_REG_EEPROMPDIACCESSSTATE, eepromAccess);

    /* Wait until EEPROM interface is not busy anymore: */
    tsStart = essHALGetTime();
    ESC_READ16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
    while (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_BUSY) {
        if ((essHALGetTime() - tsStart) > CFG_ESS_MAX_EEPROM_WAIT) {
            res = ESS_RESULT_BUSY;
            goto finished;
        }
        ESC_READ16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
    }

    do {

        /* Make sure error bits are cleared: */
        eepromCtrl &= ~REG_MASK_EEPROMCTRLSTATUS_CMDREG;
        if (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_ERRBITS)
            ESC_WRITE16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);

        /* Write address, data and cmd: */
        ESC_WRITE32(ed, ESC_REG_EEPROMADDRESS, addr >> 1);
        ESC_WRITE16(ed, ESC_REG_EEPROMDATA, MAKE_WORD(((const uint8_t*)src)[0], ((const uint8_t*)src)[1]));
        eepromCtrl |= (2 << 8); /* 2 = write cmd */
        ESC_WRITE16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
        ED_DO_EVENT(ed, ESS_EVENT_WROTEEEPROMCTRL);

        /* Wait until EEPROM interface is ready: */
        tsStart = essHALGetTime();
        ESC_READ16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
        while (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_BUSY) {
            if ((essHALGetTime() - tsStart) > CFG_ESS_MAX_EEPROM_WAIT){
                res = ESS_RESULT_BUSY;
                goto finished;
            }
            ESC_READ16(ed, ESC_REG_EEPROMCTRLSTATUS, eepromCtrl);
        }

        /* Check Err: */
        if (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_ERRWRITEENABLE) {
            ESS_DBG(("WARNING: EEPROM write access returned 'Write cmd without write enable' error!?"));
            res = ESS_RESULT_BUSY;
            goto finished;
        }

        /* Check ACK: */
        if (eepromCtrl & REG_MASK_EEPROMCTRLSTATUS_ACKCMDERR) {
            if (ackRetries++ > 10) {
                res = ESS_RESULT_MISSING_ACK;
                goto finished;
            }
            DELAY(CFG_ESS_EEPROM_ACK_WAIT);
            /* ESS_DBG(("INFO: EEPROM write access returned ACK error, retrying cmd... (%u bytes remain)", len)); */
            continue;
        }
        ackRetries = 0;

        addr += 2;
        len -= 2;
        src = (const uint8_t*)src + 2;

    } while (len > 0);

finished:
    /* Tell ECAT we have don't want control anymore: */
    eepromAccess &= ~REG_MASK_EEPROMPDIACCESSSTATE_PDIACCESS;
    ESC_WRITE8(ed, ESC_REG_EEPROMPDIACCESSSTATE, eepromAccess);

    return res;

#endif /* !CFG_ESS_EEPROM_EMULATION */
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essSetLEDState(ESS_HANDLE h, ESC_LED_TYPE led, ESC_LED_STATE state)
{
#if (CFG_ESS_SERVE_RUN_LED || CFG_ESS_SERVE_ERR_LED || CFG_ESS_SERVE_CUSTOM_LED)
    ESC_DATA* ed = (ESC_DATA*)h;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if ((led >= ESC_LED_TYPE_NUMBER_OF_LEDS) || (state > ESC_LED_STATE_ON))
        return ESS_RESULT_INVALID_ARG;

    ESS_DBG_LEDSERVE(("LEDs: Application sets LED %s to %s", ESC_LED_TYPE_TO_STRING(led), ESC_LED_STATE_TO_STRING(state)));
    ed->ledStates[led].state = state;

    return essHALSetLEDState(ed->devIdx, led, (state == ESC_LED_STATE_ON));
#else
    (void)h;
    (void)led;
    (void)state;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif
}

ESS_EXPORT void ESS_CALLTYPE essESCRead(ESS_HANDLE h, uint16_t address, void* dest, uint16_t len)
{
    _ESSHANDLE_IF_INVALID_RETURN(h);

    ESC_READ((ESC_DATA*)h, address, dest, len);
}

ESS_EXPORT void ESS_CALLTYPE essESCWrite(ESS_HANDLE h, uint16_t address, const void* src, uint16_t len)
{
    _ESSHANDLE_IF_INVALID_RETURN(h);

    ESC_WRITE((ESC_DATA*)h, address, src, len);
}

ESS_EXPORT uint8_t ESS_CALLTYPE essESCRead8(ESS_HANDLE h, uint16_t address)
{
    uint8_t res;

    _ESSHANDLE_IF_INVALID_RETURN_(h, 0);

    ESC_READ8((ESC_DATA*)h, address, res);

    return res;
}

ESS_EXPORT uint16_t ESS_CALLTYPE essESCRead16(ESS_HANDLE h, uint16_t address)
{
    uint16_t res;

    _ESSHANDLE_IF_INVALID_RETURN_(h, 0);

    ESS_ASSERT((address & 1) == 0, ("%s (0x%.4x)", "Unaligned access!", address));

    ESC_READ16((ESC_DATA*)h, address, res);

    return res;
}

ESS_EXPORT uint32_t ESS_CALLTYPE essESCRead32(ESS_HANDLE h, uint16_t address)
{
    uint32_t res;

    _ESSHANDLE_IF_INVALID_RETURN_(h, 0);

    ESS_ASSERT((address & 3) == 0, ("%s (0x%.4x)", "Unaligned access!", address));

    ESC_READ32((ESC_DATA*)h, address, res);

    return res;
}

ESS_EXPORT void ESS_CALLTYPE essESCWrite8(ESS_HANDLE h, uint16_t address, uint8_t value)
{
    ESC_WRITE8((ESC_DATA*)h, address, value);
}

ESS_EXPORT void ESS_CALLTYPE essESCWrite16(ESS_HANDLE h, uint16_t address, uint16_t value)
{
    ESS_ASSERT((address & 1) == 0, ("%s (0x%.4x)", "Unaligned access!", address));

    ESC_WRITE16((ESC_DATA*)h, address, value);
}

ESS_EXPORT void ESS_CALLTYPE essESCWrite32(ESS_HANDLE h, uint16_t address, uint32_t value)
{
    ESS_ASSERT((address & 3) == 0, ("%s (0x%.4x)", "Unaligned access!", address));

    ESC_WRITE32((ESC_DATA*)h, address, value);
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essEoESendFrame(ESS_HANDLE h, const void* data, uint16_t dataLen, void* reserved)
{
#if (CFG_ESS_SUPPORT_EOE)

    ESC_DATA* ed = (ESC_DATA*)h;
    (void)reserved;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    if ((ed->eoeSend.active) || (ed->mbxOutInterruptPending))
        return ESS_RESULT_TRY_AGAIN;

    if (ed->mbxActive == ESS_FALSE)
        return ESS_RESULT_WRONG_ECAT_STATE;

    if (dataLen > sizeof(ed->eoeSend.data))
        return ESS_RESULT_INVALID_SIZE;

    ed->eoeSend.dataSent = 0;
    ed->eoeSend.dataLen = dataLen;
    ed->eoeSend.fragmNo = 0;
    ed->eoeSend.frameNo++;

    ESS_MEMCOPY(ed->eoeSend.data, data, dataLen);
    ed->eoeSend.active = ESS_TRUE;

    return ESS_RESULT_SUCCESS;

#else
    (void)h;
    (void)data;
    (void)dataLen;
    (void)reserved;

    return ESS_RESULT_NOT_IMPLEMENTED;
#endif

}

ESS_EXPORT void* ESS_CALLTYPE essGetTag(ESS_HANDLE h)
{
    _ESSHANDLE_IF_INVALID_RETURN_(h, NULL);

    return ((ESC_DATA*)h)->config->tag;
}

ESS_EXPORT ESS_RESULT ESS_CALLTYPE essIoctl(ESS_HANDLE h, ESS_IOCTL fn, void* data, uint32_t sizeOfData)
{
    ESC_DATA* ed = (ESC_DATA*)h;

    _ESSHANDLE_IF_INVALID_RETURN_(h, ESS_RESULT_INVALID_HANDLE);

    switch (fn) {
        case 0x12d1e: /* Used to test if debugging works... just setting ed2 to NULL and accessing it (while trying to avoid static code analyzers complains :) */
            {
                ESC_DATA* volatile ed2 = ed;
                ed2 = (ESC_DATA*)(ed - ed2);
                ed2->magic++;
                return ESS_RESULT_SUCCESS;
            }

#if (!CFG_ESS_DISABLE_DEPRECATED)
        case 0x66600001: /* essSyncInputs() was added later, earlier this debug function was a quick hack for the same behavior */
            return essSyncInputs(h, (ESS_ISYN_FLAGS)(1 << ((uintptr_t)data & 0xf)));
#endif

#ifdef ESS_DEBUG_ENABLE_REGACCTEST1
        case 0x7E570002: /* Used for testing essESCRead()/essESCWrite() in ESC reg. 0x0f80..0x0fff */
            {
                static const uint16_t ADDRESS = 0x0f80;
                volatile uint32_t tWrite;
                volatile uint32_t tRead;
                uint8_t buf0[128];
                uint8_t buf1[128];
                uint32_t ts;
                int i, j;

                ESS_DBG(("DEBUG: Testing ESC%u read access...", ed->devIdx));
                ts = essHALGetTime();
                essESCRead(h, ADDRESS, buf0, sizeof(buf0));
                for (i = 0; i < 1000; ++i) {
                    essESCRead(h, ADDRESS, buf1, sizeof(buf0));

                    for (j = 0; j < sizeof(buf0); ++j)
                        if (buf0[j] != buf1[j])
                            return ESS_RESULT_UNKNOWN;
                }
                tRead = essHALGetTime() - ts;

                ESS_DBG(("DEBUG: Testing ESC%u write access...", ed->devIdx));
                ts = essHALGetTime();
                for (i = 0; i < 1000; ++i) {
                    for (j = 0; j < sizeof(buf0); ++j)
                        buf0[j] = i + j;

                    essESCWrite(h, ADDRESS, buf0, sizeof(buf0));

                    essESCRead(h, ADDRESS, buf1, sizeof(buf0));
                    for (j = 0; j < sizeof(buf0); ++j)
                        if (buf0[j] != buf1[j])
                            return ESS_RESULT_UNKNOWN;
                }
                tWrite = essHALGetTime() - ts;

                ESS_DBG(("DEBUG: ESC%u register access test succeeded. (Read took %u ms, write took %u ms)", ed->devIdx, tRead, tWrite));
            }
            return ESS_RESULT_SUCCESS;
#endif


        default:
            return essHALIoctl(ed->devIdx, fn, data, sizeOfData);
    }
}


#if (CFG_ESS_MAX_SM_COUNT > 16)
# error "CFG_ESS_MAX_SM_COUNT must not be greater than 16!"
#endif

